/* TODO: Write an appropiate introduction to this file  */

#include <gcc-plugin.h>
#include <coretypes.h>
#include <gimple.h>
#include <tree.h>
#include <tree-flow.h>
#include <tree-pass.h>

#include "debug-specprag.h"
#include <string.h>


/* Test function that only prints a text  */
void
test_print(void)
{
	printf("Inserción satisfactoria\n");
}



/* Function to get the name of the name of the variable inside a clause */
const char * 
var_name_in_clause(tree clause)
{
	/* Get the declaration of the variable in the clause */
	tree var_decl = OMP_CLAUSE_DECL (clause);
						
	/* Get the name of the variable in the clause */
	//tree t_clause_name = DECL_NAME (var_decl);
	//const char * clause_name = IDENTIFIER_POINTER (t_clause_name);
	const char * clause_name = IDENTIFIER_POINTER (DECL_NAME(var_decl));
	
	return clause_name;
}


/* Print each OpenMP clause, with the name of the variable  */
void 
print_clauses (tree clauses)
{
	tree c; /* For clauses iteration */
	
	/* Traversing each clause in the pragma */
	for (c = clauses; c; c = OMP_CLAUSE_CHAIN(c))
	{
		/* Get the tree_code_name */
		//printf("Tree code name: %s\n",tree_code_name[TREE_CODE(c)]);
		switch (OMP_CLAUSE_CODE (c))
		{
		case OMP_CLAUSE_PRIVATE:
			printf("\tPrivate Clause: '%s'\n", var_name_in_clause(c));
			break;
						
		case OMP_CLAUSE_SHARED:
			printf("\tShared clause: '%s'\n", var_name_in_clause(c));
			break;

		case OMP_CLAUSE_SPECULATIVE:
			printf("\tSpeculative clause: '%s'\n", var_name_in_clause(c));
			break;
		default:
			printf("\tAnother kind of clause\n");

		} //SWITCH OMP_CLAUSE_CODE
	} //FOR that travers clauses
	printf("\n");
}


/* Function to insert a printf. 
	PRE: TREE VARIABLE has to be a VAR_DECL tree.
*/
/* A printf() statement is change into following two gimple statements:
 * 
 * <D.XXX> = (const char * restrict) &"This a random string\n"[0]
 * printf(<D.XXX>);
 */
/* TODO: Do something with the third parameter.  */
void
insert_printf (gimple_stmt_iterator *ptr_gsi, const char * comment, tree variable)
{
	tree var_for_array_ref;

	/* If variable is a VAR_DECL : Do nothing.  */
	/* If variable is an ARRAY_REF: Adjust to use as parameters of the printf.  */
	if (TREE_CODE (variable) == ARRAY_REF) {
		
		var_for_array_ref = build_decl (UNKNOWN_LOCATION, VAR_DECL,
																			get_identifier("aux_pre_printf"),
																			integer_type_node);
		DECL_COMMON (var_for_array_ref) = 1;
		TREE_USED (var_for_array_ref) = 1;
		TREE_PUBLIC (var_for_array_ref) = 1;

		gimple variable_assign = gimple_build_assign (var_for_array_ref, variable);
		gsi_insert_after (ptr_gsi, variable_assign, GSI_NEW_STMT);

	}

	/* Generate a STRING_CST  */
	tree note = build_string (strlen(comment),comment);
	
	/* We need to set the TREE_TYPE manually  */
	TREE_TYPE(note) = build_array_type( char_type_node, 
												build_index_type(
												build_int_cst( NULL_TREE, strlen( comment)) ));
	TREE_READONLY (note) = 1;
	TREE_STATIC (note) = 1;

	
	/* GCC needs a const char * restrict type node to convert
	*  a STRING_CST from AST to Gimple  */
	tree const_char_restrict_ptr_type_node = 
		build_qualified_type(
			build_pointer_type(
				build_qualified_type(
					char_type_node,	TYPE_QUAL_CONST)), TYPE_QUAL_RESTRICT);
	
	/* In AST, to use STRING_CST, the form is like this
	* <ADDR_EXPR<STRING_CST>>,
	* but when we turn to gimple, it is like this
	* <ADDR_EXPR<ARRAY_REF<STRING_CST>>>.
	*/
	tree note_arraywrap = build4 (ARRAY_REF, char_type_node, note, 
													build_int_cst(NULL_TREE, 0), NULL_TREE, NULL_TREE);	
	tree note_addrwrap = build1 (ADDR_EXPR, 
															const_char_restrict_ptr_type_node, note_arraywrap);

	/* FIXME: Is it really necessary?  */
	//note_addrwrap = fold_convert (const_char_restrict_ptr_type_node, note_addrwrap);				

	/*
	* If is_gimple_min_invariant(note_addrwrap) is true, 
	* we build a corrent argument, otherwise the argument
	* is not suitable for gimple call 
  */
  if(!is_gimple_min_invariant(note_addrwrap))
  {
		printf("Something wrong with is_gimple_min_invariant\n");
    return ;
  }

	/* This create a temporary var  */
	tree tmp_var = build_decl (UNKNOWN_LOCATION, VAR_DECL, NULL,
															const_char_restrict_ptr_type_node);
	TREE_USED (tmp_var) = 1;
	//This following two alternatives don't work
	//tree tmp_var = make_rename_temp (const_char_restrict_ptr_type_node, "plugin_var");
	//tree tmp_var = create_tmp_var(const_char_restrict_ptr_type_node, NULL);
	
	/* Build the GIMPLE_ASSIGN statement and insert it */
	gimple tmp_gstmt = gimple_build_assign (tmp_var, note_addrwrap);
	gsi_insert_after(ptr_gsi, tmp_gstmt, GSI_NEW_STMT);
	
	/* printf is a builtin function */
	tree funcdecl = built_in_decls[BUILT_IN_PRINTF];
	
	/* Build the GIMPLE_CALL and insert it  */
	gimple gcall;
	
	if (TREE_CODE (variable) == VAR_DECL)
		gcall = gimple_build_call (funcdecl, 2, tmp_var, variable);

	else if (TREE_CODE (variable) == ARRAY_REF)
		gcall = gimple_build_call (funcdecl, 2, tmp_var, var_for_array_ref);

	else
		gcall = gimple_build_call (funcdecl, 1, tmp_var);
		
	gsi_insert_after(ptr_gsi, gcall, GSI_NEW_STMT);

}


/* Function that only prints a VAR_DECL tree that corresponds with the 
the name of the variable passed as argument.  */
void
debug_var_decl_tree (tree variable, const char * variableName)
{	
	tree varDecl = variable; //By default, varDecl is a VAR_DECL

	/* If variable is an ARRAY_REF: Extract the VAR_DECL from the ARG 0.  */
	if (TREE_CODE(varDecl) == ARRAY_REF) {
		//debug_tree(variable);
		varDecl = TREE_OPERAND(variable,0);
	}

	if (TREE_CODE(varDecl) == VAR_DECL) {
		
		/* Extract the variable name from the VAR_DECL tree.  */
		const char * nameInTree = "";
		
		/* We omit artificial variables created by the compiler.  */
		if (DECL_NAME(varDecl) == NULL)
			return;
			
		nameInTree = IDENTIFIER_POINTER (DECL_NAME (varDecl));
		
		/* Only prints the variable if its name matches with variableName.  */
		if ( strcmp( nameInTree, variableName) == 0  ) {
			printf("Printing the tree of the variable '%s'\n", nameInTree);
			debug_tree(varDecl);
			printf("\n");
		}	
	}	

}


/* TODO: REMOVE THIS FUNCTION. IT JUST A COPY NOT USED.  */
/* Given an expression tree, EXPR, that may contains SSA names, returns an
   equivalent tree with the SSA names converted to var/parm/field decls
   so that it can be used with '%E' format modifier when emitting warning
   messages.

   This function currently only supports VAR/PARM/FIELD_DECL, reference
   expressions (COMPONENT_REF, INDIRECT_REF, ARRAY_REF), integer constant,
   and SSA_NAME. If EXPR contains any other tree nodes (e.g. an arithmetic
   expression appears in array index), NULL_TREE is returned.  */

tree
get_non_ssa_expr (tree expr)
{
  switch (TREE_CODE (expr))
    {
      case VAR_DECL:
      case PARM_DECL:
      case FIELD_DECL:
        {
          if (DECL_NAME (expr))
            return expr;
          else
            return NULL_TREE;
        }
      case COMPONENT_REF:
        {
          tree base, orig_base = TREE_OPERAND (expr, 0);
          tree component, orig_component = TREE_OPERAND (expr, 1);
          base = get_non_ssa_expr (orig_base);
          if (!base)
            return NULL_TREE;
          component = get_non_ssa_expr (orig_component);
          if (!component)
            return NULL_TREE;
          /* If either BASE or COMPONENT is converted, build a new
             component reference tree.  */
          if (base != orig_base || component != orig_component)
            return build3 (COMPONENT_REF, TREE_TYPE (component),
                           base, component, NULL_TREE);
          else
            return expr;
        }
      case MEM_REF:
        {
          tree orig_base = TREE_OPERAND (expr, 0);
	  if (TREE_CODE (orig_base) == SSA_NAME)
	    {
	      tree base = get_non_ssa_expr (orig_base);
	      if (!base)
		return NULL_TREE;
	      return fold_build2 (MEM_REF, TREE_TYPE (expr),
				  base, TREE_OPERAND (expr, 1));
	    }
	  return expr;
        }
      case ARRAY_REF:
        {
          tree array, orig_array = TREE_OPERAND (expr, 0);
          tree index, orig_index = TREE_OPERAND (expr, 1);
          array = get_non_ssa_expr (orig_array);
          if (!array)
            return NULL_TREE;
          index = get_non_ssa_expr (orig_index);
          if (!index)
            return NULL_TREE;
          /* If either ARRAY or INDEX is converted, build a new array
             reference tree.  */
          if (array != orig_array || index != orig_index)
            return build4 (ARRAY_REF, TREE_TYPE (expr), array, index,
                           TREE_OPERAND (expr, 2), TREE_OPERAND (expr, 3));
          else
            return expr;
        }
      case SSA_NAME:
        {
          tree vdecl = SSA_NAME_VAR (expr);
          if (DECL_ARTIFICIAL (vdecl)
              && !gimple_nop_p (SSA_NAME_DEF_STMT (expr)))
            {
              gimple def_stmt = SSA_NAME_DEF_STMT (expr);
              if (gimple_assign_single_p (def_stmt))
                vdecl = gimple_assign_rhs1 (def_stmt);
            }
          return get_non_ssa_expr (vdecl);
        }
      case INTEGER_CST:
        return expr;
      default:
        /* Return NULL_TREE for any other kind of tree nodes.  */
        return NULL_TREE;
    }
}

