/* Handling statements functionality
	
 	2013.
 	Written by Sergio Aldea <sergio@infor.uva.es>

This file is part of specprag.  */

#include <gcc-plugin.h>
#include <coretypes.h>
#include <gimple.h>
#include <tree.h>
#include <tree-flow.h>
#include <tree-pass.h>


/* This function returns the gimple_stmt_iterator of the first 
	statement in the main function.  */
gimple_stmt_iterator
gsi_init_main(gimple_seq seq)
{
	gimple stmt;
	gimple_seq seq_child;
	gimple_stmt_iterator gsi_return;
	
	/* Get the first statement of the gimple sequence.  */
	stmt = gimple_seq_first_stmt (seq);
	
	if ( gimple_code(stmt) == GIMPLE_BIND )
	{	
		/* Get first stmt of GIMPLE_BIND  */
		seq_child = gimple_bind_body(stmt);
		gsi_return = gsi_init_main(seq_child);
	
	} else {
		gsi_return = gsi_start (seq);
	}

	return gsi_return;
}

/* This function returns the gimple_stmt_iterator of the last 
	statement in the main function.  */
gimple_stmt_iterator
gsi_last_main (gimple_seq seq)
{
	gimple stmt;
	gimple_seq seq_child;
	gimple_stmt_iterator gsi_return;

	/* Get the last statement of the gimple sequence.  */
	stmt = gimple_seq_last_stmt (seq);
	
	if ( gimple_code(stmt) == GIMPLE_BIND )
	{	
		/* Get last stmt of GIMPLE_BIND  */
		seq_child = gimple_bind_body(stmt);
		gsi_return = gsi_last_main(seq_child);
	
	} else {
		gsi_return = gsi_last (seq);
	}
	
	return gsi_return;
}


/* Function that checks whether a PARALLEL directive is associated 
		with a FOR directive. It returns 'true' if there is a FOR directive
		and 'false' in the other case.
		
		Structure is the following:
				
				GIMPLE_OMP_PARALLEL
				->	GIMPLE_BIND
					->	GIMPLE_OMP_FOR
		*/
bool
forAfterParallel (gimple body_stmt)
{
	gimple_seq body_seq;
	gimple_stmt_iterator gsi;
	gimple stmt;

	/* GIMPLE_BIND variables. */
	gimple_seq bind_body_seq;
	gimple_stmt_iterator bind_gsi;
	gimple bind_stmt;

	/* Get the OMP body of the PARALLEL directive.  */
	body_seq = gimple_omp_body(body_stmt);

	printf("Searching FOR directive associated with the PARALLEL directive...\n");
	for (gsi = gsi_start (body_seq); !gsi_end_p (gsi); gsi_next (&gsi))
	{
		stmt = gsi_stmt(gsi);
			
		if (gimple_code(stmt) == GIMPLE_BIND) {
			//printf("GIMPLE_BIND located\n");
				
			bind_body_seq = gimple_bind_body(stmt);
			bind_gsi = gsi_start(bind_body_seq);
			bind_stmt = gsi_stmt (bind_gsi);
			
			//debug_gimple_stmt(bind_stmt);
			
			if (gimple_code(bind_stmt) == GIMPLE_OMP_FOR) {
				//printf("GIMPLE_OMP_FOR located\n");
				return true;
			}
		}
	}
	
	/* If there is NOT a FOR directive associated with the PARALLEL directive
	returns FALSE.  */
	return false;
}

