/* Generation of functions which are copy of those called inside the body
	of the OpenMP FOR loops marked as speculative loops.  
	
	2013.
	Written by Sergio Aldea <sergio@infor.uva.es>

This file is part of specprag.  */ 

#include <gcc-plugin.h>
#include <coretypes.h>
#include <gimple.h>
#include <tree.h>
#include <tree-flow.h>
#include <tree-pass.h>

#include <stdio.h>
#include "handle-stmts.h"
#include "spec-functions.h"


/* Global variable used by replace_...STORE and READ functions.  */
tree value;

/* Global variable to represent the current PARAM_DECL.  */
tree current_param;

/* Global variables for the label at the end of the copied function.  */
tree labelSquash_decl;
gimple labelSquash;



/*****************************************************************************************
TODO: CURRENTLY UNUSED.
Functions to save all the functions detected in the source code, and their body.
******************************************************************************************/
void init_All_Functions (All_Functions *a) {
	memset(&(a->name[0]), 0, sizeof(a->name));	
	memset(&(a->func_decl[0]), 0, sizeof(a->func_decl));	
	memset(&(a->func_body[0]), 0, sizeof(a->func_body));	
  a->used = 0;
}

void insert_All_Functions (All_Functions *a, tree function) {
	a->name[a->used] = IDENTIFIER_POINTER (DECL_NAME (function));
	a->func_decl[a->used] = function;
	a->func_body[a->used] = gimple_body (function);
	a->used++;
}

gimple_seq recovery_function_body (All_Functions *a, tree function) {
	
	int i;
	
	for (i = 0; i < a->used; i++) {
		if (operand_equal_p (function, a->func_decl[i], OEP_PURE_SAME)) {
			return a->func_body[i];
		}
	}
	return NULL;
}
/*****************************************************************************************/

/*****************************************************************************************
 Functions to control and save the functions detected in the source code as containers
of speculative variables.
******************************************************************************************/

/* Initialize a Function.  */
void init_Function(Function *a) {
	memset(&(a->name[0]), 0, sizeof(a->name));	
	memset(&(a->copied_function[0]), 0, sizeof(a->copied_function));	
	memset(&(a->body_spec[0]), 0, sizeof(a->body_spec));	
  a->used = 0;
}

/* Insert a new function in the struct.  */
void insert_Function(Function *a, const char * original_name, tree function) {
 	
	a->name[a->used] = original_name;

	char * spec_name = NULL;
	int ss = asprintf(&spec_name, "%s%s", original_name, "_spec");
	
	a->name_spec[a->used] = spec_name;
	a->copied_function[a->used] = function;
	a->used++;
}


/* Search for a original function, and if it is found, copy its body.  */
void search_Original_Function (Function *a, tree current_function) {
	int i;
	gimple_seq original_body, copied_body;
	
	/* Get the name of the function.  */
	const char * original_name = " ";
	original_name = IDENTIFIER_POINTER (DECL_NAME(current_function));
	
	/* Search for a function with that name.  */
	for (i = 0; i < a->used; i++) {
		if (strcmp (a->name[i], original_name) == 0) {
			// Add the body of the function.
			original_body = gimple_body (current_function);
			copied_body = (gimple_seq) copy_gimple_seq_and_replace_locals (original_body);
			a->body_spec[i] = copied_body;
		}
	}
}

/* Search for a copied version of a function, and assign the body of the original.  */
bool search_Copied_Function (Function *a, tree current_function) {
	
	int i;
	bool found = false; // Initially, not found
	
	/* Get the name of the function.  */
	const char * copied_name = " ";
	copied_name = IDENTIFIER_POINTER (DECL_NAME(current_function));
	

	/* Search for a function with that name.  */
	for (i = 0; i < a->used; i++) {
		if (strcmp (a->name_spec[i], copied_name) == 0) {
			
			found = true;

			/* If we detect the copied version of the function, assign the body previously gotten.  */
			tree bl = make_node(BLOCK);
			BLOCK_SUPERCONTEXT(bl) = current_function;
			DECL_INITIAL(current_function) = bl;
			BLOCK_VARS(bl) = gimple_bind_vars (gsi_stmt (gsi_start (a->body_spec[i]))); // NULL_TREE;
			//BLOCK_VARS(bl) = NULL_TREE;
			TREE_USED(bl) = 1;
			//tree bind = build3(BIND_EXPR, void_type_node, BLOCK_VARS(bl), NULL_TREE, bl);
			//TREE_SIDE_EFFECTS(bind) = 1;
			//DECL_SAVED_TREE(current_function_decl) = bind;
			//DECL_SOURCE_LOCATION (fn_decl) = gimple_location (gimple_seq_first_stmt(body_copy));
		
			gimple_set_body (current_function, a->body_spec[i]); 
		}
	}

	return found;
}

/* Returns 'true' if the fn_call is the original version of the current function.  */
bool is_A_Recursive_Function ( Function *a, tree copied, tree original) {
	
	/* Get the name of the current function.  */
	const char * copied_name = " ";
	copied_name = IDENTIFIER_POINTER (DECL_NAME(copied));

	/* Get the name of the function detected (possibly the original function).  */
	const char * original_name = " ";
	original_name = IDENTIFIER_POINTER (DECL_NAME(original));

	int i;

	for (i = 0; i < a->used; i++) {
		if (strcmp (a->name_spec[i], copied_name) == 0 &&
				strcmp (a->name[i], original_name) == 0 ) {
			
			return true;
		}
	}
	
	return false;
}
/****************************************************************************************/




/* Function to copy the current function and create a new one with the 
same functionality (but different name, and a new parameter: CURRENT):

	original_function (int arg0, int arg1) {
		// body
	};
	
	creates ->

	original_function_spec (int arg0, int arg1, int current) {
		
		int new_value; //New variable, used by specload and specstore
		// body
	}
*/
tree
create_copy_function (tree original_function, tree current)
{
	gimple_seq body_copy;

	/* Get the name of the function and create a new one.  */
	const char * original_name = IDENTIFIER_POINTER (DECL_NAME (original_function));
	char * spec_name = NULL;
	int ss = asprintf(&spec_name, "%s%s", original_name, "_spec");

	printf("\tCreating a copy of the current function: %s\n", spec_name);
	
	/* Extract the body of the original function.  */
	gimple_seq body = gimple_body (original_function);
	
	//body_copy = gimple_seq_copy (body);
	body_copy = (gimple_seq) copy_gimple_seq_and_replace_locals (body);

	//printf("First instruction of the function body\n");
	//debug_gimple_stmt (gsi_stmt (gsi_start (body_copy)));
	
	/* Getting the vars in the block of the current function.  */
	//tree variables = gimple_bind_vars (gsi_stmt (gsi_start (body_copy)));
	//tree variables = gimple_bind_vars(gimple_seq_first_stmt(body_copy));
	//debug_tree (variables);
	
	/* Extract the tree_list of types.  */
	//tree parmlist = TYPE_ARG_TYPES (TREE_TYPE (original_function));
	//debug_tree (parmlist);

	/*******************************************************************************/
	/* Create the list of parameters, including the return type, and add the type of 
	the new parameter: 'int current'.  */
	/*******************************************************************************/
	//tree original_return_type = TREE_TYPE (DECL_RESULT (original_function));
	tree original_return_type = TREE_TYPE (TREE_TYPE (original_function));

	tree param;
	tree param_list = NULL_TREE;
	for (param = DECL_ARGUMENTS (original_function); param; param = DECL_CHAIN (param)) {
		tree arg_type = build_tree_list (NULL_TREE, TREE_TYPE (param));
		param_list = chainon (param_list, arg_type);
	}	
	
	/* Set the type of the new argument: 'int current'.  */
	param_list = chainon (param_list, build_tree_list (NULL_TREE, integer_type_node));
	
	/* Finish the chain with the return type required: 'int' (instead of the type of the original function).  */
	//param_list = chainon (param_list, build_tree_list (NULL_TREE, original_return_type));
	param_list = chainon (param_list, build_tree_list (NULL_TREE, integer_type_node));
	

	/**********************************************************************************/
	/* Create the function declation, with the corrected type.  */
	/**********************************************************************************/
	//tree fn_type = TREE_TYPE (original_function);
	//tree fn_type = build_function_type (original_return_type, param_list);
	tree fn_type = build_function_type (integer_type_node, param_list);
	
	//tree fn_decl = build_fn_decl (spec_name, fn_type);
	//tree fn_decl = build_decl (UNKNOWN_LOCATION, FUNCTION_DECL, 
	tree fn_decl = build_decl (BUILTINS_LOCATION, FUNCTION_DECL, 
															get_identifier(spec_name), fn_type);
	
	DECL_CONTEXT (fn_decl) = DECL_CONTEXT (original_function); //NULL_TREE;
  DECL_EXTERNAL (fn_decl) = DECL_EXTERNAL (original_function);
	TREE_STATIC (fn_decl) = TREE_STATIC (original_function);
	TREE_PUBLIC (fn_decl) = TREE_PUBLIC (original_function);
	TREE_ADDRESSABLE (fn_decl) = TREE_ADDRESSABLE (original_function);
	TREE_USED (fn_decl) = TREE_USED (original_function);
	//DECL_SOURCE_LOCATION (fn_decl) = DECL_SOURCE_LOCATION (original_function);
	
	/* Rewrite the type of the function:   
		This option is only needed if we create the function using the type of the original.  */
	//TYPE_ARG_TYPES (TREE_TYPE (fn_decl)) = param_list;

	SET_DECL_ASSEMBLER_NAME (fn_decl, get_identifier(spec_name));

	/* Set the return type as 'int', and not as the original.  */
	//tree return_type = TREE_TYPE (TREE_TYPE (original_function));
	tree return_type = integer_type_node;

	//tree result_decl = build_decl (UNKNOWN_LOCATION, RESULT_DECL, NULL_TREE,
	tree result_decl = build_decl (BUILTINS_LOCATION, RESULT_DECL, NULL_TREE,
                                 return_type);
	DECL_IGNORED_P (result_decl) = 1;
	DECL_CONTEXT (result_decl) = fn_decl;
	DECL_RESULT (fn_decl) = result_decl;
	//This sets the context and location as the original.
	//DECL_RESULT (fn_decl) = DECL_RESULT (original_function); 


	/************************************************************************/
	/* Set the parameters of the function, getting the original parameters.  */
	/************************************************************************/
	tree arglist = NULL_TREE;
	
	//tree param;
	for (param = DECL_ARGUMENTS (original_function); param; param = DECL_CHAIN (param)) {
		//debug_tree(param);
		tree param_name = DECL_NAME(param);
		//tree param_decl = build_decl (UNKNOWN_LOCATION, PARM_DECL, param_name, TREE_TYPE(param));
		tree param_decl = build_decl (BUILTINS_LOCATION, PARM_DECL, param_name, TREE_TYPE(param));
		//debug_tree(param_decl);

		DECL_ARG_TYPE (param_decl) = TREE_TYPE(param_decl);
		DECL_CONTEXT (param_decl) = fn_decl;
		//DECL_CONTEXT (param_decl) = DECL_CONTEXT (original_function);
		//DECL_SOURCE_LOCATION (param_decl) = DECL_SOURCE_LOCATION (original_function);
		TREE_USED (param_decl) = TREE_USED(param);
		TREE_ADDRESSABLE (param_decl) = TREE_ADDRESSABLE(param);
		TREE_STATIC (param_decl) = TREE_STATIC (param);
		TREE_PUBLIC (param_decl) = TREE_PUBLIC (param);
			
		//TODO: next stmt is doubtful of working correctly due to 'original_function'
		//DECL_ARG_TYPE (param_decl) = TREE_VALUE (TYPE_ARG_TYPES (TREE_TYPE (original_function)));
		//DECL_ARG_TYPE (param_decl) = TREE_VALUE (TYPE_ARG_TYPES (TREE_TYPE (fn_decl)));
		
		arglist = chainon (arglist, param_decl);
		//arglist = chainon (arglist, param);
		//tree copy_param = copy_node (param);
		//arglist = chainon (arglist, copy_param);
	}	
	
	/* Set the new parameter: 'int current'.  */
	//tree current_param = build_decl (UNKNOWN_LOCATION, PARM_DECL, 
	current_param = build_decl (BUILTINS_LOCATION, PARM_DECL, 
																		get_identifier("current"), integer_type_node);
	
	DECL_ARG_TYPE (current_param) = TREE_TYPE(current_param);
	DECL_CONTEXT (current_param) = fn_decl;
	//DECL_CONTEXT (current_param) = DECL_CONTEXT (original_function);
	//DECL_SOURCE_LOCATION (current_param) = DECL_SOURCE_LOCATION (original_function);
	TREE_USED (current_param) = 1;
	
	//TODO: next stmt is doubtful of working correctly due to 'original_function'
	//DECL_ARG_TYPE (current_param) = TREE_VALUE (TYPE_ARG_TYPES (TREE_TYPE (original_function)));
	//DECL_ARG_TYPE (current_param) = TREE_VALUE (TYPE_ARG_TYPES (TREE_TYPE (fn_decl)));
	
	arglist = chainon (arglist, current_param);
	
	/* Set the parameters to the function declaration.  */
	DECL_ARGUMENTS (fn_decl) = arglist;
	
	
	/*****************************************************/
	/* Fill the body of the function.  */
	/*****************************************************/
	tree stmts_list = NULL_TREE;

	DECL_INITIAL(fn_decl) = DECL_INITIAL (original_function);
	
	
	//debug_tree(DECL_SAVED_TREE (original_function));
	// This causes error probably because we need to copy the tree, not assign it without more operations.
	//DECL_SAVED_TREE (fn_decl) = DECL_SAVED_TREE (original_function);
	//debug_tree (DECL_SAVED_TREE (fn_decl));

	
/* WE DEFERRED THIS OPERATIONS UNTIL HAVING THE GIMPLIFY VERSION OF THE FUNCTION BODY.  */
/*	tree bl = make_node(BLOCK);
	BLOCK_SUPERCONTEXT(bl) = fn_decl;
	DECL_INITIAL(fn_decl) = bl;
	BLOCK_VARS(bl) = gimple_bind_vars (gsi_stmt (gsi_start (body_copy))); // NULL_TREE;
	//BLOCK_VARS(bl) = NULL_TREE;
	TREE_USED(bl) = 1;
	tree bind = build3(BIND_EXPR, void_type_node, BLOCK_VARS(bl), stmts_list, bl);
	TREE_SIDE_EFFECTS(bind) = 1;
	DECL_SAVED_TREE(fn_decl) = bind;
	//DECL_SOURCE_LOCATION (fn_decl) = gimple_location (gimple_seq_first_stmt(body_copy));
*/	

/*
	// Meaning: build_block (tree vars, tree subblocks, tree supercontext, tree chain)
	tree __func_art_block = build_block(NULL_TREE, NULL_TREE, NULL_TREE, NULL_TREE);
	DECL_INITIAL(fn_decl) = __func_art_block;
	DECL_SAVED_TREE(fn_decl) = build3(BIND_EXPR, void_type_node, NULL_TREE, stmts_list, __func_art_block);
	//DECL_SAVED_TREE (fn_decl) = NULL_TREE;
*/
	
	gimplify_function_tree (fn_decl);

	/* Once the function is gimplified, we can add the copied body from the original function.  */
	/* XXX: This is deferred, because in the point this function is called, we dont have a gimplify
	version of the body of the function that we need to copy.  */
	//gimple_set_body (fn_decl, body_copy); 
	//gimple_set_body (fn_decl, NULL); 
	
	cgraph_add_new_function (fn_decl, false);
	cgraph_finalize_function (fn_decl, true);
	
	//debug_tree (original_function);
	//debug_tree (fn_decl);
	return fn_decl;
}


/* Function to set the context of the local variable to the current function.  */
void
set_context_local_variables (tree copy_function)
{	
	// Get a tree BLOCK with VAR_DECLs and their scope 
	tree function_vars = DECL_INITIAL (copy_function);
	//debug_tree (function_vars);
	
	/* Traverse the chain of trees.  */
	tree var;
	for (var = BLOCK_VARS (function_vars); var; var = DECL_CHAIN (var)) {
		/* It seems that changing the context here is enough. */
		DECL_CONTEXT(var) = copy_function;
	}
}


static tree
print_tree (tree *tp, int *ws, void *data) {

	tree t = *tp;
	tree op2;
	//debug_tree(t);
	if (TREE_CODE(t) == ARRAY_REF) {
		op2 = TREE_OPERAND(t,1);
		//debug_tree (op2);
		if (op2 != NULL && TREE_CODE(op2) == PARM_DECL) {
			//TREE_OPERAND(t,1) = 
		}
	}

	return NULL_TREE;

}


/* Function to replace every PARM_DECL (which are still the original) from the new function
	 by new PARM_DECLs, matching with the parameters created to the new function.  

	 It is also necessary to replace the context of the local variables.
*/
void
replace_parm_decls (tree copy_function)
{
	printf("Replacing PARM_DECL from the new function created\n");

	//struct nesting_info *info; 
	//walk_body (convert_nonlocal_reference_stmt, convert_nonlocal_reference_op,
	//         info, gimple_body (copy_function));
	//walk_gimple_seq (gimple_body(copy_function), 
	//								convert_nonlocal_reference_stmt, 
	//								convert_nonlocal_reference_op,
	//								NULL);
	
	gimple_seq seq = gimple_body (copy_function);
	gimple_stmt_iterator gsi = gsi_start (seq);
	gimple function_body = gsi_stmt (gsi);
	gimple stmt;
	tree param;
	
	/* Get the number of arguments to set the length to the array of names.  */
	int num_args = 0;
	for (param = DECL_ARGUMENTS (copy_function); param; param = DECL_CHAIN (param)) {
			num_args++;
	}
	
	tree parameters[num_args]; 	// List of parameters
	const char * body_param_name;	// Name of the parameter in the body
	const char * function_param_name[num_args]; // List with the name of the parameters in FUNC_DECL
	
	
	/* Get the parameters to obtain their names.  */
	int i;
	for (param = DECL_ARGUMENTS (copy_function), i = 0; param; param = DECL_CHAIN (param), i++) {
		function_param_name[i] = IDENTIFIER_POINTER (DECL_NAME (param));
		parameters[i] = param;
		printf("|->Saving the parameter: %s\n", function_param_name[i]);
	}	

	/* Traverse the function body looking for PARM_DECL.  */
	for (gsi = gsi_start (gimple_bind_body(function_body)); !gsi_end_p (gsi); gsi_next (&gsi))
	{
		stmt = gsi_stmt (gsi);
		//printf("Working on statement...:\n");
		//debug_gimple_stmt(stmt);

		
		if (gimple_num_ops(stmt) != 0) {
				
			/* Get each operand to search the PARM_DECL.  */
			int i;
			tree op;
			for (i=0; i<gimple_num_ops(stmt); i++) {
				
				op = gimple_op (stmt,i);
				//debug_tree (op);
				//walk_tree (&op, print_tree, NULL, NULL);

				if (op != NULL) { // To avoid operands NULL, as in GIMPLE_CALL without LHS.
					
					if (TREE_CODE(op) == PARM_DECL) {
						body_param_name = IDENTIFIER_POINTER (DECL_NAME (op));
					
						// for each parameter
						int k;
						for (k = 0; k < num_args; k++)
							if (strcmp (body_param_name, function_param_name[k]) == 0)
								gimple_set_op (stmt, i, parameters[k]);
					} // if == PARM_DECL
					
				
					if (TREE_CODE(op) != PARM_DECL) {
						int num_ops = TREE_OPERAND_LENGTH(op);
						//printf("Numero de operandos : %d\n", num_ops);
						//debug_tree(op);

						// for each operand, replace it by the one defined in the function
						int j;
						for (j = 0; j < num_ops; j++) {
							
							/*********************************************************/
							/* TODO: Do it RIGHT. Consider multidimensional arrays   */
							/* Right now, only operates well with one or two DIM     */
							if (TREE_OPERAND (op,j) != NULL && TREE_CODE(TREE_OPERAND(op,j)) == ARRAY_REF) {
								//debug_tree (TREE_OPERAND (op,j));
								//debug_tree (TREE_OPERAND(TREE_OPERAND (op,j),1));
								
								// Omit artificial variables.
								if (DECL_NAME (TREE_OPERAND(TREE_OPERAND(op,j),1)) != NULL) {

									body_param_name = IDENTIFIER_POINTER (DECL_NAME (TREE_OPERAND(TREE_OPERAND(op,j),1)));			
								
									// for each parameter
									int k;
									for (k = 0; k < num_args; k++)
										if (strcmp (body_param_name, function_param_name[k]) == 0) 
											TREE_OPERAND (TREE_OPERAND (op,j),1) = parameters [k];
								}
							}
							
							/*********************************************************/
							
							if (TREE_OPERAND(op,j) != NULL && TREE_CODE (TREE_OPERAND(op,j)) == PARM_DECL) {
								
								/* Omit artificial variables.  */
								if (DECL_NAME (TREE_OPERAND(op,j)) != NULL) {

									body_param_name = IDENTIFIER_POINTER (DECL_NAME (TREE_OPERAND(op,j)));
									// for each parameter
									int k;
									for (k = 0; k < num_args; k++)
										if (strcmp (body_param_name, function_param_name[k]) == 0) 
											TREE_OPERAND (op,j) = parameters[k];
								}		
							}
						} // num_ops
					} // if != PARM_DECL
				
				
				} // op != NULL


			} //for gimple_num_ops
		} // if gimple_num_ops 

	} // for gsi

}


/* Function to create the variables needed inside the function which contains 
the speculative read and stores.  */
void
prepare_copied_function (tree copy_function)
{	
	
	/* Get the pointer to the last statement of the FOR loop.  */
	gimple_seq body_bind; // Body inside a GIMPLE_BIND
	gimple_seq body;			// Body without the bind
	
	body_bind = gimple_body (copy_function);
	body = gimple_bind_body (gimple_seq_first_stmt (body_bind));

	gimple_stmt_iterator gsi_last_stmt = gsi_last (body);

	/***********/
	/* VALUE.  */
	/***********/
	value = build_decl (UNKNOWN_LOCATION, VAR_DECL,
														get_identifier("value_function"),
														integer_type_node);
	DECL_COMMON (value) = 1;
	TREE_USED (value) = 1;
	TREE_PUBLIC (value) = 1;
	
	// This option is mandatory. If not -> ERROR at using '&value'.
	// However, this makes imposible to use 'value' in the IF condition.
	TREE_STATIC (value) = 1; 
	// This option also makes imposible to use 'value' in the IF condition.
	TREE_ADDRESSABLE (value) = 1;	
	
	layout_decl(value, false);
	rest_of_decl_compilation(value, 1, 0);

	/*********************************************/
	/* Label Squash at the end of the function.  */
	/*********************************************/
	labelSquash_decl = create_artificial_label (UNKNOWN_LOCATION);
	labelSquash = gimple_build_label (labelSquash_decl);

	/* Build 'return 0;' and 'return 1;'.  */
	gimple return_0 = gimple_build_return (build_int_cst (integer_type_node, 0));
	gimple return_1 = gimple_build_return (build_int_cst (integer_type_node, -1));
	
	/* Insert the returns and label at the end of the function.  */
	gsi_insert_after (&gsi_last_stmt, return_0, GSI_NEW_STMT);
	gsi_insert_after (&gsi_last_stmt, labelSquash, GSI_NEW_STMT);
	gsi_insert_after (&gsi_last_stmt, return_1, GSI_NEW_STMT);
	//gsi_insert_after (&gsi_last_stmt, gimple_build_nop(), GSI_NEW_STMT);
}


/* Function to replace a reading from an speculative variable for a 
*  speculative load operation.  */

void
replace_speculative_load_in_function (gimple_stmt_iterator *ptr_gsi, tree lhs, tree variable, tree current_t_alone)
{	

	tree specload, earlySquash;
	tree size;
	tree int_ptr_variable;
	tree label_true_decl, label_false_decl;


	/* Create the type (unsigned char *).  */
	tree unsigned_char_ptr_type_node = build_pointer_type (unsigned_char_type_node);
	
	/* Create the parameters:
		- current
		- value 
		- sizeof(variable)
		- (unsigned char*) &value
		- (unsigned char*) &variable
	*/

	/***********/
	/* current */
	/***********/
	/* It is already created at the beginning of the plugin as global variable.  */
	
	/*********/
	/* value */
	/*********/
	/* It is already created at the beginning of the plugin as global variable.  */
	
	/***************/
	/* pre_if_var */
	/***************/
	tree pre_if_var = build_decl (UNKNOWN_LOCATION, VAR_DECL, 
																		get_identifier ("pre_if_var"),
																		//NULL,
																		integer_type_node);
	DECL_IGNORED_P (pre_if_var) = 1;

	/********************/
	/* sizeof(variable)	*/
	/********************/
	/* Get the size of the variable for the second argument of specload_pointer.  */
	/* TODO: Distinguish between arrays and single variables.  */
	/* In arrays we search for the size of an array element. 'variable' is an 'array_ref' */
	tree variable_element_size = TYPE_SIZE_UNIT( TREE_TYPE (variable));
	
	/*********************************/
	/* (unsigned char *) &variable.  */
	/*********************************/
	/* TODO: If the variable is not an array, probably this stuff is not correct.  */

	// These lines solves cases in which the speculative 
	//variable doesnt have the ADDRESSABLE bit activated.
	if ((TREE_CODE(variable) == VAR_DECL) &&
			(TREE_ADDRESSABLE(variable)==0) ) 
	{
		TREE_ADDRESSABLE (variable) = 1;
	}	
	else if ( (TREE_CODE(variable) == ARRAY_REF) && 
						(TREE_ADDRESSABLE ( TREE_OPERAND (variable, 0)) == 0) )
	{
		TREE_ADDRESSABLE ( TREE_OPERAND(variable, 0) ) = 1; 
	}

	/* If the variables is not recognized as VAR_DECL, or ARRAY_REF, exit.  */
	if ((TREE_CODE(variable) != VAR_DECL && TREE_CODE(variable) != ARRAY_REF))
	{
		printf("Speculative variable is not an ARRAY_REF or a VAR_DECL\n");
		debug_tree(variable);
		exit(-1);
	}
	
	tree variable_addr = build_fold_addr_expr (variable);
	tree pre_unsigned_char_ptr_var = fold_convert (unsigned_char_ptr_type_node, variable_addr);
	
	
	/* TODO: Check if this patch is correct. It is only necessary when the speculative variable is 
	* a global variable. Other solution is to not doing the fold_convert, and directly 
	* build the assignment with 'variable_addr' .  */ 
	/* This macro is necessary for global variables. Local variables are not affected.  */
	STRIP_USELESS_TYPE_CONVERSION (pre_unsigned_char_ptr_var);

	tree unsigned_char_ptr_var = build_decl (UNKNOWN_LOCATION, VAR_DECL, 
																								//get_identifier("specload_address"),
																								NULL, 
																								unsigned_char_ptr_type_node);
	
	TREE_USED (unsigned_char_ptr_var) = 1;
	DECL_IGNORED_P (unsigned_char_ptr_var) = 1;

	gimple aux_stmt = gimple_build_assign (unsigned_char_ptr_var, pre_unsigned_char_ptr_var);
	gsi_insert_before (ptr_gsi, aux_stmt, GSI_SAME_STMT);

	
	/****************************/
	/* (unsigned char *) &value */
	/****************************/
	tree value_addr = build_fold_addr_expr (value);
	tree pre_unsigned_char_ptr_var2 = fold_convert (unsigned_char_ptr_type_node, value_addr);
	tree unsigned_char_ptr_var2 = build_decl (UNKNOWN_LOCATION, VAR_DECL, 
																								//get_identifier("specload_value"),
																								NULL, 
																								unsigned_char_ptr_type_node);
	TREE_USED (unsigned_char_ptr_var2) = 1;
	DECL_IGNORED_P (unsigned_char_ptr_var2) = 1;
	
	gimple aux_stmt2 = gimple_build_assign (unsigned_char_ptr_var2, pre_unsigned_char_ptr_var2);
	gsi_insert_before (ptr_gsi, aux_stmt2, GSI_SAME_STMT);



	/* Build the function prototype for specload:
	* 		int specload_pointer (unsigned char *, sizeof, int, unsigned char *)    */
	tree specload_pointer_type = build_function_type_list (integer_type_node,
																unsigned_char_ptr_type_node,
																integer_type_node,
																integer_type_node,
																unsigned_char_ptr_type_node,
																NULL_TREE);

	/* Create the function declaration, the function call and a temporal variable:
		
			D.xxxx = specload_pointer(..);
			if ( D.xxxx == -1 )
	*/
	tree specload_pointer = build_fn_decl ("specload_pointer", specload_pointer_type);
	TREE_ADDRESSABLE (specload_pointer) = 1;
	TREE_USED (specload_pointer) = 1;
	TREE_PUBLIC (specload_pointer) = 1;
	TREE_STATIC (specload_pointer) = 1;
	TREE_NOTHROW (specload_pointer) = 0;

	gimple specload_pointer_call = gimple_build_call (specload_pointer, 4, 
																	unsigned_char_ptr_var,
																	variable_element_size,
																	current_param,
																	unsigned_char_ptr_var2);

	
	/* Insert 'current.x = current' before the function call.  */
	//gimple current_assign0 = gimple_build_assign (current_t_alone, current_t);
	//gsi_insert_before (ptr_gsi, current_assign0, GSI_SAME_STMT);
	/* D.xxxx = specload_pointer(...);.  */
	gimple_call_set_lhs (specload_pointer_call, pre_if_var);
	gsi_insert_after (ptr_gsi, specload_pointer_call, GSI_SAME_STMT);
	
	/* Remove the original reading into the speculative variables.  */
	gsi_remove(ptr_gsi, true);

	
	/* Now, create and insert the if condition.  
			D.xxxx = specload_pointer();
			if ( D.xxxx == -1 )
				goto labelSquash; 
			lhs = value	
	*/
	
	/* Build the if condition and the labels.  */
	label_true_decl = create_artificial_label (UNKNOWN_LOCATION);
	gimple label_true = gimple_build_label (label_true_decl);
	
	label_false_decl = create_artificial_label (UNKNOWN_LOCATION);
	gimple label_false = gimple_build_label (label_false_decl);
	
	gimple if_stmt = gimple_build_cond (EQ_EXPR, pre_if_var, 
																			build_int_cst (integer_type_node, -1),
																			labelSquash_decl, label_false_decl);
	
	/* Create the function earlySquash().  */	
	/* TODO: Remove the early_Squash function.  */
	//tree earlySquash_type = build_function_type_list (void_type_node, 
	//													integer_type_node, NULL_TREE);
	//earlySquash = build_fn_decl ("early_Squash", earlySquash_type);
	//gimple earlySquash_call = gimple_build_call (earlySquash, 1, 
	//															build_int_cst (integer_type_node,1));
	

	/* Build LHS = value.  */
	gimple assign_stmt = gimple_build_assign (lhs, value);
	
	/* Insert if (D.xxxx == -1).  */
	gsi_insert_after(ptr_gsi, if_stmt, GSI_NEW_STMT);
	
	/* Insert label_true.  */
	//gsi_insert_after (ptr_gsi, label_true, GSI_NEW_STMT);
	/* Insert earlySquash(1);.  */
	//gsi_insert_after (ptr_gsi, earlySquash_call, GSI_NEW_STMT);
	
	/* Insert label_false.  */
	gsi_insert_after (ptr_gsi, label_false, GSI_NEW_STMT);
	/* Insert LHS = value;  */
	gsi_insert_after (ptr_gsi, assign_stmt, GSI_NEW_STMT);
	
	// TODO: Remove this line
	//insert_printf(ptr_gsi, "[spec] Valor de value = %d\n", lhs);

}



/* Function to replace a writing into an speculative variable for a 
*  speculative store operation.  */

void
replace_speculative_store_in_function (gimple_stmt_iterator *ptr_gsi, tree variable, tree rhs, tree current_t_alone)
{
	tree specstore_pointer, funcdecl;
	
	/* Create the type (unsigned char *).  */
	tree unsigned_char_ptr_type_node = build_pointer_type (unsigned_char_type_node);

	/* Create some of the parameters:
		- current
		- value   
		- sizeof(variable)
		- (unsigned char *) &value
		- (unsigned char *) &variable
	*/
	
	/***********/
	/* current */
	/***********/
	/* It is already created at the beginning of the plugin as global variable.  */

	/*********/
	/* value */
	/*********/
	/* It is already created at the beginning of the plugin as global variable.  */
	

	
	/********************/
	/* sizeof(variable)	*/
	/********************/
	/* Get the size of the variable for the second argument of specload_pointer.  */
	/* TODO: Distinguish between arrays and single variables.  */
	/* In arrays we search for the size of an array element. 'variable' is an 'array_ref' */
	tree variable_element_size = TYPE_SIZE_UNIT( TREE_TYPE(variable));


	/*********************************/
	/* (unsigned char *) &variable.  */
	/*********************************/
	/* TODO: If the variable is not an array, probably this stuff is not correct.  */
	
	// These lines solves cases in which the speculative 
	//variable doesnt have the ADDRESSABLE bit activated.
	if ((TREE_CODE(variable) == VAR_DECL) &&
			(TREE_ADDRESSABLE(variable)==0) ) 
	{
		TREE_ADDRESSABLE (variable) = 1;
	}	
	else if ( (TREE_CODE(variable) == ARRAY_REF) && 
						(TREE_ADDRESSABLE ( TREE_OPERAND (variable, 0)) == 0) )
	{
		TREE_ADDRESSABLE ( TREE_OPERAND(variable, 0) ) = 1; 
	}
	
	/* If the variables is not recognized as VAR_DECL, or ARRAY_REF, exit.  */
	if ((TREE_CODE(variable) != VAR_DECL && TREE_CODE(variable) != ARRAY_REF))
	{
		printf("Speculative variable is not an ARRAY_REF or a VAR_DECL\n");
		debug_tree(variable);
		exit(-1);
	}

	tree variable_addr = build_fold_addr_expr (variable);
	tree pre_unsigned_char_ptr_var = fold_convert (unsigned_char_ptr_type_node, variable_addr);
	
	/* TODO: Check if this patch is correct. It is only necessary when the speculative variable is 
	* a global variable. Other solution is to not doing the fold_convert, and directly 
	* build the assignment with 'variable_addr' .  */ 
	/* This macro is necessary for global variables. Local variables are not affected.  */
	STRIP_USELESS_TYPE_CONVERSION (pre_unsigned_char_ptr_var);

	tree unsigned_char_ptr_var = build_decl (UNKNOWN_LOCATION, VAR_DECL, 
																								//get_identifier ("specstore_address"),
																								NULL,
																								unsigned_char_ptr_type_node);
	TREE_USED (unsigned_char_ptr_var) = 1;
	DECL_IGNORED_P (unsigned_char_ptr_var) = 1;

	gimple aux_stmt = gimple_build_assign (unsigned_char_ptr_var, pre_unsigned_char_ptr_var);
	gsi_insert_before (ptr_gsi, aux_stmt, GSI_SAME_STMT);
	
	/****************************/
	/* (unsigned char *) &value */
	/****************************/
	tree value_addr = build_fold_addr_expr (value);
	tree pre_unsigned_char_ptr_var2 = fold_convert (unsigned_char_ptr_type_node, value_addr);
	tree unsigned_char_ptr_var2 = build_decl (UNKNOWN_LOCATION, VAR_DECL, 
																								//get_identifier ("specstore_value"),
																								NULL, 
																								unsigned_char_ptr_type_node);
	TREE_USED (unsigned_char_ptr_var2) = 1;
	DECL_IGNORED_P (unsigned_char_ptr_var2) = 1;
	
	gimple aux_stmt2 = gimple_build_assign (unsigned_char_ptr_var2, pre_unsigned_char_ptr_var2);
	gsi_insert_before (ptr_gsi, aux_stmt2, GSI_SAME_STMT);


	/* value = rhs;  */
	/* Assign RHS to 'value' variable.  */
	gimple value_assign = gimple_build_assign (value, rhs);
	gsi_insert_before (ptr_gsi, value_assign, GSI_SAME_STMT);


	/* Build the function prototype:
	* 		void specstore(unsigned char*, int, int, unsigned char*)    */
	tree specstore_pointer_type = build_function_type_list (void_type_node,
																	unsigned_char_ptr_type_node,
																	integer_type_node,
																	integer_type_node,
																	unsigned_char_ptr_type_node,
																	NULL_TREE);
																
	/* Create the function declaration and insert it */
	specstore_pointer = build_fn_decl ("specstore_pointer", specstore_pointer_type);
	gimple specstore_pointer_call = gimple_build_call (specstore_pointer, 4,
																			unsigned_char_ptr_var,
																			variable_element_size, 
																			current_param,
																			unsigned_char_ptr_var2);
	
	/* Insert 'current.x = current' before the function call.  */
	//gimple current_assign0 = gimple_build_assign (current_t_alone, current_t);
	//gsi_insert_before (ptr_gsi, current_assign0, GSI_SAME_STMT);
	
	gsi_insert_after (ptr_gsi, specstore_pointer_call, GSI_SAME_STMT);
	
	/* Remove the original writing into the speculative variables.  */
	gsi_remove (ptr_gsi, true);

	
	/* DEBUG: To check the value of the speculative variable after the writing.  */
	//insert_printf(ptr_gsi, "Speculative variable has a value of = %d\n", variable);
	//printf("Printing speculative variable after printf\n");
	//debug_tree(TREE_OPERAND(variable,0));
	//printf("\n\n");
	//debug_tree(variable);
}

