/* Generation of functions which are copy of those called inside the body
	of the OpenMP FOR loops marked as speculative loops.  
	
	2013.
	Written by Sergio Aldea <sergio@infor.uva.es>

This file is part of specprag.  */ 

/* Copy the current function and create a new one with the 
same functionality.  */
extern tree create_copy_function (tree, tree);

/* Replace speculative loads and stores in the current function.  */
extern void replace_load_store_in_function (void);

/* Set the context of the local variable to the current function.  */
extern void set_context_local_variables (tree);

/* Replace every PARM_DECL (which are still the original) from the new function
	 by new PARM_DECLs, matching with the parameters created to the new function.  */
extern void replace_parm_decls (tree);

extern void replace_speculative_store_in_function (gimple_stmt_iterator *, tree, tree, tree);
extern void replace_speculative_load_in_function (gimple_stmt_iterator *, tree, tree, tree);

/* Create the variables needed inside the function which contains the spec-read and -stores.  */
extern void prepare_copied_function (tree);


/* Struct to save a maximum of 20 function calls with speculative variables,
	inside a speculative loop.  */
typedef struct {
  const char *name[20]; // Name of the original function 
	const char *name_spec[20]; // Name of the speculative version of the function
	tree copied_function[20]; // TREE object representing the copied function
	gimple_seq body_spec[20]; // GIMPLE body of the copied function (before processing)
  size_t used;	// Number of functions already copied.
} Function;


extern void init_Function(Function *a);
extern void insert_Function(Function *a, const char * original_name, tree function);
extern void search_Original_Function (Function *a, tree current_function);
extern bool search_Copied_Function (Function *a, tree current_function);
extern bool is_A_Recursive_Function (Function *a, tree copied_function, tree original_function);


/* TODO: CURRENTLY UNUSED **********************************************/
/* Struct to save (maximum 50) ALL the function calls detected.  */
typedef struct {
	const char *name[50]; // Name of the function
	tree func_decl[50]; // TREE object representing the function
	gimple_seq func_body[50]; // GIMPLE body of the function
	size_t used;	// Number of functions copied in this structure
} All_Functions;

extern void init_All_Functions (All_Functions *a);
extern void insert_All_Functions (All_Functions *a, tree function);
extern gimple_seq recovery_function_body (All_Functions *a, tree function);
/****************************************************************************/
