/********************************************************************
** @version BRA01
** Aggressive window.
** Supports forwarding.
** Dependence violation checks on stores.
** B: Inclusive Squash
** R: Exclusive Squash
** A: Exceptions Management and unexpected ending loops
** @author Alvaro Garcia
** @date 24th March 2011
*********************************************************************/

//  Main header file with all common variables
    #include "appcommon.h"
    #include "appcommon2.h"
    #include "constants.h"
//  speccode: OpenMP header file included
    #include <omp.h>
    #include <stdio.h>
    #include <signal.h>

// speccode: to measure time
// This function should be defined as global variable of the application
extern double spec_time;

//*********************************************************************
//   specredadd(): speculative ADD reduction operation
//*********************************************************************
  void specredadd(int index, int current, data_spec value, int *my_IV_tail)
  {
    int tidaux;

//  speccode: debug
//  tid = omp_get_thread_num();
//  printf("specredadd: tid = %d, current = %d Beginning ADD reduction\n",tid,current);
//  speccode: end debug

    tidaux = modvector[current];
  
    #pragma omp flush(AV)
    if (AV[tidaux][index] == NOTACC) 
    {

//  speccode: debug
//  printf("specredadd: tid = %d, current = %d Storing the value in shadow[]\n",tid,current);
//  speccode: end debug


//  If the corresponding cell of shadow[] is NOTACC,
//  then we simply store the value there.
      shadow[tidaux][index] = value;
      #pragma omp flush(shadow)

//  speccode: debug
//  printf("specredadd: tid = %d, current = %d Updating IV\n",tid,current);
//  speccode: end debug

      IV[tidaux][*my_IV_tail] = index;
      *my_IV_tail = *my_IV_tail + 1;
      #pragma omp flush(IV)

//    Then, mark this value as being reduced:

//  speccode: debug
//  printf("specredadd: tid = %d, current = %d Marking the value as reduced\n",tid,current);
//  speccode: end debug

      AV[tidaux][index] = REDADD;
      #pragma omp flush(AV)
    }
    else
    {
//  We assume here that the state is REDADD, thus the value,
//  IV and AV have been updated previously.

//  speccode: debug
//  printf("specredadd: tid = %d, current = %d Simply accumulating the value\n",tid,current);
//  speccode: end debug

//  Thus, we only accumulate the value

      shadow[tidaux][index] = shadow[tidaux][index] + value;
      #pragma omp flush(shadow)
    }

//  speccode: debug
//  printf("specredadd: tid = %d, current = %d Ending ADD reduction\n",tid,current);
//  speccode: end debug

  } // END specredadd

//*********************************************************************
//   specredmax(): speculative MAX reduction operation
//*********************************************************************

  void specredmax(int index, int current, data_spec value, int *my_IV_tail)
  {

    int tidaux;

//  speccode: debug
//  tid = omp_get_thread_num()
//  printf("specredmax: tid = %d, current = %d Beginning MAX reduction\n",tid,current);
//  speccode: end debug

    tidaux = modvector[current];

    #pragma omp flush(AV)
    if (AV[tidaux][index] == NOTACC) 
    {

//  speccode: debug
//  printf("specredmax: tid = %d, current = %d Storing the value in shadow[]\n",tid,current);
//  speccode: end debug

//  If the corresponding cell of shado[] is NOTACC
//  the we simply store the value there.
      shadow[tidaux][index] = value;
      #pragma omp flush(shadow)

//  speccode: debug
//  printf("specredmax: tid = %d, current = %d Updating IV\n",tid,current);
//  speccode: end debug

      IV[tidaux][*my_IV_tail] = index;
      *my_IV_tail = *my_IV_tail + 1;
      #pragma omp flush(IV)

//  Then mark this value as being reduced:

//  speccode: debug
//  printf("specredmax: tid = %d, current = %d Marking the value as reduced\n",tid,current);
//  speccode: end debug

      AV[tidaux][index] = REDMAX;
      #pragma omp flush(AV)
    
    }
    else
    {

//  We assume here that the state is REDMAX, thus the value,
//  IV and AV have been updated previously.

//  speccode: debug
//  printf("specredmax: tid = %d, current = %d Simply perform MAX operation\n",tid,current);
//  speccode: end debug

//  Thus, we perform MAX operation
//  shadow[index,tidaux] = max(shadow[index,tidaux],value)
      if (shadow[tidaux][index] < value)
        shadow[tidaux][index] = value;   
      #pragma omp flush(shadow)
      
    }

//  speccode: debug
//  printf("specredmax: tid = %d, current = %d Ending MAX reduction\n",tid,current);
//  speccode: end debug

} // END specredmax

/** 
** @version specload-B: inclusive squash
** @author alvaro garcia
**/
//*********************************************************************
//   specload(): speculative load operation with forwarding
//*********************************************************************
#ifdef inclusive
  int specload(int index, int current, data_spec *value, data_spec *maincopy, int *my_IV_tail)
  {

    char ea;
    int aux;
    int specaux1;

//  speccode: debug
//  tid = omp_get_thread_num()
//  printf("specload: tid = %d, current = %d Beginning load\n",omp_get_thread_num(),current);
//  speccode: end debug

    aux = modvector[current];
  //printf("specload: tid = %d, current = %d aux = %d\n",omp_get_thread_num(),current,aux);

#ifdef exceptions
/***********************************************************************
**  If the index is outside the boundaries of the speculative zone then
**  return -1
**  @author Alvaro Garcia Yaguez
**  @date 2nd February 2011
************************************************************************/
  if ( index < 0 || index >= shared_size )
  //if ( boundSpeculative(index) == -1)
  {
    //printf("SPECLOAD: INDEX %d CURRENT %d\n",index,current);
    raise(SIGSEGV);
  }
#endif

//  New for "earlysquash" versions: If I was squashed, goodbye...
  if (wheel[aux] == SQUASHED)
  {
    return -1;
  }

//  If value not yet used by thread, then search for version
    if (AV[aux][index] == NOTACC)
    {
//  First update AV, to notify any predecessor store that we
//  may be violating a dependecy.      

      gELV[index] = 'T';
      AV[aux][index] = EXPLD;
      #pragma omp flush(AV, gELV)
//    Update IV
//      #pragma OMP FLUSH(IV,IV_tail)
//      IV_tail[tidaux] = IV_tail[tidaux] + 1;
//      IV[IV_tail[tidaux],tidaux] = index;
//      #pragma OMP FLUSH(IV,IV_tail)
//    Version 137 and above: use a local my_IV_tail instead and
//    later move changes to IV_tail just after finishing this thread.

      IV[aux][*my_IV_tail] = index;
      *my_IV_tail = *my_IV_tail + 1;

//    If I am the non-spec, go to main copy directly.

      #pragma omp flush(wheel_ns)
      if(current == wheel_ns)
      {
        #pragma omp flush(maincopy)
        shadow[aux][index] = maincopy[index];
        #pragma omp flush(shadow)

//      speccode: debug
//      printf("specload: tid = %d, current = %d I am the ns: getting value from main copy\n",omp_get_thread_num(),current);
//      speccode: end debug

      }
      else
      {

//    Forward the most recent version that has been UPDATED  
/*
        #pragma omp flush(AV, wheel_ns, shadow)
        for ( specaux1 = current-1; specaux1 >= wheel_ns ; specaux1--)
                {
          ea = AV[modvector[specaux1]][index];
          if ( (ea == ELUP) || (ea == UPDATE) )
          {
//          We get the value from predecessor.
            shadow[aux][index] = shadow[modvector[specaux1]][index];
            #pragma omp flush(shadow)
                        break;
                  }
                  #pragma omp flush(AV, wheel_ns, shadow)
                }
                if ( specaux1 < wheel_ns )
                {
//        Too late. We get reference value.
          #pragma omp flush(maincopy)
          shadow[aux][index] = maincopy[index];
          #pragma omp flush(shadow)

                }
      } // END current == wheel_ns
*/

        specaux1 = current - 1;

e987:
//    El compilador no deja poner etiquetas justo antes de una instruccion omp
        if(1)
        {
          #pragma omp flush(AV, wheel_ns, shadow)
        }
        ea = AV[modvector[specaux1]][index];
        if ( (ea == ELUP) || (ea == UPDATE) )
        {

          if ( specaux1 >= wheel_ns )
          {

//        We get the value from predecessor.
            shadow[aux][index] = shadow[modvector[specaux1]][index];
            #pragma omp flush(shadow)

          }
          else
          {
//        Too late. We get reference value.
            #pragma omp flush(maincopy)
            shadow[aux][index] = maincopy[index];
            #pragma omp flush(shadow)
          }  // END specaux1 >= wheel_ns        
        }
        else
        {
//      specaux1 does not have an updated value. Go to predecessor.

          specaux1 = specaux1 - 1;
          if ( specaux1 >= wheel_ns )
            goto e987;

//      If we are here, we are outside boundary. 
//      We get reference value.

          #pragma omp flush(maincopy)
          shadow[aux][index] = maincopy[index];
          #pragma omp flush(shadow)
        } //END ea == ELUP || ea == UPDATE

      } // END current == wheel_ns

//      speccode: debug
//      printf("specload: tid = %d, current = %d And I dont forget to update AV and IV\n",tid,current);
//      speccode: end debug

    } // END AV[index, tidaux] == NOTACC

//  Value now here, perform the load

    *value = shadow[aux][index];

//      speccode: debug
//      printf("specload: tid = %d, current = %d Load finished. Value read: %d\n",tid,current,*value);
//      speccode: end debug
    return 0;

//e550:
  } // END specload
#endif


/** 
** @version specload-R: exclusive squash
** @author alvaro garcia
**/
//*********************************************************************
//   specload(): speculative load operation with forwarding
//*********************************************************************
#ifdef exclusive
  int specload(int index, int current, data_spec *value, data_spec *maincopy, int *my_IV_tail)
  {

    char ea;
    int aux;
    int specaux1;
    //ASversion 
    char dataforward='T';

//  speccode: debug
//  tid = omp_get_thread_num()
//  printf("specload: tid = %d, current = %d Beginning load\n",omp_get_thread_num(),current);
//  speccode: end debug

// Statistics
   //access_specload++;

    aux = modvector[current];

#ifdef exceptions
/***********************************************************************
**  If the index is outside the boundaries of the speculative zone then
**  return -1
**  @author Alvaro Garcia Yaguez
**  @date 2nd February 2011
************************************************************************/
  if ( index < 0 || index >= shared_size )
  //if ( boundSpeculative(index) == -1)
  {
    //printf("SPECLOAD: INDEX %d CURRENT %d\n",index,current);
    raise(SIGSEGV);
  }
#endif

//  New for "earlysquash" versions: If I was squashed, goodbye...
  if (wheel[aux] == SQUASHED)
  {
    return -1;
  }

//  If value not yet used by thread, then search for version
     
    if (AV[aux][index] == NOTACC)
    {
//  First update AV, to notify any predecessor store that we
//  may be violating a dependecy.      

      gELV[index] = 'T';
      AV[aux][index] = EXPLD;
      #pragma omp flush(AV, gELV)

//    Update IV
//      #pragma OMP FLUSH(IV,IV_tail)
//      IV_tail[tidaux] = IV_tail[tidaux] + 1;
//      IV[IV_tail[tidaux],tidaux] = index;
//      #pragma OMP FLUSH(IV,IV_tail)
//    Version 137 and above: use a local my_IV_tail instead and
//    later move changes to IV_tail just after finishing this thread.
   
      IV[aux][*my_IV_tail] = index;
      *my_IV_tail = *my_IV_tail + 1;

//    If I am the non-spec, go to main copy directly.
    
      #pragma omp flush(wheel_ns)
      if(current == wheel_ns)
      {
        #pragma omp flush(maincopy)
        shadow[aux][index] = maincopy[index];
        #pragma omp flush(shadow)

//      speccode: debug
//      printf("specload: tid = %d, current = %d I am the ns: getting value from main copy\n",omp_get_thread_num(),current);
//      speccode: end debug

      }
      else
      {
        specaux1 = current - 1;
        dataforward = 'F';
        #pragma omp flush(wheel_ns)
        while( (specaux1 >= wheel_ns) && (dataforward == 'F') )
        {
          #pragma omp flush(AV, wheel, shadow)
          if ( ( wheel[modvector[specaux1]] != SQUASHED) && ( wheel[modvector[specaux1]] != FREE ) ) 
          {
            ea = AV[modvector[specaux1]][index];
            // Posible cambio
            if ( (ea == ELUP) || (ea == UPDATE) )
            //if ( (ea == ELUP) || (ea == UPDATE) || (ea == EXPLD) )
            {
              #pragma omp flush(consumer_list)
              consumer_list[aux][modvector[specaux1]] = 'T';
              #pragma omp flush(consumer_list, wheel)

              //if ( ( specaux1 >= wheel_ns ) && ( wheel[modvector[specaux1]] != SQUASHED ) && ( wheel[modvector[specaux1]] != FREE ) ) 
              if ( ( wheel[modvector[specaux1]] != SQUASHED ) && ( wheel[modvector[specaux1]] != FREE ) ) 
              {
//        We get the value from predecessor.
                #pragma omp flush(shadow)
                shadow[aux][index] = shadow[modvector[specaux1]][index];
                dataforward = 'T';
                #pragma omp flush(shadow)
              }
            }
          }
          specaux1 = specaux1 - 1;
          #pragma omp flush(wheel_ns)
        }
        if ( dataforward == 'F' ) 
        {
//        Too late. We get reference value.
          #pragma omp flush(maincopy)
          shadow[aux][index] = maincopy[index];
          #pragma omp flush(shadow)
        }  
      } // END current == wheel_ns
    } // END AV[index, tidaux] == NOTACC

//  Value now here, perform the load

    *value = shadow[aux][index];  

//      speccode: debug
//      printf("specload: tid = %d, current = %d Load finished. Value read: %d\n",tid,current,*value);
//      speccode: end debug

    return 0;

//e550:
  } // END specload
#endif

/** 
** @version specstore-B: inclusive squash
** @author alvaro garcia
**/
#ifdef inclusive
//*********************************************************************
//   specstore(): speculative store operation with check for violations 
//*********************************************************************
  void specstore(int index, int current, data_spec value,int *my_IV_tail)
  {

    int squa;
        int aux;
        int specaux1;
        #ifdef logical
      boolean localsquash;
        #else
      char localsquash;
        #endif

//  speccode: debug
//  tid = omp_get_thread_num()
//  printf("specstore: tid = %d, current = %d Beginning store\n",omp_get_thread_num(),current);
//  speccode: end debug

#ifdef exceptions
/***********************************************************************
**  If the index is outside the boundaries of the speculative zone then
**  return -1
**  @author Alvaro Garcia Yaguez
**  @date 2nd February 2011
************************************************************************/
  //if ( boundSpeculative(index) == -1)
  if ( index < 0 || index >= shared_size )
  {
    //printf("SPECSTORE: INDEX %d CURRENT %d\n",index,current);
    raise(SIGSEGV);
  }
#endif

    aux = modvector[current];

//  First of all, store the value

    shadow[aux][index] = value;  //LOCALSTORE

//  #pragma OMP FLUSH(shadow)
//  #pragma OMP FLUSH(AV)

    if ( AV[aux][index] == EXPLD )
    {

//  Mark this value as ELUP asap...  
      AV[aux][index] = ELUP;
      #pragma omp flush(shadow, AV)
    }
    else
    {
      if ( AV[aux][index] == NOTACC )
      {
//    Mark this value as modified:
        AV[aux][index] = UPDATE;
        #pragma omp flush(shadow, AV)

//      Update IV: the value was not accessed by this thread 
//        #pragma OMP FLUSH(IV_tail)
//        IV_tail[tidaux] = IV_tail[tidaux] + 1;
//        IV[IV_tail[tidaux]][tidaux] = index;
//        #pragma OMP FLUSH(IV,IV_tail)
//      Version 137 and above: use a local my_IV_tail instead and
//      later move changes to IV_tail just after finishing this thread.
        IV[aux][*my_IV_tail] = index;
        *my_IV_tail = *my_IV_tail + 1;
        #pragma omp flush(IV)

//      If it is already ELUP or UPDATE, nothing to do
      }
    }

//  Look for possible RAW dependences.
//  New since version 140: RAW dependences should be searched
//  even if the thread is ms, since this situation could
//  change during execution of this loop.

    #pragma omp flush(gELV, wheel_ms)

    if ( gELV[index] == 'T' )
    {

//  speccode: debug
//  printf("specstore: tid = %d, current = %d gELV is %c\n",tid,current,gELV[index]);
//  speccode: end debug

      localsquash = 'F';
//  speccode: debug
//  printf("Checking for squashes between %d and %d\n",current+1,wheel_ms);
//  speccode: end debug

      specaux1 = current + 1;

e1122:
//    El compilador no deja poner etiquetas justo antes de una instruccion omp
      if(1)
      {
        #pragma omp flush(wheel_ms, AV)
          }
      if ( specaux1 > wheel_ms )
        goto e2233;
      else
      {

        if ( AV[modvector[specaux1]][index] == UPDATE )
                {

//  speccode: debug
//  printf("specstore: tid = %d, current = %d Shield found. Exiting DO loop\n",tid,current,gELV[index]);
//  speccode: end debug

          goto e2233;
                }
        if ( AV[modvector[specaux1]][index] != NOTACC )
        {
//      SQUASH!    

//***********************************************************************

          #pragma omp critical
          {
            // Statistics
            SC_specstore++;
            if ( ( wheel[aux] == RUNNING ) && ( specaux1 <= wheel_ms ) )
            {
//          begin squash critical section

//          We increment the spec_violations counter here
              spec_violations = spec_violations + 1;

//          Squash "specaux1" and all its successors.
//          If a sucessor is RUNNING, this is set as SQUASHED.
//          If a sucessor is DONE, then is set as FREE.

              #pragma omp flush(wheel_ms)

              for ( squa = specaux1 ; squa <= wheel_ms ; squa++ )
              {
//            We increment the number of squashes performed.
                spec_squashes = spec_squashes + 1;

                #pragma omp flush(wheel)

                if ( wheel[modvector[squa]] == RUNNING )
                                {

//                speccode: debug
//                printf("Consumer %d RUNNING->SQUASHED\n", squa);
//                speccode: end debug

                  wheel[modvector[squa]] = SQUASHED;

                                }

                if ( wheel[modvector[squa]] == DONE )
                                {

//                speccode: debug
//                printf("Consumer %d DONE->FREE\n", squa);
//                speccode: end debug
                  wheel[modvector[squa]] = FREE;

                                }

                #pragma omp flush(wheel, wheel_ms)
              } // END FOR

//            speccode: debug
//            printf("Last consumer squashed= %d\n", wheel_ms);
//            speccode: end debug

              wheel_ms = specaux1 - 1;

//            speccode: debug
//            printf("New most-spec thread= %d", wheel_ms);
//            speccode: end debug

              localsquash = 'T';


            } //END IF((wheel[tidaux]==RUNNING)&&(specaux1<=wheel_ms)) 

          } // END OMP CRITICAL

//***********************************************************************

// speccode: end squash critical section.

        } // END IF(AV[index,modvector[specaux1]] != NOTACC)
        //else
          // PARA PONER UN DEBUG          

        if (localsquash == 'T')
        {
//      The squashes has been performed.
//      We should exit the DO loop.
          goto e2233;
        }


      } // END IF(specaux1 >= wheel_ms)

//    v140 :  Now, we try to repeat the trick.
      specaux1 = specaux1 + 1;
      goto e1122;

//    Once the searche is finished, label "e2233" is reached.
e2233: ;
//   continue;
    } // END IF(gELV[index])

//  speccode: debug
//  printf("specstore: tid= %d, current= %d Store finished. Value stored: %d", tid, current,value);
//  speccode: end debug

  } // END specstore
#endif

#ifdef exclusive
/** 
** @version specstore-R: exclusive squash
** @author alvaro garcia
**/
//*********************************************************************
//   specstore(): speculative store operation with check for violations 
//*********************************************************************
  void specstore(int index, int current, data_spec value,int *my_IV_tail)
  {

    int squa, squa2;
    int aux;
    int specaux1;
    int i;
	#ifdef logical
      boolean localsquash;
	#else
      char localsquash;
	#endif

// AS-version
//    int SquashList[wsize];
//    int Tail=0;

//  speccode: debug
//  tid = omp_get_thread_num()
//  printf("specstore: tid = %d, current = %d Beginning store\n",omp_get_thread_num(),current);

//  Statistics
    //access_specstore++;

    aux = modvector[current];

#ifdef exceptions
/***********************************************************************
**  If the index is outside the boundaries of the speculative zone then
**  return -1
**  @author Alvaro Garcia Yaguez
**  @date 2nd February 2011
************************************************************************/
  //if ( boundSpeculative(index) == -1)
  if ( index < 0 || index >= shared_size )
  {
    //printf("SPECSTORE: INDEX %d CURRENT %d \n",index,current);
    raise(SIGSEGV);
  }
#endif

//  First of all, store the value

    shadow[aux][index] = value;  //LOCALSTORE

  #pragma OMP FLUSH(shadow)
//  #pragma OMP FLUSH(AV)

    if ( AV[aux][index] == EXPLD )
    {

//  Mark this value as ELUP asap...  
      AV[aux][index] = ELUP;
      #pragma omp flush(shadow, AV)
    }
    else
    {
      if ( AV[aux][index] == NOTACC )
      {
//    Mark this value as modified:
        AV[aux][index] = UPDATE;        
        #pragma omp flush(shadow, AV)

//      Update IV: the value was not accessed by this thread 
//        #pragma OMP FLUSH(IV_tail)
//        IV_tail[tidaux] = IV_tail[tidaux] + 1;
//        IV[IV_tail[tidaux]][tidaux] = index;
//        #pragma OMP FLUSH(IV,IV_tail)
//      Version 137 and above: use a local my_IV_tail instead and
//      later move changes to IV_tail just after finishing this thread.
        IV[aux][*my_IV_tail] = index;
        *my_IV_tail = *my_IV_tail + 1;
        #pragma omp flush(IV)

//      If it is already ELUP or UPDATE, nothing to do
      }
    } 

//  Look for possible RAW dependences.
//  New since version 140: RAW dependences should be searched
//  even if the thread is ms, since this situation could
//  change during execution of this loop.
              
    #pragma omp flush(gELV, wheel_ms)

    if ( gELV[index] == 'T' )
    {

//  speccode: debug
//  printf("specstore: tid = %d, current = %d gELV is %c\n",tid,current,gELV[index]);
//  speccode: end debug

      localsquash = 'F';

//  speccode: debug
//  printf("Checking for squashes between %d and %d\n",current+1,wheel_ms);
//  speccode: end debug

      specaux1 = current + 1;

e1122: 
//    Error compiler: we can not put a label before an OMP sentence
      if(1)
      {	
        #pragma omp flush(wheel_ms, AV, wheel)
      }
      if ( specaux1 > wheel_ms )
        goto e2233;
      else
      {
        if ( ( wheel[modvector[specaux1]] == SQUASHED) || ( wheel[modvector[specaux1]] == FREE ) )
        {
//      This thread have already been squashed
          specaux1 = specaux1 + 1 ;
          goto e1122;
        }

        if ( AV[modvector[specaux1]][index] == UPDATE )
	{

//  speccode: debug
//  printf("specstore: tid = %d, current = %d Shield found. Exiting DO loop\n",tid,current,gELV[index]);
//  speccode: end debug
          specaux1 = specaux1 + 1 ;
          goto e1122;
          //goto e2233;
	}
        if ( AV[modvector[specaux1]][index] != NOTACC )
        {
//      SQUASH!    
     
//***********************************************************************

          #pragma omp critical
          {

            //Statistics
            SC_specstore++;
            //if ( ( wheel[aux] == RUNNING ) && ( specaux1 <= wheel_ms ) ) 
            squa = specaux1;
            if ( ( wheel[modvector[specaux1]] != SQUASHED) && ( wheel[modvector[specaux1]] != FREE ) )
            {
//            Store the number of thread which have just been squashed
//            As of this thread we will check the Consumer List afterwards.

//            Add to the list of squashes threads
//              SquashList[Tail] = specaux1;
//              Tail++;

//            We increment the spec_violations counter here
              spec_violations = spec_violations + 1;          

//            Squash "specaux1" 
//            If the thread is RUNNING, this is set as SQUASHED.
//            If the thread is DONE, then is set as FREE.

//            We increment the number of squashes performed.
              spec_squashes = spec_squashes + 1;
               
              #pragma omp flush(wheel)
              if ( wheel[modvector[specaux1]] == RUNNING )
	      {
//              speccode: debug
//              printf("Consumer %d RUNNING->SQUASHED\n", squa);
//              speccode: end debug
                wheel[modvector[specaux1]] = SQUASHED;
              }
              if ( wheel[modvector[specaux1]] == DONE )
	      {
//              speccode: debug
//              printf("Consumer %d DONE->FREE\n", squa);
//              speccode: end debug
                bubbles = bubbles + 1;
                wheel[modvector[specaux1]] = FREE;
	      }

            } //END IF 

            #pragma omp flush(wheel,wheel_ms,bubbles)


// Keep on checking for squashes up to wheel_ms
            specaux1 = specaux1 + 1;
            //#pragma omp flush(wheel_ms)
            while ( specaux1 <= wheel_ms )
            {
              #pragma omp flush(wheel)
              if ( ( wheel[modvector[specaux1]] == SQUASHED) || ( wheel[modvector[specaux1]] == FREE ) )
              {
//              This thread have already been squashed
                specaux1 = specaux1 + 1;
                continue;
              }
              if ( AV[modvector[specaux1]][index] == UPDATE )
              {
//              This thread have already been squashed
                specaux1 = specaux1 + 1;
                continue;
                //break;
              }
              if ( AV[modvector[specaux1]][index] != NOTACC )
              {
//              SQUASH!    
                // Pregunta: Los siguientes los considero tambien violaciones o no
                //spec_violations = spec_violations + 1;          
                spec_squashes = spec_squashes + 1;

//              Add to the list of squashes threads
//                SquashList[Tail] = specaux1;
//               Tail++;
               
                #pragma omp flush(wheel)
                if ( wheel[modvector[specaux1]] == RUNNING )
                  wheel[modvector[specaux1]] = SQUASHED;
                if ( wheel[modvector[specaux1]] == DONE )
                {
                  wheel[modvector[specaux1]] = FREE;
                  bubbles = bubbles + 1;
                }
                #pragma omp flush(wheel,bubbles)
              }
              specaux1 = specaux1 + 1;
              //#pragma omp flush(wheel_ms)
            } // END WHILE

// Checking for squashes in Consumer List of Squashed threads
            for ( ; squa <= wheel_ms ; squa++ )
            //for ( int i = 0 ; i < Tail ; i++ )
            {
              //squa = SquashList[Tail];
              if ( ( wheel[modvector[squa]] == SQUASHED ) || ( wheel[modvector[squa]] == FREE ) )
              {
                for ( i=0 ; i < wsize ; i++ )
                  consumer_list[modvector[squa]][i] = 'F';
                #pragma omp flush(consumer_list)
                for ( squa2 = squa+1 ; squa2 <= wheel_ms ; squa2++ )
                {
                  if( consumer_list[modvector[squa2]][modvector[squa]] == 'T' )
                  {
                   #pragma omp flush(wheel)
                    if ( wheel[modvector[squa2]] == RUNNING )
                    {
//                    speccode: debug
//                    printf("Consumer %d RUNNING->SQUASHED\n", squa);
//                    speccode: end debug
                      spec_squashes = spec_squashes + 1;
                      wheel[modvector[squa2]] = SQUASHED;
                    }
                    if ( wheel[modvector[squa2]] == DONE )
                    {
//                    speccode: debug
//                    printf("Consumer %d DONE->FREE\n", squa);
//                    speccode: end debug
                      spec_squashes = spec_squashes + 1;
                      wheel[modvector[squa2]] = FREE;
                      bubbles = bubbles + 1;
                    }
                    consumer_list[modvector[squa2]][modvector[squa]] = 'F';
                    #pragma omp flush(wheel,consumer_list,bubbles)
                  }
                }
              } // END IF
              #pragma omp flush(wheel_ms,wheel,bubbles)
            } // END FOR

            localsquash = 'T';


            /*
            if ( ( wheel[aux] == RUNNING ) && ( specaux1 <= wheel_ms ) ) 
            {
//          We increment the spec_violations counter here
              spec_violations = spec_violations + 1;          
              #pragma omp flush(wheel_ms)
              for ( squa = specaux1 ; squa <= wheel_ms ; squa++ )
              {
//            We increment the number of squashes performed.
                spec_squashes = spec_squashes + 1;
                #pragma omp flush(wheel)
                if ( wheel[modvector[squa]] == RUNNING )
		{
                  wheel[modvector[squa]] = SQUASHED;
		}
                if ( wheel[modvector[squa]] == DONE )
		{
                  wheel[modvector[squa]] = FREE;
                  #pragma omp flush(bubbles )
                  bubbles = bubbles + 1;
		}
                #pragma omp flush(wheel, wheel_ms, bubbles )
              } // END FOR
//              wheel_ms = specaux1 - 1;
              localsquash = 'T';
            } //END IF((wheel[tidaux]==RUNNING)&&(specaux1<=wheel_ms)) 
            */

          } // END OMP CRITICAL
//***********************************************************************

// speccode: end squash critical section.

        } // END IF(AV[index,modvector[specaux1]] != NOTACC)

        if (localsquash == 'T')
        {
//      The squashes has been performed.
//      We should exit the DO loop.
          goto e2233;     
        }

      
      } // END IF(specaux1 >= wheel_ms)

//    v140 :  Now, we try to repeat the trick.
      specaux1 = specaux1 + 1;
      goto e1122;

//    Once the searche is finished, label "e2233" is reached.
e2233: ;
//   continue;
    } // END IF(gELV[index])

//  speccode: debug
//  printf("specstore: tid= %d, current= %d Store finished. Value stored: %d", tid, current,value);
//  speccode: end debug

  } // END specstore
#endif

//***********************************************************************
//   specstart(): operation to initialize speculation before every
//                speculative section
//***********************************************************************
//    No need to use FLUSH: this is executed outside parallel region
//***********************************************************************

  void specstart(int iterations)
  {
//  Main header file with all common variables
//    #include "appcommon.h"
    int specaux1,specaux2;
    int i,j;

//  Initialize the wheel structure
    wheel_ns = 0;
    wheel_ms = -1;

    endLoop = 0;
    endReturn = 0;
    nonSpeculative = 0;

/**
**  initialize the number of iterations to execute in the loop
*/
    upper_limit = iterations;


    /*for ( specaux1 = 0 ; specaux1 <= (threads-1) ; specaux1++ )
    {
      wheel[specaux1] = RUNNING;
    }*/

    for ( specaux1 = 0 ; specaux1 <= (wsize-1) ; specaux1++ )
      wheel[specaux1] = FREE;

//  Initialize AV and IV
    avclear_(AV, shared_size*wsize);

    //for ( specaux1 = 0; specaux1 <= (threads-1) ; specaux1++ )
    for ( specaux1 = 0; specaux1 < wsize ; specaux1++ )
    {
      IV[specaux1][0] = -1;
      IV_tail[specaux1] = 0;
    }

	//printf("wsize=%d\n",wsize);
	//for(specaux1=0;specaux1 < wsize;specaux1++)
	//  printf("%d\n",IV_tail[specaux1]);

//  Initialize the global forward vector
//  El *4 es por es tipo int -> 4bytes
    #ifdef logical
      gelvclear_(gELV, shared_size*4);
	#else
      gelvclear_(gELV, shared_size);
	#endif

//ASversion
#ifdef exclusive
    bubbles = 0;
    for ( i = 0 ; i < wsize ; i++ )
      for ( j = 0 ; j < wsize ; j++)
        consumer_list[i][j]='F';
#endif

//  Initialize the partial loop bounds 
    lower_bound = 1;
    upper_bound = wsize*blk;
    
  }

//***********************************************************************
//   initializeCounters(): operation to initialize counters
//***********************************************************************
//    No need to use FLUSH: this is executed outside parallel region
//***********************************************************************
   void initializeCounters()
   {
    spec_violations = 0;
    spec_squashes = 0;
#ifdef exceptions
    countExceptions = 0;
    exceptionsBlocksRestart = 0;
#endif
    SC_commit = 0;
    SC_threadend1 = 0;
    SC_threadend2 = 0;
    SC_specstore = 0;
    //access_specload = 0;
  }

  void specinit()
  {

//  Main header file with all common variables
//    #include "appcommon.h"

/*  Pendiente variables para bloques dinamicos
    int px, py, i;
    
    int blockfile, ios; */

//  real*8   IEEE double-precision floating-point (Eight-byte) 
    double dummy, logi, logtotal;
	int specaux1;

//  Initialize the modvector structure (only once).

    upper_limit = cur_upper_limit;

    for ( specaux1 = 0 ; specaux1 <= (max_upper_limit+8-1) ; specaux1++ )
      modvector[specaux1] = ((specaux1) % (wsize));

//  Initialize squash counters: they will hold squash information
//  for all spec sections, not only for one: the average number should 
//  be calculated.
    
/*    spec_violations = 0;
    spec_squashes = 0;
#ifdef exceptions
    countExceptions = 0;
    exceptionsBlocksRestart = 0;
#endif
    SC_commit = 0;
    SC_threadend1 = 0;
    SC_threadend2 = 0;
    SC_specstore = 0;
    //access_specload = 0;
*/
//  Depending on the value of blk, generate a fixed-block scheduling
//  or read a variable one from disk.

    if ( blk > 0 )
    {
//  Fixed block size: initialize vector with it.
//  upper_limit & blk al ser enteros la division es entera 
//  for ( maxblock = 0; maxblock <= (upper_limit/blk)-1 ; maxblock++ )
//  Lo sustituimos para que haga una iteracion mas y en caso de no ser
//  exacta la division en el siguiente bloque estaran las restantes iteraciones 
//  y algunas ms que nose ejecutaran por las restricciones del motor
//  Solo se ejecuta hasta max_upper_limit iteraciones
      for ( maxblock = 0; maxblock <= (upper_limit/blk) ; maxblock++ )
      {

//    First element: iteration to start
        varblock[0][maxblock] = blk*(maxblock);

//    Second element: iteration to finish
        varblock[1][maxblock] = blk*(maxblock+1)-1;

//    Third element: Number of attempts for this block
        varblock[2][maxblock] = 0;

    //printf("New block: from %d to %d\n",varblock[0][maxblock],varblock[1][maxblock]);

      }

    }
//  else
//  pendiente de hacer es para tamaos de bloque obtenidos de ficheros
//  y para bloques dinamicos

  }

/**
** 
** @author alvaro garcia
** @arg current block number
** @arg nonSpeculative flag to indicate if the thread in the non speculative
** @arg my_IV_tail index to IV structure
** @return 0: there is not any block left
           1: new block assigned
*/
  int getFirstBlock(int *current,int *nonSpeculative, int *my_IV_tail)
  {
    *my_IV_tail = 0;
    int aux=0;
    int tid = omp_get_thread_num();

//  speccode: debug
    //printf("getFirstBlock: Hola soy: %d , numero de threads %d\n",omp_get_thread_num(),omp_get_num_threads());
//  speccode: end_debug

//***********************************************************************
    #pragma omp critical
    {
	// There is no more work to do
	if ( varblock[0][wheel_ms+1] > upper_limit-1 )
              aux=0;
        else
        {
//          Only if the ms pointer plus one is pointing to a
//          FREE cell we can proceed
              #pragma omp flush(wheel, wheel_ms)

              if ( wheel[modvector[wheel_ms+1]] == FREE )
              {
//            Clean the access vector of the cell to be used
                cleanaccessv(wheel_ms);

//            Modify the status of the window cell to be used 
                wheel[modvector[wheel_ms+1]] = RUNNING;

                #pragma omp flush(wheel, wheel_ms)
//            Advance the wheel_ms pointer
                wheel_ms = wheel_ms + 1;

		//#ifdef exceptions
		/** @author Alvaro Garcia 
                **  @version Excepcions
                **  nonSpeculative will be 1 if you are the thread
                **  non speculative before starting the job. Otherwise 
                **  the value will be 0 **/
                *nonSpeculative = (wheel_ms == wheel_ns);
		//#endif

                #pragma omp flush(wheel_ms)

//            Prepare the chunk for running
                *current = wheel_ms;
                *my_IV_tail = 0;
		aux=1;

//            Calculate next chunk size        
                //dynscheduler(*current);
              }  // END if ( wheel[modvector[wheel_ms+1]] == FREE )
            } // END else ( varblock[1,wheel_ms] >= upper_limit )
     }
     return aux;
//***********************************************************************
 }


/** 
** @version specload-B: inclusive squash
** @author alvaro garcia
**/
//***********************************************************************
//  threadend(): end of thread operation with aggressive slidding window 
//***********************************************************************
#ifdef inclusive
  char threadend(int *current,int *nonSpeculative, data_spec *maincopy, int *my_IV_tail)
  {

//  Main header file with all common variables
//    #include "appcommon.h"
//  speccode: OpenMP header file included
//    #include <omp.h>

//  Local work flags
    char auxdone;
    char auxjobtodo;
        int aux;
        char flag;
        int i;

//  speccode: debug
//  printf("threadend: tid = %d, current = %d %dBeginning thread end\n",omp_get_thread_num(),*current,wheel[modvector[*current]]);
//  speccode: end debug

//  Set work flags
    auxdone = 'F';
    auxjobtodo = 'F';
    flag = JOBTODO;

//***********************************************************************
    #pragma omp critical
    {

//  Statistics
    SC_threadend1++;

//  speccode: debug
//  printf("threadend: tid = %d, current = %d entering critical section\n",tid,current);
//  printf("auxdone= %c, auxjobtodo= %c",auxdone, auxjobtodo);
//  speccode: end debug
//  speccode: debug
//  printf("Beginning critical section\n");
//  speccode: end debug

//  New for v137: Publish changes on my_IV_tail and reset local value
    //for(i=0;i<wsize;i++)
        //printf("i=%d::::specaux1= %d\n",i,IV_tail[i]);
//  printf("threadend:i:::%d IV_tail=%d\n",modvector[*current],IV_tail[modvector[*current]]);
      IV_tail[modvector[*current]] = *my_IV_tail;
      *my_IV_tail = 0;
    //for(i=0;i<wsize;i++)
//      printf("i=%d::::specaux1= %d\n",i,IV_tail[i]);

      #pragma omp flush(wheel, IV_tail)

//  First check if not squashed
      if ( wheel[modvector[*current]] != SQUASHED )
      {
        #pragma omp flush(wheel_ns, wheel_ms)

//  speccode: debug
//  printf("threadend: tid= %d, current= %d I was not squashed\n",omp_get_thread_num(), *current);
//  speccode: end debug

//      Now check if non-speculative
        if ( *current == wheel_ns )
        {
//      No-speculative thread finished

//  speccode: debug
//  printf("threadend: tid= %d, current= %d I am ns! Calling commit()\n",omp_get_thread_num(), *current);
//  speccode: end debug

          aux = commit(maincopy);
          wheel_ns = aux + 1;
  //printf("TID: %d, Actualizando wheel_ns: %d\n",omp_get_thread_num(),wheel_ns);
//      Was this the last one?

          if ( varblock[1][aux] >= upper_limit-1 )
          {
//        Mark this thread as done
            auxdone = 'T';
          }
          else
          {
//        Move the non-speculative pointer
            //wheel_ns = aux + 1;
            #pragma omp flush(wheel_ns)

//        If the block being executed by the most speculative
//        thread has an upper iteration that is greater than or
//        equal to the last iteration to perform, then we finish
            if ( varblock[1][wheel_ms] >= upper_limit-1 )
              auxdone = 'T';
            else
            {
//          Only if the ms pointer plus one is pointing to a
//          FREE cell we can proceed
              #pragma omp flush(wheel, wheel_ms)

              if ( wheel[modvector[wheel_ms+1]] == FREE )
              {
//            Clean the access vector of the cell to be used
                //printf("ANTES1:%d\n",IV_tail[modvector[wheel_ms+1]]);
                cleanaccessv(wheel_ms);
    //for(i=0;i<wsize;i++)
        //printf("EN el 1 i=%d::::specaux1= %d\n",i,IV_tail[i]);

//            Modify the status of the window cell to be used 
                wheel[modvector[wheel_ms+1]] = RUNNING;

                #pragma omp flush(wheel, wheel_ms)
//            Advance the wheel_ms pointer
                wheel_ms = wheel_ms + 1;

		//#ifdef exceptions
		/** @author Alvaro Garcia 
                **  @version Excepcions
                **  nonSpeculative will be 1 if you are the thread
                **  non speculative before starting the job. Otherwise 
                **  the value will be 0 **/
                *nonSpeculative = (wheel_ms == wheel_ns);
		//#endif

                #pragma omp flush(wheel_ms)

//            Prepare the chunk for running
                auxjobtodo = 'T';
                *current = wheel_ms;

//            Calculate next chunk size        
                dynscheduler(*current);
              }  // END if ( wheel[modvector[wheel_ms+1]] == FREE )
            } // END else ( varblock[1,wheel_ms] >= upper_limit )
          } // END else ( varblock[1,tidaux] >= upper_limit )
        } // END if ( *current == wheel_ns )
        else
        {

//      Speculative thread finished

//  speccode: debug
//  printf("threadend: tid= %d, current= %d I am speculative\n",tid, current);
//  speccode: end debug

//      Set the current iteration as DONE
          wheel[modvector[*current]] = DONE;

          #pragma omp flush(wheel)

//      If the block being executed by the most speculative
//      thread has an upper iteration that is greater than or
//      equal to the last iteration to perform, then we finish

          if ( varblock[1][wheel_ms] >= upper_limit-1)
            auxdone = 'T';
          else
          {
//        Check if the cell pointed to by wheel_ms+1 is free
            #pragma omp flush(wheel)

            if ( wheel[modvector[wheel_ms+1]] == FREE )
            {
//          Clean the access vector of the cell to be used
              //  printf("ANTES2:%d\n",IV_tail[modvector[wheel_ms+1]]);
              cleanaccessv(wheel_ms);
    //for(i=0;i<wsize;i++)
        //printf("En el 2i=%d::::specaux1= %d\n",i,IV_tail[i]);

//          Modify the status of the window cell to be used
              wheel[modvector[wheel_ms+1]] = RUNNING;

              #pragma omp flush(wheel)

//          Increment the most speculative pointer
              wheel_ms = wheel_ms + 1;

		//#ifdef exceptions
                /** @author Alvaro Garcia 
                **  @version Excepcions
                **  nonSpeculative will be 1 if you are the thread
                **  non speculative before starting the job. Otherwise 
                **  the value will be 0 **/
                *nonSpeculative = (wheel_ms == wheel_ns);
		//#endif

              #pragma omp flush(wheel_ms)

//          Prepare the chunk for running
              auxjobtodo = 'T';
              *current = wheel_ms;

//          Calculate next chunk size
              dynscheduler(*current);
            } // END  if ( wheel[modvector[wheel_ms+1]] == FREE )
          } // END else ( varblock[1,wheel_ms] >= upper_limit)
        } // END else (current == wheel_ns)
      } // END  if ( wheel[modvector[current]] != SQUASHED )
      else
      {
//    Squashed thread

//  speccode: debug
//  printf("threadend: tid= %d, current= %d I was squashed :-(\n",omp_get_thread_num(), *current);
//  speccode: end debug
//    Mark the current chunk as FREE
        wheel[modvector[*current]] = FREE;

        auxdone = 'F';
        auxjobtodo = 'F';

        #pragma omp flush(wheel)

//      (Note that the ms pointer was moved backwards by the thread that
//      performs the squash operation.)
      } // END else ( wheel[modvector[*current]] != SQUASHED )

//  speccode: debug
//  printf("threadend: tid= %d, current= %d leaving critical section\n",tid, current);
//  printf("auxdone= %c, auxjobtodo= %c", auxdone,auxjobtodo);
//  speccode: end debug

    } // END  #pragma OMP CRITICAL

//***********************************************************************

//  Now check what to do next
    if ( auxdone == 'T' )
    {
      flag = JOBDONE;
      return flag;
    }

    if (auxjobtodo == 'T' )
    {
      flag = JOBTODO;
      return flag;
    }

//  If we are here, we cannot start new thread, so spin-wait

//  speccode: debug
//  printf("threadend: tid= %d, current= %d entering spin wait\n",tid, current);
//  printf("auxdone= %c, auxjobtodo= %c", auxdone,auxjobtodo);
//  speccode: end debug

    #pragma omp flush(wheel, wheel_ms)

e456:
    if ( wheel[modvector[wheel_ms+1]] != FREE )  // SPIN1
    {
      #pragma omp flush(wheel, wheel_ms)  //SPIN2
//     printf("threadend: tid= %d, current= %d \n",omp_get_thread_num(), *current);
      goto e456; // SPIN3
    }

//  speccode: debug
//  printf("threadend: tid= %d, current= %d leaving spin wait\n",tid, current);
//  printf("auxdone= %c, auxjobtodo= %c", auxdone,auxjobtodo);
//  speccode: end debug

//***********************************************************************
    #pragma omp critical 
    {
//  Statistics
      SC_threadend2++;

//  Window moved, can start new thread now

      auxdone = 'F';

//  Double check that the new cell is really free and
//  that there are more iterations to run

      if ( (wheel[modvector[wheel_ms+1]] == FREE ) && ( varblock[1][wheel_ms] < upper_limit-1 ) )
      {
//    Clean the access vector of the cell to be used      
        //        printf("ANTES3:%d\n",IV_tail[modvector[wheel_ms+1]]);
        cleanaccessv(wheel_ms);
//    Increment the most speculative pointer
        wheel_ms = wheel_ms + 1;

		//#ifdef exceptions
		/** @author Alvaro Garcia 
                **  @version Excepcions
                **  nonSpeculative will be 1 if you are the thread
                **  non speculative before starting the job. Otherwise 
                **  the value will be 0 **/
                *nonSpeculative = (wheel_ms == wheel_ns);
		//#endif

        #pragma omp flush(wheel_ms)  

//    Prepare the chunk for running
        auxjobtodo = 'T';
        *current = wheel_ms;

//    Calculate next chunk size
        dynscheduler(*current);

        wheel[modvector[wheel_ms]] = RUNNING;

        #pragma omp flush(wheel) 
      } // END if
      else
      {
//    The new cell is not actually free, check out if done

//  speccode: debug
//  printf("threadend: tid= %d, new cell not free...\n",tid);
//  printf("auxdone= %c, auxjobtodo= %c", auxdone,auxjobtodo);
//  speccode: end debug

//    If the block being executed by the most speculative
//    thread has an upper iteration that is greater than or
//    equal to the last iteration to perform, then we finish

        if ( varblock[1][wheel_ms] >= upper_limit-1)
        {
          auxdone = 'T';
          auxjobtodo = 'T';
        }
        else
        {
          auxjobtodo = 'F';
          auxdone = 'F';
        }
      }
    } // END #pragma OMP CRITICAL

//***********************************************************************

//  Spin wait if no job to run

    if (auxjobtodo == 'F' )
    {
      #pragma omp flush(wheel, wheel_ms) 
//  speccode: debug
//  printf("threadend: tid= %d, current= %d re-entering spin wait\n",tid,current);
//  printf("auxdone= %c, auxjobtodo= %c", auxdone,auxjobtodo);
//  speccode: end debug
      goto e456;
    }

//  Jump to end if iterations finished

    if ( auxdone == 'T')
    {
      flag = JOBDONE;
      return flag;
    }

//  GO to work
    flag = JOBTODO;
    return flag;

  } // END threadend()
#endif

/** 
** @version specload-R: exclusive squash
** @author alvaro garcia
**/
//***********************************************************************
//  threadend(): end of thread operation with aggressive slidding window 
//***********************************************************************
#ifdef exclusive
  char threadend(int *current, int *nonSpeculative, data_spec *maincopy, int *my_IV_tail)
  {

//  Main header file with all common variables
//    #include "appcommon.h"
//  speccode: OpenMP header file included
//    #include <omp.h>

//  Local work flags
    char auxdone;
    char auxjobtodo;
	int aux;
	char flag;
	int i;

//  speccode: debug
//  printf("threadend: tid = %d, current = %d Beginning thread end\n",omp_get_thread_num(),*current);
//  speccode: end debug

//  Set work flags
    auxdone = 'F';
    auxjobtodo = 'F';
    flag = JOBTODO;

//***********************************************************************
    #pragma omp critical
    {

//  Statistics
    SC_threadend1++;

      IV_tail[modvector[*current]] = *my_IV_tail;
      *my_IV_tail = 0;

      #pragma omp flush(wheel, IV_tail)

//  First check if not squashed
      if ( wheel[modvector[*current]] != SQUASHED ) 
      {
        #pragma omp flush(wheel_ns, wheel_ms)

//  speccode: debug
//  printf("threadend: tid= %d, current= %d I was not squashed\n",omp_get_thread_num(), *current);
//  speccode: end debug

//      Now check if non-speculative
        if ( *current == wheel_ns )
        {
//      No-speculative thread finished

//  speccode: debug
//  printf("threadend: tid= %d, current= %d I am ns! Calling commit()\n",omp_get_thread_num(), *current);
//  speccode: end debug

          aux = commit(maincopy);

//      Was this the last one?

          if ( varblock[1][aux] >= upper_limit-1 )
          {
//        Mark this thread as done
            auxdone = 'T';
          }
          else
          {
//        Move the non-speculative pointer
            wheel_ns = aux + 1;
            //#pragma omp flush(wheel_ns)

              #pragma omp flush(wheel, wheel_ms, bubbles, wheel_ns )
                int i = wheel_ns;
                while ( ( i <= wheel_ms ) && ( auxjobtodo == 'F' ) )
                {
                  if ( wheel[modvector[i]] == FREE )
                  {
                    cleanaccessv(i-1);
                    wheel[modvector[i]] = RUNNING;
                    #pragma omp flush(wheel, bubbles)
                    bubbles = bubbles - 1;
                    #pragma omp flush(bubbles)
                    auxjobtodo = 'T';
                    *current = i;
                  }
                  i = i + 1;
                }
                if ( auxjobtodo == 'F' )
                {
                  if ( varblock[1][wheel_ms] >= upper_limit-1 )
                    auxdone = 'T';
                  else if ( wheel[modvector[wheel_ms+1]] == FREE )
                  {
                    cleanaccessv(wheel_ms);
                    wheel[modvector[wheel_ms+1]] = RUNNING;
                    #pragma omp flush(wheel, wheel_ms)
                    wheel_ms = wheel_ms + 1;
		//#ifdef exceptions
		/** @author Alvaro Garcia 
                **  @version Excepcions
                **  nonSpeculative will be 1 if you are the thread
                **  non speculative before starting the job. Otherwise 
                **  the value will be 0 **/
                *nonSpeculative = (wheel_ms == wheel_ns);
		//#endif
                    #pragma omp flush(wheel_ms)
                    auxjobtodo = 'T';
                    *current = wheel_ms;
                  }
                }
          } // END else ( varblock[1,tidaux] >= upper_limit )
        } // END if ( *current == wheel_ns )
        else
        {

//      Speculative thread finished

//      Set the current iteration as DONE
          wheel[modvector[*current]] = DONE;

              #pragma omp flush(wheel, wheel_ms, bubbles, wheel_ns )
                int i = wheel_ns;
                while ( ( i <= wheel_ms ) && ( auxjobtodo == 'F' ) )
                {
                  if ( wheel[modvector[i]] == FREE )
                  {
                    cleanaccessv(i-1);
                    wheel[modvector[i]] = RUNNING;
                    #pragma omp flush(wheel, bubbles)
                    bubbles = bubbles - 1;
                    #pragma omp flush(bubbles)
                    auxjobtodo = 'T';
                    *current = i;
		//#ifdef exceptions
		/** @author Alvaro Garcia 
                **  @version Excepcions
                **  nonSpeculative will be 1 if you are the thread
                **  non speculative before starting the job. Otherwise 
                **  the value will be 0 **/
                *nonSpeculative = (wheel_ms == wheel_ns);
		//#endif
                  }
                  i = i + 1;
                }
                if ( auxjobtodo == 'F' )
                {
                  if ( varblock[1][wheel_ms] >= upper_limit-1 )
                    auxdone = 'T';
                  else if ( wheel[modvector[wheel_ms+1]] == FREE )
                  {
                    cleanaccessv(wheel_ms);
                    wheel[modvector[wheel_ms+1]] = RUNNING;
                    #pragma omp flush(wheel, wheel_ms)
                    wheel_ms = wheel_ms + 1;
		//#ifdef exceptions
		/** @author Alvaro Garcia 
                **  @version Excepcions
                **  nonSpeculative will be 1 if you are the thread
                **  non speculative before starting the job. Otherwise 
                **  the value will be 0 **/
                *nonSpeculative = (wheel_ms == wheel_ns);
		//#endif
                    #pragma omp flush(wheel_ms)
                    auxjobtodo = 'T';
                    *current = wheel_ms;
                  }
                }

        } // END else (current == wheel_ns)
      } // END  if ( wheel[modvector[current]] != SQUASHED )
      else
      {
//    Squashed thread

//    Mark the current chunk as FREE
        wheel[modvector[*current]] = FREE;

        auxdone = 'F';
        auxjobtodo = 'F';

        #pragma omp flush(wheel)

      } // END else ( wheel[modvector[*current]] != SQUASHED )

    } // END  #pragma OMP CRITICAL
	
//***********************************************************************

//  Now check what to do next
    if ( auxdone == 'T' )
    {
      flag = JOBDONE;
      return flag;
    }

    if (auxjobtodo == 'T' )
    {
      flag = JOBTODO;
      return flag;
    }

//  If we are here, we cannot start new thread, so spin-wait

//  speccode: debug
//  printf("threadend: tid= %d, current= %d entering spin wait\n",tid, current);
//  printf("auxdone= %c, auxjobtodo= %c", auxdone,auxjobtodo);
//  speccode: end debug

    #pragma omp flush(wheel, wheel_ms, bubbles)

e456:  
    if ( (wheel[modvector[wheel_ms+1]] != FREE) && ( bubbles <= 0 ) )  // SPIN1
    {
      #pragma omp flush(wheel, wheel_ms, bubbles)  //SPIN2
//     printf("threadend: tid= %d, current= %d \n",omp_get_thread_num(), *current);
      goto e456; // SPIN3
    }

//  speccode: debug
//  printf("threadend: tid= %d, current= %d leaving spin wait\n",tid, current);
//  printf("auxdone= %c, auxjobtodo= %c", auxdone,auxjobtodo);
//  speccode: end debug

//***********************************************************************
    #pragma omp critical 
    {
//  Statistics
      SC_threadend2++;

//  Window moved, can start new thread now

      auxdone = 'F';

              #pragma omp flush(wheel, wheel_ms, bubbles, wheel_ns )
                int i = wheel_ns;
                while ( ( i <= wheel_ms ) && ( auxjobtodo == 'F' ) )
                {
                  if ( wheel[modvector[i]] == FREE )
                  {
                    cleanaccessv(i-1);
                    wheel[modvector[i]] = RUNNING;
                    #pragma omp flush(wheel, bubbles)
                    bubbles = bubbles - 1;
                    #pragma omp flush(bubbles)
                    auxjobtodo = 'T';
                    *current = i;
		//#ifdef exceptions
		/** @author Alvaro Garcia 
                **  @version Excepcions
                **  nonSpeculative will be 1 if you are the thread
                **  non speculative before starting the job. Otherwise 
                **  the value will be 0 **/
                *nonSpeculative = (wheel_ms == wheel_ns);
		//#endif
                  }
                  i = i + 1;
                }
                if ( auxjobtodo == 'F' )
                {
                  if ( varblock[1][wheel_ms] >= upper_limit-1 )
                  {
                    auxdone = 'T';
                    auxjobtodo = 'T';
                  }
                  else if ( wheel[modvector[wheel_ms+1]] == FREE )
                  {
                    cleanaccessv(wheel_ms);
                    wheel[modvector[wheel_ms+1]] = RUNNING;
                    #pragma omp flush(wheel, wheel_ms)
                    wheel_ms = wheel_ms + 1;
		//#ifdef exceptions
		/** @author Alvaro Garcia 
                **  @version Excepcions
                **  nonSpeculative will be 1 if you are the thread
                **  non speculative before starting the job. Otherwise 
                **  the value will be 0 **/
                *nonSpeculative = (wheel_ms == wheel_ns);
		//#endif
                    #pragma omp flush(wheel_ms)
                    auxjobtodo = 'T';
                    *current = wheel_ms;
                  }
                }

    } // END #pragma OMP CRITICAL
	
//***********************************************************************

//  Spin wait if no job to run
    
    if (auxjobtodo == 'F' )
    {
      #pragma omp flush(wheel, wheel_ms) 
//  speccode: debug
//  printf("threadend: tid= %d, current= %d re-entering spin wait\n",tid,current);
//  printf("auxdone= %c, auxjobtodo= %c", auxdone,auxjobtodo);
//  speccode: end debug
      goto e456;
    }

//  Jump to end if iterations finished

    if ( auxdone == 'T')
    {
      flag = JOBDONE;
      return flag;
    }

//  GO to work
    flag = JOBTODO;
    return flag;

  } // END threadend()
#endif


//***********************************************************************
//  cleanaccessv(): reset AV and IV for the (freedcell+1) window cell 
//***********************************************************************

  void cleanaccessv(int freedcell)
  {
//  Main header file with all common variables
//  #include "appcommon.h"
    int aux;
    int specaux1;
    int i;

//  speccode: debug
//  printf("cleanaccessv: Cleaning space for iteration %d",freedcell+1);
//  speccode: end debug

    aux = modvector[freedcell + 1];

    #pragma omp flush(IV_tail, IV)

    //for(i=0;i<wsize;i++)
	//printf("i=%d::::specaux1= %d\n",i,IV_tail[i]);
	//  printf("IV_tail[6]= %d y IV_tail[aux]=%d\n",IV_tail[6],IV_tail[aux]);
    for ( specaux1 = 0 ; specaux1 < IV_tail[aux] ; specaux1++ )
    //for ( specaux1 = 0 ; specaux1 <= IV_tail[aux] ; specaux1++ )
    {
	//  printf("specaux1:%d IV= %d\n",specaux1,IV[specaux1][aux]);
      AV[aux][IV[aux][specaux1]] = NOTACC;
    }
	  //printf("IV_tail[6]= %d\n",IV_tail[6]);
    #pragma omp flush(AV)
    //for(i=0;i<wsize;i++)
	//printf("EN EL i=%d::::specaux1= %d\n",i,IV_tail[i]);

//ASversion
#ifdef exclusive
    for (i=0; i<wsize; i++)
      consumer_list[i][aux] = 'F';
    for (i=0; i<wsize; i++)
      consumer_list[aux][i] = 'F';
    #pragma omp flush(consumer_list)
#endif

    IV[aux][0] = -1;
    IV_tail[aux] = 0;
    #pragma omp flush(IV, IV_tail)
 
  } // END cleanaccessv


//***********************************************************************
//  commit(): commit operation for aggressive sliding window, without
//            checking for violations and with reduction support 
//***********************************************************************

  int commit(data_spec *maincopy)
  {
//  Main header file with all common variables
//    #include "appcommon.h"
//  speccode: OpenMP header file included
//    #include <omp.h>

//  Auxiliary variables
// mytid: en el fortran esta declarada pero no se usa, ahora la voy a usar
// en lugar de usar tidaux, que solo lo usare 
    int localindex, localtid, mytid;
	int specaux1, specaux2;

//  speccode: debug
//  printf("commit: %d tid=%d I was called\n",current,omp_get_thread_num());
//  speccode: end debug

//  Statistics
    SC_commit++;

//  Mark the iteration as DONE
    wheel[modvector[wheel_ns]] = DONE;
    #pragma omp flush(wheel)

    

//  Scan wheel to find most advanced DONE iteration
    for ( mytid = wheel_ns ; mytid <= (wheel_ms - 1) ; mytid++ )
    { 
      if ( ( wheel[modvector[mytid]] == DONE ) && ( wheel[modvector[mytid + 1]] != DONE ) )
      {
        break;
      }
    }

//  Check for corner case: the most speculative can also be DONE
    if ( ( mytid == (wheel_ms - 1) ) && (wheel[modvector[wheel_ms]] == DONE) )
      mytid = mytid + 1;



//  Using IV, perform a commit of modified elements to maincopy

//  speccode: debug
//  printf("commit: %d Committing iterations %d to %d",current, wheel_ns, mytid);
//  speccode: end debug

    #pragma omp flush(wheel_ns)
	
    for ( specaux1 = wheel_ns ; specaux1 <= mytid ; specaux1++ )
    {
      localtid = modvector[specaux1];
 
      #pragma omp flush(shadow, IV, IV_tail)

      for ( specaux2 = 0; specaux2 < IV_tail[localtid] ; specaux2++ )
      //for ( specaux2 = 0; specaux2 <= IV_tail[localtid] ; specaux2++ )
      {
        localindex = IV[localtid][specaux2];

//      The element on this list can only be different to NOTACC

        #pragma omp flush(AV)
        if ( ( AV[localtid][localindex] == ELUP ) || (AV[localtid][localindex] == UPDATE ) ) 
        {
//      Commit the updated version

//  speccode: debug
//  printf("commit: %d Committing element %d of thread %d",current, localindex, specaux1);
//  printf("Old value: %d",maincopy[localindex]);
//  speccode: end debug

          maincopy[localindex] = shadow[localtid][localindex];
          #pragma omp flush(maincopy)

//  speccode: debug
//  printf("New value: %d",maincopy[localindex]);
//  speccode: end debug
        }
        else
        { 
//        ADD reduction operation?
          if ( AV[localtid][localindex] == REDADD )
          {
            #pragma omp flush(shadow)
            maincopy[localindex] = maincopy[localindex] + shadow[localtid][localindex];
            #pragma omp flush(maincopy)
          }
          else
          {
//          MAX reduction operation?
            if ( AV[localtid][localindex] == REDMAX )
            {
              #pragma omp flush(shadow)
//            maincopy[localindex] = max(maincopy[localindex],shadow[localindex,localtid])
              if ( maincopy[localindex] < shadow[localtid][localindex] )
                maincopy[localindex] = shadow[localtid][localindex];
              #pragma omp flush(maincopy)
            }
          }
        }
      }  // END for ( specaux2= 0 ) 

//    Mark the freed cell as FREE
      wheel[localtid] = FREE;
      #pragma omp flush(wheel)

//    Done with this thread, proceed with next.
    } // END for ( specaux1 = wheel_ns )   
	return mytid;

  }  // END commit()
 
 
//***********************************************************************
//  specepilogue(): Show total number of violations and squashes.
//            (should be called as last instruction of the application) 
//***********************************************************************
  void specepilogue(int level)
  {
//  Main header file with all common variables
//    #include appcommon.h
    #include <stdio.h>

    int i;

    if (level > 0) {
       fprintf(stderr,"------------------------------------------\n");
       fprintf(stderr,"Speculation summary\n");

       fprintf(stderr,"Total violations: %d ; Total squashes: %d\n", spec_violations, spec_squashes);
       //printf("Access specstore SC: %d \n", SC_specstore);
       //printf("Access threadend1 SC: %d \n", SC_threadend1);
       //printf("Access threadend2 SC: %d \n", SC_threadend2);
       //printf("Access commit SC: %d \n", SC_commit);
       fprintf(stderr,"Accumulated time of speculative sections: %lf\n",spec_time);

       if (level > 1) {
          fprintf(stderr,"------------------------------------------\n");
          fprintf(stderr,"Block number, block size, number of executions\n");

          for ( i = 0; i <= maxblock+1 ; i++ )
              fprintf(stderr,"%d %d %d\n",i, varblock[1][i]-varblock[0][i]+1, varblock[2][i]);
       }
    }

  } // END specepilogue()

/***********************************************************************
**  exceptionsEpilogue(): squash all threads from current to wheel_ms 
**                   this function is called by exception.h while a signal
**                   is handled.
**  @author Alvaro Garcia Yaguez
**  @date 26th January 2011
************************************************************************/
  void exceptionsEpilogue()
  {
    //printf("RESULTS: \n");
    //printf("\tTotal violations: %d\n", spec_violations);
    //printf("\tTotal blocks restarted by violations: %d\n",spec_squashes);
    //printf("\tTotal excepcions: %d\n",countExceptions);
    //printf("\tTotal blocks restarted by excepcions: %d\n",exceptionsBlocksRestart);
    //printf("\ttotal blocks restarted: %d\n",exceptionsBlocksRestart+spec_squashes);
    //printf("Access specstore SC: %d \n", SC_specstore);
    //printf("Access threadend1 SC: %d \n", SC_threadend1);
    //printf("Access threadend2 SC: %d \n", SC_threadend2);
    //printf("Access commit SC: %d \n", SC_commit);
  }



//***********************************************************************
//  dynscheduler(): Function that calculates next chunk size
//***********************************************************************

  void dynscheduler(int chunk)
  {
//  Main header file with all common variables
//    #include appcommon.h
 
//    int i;
//    double dummy1, dummy2, dummy3, logi, meanKshed, logtotal;

//  Dynamic scheduling.
 //   if ( ( blk <= -100 ) && ( blk >= -199 ) )
  //  {
//  Dynamic, with function "log(i)^2 / mean of Ksched squashes"
//  We should use "log(i)^2 / mean of Ksched squashes"
//  If varblock[1,chunk] is zero, calculate both first and last
//  fields and return them. If not, simply return.

   //   if ( varblock[0][chunk] != 0) return;

//  Hacer
 //   }

  } // END dynscheduler()

/***********************************************************************
**  freeWheel(): releases a slot from wheel
**  @author Alvaro Garcia Yaguez
**  @date 18th March 2011
************************************************************************/
  void freeWheel(int current)
  {
    #pragma omp critical
    {
      wheel[modvector[current]] = FREE;
    }
  }

/***********************************************************************
**  breakEnd(): fires the endLoop flag, squash all successors and commit
                its block 
**  @author Alvaro Garcia Yaguez
**  @date 18th March 2011
************************************************************************/
  void breakEnd(int current,data_spec *maincopy, int my_IV_tail)
  {
    int squa;

    #pragma omp critical
    {
      #pragma omp flush(wheel_ms)
      endLoop=1;
      for ( squa = current+1 ; squa <= wheel_ms ; squa++ )
      {
        #pragma omp flush(wheel)
        if ( wheel[modvector[squa]] == RUNNING )
          wheel[modvector[squa]] = SQUASHED;
        if ( wheel[modvector[squa]] == DONE )
          wheel[modvector[squa]] = FREE;
        #pragma omp flush(wheel, wheel_ms)
      } // END FOR
      /* Begin Squash Successors */
      wheel_ms = current;
      #pragma omp flush(wheel_ms, endLoop)
      /* End Squash Successors */
      IV_tail[modvector[current]] = my_IV_tail;
      //*my_IV_tail = 0;
      commit(maincopy);
    }
  }

/***********************************************************************
**  returnEnd(): fires the endReturn flag, squash all successors and commit
                its block 
**  @author Alvaro Garcia Yaguez
**  @date 21th March 2011
************************************************************************/
  void returnEnd(int current,data_spec *maincopy, int my_IV_tail)
  {
    int squa,aux;

    #pragma omp critical
    {
      #pragma omp flush(wheel_ms)
      endReturn=1;
      for ( squa = current+1 ; squa <= wheel_ms ; squa++ )
      {
        #pragma omp flush(wheel)
        if ( wheel[modvector[squa]] == RUNNING )
          wheel[modvector[squa]] = SQUASHED;
        if ( wheel[modvector[squa]] == DONE )
          wheel[modvector[squa]] = FREE;
        #pragma omp flush(wheel, wheel_ms)
      } // END FOR
      /* Begin Squash Successors */
      wheel_ms = current;
      #pragma omp flush(wheel_ms, endReturn)
      /* End Squash Successors */
      IV_tail[modvector[current]] = my_IV_tail;
      //*my_IV_tail = 0;
      aux = commit(maincopy);
      wheel_ns = aux + 1;
      //printf("RETURN END: TID: %d, Actualizando wheel_ns: %d\n",omp_get_thread_num(),wheel_ns);
    }
  }

/***********************************************************************
**  squashThreads(): squash all threads from current to wheel_ms 
**  @author Alvaro Garcia Yaguez
**  @date 24th January 2011
************************************************************************/
  void squashThreads(int current)
  {
        //int current = currentBlock[omp_get_thread_num()];
        int squa;


          #pragma omp critical
          {
            countExceptions++;
            if ( ( wheel[modvector[current]] == RUNNING ) && ( current <= wheel_ms ) )
            {
//          We increment the spec_violations counter here
//            spec_violations = spec_violations + 1;          

//          Squash "specaux1" and all its successors.
//          If a sucessor is RUNNING, this is set as SQUASHED.
//          If a sucessor is DONE, then is set as FREE.

              #pragma omp flush(wheel_ms)
        //printf("squashThreads: tid = %d, current = %d, wheel_ns %d, wheel_ms %d\n",omp_get_thread_num(),current,wheel_ns,wheel_ms);

              for ( squa = current ; squa <= wheel_ms ; squa++ )
              {
//            We increment the number of squashes performed
//              spec_squashes = spec_squashes + 1;
                exceptionsBlocksRestart = exceptionsBlocksRestart + 1;
                //printf("squashThreads %d \n",squa);

                #pragma omp flush(wheel)

                if ( wheel[modvector[squa]] == RUNNING )
                {
                  wheel[modvector[squa]] = SQUASHED;
                }

                if ( wheel[modvector[squa]] == DONE )
                {
                  wheel[modvector[squa]] = FREE;
                }

                #pragma omp flush(wheel, wheel_ms)
              } // END FOR

              wheel_ms = current - 1;
            }

          } // END OMP CRITICAL
  }

/***********************************************************************
**  boundSpeculative(): it is used to give us information whether a index
**                      is inside the limits a speculative zone, returning
**                      0 or it is outside, returning -1.
**  @author Alvaro Garcia Yaguez
**  @date 2nd February 2011
************************************************************************/
  int boundSpeculative(int index)
  {
    if ( index < 0 || index >= shared_size )
    {
        //printf("TID: %d, Index: %d \n",omp_get_thread_num()index);
        return -1;
    }
    return 0;
  }
