
#include <stdio.h>
#include <stdlib.h>

//***********************************************************************
// speccode: OpenMP header file included
#include <omp.h>
//***********************************************************************
//***********************************************************************
// speccode: Main header file with all common variables
#include "specEngine.h"
//************************************************************************

/*******************************************/
/* User parameters set by the programmer:  */
/*******************************************/
// Array size 
#define MAX 100
// Number of iterations of the parallel loop
#define NITER 1000000
//#define NITER 1

// Global data structure
int array[MAX];

/*--------------------------------------------------------------------------*/

int main() {

  /* Local variables:
   * 	P represents the number of iteration
   *		Q represents an the array index
   */
  int P, Q, aux, i;
  FILE *file;
  int sum_parallel = 0;
  int sum_sequential = 0;
  int pre_sum = 0;
  int array_sequential[MAX];

  /* Open and read the file.  */
  if ((file = fopen("rand1000000.in", "r")) == NULL) {
    printf("Error opening the file\n ");
    exit(0);
  } else {
    fscanf(file, "%d", &aux);
    for (i = 0; i < MAX; i++) {
      array[i] = aux;
      array_sequential[i] = aux;
      fscanf(file, "%d", &aux);
    }
    fclose(file);
  }

  /* Sum the elements of the array.  */
  for (i = 0; i < MAX; i++)
    pre_sum = pre_sum + array[i];
  printf(" Sum of the elements of the array (before processing) = %d\n", pre_sum);

  /* Print number of threads to be set  */
  fprintf(stderr," Execution with %d threads.\n\n", threads);

  // The loop is not parallelizable because indexes of 
  // the array elements depends on the input values from a file

  //***********************************************************************
  // speccode: OMP threading directive
  omp_set_num_threads(threads);
  // Initialize data structures needed to speculative parallelization
  specbegin(NITER);
  //specinit();
  //specstart(NITER);
  //***********************************************************************

  // speccode: Speculative loop
  int iteration, ini;
  int value = 0;
  int itera_parallel = 0;

  /************************************************************************/
  // speccode: Start profiling
  double spec_time=0;
  double spec_start, spec_stop;
  spec_start = get_time();
  /***********************************************************************/

#pragma omp parallel default(none)	\
  private(Q,P,aux, iteration, ini, value, \
      current, tid, retflag)    \
  shared(array, wheel_ns, wheel_ms,    \
      wheel, upper_limit, varblock) \
  reduction(+:itera_parallel)
  {

#pragma omp for \
    schedule(static)

    //for ( P = 0 ; P < NITER ; P++ )
    for (tid = 0; tid <= threads - 1; tid++) {

      ini = 0;
      current = tid;
      iteration = varblock[0][current]+ini;


      // If there is not work to this thread, go out.
      if (iteration > upper_limit - 1)
        goto labelSquash_1;
      // Increment the number of attempts for this iteration.
      varblock[2][current] = varblock[2][current]+1;

      // Diego, Fri Feb  1 11:03:39 CET 2013: fin de los cambios.
      // ======================================

labelStartIteration_1: 
      //  For loop body code

      itera_parallel++;
      Q = iteration % (MAX) + 1;

      /**********************************************************************/ 
      // speccode: speculative load. Original line:
      //    aux = array[Q-1];
      //
      // linear = (Q - 1);
      if (specload_pointer((void *) &(array[Q-1]), sizeof (array[Q-1]), current, (void *) &value) == -1)
        earlySquash(1);
      aux = value;
      /***********************************************************************/

      Q = (4 * aux) % (MAX) + 1;

      /***********************************************************************/
      // speccode: speculative store. Original line:
      //     array[Q-1] = aux;
      //
      //value = aux;
      specstore_pointer((void *) &(array[Q-1]), sizeof (array[Q-1]), current, (void *) &value);
      /***********************************************************************/

      /************************************************/
      /* speccode: end of thread and commit						*/
      /************************************************/

      //labelEndIteration_1:
      if ((iteration != varblock[1][current] + ini) && (iteration < NITER - 1)) {
        iteration = iteration + 1;
        goto labelStartIteration_1;
      }

      /* Thread done, perform the commit */
labelSquash_1:

      // ======================================
      // Diego, Fri Feb  1 11:03:39 CET 2013: todo esto es viejo. 
      // threadend_pointer sólo tiene un parámetro de entrada:
      // retflag = threadend_pointer(&current, &nonSpeculative);
      retflag = threadend_pointer(&current);
      // Diego: fin de cambios.
      // ======================================

      /* End if job done */
      if (retflag == JOBDONE) goto labelEndLoop_1;
      if (retflag == JOBTODO) {

        /*  Set loop variable */
        iteration = varblock[0][current] + ini;
        varblock[2][current] = varblock[2][current] + 1;
        goto labelStartIteration_1;
      }

labelEndLoop_1:
      ; /*freeWheel(current);*/

    } /*End of the FOR loop*/

  } // #pragma omp parallel

  /************************************************************************/
  // speccode: Stop measuring time
  spec_stop = get_time();
  spec_time += spec_stop - spec_start;
  /************************************************************************/

  //	speccode: free speculation structures
  specend();

  /*****************************************************/
  /* Executing the loop in sequential to compare times */
  /*****************************************************/
  double seq_init = 0;
  double seq_end = 0;
  double seq_time = 0;
  int itera_sequential = 0;

  seq_init = get_time();
  for (P = 0; P < NITER; P++) {

    itera_sequential++;
    Q = P % (MAX) + 1;
    aux = array_sequential[Q - 1];

    Q = (4 * aux) % (MAX) + 1;
    array_sequential[Q - 1] = aux;
  }

  seq_end = get_time();
  seq_time += seq_end - seq_init;


  /****************************/
  /* Printing execution times */
  /****************************/
  fprintf(stderr,"Sequential time: %f seg. ", seq_time);
  fprintf(stderr,"Speculative time: %lf seconds.\n", spec_time);
  fprintf(stderr,"Loss rate: %f times longer than sequential time\n\n", (spec_time/seq_time));

  /********************/
  /* Writing results  */
  /********************/
  sum_parallel=0;
  for (i = 0; i < MAX; i++)
    sum_parallel = sum_parallel + array[i];
  printf(" Sum of the elements of the array (after execution) = %d\n", sum_parallel);

  sum_sequential=0;
  for (i = 0; i < MAX; i++)
    sum_sequential = sum_sequential + array_sequential[i];
  fprintf(stderr," Sum of the elements of the array (after sequential execution) = %d\n", sum_sequential);

  if (sum_parallel != sum_sequential) {
    printf("ERROR:");
    for (i = 0; i < MAX; i++) {
      if (array[i] != array_sequential[i])
        printf("array[%d]=%d, array_sequential[%d]=%d", i, array[i], i, array_sequential[i]);
    }
    printf("\n");
  }

  /* Printing array elements.  */
  //for (i=0; i < MAX; i++)
  //	printf("array[%d]=%d\n", i, array[i]);


  fprintf(stderr,"Iterations done in parallel: %d\n", itera_parallel);
  fprintf(stderr,"Iterations done in sequential: %d\n", itera_sequential);

#ifdef TIME_SPECLOAD
  printSpecloadTime();
#endif
#ifdef TIME_SPECSTORE
  printSpecstoreTime();
#endif

#ifdef MEASURE_COUNTERS
  specepilogue();
  // printAccesses();
#endif

  return 0;
}
