#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

/* Funtion to get the current time.  */
double gettime()
{
  struct timeval tv;
  gettimeofday(&tv, (void *)0);
  return (tv.tv_sec + 1.0e-6*tv.tv_usec);
}

/*******************************************/
/* User parameters set by the programmer:  */
/*******************************************/
// Array size 
#define MAX 4
// Number of iterations of the loop
#define NITER 180000

/*--------------------------------------------------------------------------*/

int main() {

  /* Start profiling.  */
  double seq_init = 0;
  double seq_end = 0;
  double seq_time = 0;

  seq_init = gettime();

  /* Local variables:*/
  int i, aux;
  int j,k,l;
  int array[MAX];
  int pre_sum = 0;
  int sum = 0;
  FILE *file;

  /* Open and read the file.  */
  if ((file = fopen("data.in", "r")) == NULL) {
    printf("Error opening the file\n ");
    exit(0);
  } else {
    j = fscanf(file, "%d", &aux);
    for (i = 0; i < MAX; i++) {
      array[i] = aux;
      j = fscanf(file, "%d", &aux);
    }
    fclose(file);
  }


  /* Sum the elements of the array.  */
  for (i = 0; i < MAX; i++){
    pre_sum = pre_sum + array[i];
    printf("BEFORE:array[%d]=%d\n",i,array[i]);
  }

  printf(" Sum of the elements of the array (before processing) = %d\n", pre_sum);
  l=0;
  j=0;
  k=0;

  /* Start profiling.  */
  //seq_init = gettime();

#ifndef SEQUENTIAL	
  specbegin(NITER);
#endif	
  // The loop is not parallelizable because indexes of 
  // the array elements depends on the input values from a file
#pragma omp parallel for default(none) schedule(static) \
  private(i, k, aux) \
  shared(array) \
  speculative(j, l)
  for ( i = 0 ; i < NITER ; i++ ) {

    if (i == 18000) {
      k=l;
    }

    if (i == 3600) {
      k = j;
    }

    for (k = 0; k<array[i%MAX]+NITER; k++) {

      if (k >= 179900) {
        l = k+array[(i+k)%MAX];
      }

      if (k <= 1200) {
        j = array[i%MAX];
      }
      aux = (k + NITER) % 100000;
    }

    if (i == NITER-1) {
      l = j;
    }
  }

  /* Stop profiling.  */
  //seq_end = gettime();
  //seq_time += seq_end - seq_init;

#ifdef MEASURE_COUNTERS
  printf("HOLA\n");
  specepilogue();
  printAccesses();
#endif

  /********************/
  /* Writing results  */
  /********************/
  printf(" J = %d,L= %d, \n", j,l);
  sum = 0;
  for (i = 0; i < MAX; i++){
    sum = sum + array[i];
    printf("AFTER:array[%d]=%d\n",i,array[i]);
  }
  printf(" Sum of the elements of the array =  %d\n", sum);

  /* Stop profiling.  */
  seq_end = gettime();
  seq_time += seq_end - seq_init;

  /****************************/
  /* Printing execution times */
  /****************************/
  fprintf(stderr,"Execution time: %f seg. \n", seq_time);


  return 0;
}


