
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

//***********************************************************************
// speccode: OpenMP header file included
#include <omp.h>
//***********************************************************************
//***********************************************************************
// speccode: Main header file with all common variables
#include "specEngine.h"
//************************************************************************

/*******************************************/
/* User parameters set by the programmer:  */
/*******************************************/
// Array size 
#define MAX 4
// Number of iterations of the parallel loop
#define NITER 180000

/* Funtion to get the current time.  */
double gettime()
{
	 struct timeval tv;
	 gettimeofday(&tv, (void *)0);
	 return (tv.tv_sec + 1.0e-6*tv.tv_usec);
}

/*--------------------------------------------------------------------------*/

int main() {
	/* Local variables:*/
	int array[MAX];
	int i, aux;
 	int j,k,l;
	int sum_parallel = 0;
	int pre_sum = 0;
	FILE *file;

	/* Open and read the file.  */
	if ((file = fopen("data.in", "r")) == NULL) {
		printf("Error opening the file\n ");
		exit(0);
	} else {
		j = fscanf(file, "%d", &aux);
		for (i = 0; i < MAX; i++) {
			array[i] = aux;
			j = fscanf(file, "%d", &aux);
		}
		fclose(file);
	}


	/* Sum the elements of the array.  */
	for (i = 0; i < MAX; i++){
		pre_sum = pre_sum + array[i];
		printf("BEFORE:array[%d]=%d\n",i,array[i]);
	}
	printf(" Sum of the elements of the array (before processing) = %d\n", pre_sum);
		
	l=0;
	j=0;
	k=0;


	//***********************************************************************
	// speccode: OMP threading directive
	omp_set_num_threads(threads);
	// Initialize data structures needed to speculative parallelization
	specbegin(NITER);
	//***********************************************************************
	
	// speccode: Speculative loop
	int iteration, ini, value;
	int itera_parallel = 0;

	/************************************************************************/
	// speccode: Start profiling
	double spec_time=0;
	double spec_start, spec_stop;
	spec_start = gettime();
	/***********************************************************************/

#pragma omp parallel default(none)	\
	private(k,aux ,iteration, ini, current, tid, retflag, value) \
	shared(array,j,l,wheel_ns, wheel_ms, wheel, upper_limit, varblock) \
	reduction(+:itera_parallel)
{
#pragma omp for \
	schedule(static)
	
	//for ( i = 0 ; i < NITER ; i++ )
	for (tid = 0; tid <= threads - 1; tid++) {
 
		// ======================================
		// Diego, Fri Feb  1 11:03:39 CET 2013: todo esto es viejo. 
		// ini = 0;
		//	
		//if (getFirstBlock(&current, &nonSpeculative) == 0) {
		//    goto labelEndLoop_1;
		// }
		// iteration = varblock[0][current] + ini;
		// varblock[2][current] = varblock[2][current] + 1;
		// ======================================
		
		// Diego, Fri Feb  1 11:03:39 CET 2013: todo esto es nuevo. 
		// La variable ini indica la primera iteración a ejecutarse. En
		// esta aplicación es la cero.
		ini = 0;
		current = tid;
		iteration = varblock[0][current]+ini;
		// If there is not work to this thread, go out.
		if (iteration > upper_limit - 1)
			goto labelSquash_1;
		// Increment the number of attempts for this iteration.
		varblock[2][current] = varblock[2][current]+1;
		
		// Diego, Fri Feb  1 11:03:39 CET 2013: fin de los cambios.
		// ======================================
 
labelStartIteration_1: 
#ifdef MEASURE_COUNTERS
		itera_parallel++;
#endif

		//  For loop body code
		if (iteration == 18000) {
			// k = l;
			if (specload_pointer((unsigned char *) &(l), sizeof (l), current, (unsigned char *) &k) == -1)
				earlySquash(1);
		}
		if (iteration == 3600) {
			// k = j;
			if (specload_pointer((unsigned char *) &(j), sizeof (j), current, (unsigned char *) &k) == -1)
				earlySquash(1);
		}
		for (k = 0; k<array[iteration%MAX]+NITER; k++) {
			if (k >= 179900) {
				value = k+array[(iteration+k)%MAX];
				specstore_pointer((unsigned char *) &(l), sizeof (l), current, (unsigned char *) &value);
			}
			if (k <= 1200) {
				//j = array[i%MAX];
				specstore_pointer((unsigned char *) &(j), sizeof (j), current, (unsigned char *) &array[iteration%MAX]);
			}
			aux = (k + NITER) % 100000;
		}
		if (iteration == NITER-1) {
			// l = j;
			if (specload_pointer((unsigned char *) &(j), sizeof (j), current, (unsigned char *) &value) == -1)
				earlySquash(1);
			specstore_pointer((unsigned char *) &(l), sizeof (l), current, (unsigned char *) &value);
		}
		/***********************************************************************/


//labelEndIteration_1:
		if ((iteration != varblock[1][current] + ini) && (iteration < NITER - 1)) {
			iteration = iteration + 1;
			goto labelStartIteration_1;
		}
		
		/* Thread done, perform the commit */
labelSquash_1:

		// ======================================
		// Diego, Fri Feb  1 11:03:39 CET 2013: todo esto es viejo. 
								// threadend_pointer sólo tiene un parámetro de entrada:
		// retflag = threadend_pointer(&current, &nonSpeculative);
		retflag = threadend_pointer(&current);
		// Diego: fin de cambios.
		// ======================================
		
		/* End if job done */
		if (retflag == JOBDONE) goto labelEndLoop_1;
		if (retflag == JOBTODO) {
		
			/*  Set loop variable */
			iteration = varblock[0][current] + ini;
			varblock[2][current] = varblock[2][current] + 1;
			goto labelStartIteration_1;
		}

labelEndLoop_1:
		; /*freeWheel(current);*/
	
	} /*End of the FOR loop*/

} // #pragma omp parallel

	/************************************************************************/
	// speccode: Stop measuring time
	spec_stop = gettime();
	spec_time += spec_stop - spec_start;
	/************************************************************************/

//	speccode: free speculation structures
//	specend();

#ifdef TIME_SPECLOAD
 printSpecloadTime();
#endif
#ifdef TIME_SPECSTORE
 printSpecstoreTime();
#endif

#ifdef MEASURE_COUNTERS
 specepilogue();
 printAccesses();
#endif

	/********************/
	/* Writing results  */
	/********************/
	printf(" J = %d,L= %d, \n", j,l);
	sum_parallel=0;
	for (i = 0; i < MAX; i++){
		sum_parallel = sum_parallel + array[i];
		printf("AFTER:array[%d]=%d\n",i,array[i]);
	}
	printf(" Sum of the elements of the array =  %d\n", sum_parallel);

#ifdef MEASURE_COUNTERS
	printf("Iterations done in parallel: %d\n", itera_parallel);
#endif

	fprintf(stderr,"Execution time: %f seg. \n", spec_time);
	return 0;
}
