#!/bin/bash

APP=incr-2d

if test $# != 2
then
   echo "$0: format: $0 threads PWD-INPUT-SET"
   exit -1
fi

THREADS=$1
INPUTSET=$2

echo "-------------------------------------------------------"
echo "$0: running experiment with $THREADS threads."
echo 

# Generate "variables.h"
#cp fixed-vars.h variables.h
#echo "// Constant automatically added by $0" >> variables.h
#echo "#define threads $THREADS" >> variables.h
#make
# Now, we have to make_with_parameters
make clean
#make maxiter=9999997 blocksize=500 maxpointer=2000 threads=$THREADS

make maxiter=9999997 blocksize=1250 maxpointer=50 threads=$THREADS mask=63
#make maxiter=9999997 blocksize=-7 maxpointer=50 threads=$THREADS mask=63

# Generate "script"
echo "#! /bin/bash" > script
echo "#$ -cwd" >> script
echo "#$ -pe threads $THREADS" >> script
echo "/usr/bin/time -p ./incr-2d $INPUTSET" >> script

# Submit "script"
rm -rf script.*
RUN=`qsub script | cut -d " " -f 3`

control_c()
# run if user hits control-c
{
  echo -en "\n*** Ouch! Exiting ***\n"
  qdel $RUN
  exit $?
}
       
# trap keyboard interrupt (control-c)
trap control_c SIGINT



# Waiting for the script.o* file to be created
echo "Experiment $RUN queued."
while ! test -e script.o*
do
   sleep 1 ;
done

despues=$(date --date='+2 min' +%s)
# Waiting for the script.o* file to be created
echo "Experiment $RUN running."
qstat | grep $RUN > tt
while test -s tt
do
   sleep 1 ;
   qstat | grep $RUN > tt
   ahora=$(date +%s)
   if [ ${ahora} -ge ${despues} ]
   then
     echo -en "\n*** Ouch! Exiting *** Time exceeded: Experiment with $THREADS threads, with the input set $INPUTSET carried more than 2 minutes, results differ\n"
     qdel $RUN
     exit $?
   fi
 done
rm tt

# Compare results
AUX=`echo "$INPUTSET" | cut -d "/" -f 6`
echo "Experiment $RUN finished."
if ! diff script.o* ../01*/Results/incr-2d-seq-${AUX}.txto
then
   echo "$0: Results differ." 
#exit -1
fi

APP2=`echo $INPUTSET | awk -F / '{print $(NF)}'`
if [ $THREADS -lt 10 ]
then 
   FILENAME=$APP-0$THREADS-$APP2
else
   FILENAME=$APP-$THREADS-$APP2
fi 

mv script.e* Results/$FILENAME.txt
mv script.o* Results/o$FILENAME.txt

