/*
* incr-2d: C version of 2-dimensional, incremental convex hull
* Alvaro Garcia-Yaguez
* University of Valladolid
* Started in Sat Oct 9 16:15:00 CET 2010
*/

#include <stdio.h>
#include <stdlib.h>
// Header file for measuring performance with Sun compiler:
// #include "libcollector.h"


// N: total number of points  (min. 3)
// For 2D-Hull, Disc, 10M = 10.000.000
#define N 1000000

#define source 1
#define target 2
#define succ   3
#define pred   4
#define inch   5

#define x 1
#define y 2

// TOP: Number of points in the Convex Hull
// For 2D-Hull, Disc, 10M = 4393
#define TOP 4400

// speccode: time variable and function
double time_spec=0;
void gettime(double *);


int hull[TOP+1][5+1];
double input[2+1][N+1];

int i,j,k;
int p, psource, ptarget;
int auxinch;
int lastedge;
int e, eprev,enext,ea;
int first;
double px,py,qx,qy,dx,dy;
double determ;
int outside;

// Function
double rightturn(double dx, double dy, double px, double py, double qx, double qy);

int main() {
FILE *inputfile;

/************************************************************************/
// seqcode: Stop profiling with Sun Compiler
    // collector_pause();
/************************************************************************/
	
//  Opens the input file
    int not_used;
    if (inputfile = fopen("disc.in", "r"))
    {
      for(i = 1; i < N+1; i++)
      {
        not_used = fscanf(inputfile, "%lf %lf", &input[x][i], &input[y][i]);
        //printf("New point: %lf %lf\n",input[x][i],input[y][i]);
      }

      fclose(inputfile);
    }
    else
    {
      printf ( "Error reading from file \n " );
      exit(-1);
    }

/************************************************************************/
// seqcode: Start profiling with Sun Compiler
    // collector_resume();
/************************************************************************/

    // Calculate the first determinant
    dx = input[x][1];
    dy = input[y][1];
    px = input[x][2];
    py = input[y][2];
    qx = input[x][3];
    qy = input[y][3];

    determ = rightturn(dx,dy,px,py,qx,qy);

    if ( determ == 0 )
    {
      printf("incremental: The first three points are aligned.");
      printf("incremental: We aree too lazy to proceed. Stop.");
      exit(-1);
    }

    // If determinant < 0, reorder points two and three
    if ( determ < 0) 
    {
      input[x][2] = qx;
      input[y][2] = qy;
      input[x][3] = px;
      input[y][3] = py;
    }

    // Links the first edge
    hull[1][source] = 1;
    hull[1][target] = 2;
    hull[1][pred] = 3;
    hull[1][succ] = 2;
    hull[1][inch] = 1;

    // Links the second edge
    hull[2][source] = 2;
    hull[2][target] = 3;
    hull[2][pred] = 1;
    hull[2][succ] = 3;
    hull[2][inch] = 1;

    // Links the third edge
    hull[3][source] = 3;
    hull[3][target] = 1;
    hull[3][pred] = 2;
    hull[3][succ] = 1;
    hull[3][inch] = 1;

    hull[TOP][1] = 1;
    hull[TOP][5] = 3;
    // Put -1 in all remaining elements of hull, just in case.
    for ( i = 4 ; i < TOP ; i++)
    {
      for ( j = 1 ; j < 6 ; j++ )
      {
        hull[i][j] = -1;
      }
    }
    //printf("INI: %d\n",i);
    //printf("TOP: %d\n",hull[TOP][5]);

    // -------------------------------------------------
    // speccode: time measurement
    double time_start, time_stop;
    gettime(&time_start);
    // -------------------------------------------------
		
    // Main loop (iterating over remaining points)
    for ( i = 4 ; i < N+1 ; i++)
    {
      outside = 0;
      for ( e = 1 ; e < 4 ; e++)
      {  
        psource = hull[e][source];
        dx = input[x][psource];
        dy = input[y][psource];
        
        ptarget = hull[e][target];
        px = input[x][ptarget];
        py = input[y][ptarget];
  
        qx = input[x][i];
        qy = input[y][i];

        determ = rightturn(dx,dy,px,py,qx,qy);
        if ( determ < 0 )
        {
          // The point is outside
          outside = 1;
          break;
        }
      }

e500:
      auxinch = hull[e][inch]; 
      if ( ( auxinch == 0 ) && (outside) )
      {
        eprev = hull[e][pred];

        psource = hull[eprev][source];        
        dx = input[x][psource];
        dy = input[y][psource];

        ptarget = hull[eprev][target];
        px = input[x][ptarget];
        py = input[y][ptarget];

        qx = input[x][i];
        qy = input[y][i];

        determ = rightturn(dx,dy,px,py,qx,qy);
       
        if ( determ < 0 )
          e = eprev;
        else 
        {
          enext = hull[e][succ];
           
          psource = hull[enext][source];
          dx = input[x][psource];
          dy = input[y][psource];

          ptarget = hull[enext][target];
          px = input[x][ptarget];
          py = input[y][ptarget];

          qx = input[x][i];
          qy = input[y][i];

          determ = rightturn(dx,dy,px,py,qx,qy);

          if ( determ >= 0 )
            outside = 0;
          else 
            e = enext;
        }
        goto e500;
      }
      if ( outside)
      {
        //Compute upper tangent

        enext = hull[e][succ];
e600:
        psource = hull[enext][source];
        dx = input[x][psource];
        dy = input[y][psource];

        ptarget = hull[enext][target];
        px = input[x][ptarget];
        py = input[y][ptarget];

        qx = input[x][i];
        qy = input[y][i];

        determ = rightturn(dx,dy,px,py,qx,qy);

        lastedge = hull[TOP][5];
 
        if (determ <= 0)
        {
          hull[enext][inch] = 0;
          hull[enext][pred] = lastedge+1;
          ea = hull[enext][succ];
          hull[enext][succ] = lastedge+2;
          enext = ea;
          goto e600;
        }
        // Compute lower tangent
         
        eprev = hull[e][pred];
e700:
        dx = input[x][hull[eprev][source]];
        dy = input[y][hull[eprev][source]];
        px = input[x][hull[eprev][target]];
        py = input[y][hull[eprev][target]];
        qx = input[x][i];
        qy = input[y][i];

        determ = rightturn(dx,dy,px,py,qx,qy);

        if ( determ <= 0)
        {
          hull[eprev][inch] = 0;
          ea = hull[eprev][pred];
          hull[eprev][pred] = lastedge+1;
          hull[eprev][succ] = lastedge+2;
          eprev = ea; 
          goto e700;
        }
        hull[e][inch] = 0;
        hull[e][pred] = lastedge+1;
        hull[e][succ] = lastedge+2;

        // Add new edges to convex hull
         
        lastedge = lastedge + 1;
        hull[TOP][5] = lastedge;
        hull[lastedge][source] = hull[eprev][target];
        hull[lastedge][target] = i;
        hull[lastedge][succ] = lastedge + 1;
        hull[lastedge][pred] = eprev;
        hull[lastedge][inch] = 1;

        lastedge = lastedge + 1;
        hull[TOP][5] = lastedge;
        hull[lastedge][source] = i;
        hull[lastedge][target] = hull[enext][source];
        hull[lastedge][succ] = enext;
        hull[lastedge][pred] = lastedge - 1;
        hull[lastedge][inch] = 1;

        // Connect to old convex hull
        hull[enext][pred] = lastedge;
        hull[eprev][succ] = lastedge - 1;

        //Update first
        hull[TOP][1] = lastedge;
      }
    // End of process, proceed with next point
    }
    //printf("Fin: %d\n",i);

    // -------------------------------------------------
    // speccode: time measurement and display
    gettime(&time_stop);
    time_spec += time_stop - time_start;
    // -------------------------------------------------

/************************************************************************/
// seqcode: Stop profiling with Sun Compiler
    // collector_pause();
/************************************************************************/

    // Show the results

    e = hull[TOP][1];
    p = hull[e][source];
e400:
    printf("%f %f\n",input[x][p],input[y][p]);
    e = hull[e][succ];
    p = hull[e][source];
    
    if ( e != hull[TOP][1] ) 
       goto e400;

    printf("Total number of edges: %d\n", hull[TOP][5]);
		
		// Print time results
		fprintf(stderr,"------------------------------------------------\n");
    fprintf(stderr,"speccode: Accumulated time for target loop: %lf.\n",time_spec);

    return 0;
}

double rightturn(double dx, double dy, double px, double py, double qx, double qy)
{
  return (dx*py + dy*qx + px*qy -qx*py -dx*qy-px*dy);
}

// ----------------------------------------------------------------------------
// speccode: time functions

void gettime(double *t)
{
struct timeval tv;
   gettimeofday(&tv, (void *)0);
   *t = tv.tv_sec + 1.0e-6*tv.tv_usec;
}

