
#include <stdio.h>

#include "macros.h"
#include "user_parameters.h"


/********************************************************/
/* All the constants definition for speculative engine  */
/********************************************************/
// work states
#define JOBDONE '0'
#define JOBTODO '1'

// wheel state definitions
#define RUNNING 1
#define DONE 2
#define FREE 3

// The numerical values used for SQUASHED and PENDING_SQUASH
// are important in vA16 and above and should not be changed.
#define SQUASHED 5
// New parameter, useful in vAxx series and harmless in others.
#define PENDING_SQUASH 4

// local Access Vector states
#define NOTACC 0 // State not used, could be eliminated
#define EXPLD 1
#define ELUP 2
#define UPDATE 3
#define REDADD 4
#define REDMAX 5
#define REDMIN 6

#define true 1
#define false 0


/**********************************************/
/* Specific variables from speculative engine */
/**********************************************/

int current;

char retflag;

#ifdef MEASURE_COUNTERS
/** @param spec_violations
                count the total number a violation is produced
 **/
int spec_violations;

/** @param spec_squashes
                count the total number a squash is produced
 **/
int spec_squashes;

/** @param SC_threadend1
                number of access to first Critical Section in threadend function
 **/
int SC_threadend1;

/** @param SC_threadend2
                number of access to second Critical Section in threadend function
 **/
int SC_threadend2;

/** @param SC_specstore
                number of access to Critical Section in specstore function
 **/
int SC_specstore;

/** @param SC_commit
                number of access to commit function
 **/
int SC_commit;
#endif

// Definition of base type for 32- and 64-bit arquitectures.
#ifdef __LP64__
typedef unsigned long long int baseType;
#else
typedef unsigned long int baseType;
#endif

// upperlimit used in the following invocation
int upper_limit;

// Alvaro: 17-12-2013 MASK definido como parametro de compilacion
//#define MASK 127

// New cell definition to save values with the use of hash (datacell project)

typedef struct datacell {
    void * origPointer;
#ifdef COMPLEX_STRUCTURES
    unsigned int copyOffset; // Replace old speculativePointer
#else
    //baseType speculativeCopy;
    double speculativeCopy;
#endif
    short unsigned int size;
    short unsigned int state;
} datacell;

// New datacell matrix definition

typedef struct threadCopy_structure {
    datacell matriz[MASK + 1][MAX_POINTER];
} threadCopy_structure;

// Sliding window definition
int wheel[wsize];

#ifdef COMPLEX_STRUCTURES
#define MAX_OFFSET 2000000
#endif
// Definicion de los datos de un slot de la ventana

typedef struct wheel_structure {
    threadCopy_structure *pointer;
#ifdef COMPLEX_STRUCTURES
    void *temporalData;
    unsigned int actualOffset;
#endif
    unsigned int maxData;
} wheel_structure;

wheel_structure wheel_data[wsize];

int wheel_ns;
int wheel_ms;

// Vector with variable block size
//int varblock[3][max_upper_limit + 8];
// dynScheduler():
// Vector to hold block sizes
//    Column 0: Start iteration.
//    Column 1: End iteration.
//    Column 2: Number of re-executions during the *current* execution of the loop. 
//    Column 3: Number of accumulated re-executions during the executions of the loop. 
int varblock[4][max_upper_limit+wsize];

// thread id's
int tid;

int IV[MASK + 1][wsize][MAX_POINTER];
int IV_tail[MASK + 1][wsize];

// Array for storing MODULO indexes.
// This array will be filled with data in specbegin()
int modvector[max_upper_limit + 8];

// Functions declarations
int specload_pointer(void* address, unsigned int byteSize, unsigned int current, void* value);
void specstore_pointer(void* address, unsigned int byteSize, unsigned int current, void* value);
char threadend_pointer(int *current);
int commit();
void specredadd(void* address, unsigned int byteSize, unsigned int current, void* value);
void specredmax(void* address, unsigned int byteSize, unsigned int current, void* value);
void specredmin(void* address, unsigned int byteSize, unsigned int current, void* value);
void specbegin(int max_iterations);
void specend();

#ifdef MEASURE_COUNTERS
void initializeCounters();
void specepilogue();
#endif

// time vectors
#ifdef TIME_SPECLOAD
double time_specload[threads];
#endif
#ifdef TIME_SPECSTORE
double time_specstore[threads];
#endif
#ifdef TIME_THREADEND
double time_threadend[threads];
#endif
#ifdef TIME_COMMIT
double time_commit[threads];
#endif
#ifdef TIME_DYNSCHEDULER
double time_dynscheduler[threads];
#endif

double get_time();
#ifdef TIME_SPECLOAD
void printSpecloadTime();
#endif
#ifdef TIME_SPECSTORE
void printSpecstoreTime();
#endif
#ifdef TIME_THREADEND
void printThreadendTime();
#endif
#ifdef TIME_COMMIT
void printCommitTime();
#endif
#ifdef TIME_DYNSCHEDULER
void printDynSchedulerTime();
#endif

#ifdef MEASURE_COUNTERS
long long int totalAccessesSpecload[threads];
long long int totalCallsSpecload[threads];
long long int totalAccessesSpecstore[threads];
long long int totalCallsSpecstore[threads];
long long int totalAccessesSpecredadd[threads];
long long int totalCallsSpecredadd[threads];
long long int totalAccessesSpecredmax[threads];
long long int totalCallsSpecredmax[threads];
long long int totalAccessesSpecredmin[threads];
long long int totalCallsSpecredmin[threads];
void printAccesses();
#endif

// --------------------------------------------------------------------------
// Definitions for dynScheduling:
void dynScheduler(int chunkIndex);
double slopeEconomia(int min, int max);
int findZone(float x, float y, float maxMeanH);
float specRightTurn(float dx, float dy, float px, float py, float qx, float qy);
void meseta(int totalIterations, int percentageStart, int percentageEnd);

// Scheduling alternatives:
//#define DYN_TEST -1
#define MESETA -1

#define DYN_PDP1 -2
#define ADA_PDP1 -3

#define DYN_PDP2 -4
#define ADA_PDP2 -5

#define DYN_TENT -6
#define ADA_TENT -7

// percentages per meseta starting and ending zones
#define INITIAL_ZONE 20
#define FINAL_ZONE 10
// highest number of iterations assigned with meseta scheduling
#define MAX_ITER 3000

// Triangle definitions for TENT function (Scheduling 2012):
#define T128 1
#define T028 2
#define T023 3
#define T034 4
#define T046 5
#define T456 6
#define T078 7
#define T067 8
#define T35inf 9

#ifdef MEASURE_COUNTERS
// Vector to store accesses to zones of TENT
 int tentZone[9];
#endif

/** @param blk block size

Options: (as defined in appcommon.h)

  - A positive number: Fixed-size-chunking

  - DYN_TEST: Test for dynamic scheduling framework (to be removed later), with fixed block size equal to 1000

  - DYN_PDP1: Dynamic scheduling, function (1) of PDP 2008:
              "log(i).log(n) / mean of Ksched squashes"
              First Ksched blocks: also with function (1)
              Useful in loops with few iterations.

  - ADA_PDP1: Adaptive scheduling, function (1) of PDP 2008:
              "log(i).log(n) / mean of Ksched squashes"
              First Ksched blocks: also with function (1)
              Useful in loops with few iterations.
              Warning: THIS CAN NOT BE USED TOGETHER WITH EXCLUSIVE SQUASHING,
              because re-execution of bubbles should be carried out with the same chunk limits.

  - DYN_PDP2: Dynamic scheduling, function (2) of PDP 2008:
              "log(i).log(i).log(n) / mean of Ksched squashes"
              First Ksched blocks: also with function (2)
              Useful in loops with few iterations.

  - ADA_PDP2: Adaptive scheduling, function (2) of PDP 2008:
              "log(i).log(i).log(n) / mean of Ksched squashes"
              First Ksched blocks: also with function (2)
              Useful in loops with few iterations.
              Warning: THIS CAN NOT BE USED TOGETHER WITH EXCLUSIVE SQUASHING,
              because re-execution of bubbles should be carried out with the same chunk limits.

  - DYN_TENT: Dynamic scheduling, "tent" function defined in Scheduling 2012. Watch out the 
              ALPHA, BETA and FIRSTCHUNK parameters.

  - ADA_TENT: Adaptive scheduling, "tent" function defined in Scheduling 2012. Watch out the
              ALPHA, BETA and FIRSTCHUNK parameters.

  */

#include "dynScheduler.h"
