
/* =====================================================================
   mec-2d: C version of 2-dimensional Minimun Enclosing Circle.
   Sergio Aldea, Alvaro Estebanez
   Original Fortran version: Belen Palop, Diego R. Llanos (members of the GeoPar group)
   University of Valladolid
Revision: January 2013
===================================================================== */

// =====================================================================

#include <stdio.h>
#include <stdlib.h>
// library use to timing K loop
#include "sys/time.h"

#include <string.h>

// =====================================================================
//      constants

//      N: total number of points (min. 1)
//#define N 10000000 //10.000.000
#define N 5 //10.000.000

//      EPS: Absolute precision threshold
#define EPS 0.001


// =====================================================================
//      Functions
double cuad_dist (double px, double py, double qx, double qy);

// =====================================================================
// speccode: time variable and function
void gettime (double *);

// =====================================================================

// =====================================================================
//      Data structures and variables
double centrox, centroy, cuad_radio;
double solpx, solpy, solqx, solqy, solrx, solry;
double puntdef;

// Input data matrix
double input[2][N];


int main (int argc, char *argv[])
{

  // Time counters
  double time_spec = 0;
  double time_start, time_stop;

  // Loop counters
  int i, j, k;

  // Auxiliar variables to stores values
  double spec_centrox, spec_centroy, spec_cuad_radio;
  double spec_solpx, spec_solpy;
  double spec_solqx, spec_solqy;
  double spec_solrx, spec_solry;
  double spec_puntdef;

  // File descriptor
  FILE *inputfile;
  char inputFileName[1024];

  // Auxiliar variables
  double px, py, qx, qy, rx, ry;
  double yqr, ypr, xqr, xpr, xrp, xqp, xpq, ypq;
  double lambda1, lambda2, denom;
  int pointsRead;
  int x, y;

  if (argc > 1) {
    strcpy(inputFileName, argv[1]);
  } else {
    strcpy(inputFileName, "disc.in");
  }



  // =====================================================================
  //      Open and read input file 

  // if (inputfile = fopen("/home/Points/input/disc-mec.in", "r")){
  // if (inputfile = fopen ("disc.in", "r"))
  if ( (inputfile = fopen(inputFileName, "r")) )
  {
    for (pointsRead = 0; pointsRead < N; pointsRead++)
      {
      if (fscanf (inputfile, "%d %d", &x, &y) < 0)
      {
        printf ("Error reading numbers from file\n");
        exit (-1);
      }
      input[0][pointsRead] = x;
      input[1][pointsRead] = y;
    } //for
  }
  else
  {
    printf ("Error reading from file\n");
    exit (-1);
  }

  fclose (inputfile);


  // -------------------------------------------------
  // speccode: time measurement
  gettime (&time_start);
  // -------------------------------------------------

  // --------------------------------------------------------------------
  //      Initializations.
  centrox = 0;
  centroy = 0;
  cuad_radio = 0;

  // --------------------------------------------------------------------
  //      Beginning of process 
  for (i = 0; i < N; i++)
  {
    px = input[0][i];
    py = input[1][i];

    if (cuad_dist (px, py, centrox, centroy) - cuad_radio > EPS)
    {

      // Debug 
      //  printf("mec-2d: Nuevo punto de fallo: %d\n", i);
      // End Debug

      // Inlined version of circunf1p() follows 
      centrox = px;
      centroy = py;
      cuad_radio = 0;
      solpx = px;
      solpy = py;
      puntdef = 1;

      // print *,"New upper limit for middle loop: ",(i-1)

      for (j = 0; j < (i - 1); j++)
      {
        qx = input[0][j];
        qy = input[1][j];

        if (cuad_dist (qx, qy, centrox, centroy) - cuad_radio > EPS)
        {

          // Inlined version of circunf2p() follows 
          centrox = (px + qx) / 2;
          centroy = (py + qy) / 2;
          cuad_radio = cuad_dist (px, py, qx, qy) / 4;
          solpx = px;
          solpy = py;
          solqx = qx;
          solqy = qy;
          puntdef = 2;

          // Debug: Check if result is correct 
          /*if (cuad_dist (qx, qy, centrox, centroy) - cuad_radio > EPS)
            {
            printf ("Error in second loop!\n");
            printf ("Center is ('%f', '%f')\n", centrox, centroy);
            printf ("Squared radius is %f", cuad_radio);
            exit (-1);
            }*/

          // start time counter
          // startTime();

          // -------------------------------------------------
          // speccode: time measurement
          //gettime (&time_start);
          // -------------------------------------------------

          for (k = 0; k < (j - 1); k++)
          {
            rx = input[0][k];
            ry = input[1][k];


            spec_centrox = centrox;
            spec_centroy = centroy;
            spec_cuad_radio = cuad_radio;

            if (cuad_dist (rx, ry, spec_centrox, spec_centroy) - spec_cuad_radio > EPS)
            {
              // Inlined version of circunf3p() follows 
              xpq = px - qx;
              xqr = qx - rx;
              xpr = px - rx;
              ypq = py - qy;
              yqr = qy - ry;
              ypr = py - ry;

              denom = 2 * (xpq * ypr - ypq * xpr);

              if (denom == 0)
              {
                printf ("Error: 3 aligned points.\n");
                exit (-1);
              }

              if (py != ry)
              {
                lambda1 = (yqr * ypr + xpr * xqr) / denom;
                spec_centrox =	(px + qx) / 2 + lambda1 * (py - qy);
                spec_centroy = (py + qy) / 2 + lambda1 * (qx - px);
              }
              else
              {
                if (py != qy)
                {
                  lambda2 = (yqr * ypq + xqr * xpq) / denom;
                  spec_centrox = (px + rx) / 2 + lambda2 * (py - ry);
                  spec_centroy = (py + ry) / 2 + lambda2 * (rx - px);
                }
              }

              spec_cuad_radio = cuad_dist (px, py, spec_centrox, spec_centroy);
              centrox = spec_centrox;
              centroy = spec_centroy;
              cuad_radio = spec_cuad_radio;

              spec_solrx = rx;
              spec_solry = ry;
              spec_puntdef = 3;

              solrx = spec_solrx;
              solry = spec_solry;
              puntdef = spec_puntdef;


              /*solpx = px;
                solpy = py;
                solqx = qx;
                solqy = qy;
                solrx = rx;
                solry = ry;
                puntdef = 3;*/

              // Debug: Check if result is correct 
              /*if (cuad_dist (qx, qy, centrox, centroy) - cuad_radio > EPS)
                {
                printf ("Error in third loop, A!\n");
                printf ("Center is ('%f', '%f')\n", centrox, centroy);
                printf ("Squared radius is %f\n",	cuad_radio);
                exit (-1);
                }

              // Debug: Check if result is correct 
              if (cuad_dist (rx, ry, centrox, centroy) - cuad_radio > EPS)
              {	
              printf ("Error in third loop, B!\n");
              printf ("Center is ('%f', '%f')\n",	centrox, centroy);
              printf ("Squared radius is %f\n", cuad_radio);
              exit (-1);
              }*/
              //printf("Iteracion [%d,%d,%d]\n", i, j, k);
              //fflush(stdout);

            }	// end if
          }		// end for k
          // End of the time from this loop
          //stopTime();
          // -------------------------------------------------
          // speccode: time measurement and display
          //gettime (&time_stop);
          //time_spec += time_stop - time_start;
          // -------------------------------------------------
          //specend();

        }		// end if
      }			// end for j
    }			// end if
  }				// end for i
  //      End of process 

  // --------------------------------------------------------------------
  // -------------------------------------------------
  // speccode: time measurement and display
  gettime (&time_stop);
  time_spec += time_stop - time_start;
  //      Check results 

  for (i = 0; i < N; i++)
  {
    px = input[0][i];
    py = input[1][i];
    if (cuad_dist (px, py, centrox, centroy) - cuad_radio > EPS)
      printf ("mec-2d: Fail!\n");
  }

  // --------------------------------------------------------------------
  //      Show results 

  printf ("mec-2d: Points processed: %d\n", N);
  printf ("mec-2d: Center of MEC: ('%f', '%f')\n", centrox, centroy);
  printf ("mec-2d: Squared radius of MEC: %f\n", cuad_radio);
  printf ("mec-2d: Number of points that define the solution: %f\n", puntdef);
  if (puntdef >= 1)
    printf ("mec-2d: Point one: ('%f', '%f')\n", solpx, solpy);
  if (puntdef >= 2)
    printf ("mec-2d: Point two: ('%f', '%f')\n", solqx, solqy);
  if (puntdef >= 3)
    printf ("mec-2d: Point three: ('%f', '%f')\n", solrx, solry);

  fprintf (stderr, "------------------------------------------------\n");
  fprintf (stderr, "Accumulated Time for target loop: %lf \n", time_spec);

  return (0);
}	// main

// =====================================================================

// =====================================================================

  double
cuad_dist (double px, double py, double qx, double qy)
{
  double res1, res2;

  res1 = (px - qx);
  res1 = res1 * res1;
  res2 = (py - qy);
  res2 = res2 * res2;

  return (res1 + res2);
}

// =====================================================================

// ----------------------------------------------------------------------------
// speccode: time functions

  void
gettime (double *t)
{
  struct timeval tv;
  gettimeofday (&tv, (void *) 0);
  *t = tv.tv_sec + 1.0e-6 * tv.tv_usec;
}
