
/* =====================================================================
   mec-2d: C version of 2-dimensional Minimun Enclosing Circle.
   Sergio Aldea, Alvaro Estebanez Lopez
   Original Fortran version: Belen Palop, Diego R. Llanos (members of the GeoPar group)
   University of Valladolid
Revision: January 2013
===================================================================== */

// =====================================================================

#include <stdio.h>
#include <stdlib.h>

//***********************************************************************
// speccode: OpenMP header file included
#include <omp.h>
//***********************************************************************
//***********************************************************************
// speccode: Main header file with all common variables
#include "specEngine.h"
//***********************************************************************

// =====================================================================
//      constants

//      N: total number of points (min. 1)
#define N 10000000		//10.000.000

//      EPS: Absolute precision threshold
#define EPS 0.001

int number_calls=0;
int max_iterations=0;

// =====================================================================
//      Data structures and variables
double centrox, centroy, cuad_radio;
double solpx, solpy, solqx, solqy, solrx, solry;
double puntdef;

//      Input data matrix
double input[2][N];

// =====================================================================
//      Functions

double cuad_dist (double px, double py, double qx, double qy);

// =====================================================================

int main (int argc, char *argv[])
{

  // Loop counters
  int i, j, k;
  int iteration, ini;//, nIterations;


  //      Old Result variables (now used to temporarily hold result[] values)
  double spec_centrox, spec_centroy, spec_cuad_radio;
  double spec_solrx, spec_solry;
  double spec_puntdef;

  //      File descriptor
  FILE *inputfile;

  //      Auxiliar variables
  double px, py, qx, qy, rx, ry;
  double yqr, ypr, xqr, xpr, xpq, ypq;
  double lambda1, lambda2, denom;
  int pointsRead;
  int x, y;

  // Time measurement variables
  double spec_time = 0;
  double spec_start, spec_stop;

  // =====================================================================
  //      sepccode: Temporal values
  double value;

  // =====================================================================
  //      Open and read input file 
  if ( (inputfile = (fopen ("disc.in", "r"))) )
  {
    for (pointsRead = 0; pointsRead < N; pointsRead++)
    {
      if (fscanf (inputfile, "%d %d", &x, &y) < 0)
      {
        printf ("Error reading from file\n");
        exit (-1);
      }

      input[0][pointsRead] = x;
      input[1][pointsRead] = y;
    }//for

  }
  else
  {
    printf ("Error reading from file\n");
    exit (-1);
  }

  fclose (inputfile);

  //***********************************************************************
  // speccode: Stop profiling
  // collector_pause();
  //***********************************************************************

  // --------------------------------------------------------------------
  //      Initializations.
  centrox = 0;
  centroy = 0;
  cuad_radio = 0;

  // --------------------------------------------------------------------
  //      Beginning of process 
  for (i = 0; i < N; i++)
  {
    px = input[0][i];
    py = input[1][i];

    if (cuad_dist (px, py, centrox, centroy) - cuad_radio > EPS)
    {

      // Debug 
      //  printf("mec-2d: Nuevo punto de fallo: %d\n", i);
      // End debug

      // Inlined version of circunf1p() follows 
      centrox = px;
      centroy = py;
      cuad_radio = 0;
      solpx = px;
      solpy = py;
      puntdef = 1;

      // print *,"New upper limit for middle loop: ",(i-1)

      for (j = 0; j < (i-1); j++)
      {
        qx = input[0][j];
        qy = input[1][j];

        if (cuad_dist (qx, qy, centrox, centroy) - cuad_radio > EPS)
        {

          // Inlined version of circunf2p() follows 
          centrox = (px + qx) / 2;
          centroy = (py + qy) / 2;
          cuad_radio = cuad_dist (px, py, qx, qy) / 4;
          solpx = px;
          solpy = py;
          solqx = qx;
          solqy = qy;
          puntdef = 2;

          /* 
          // Debug: Check if result is correct 
          if ( cuad_dist(qx, qy, centrox, centroy) - cuad_radio > EPS){
          printf("Error in second loop!\n");
          printf("Center is ('%f', '%f')\n",centrox, centroy);
          printf("Squared radius is %f", cuad_radio);
          exit(-1);
          }
          */
#ifdef MEASURE_COUNTERS
          number_calls++;
          //if (j-1 > max_iterations){
          //	max_iterations = j-1;
          //	fprintf(stderr,"new max = %d\n",max_iterations);
          //}
          //	fprintf(stderr,"j-1 = %d\n",j-1);
#endif
          //***********************************************************************
          // speccode: OMP threading directive
          omp_set_num_threads (threads);
          // speccode: initialize speculation structures
          specbegin (j-1);
          // speccode: variable upper limit inicialization
          //upper_limit = j-1;
          //***********************************************************************
          //***********************************************************************
          // speccode: Start profiling
          // collector_resume();
          //***********************************************************************

          // speccode: Start profiling
          spec_start = get_time();

          //***********************************************************************
          // speccode: Speculative loop
#pragma omp parallel default(none)                                  \
          private(rx,ry,k, iteration, ini, /*nIterations,*/ \
              yqr, ypr, xqr, xpr, xpq, ypq, \
              lambda1, lambda2,denom, spec_centrox, spec_centroy, \
              spec_cuad_radio, \
              spec_solrx,spec_solry,spec_puntdef, \
              value,current, tid, retflag)    \
          shared(px,py,qx,qy,i,j,input, wheel_ns, wheel_ms, \
              wheel,upper_limit, varblock, \
              centrox, centroy, cuad_radio, \
              solrx, solry, puntdef)
          {

#pragma omp for \
            schedule(static)

            for (tid = 0; tid <= threads - 1; tid++)
            {

              ini = 0;
              current = tid;
              iteration = varblock[0][current] + ini;
              //nIterations = j - 1;

              if (iteration > upper_limit - 1)
                goto labelSquash_1;

              varblock[2][current] = varblock[2][current] + 1;
              //printf("\t\tvarblock[0][current]:%d,current:%d,iteration:%d\n",varblock[0][current],current,iteration);
labelStartIteration_1:	//printf("ENDLOOP,iteration:%d,current:%d\n",iteration,current);
              //  Codigo del bucle
              k = iteration;

              rx = input[0][k];
              ry = input[1][k];

              //***********************************************************************
              // speccode: speculative load. Original line:
              //     spec_centrox = centrox;
              if (specload_pointer
                  ((void *) &(centrox), sizeof (centrox),
                   current, (void *) &value) == -1)
                earlySquash (1);
              spec_centrox = value;
              //***********************************************************************

              //***********************************************************************
              // speccode: speculative load. Original line:
              //     spec_centroy = centroy;
              if (specload_pointer
                  ((void *) &(centroy), sizeof (centroy),
                   current, (void *) &value) == -1)
                earlySquash (1);
              spec_centroy = value;
              //***********************************************************************

              //***********************************************************************
              // speccode: speculative load. Original line:
              //     spec_cuad_radio = cuad_radio;
              if (specload_pointer
                  ((void *) &(cuad_radio),
                   sizeof (cuad_radio), current,
                   (void *) &value) == -1)
                earlySquash (1);
              spec_cuad_radio = value;
              //***********************************************************************

              if (cuad_dist (rx, ry, spec_centrox, spec_centroy) -
                  spec_cuad_radio > EPS)
              {
                // Inlined version of circunf3p() follows 
                xpq = px - qx;
                xqr = qx - rx;
                xpr = px - rx;
                ypq = py - qy;
                yqr = qy - ry;
                ypr = py - ry;
                denom = 2 * (xpq * ypr - ypq * xpr);

                if (denom == 0)
                {
                  printf ("Error: 3 aligned points.\n");
                  printf ("i = %d\n", i);
                  printf ("j = %d\n", j);
                  printf ("k = %d\n", k);
                  printf ("px = %f , py = %f\n", px, py);
                  printf ("qx = %f , qy = %f\n", qx, qy);
                  printf ("rx = %f , ry = %f\n", rx, ry);
                  exit (-1);
                }

                if (py != ry)
                {
                  lambda1 = (yqr * ypr + xpr * xqr) / denom;

                  //***********************************************************************
                  // speccode: speculative store (DEFERRED). Original line:
                  //     centrox = (px+qx)/2 + lambda1*(py-qy);
                  spec_centrox =
                    (px + qx) / 2 + lambda1 * (py - qy);
                  // speccode: end of speculative store
                  //***********************************************************************

                  //***********************************************************************
                  // speccode: speculative store (DEFERRED). Original line:
                  //     centroy = (py+qy)/2 + lambda1*(qx-px);
                  spec_centroy =
                    (py + qy) / 2 + lambda1 * (qx - px);
                  // speccode: end of speculative store
                  //***********************************************************************

                }
                else
                {
                  if (py != qy)
                  {
                    lambda2 = (yqr * ypq + xqr * xpq) / denom;

                    //***********************************************************************
                    // speccode: speculative store (DEFERRED). Original line:
                    //     centrox = (px+rx)/2 + lambda2*(py-ry);
                    spec_centrox =
                      (px + rx) / 2 + lambda2 * (py - ry);
                    // speccode: end of speculative store
                    //***********************************************************************

                    //***********************************************************************
                    // speccode: speculative store. Original line:
                    //     centroy = (py+ry)/2 + lambda2*(rx-px);
                    spec_centroy =
                      (py + ry) / 2 + lambda2 * (rx - px);
                    // speccode: end of speculative store
                    //***********************************************************************

                  }
                }

                // speccode: speculative store (DEFERRED)
                spec_cuad_radio =
                  cuad_dist (px, py, spec_centrox, spec_centroy);
                // speccode: end of speculative store

                // speccode: speculative store (DEFERRED)
                spec_solrx = rx;
                // speccode: end of speculative store

                // speccode: speculative store (DEFERRED)
                spec_solry = ry;
                // speccode: end of speculative store

                // speccode: speculative store (DEFERRED)
                spec_puntdef = 3;
                // speccode: end of speculative store

                // speccode: Before processing next point k, publish changes on speculative data

                //***********************************************************************
                // speccode: speculative store.
                value = spec_centrox;
                specstore_pointer ((void *) &(centrox),
                    sizeof (centrox), current,
                    (void *) &value);
                // speccode: end of speculative store
                //***********************************************************************

                //***********************************************************************
                // speccode: speculative store.
                value = spec_centroy;
                specstore_pointer ((void *) &(centroy),
                    sizeof (centroy), current,
                    (void *) &value);
                // speccode: end of speculative store
                //***********************************************************************

                //***********************************************************************
                // speccode: speculative store.
                value = spec_cuad_radio;
                specstore_pointer ((void *)
                    &(cuad_radio),
                    sizeof (cuad_radio), current,
                    (void *) &value);
                // speccode: end of speculative store
                //***********************************************************************

                //***********************************************************************
                // speccode: speculative store.
                value = spec_solrx;
                specstore_pointer ((void *) &(solrx),
                    sizeof (solrx), current,
                    (void *) &value);
                // speccode: end of speculative store
                //**********************************************************************

                //***********************************************************************
                // speccode: speculative store.
                value = spec_solry;
                specstore_pointer ((void *) &(solry),
                    sizeof (solry), current,
                    (void *) &value);
                // speccode: end of speculative store
                //***********************************************************************

                //***********************************************************************
                // speccode: speculative store.
                value = spec_puntdef;
                specstore_pointer ((void *) &(puntdef),
                    sizeof (puntdef), current,
                    (void *) &value);
                // speccode: end of speculative store
                //***********************************************************************

              }	// end if

              //labelEndIteration_1:
              if ((iteration != varblock[1][current] + ini)
                  && (iteration < /*nIterations*/(j-1) - 1))
              {
                iteration = iteration + 1;
                goto labelStartIteration_1;
              }
              /* Thread done, perform the commit */
labelSquash_1:
              retflag = threadend_pointer (&current);
              /* End if job done */
              if (retflag == JOBDONE)
                goto labelEndLoop_1;
              if (retflag == JOBTODO)
              {
                /*  Set loop variable */
                iteration = varblock[0][current] + ini;
                varblock[2][current] = varblock[2][current] + 1;
                goto labelStartIteration_1;
              }
labelEndLoop_1:
              ;	/*freeWheel(current); */
            }		/*End for initLoop() */
          }		// #pragma omp parallel

          /************************************************************************/
          // speccode: Stop measuring time
          spec_stop = get_time();
          spec_time += spec_stop - spec_start;
          /************************************************************************/
        }		// end if
      }			// end for j
    }			// end if
  }				// end for i

  // End of process 

  // speccode: free speculation structures
  specend ();


  // --------------------------------------------------------------------
  //      Check results 

  for (i = 0; i < N; i++)
  {
    px = input[0][i];
    py = input[1][i];
    if (cuad_dist (px, py, centrox, centroy) - cuad_radio > EPS)
      printf ("mec-2d: Fail!\n");
  }

  // --------------------------------------------------------------------
  //      Show results 

  printf ("mec-2d: Points processed: %d\n", N);
  printf ("mec-2d: Center of MEC: ('%f', '%f')\n", centrox, centroy);
  printf ("mec-2d: Squared radius of MEC: %f\n", cuad_radio);
  // printf("mec-2d: i counter value: %d\n", cont_i);
  // printf("mec-2d: j counter value: %d\n", cont_j);
  // printf("mec-2d: k counter value: %d\n", cont_k);
  printf ("mec-2d: Number of points that define the solution: %f\n", puntdef);
  if (puntdef >= 1)
    printf ("mec-2d: Point one: ('%f', '%f')\n", solpx, solpy);
  if (puntdef >= 2)
    printf ("mec-2d: Point two: ('%f', '%f')\n", solqx, solqy);
  if (puntdef >= 3)
    printf ("mec-2d: Point three: ('%f', '%f')\n", solrx, solry);
  // printf("\n\nmec2d: Timing K loop: %f seconds\n",counter);

#ifdef TIME_SPECLOAD
  printSpecloadTime();
#endif
#ifdef TIME_SPECSTORE
  printSpecstoreTime();
#endif
#ifdef MEASURE_COUNTERS
  specepilogue();
  // printAccesses();
#endif

  // -----------------------------------------------------------
  // speccode: print time results
  fprintf (stderr, "--------------------------------------------------\n");
  fprintf (stderr, "speccode: Time elapsed: %lf seconds.number_calls=%d\n", spec_time,number_calls);
  // -----------------------------------------------------------

  return (0);
}				// main

// =====================================================================

// =====================================================================

  double
cuad_dist (double px, double py, double qx, double qy)
{

  double res1, res2;

  res1 = (px - qx);
  res1 = res1 * res1;
  res2 = (py - qy);
  res2 = res2 * res2;

  return (res1 + res2);

}

// =====================================================================
