
/************************************************************************/
// Constants 
//
// N: total number of points (min. 3)
#define N 1000000 //1.000.000
#define NITER 995000 // (N - MAX_ANCHORS) = 95000

// MAX_ANCHORS: total number of anchors (min. 1)
// For 5M: 1990.
// For 1M: 850.
#define MAX_ANCHORS 5000
//#define MAX_ANCHORS 850

// MAX_TRIANGLES: Maximum number of triangles.
// The number of triangles on the final triangulation is well known
// (2*N-2, N = #points). We need one additional triangle to start.
// We use 2*N+10, just in case.
#define MAX_TRIANGLES 2*N+10

// NULL element
#define NULO -1

/************************************************************************/
// Data structures
//
// triangles: Table with triangles
// We also use the first element of the last MAX_ANCHORS rows of "triangles" to hold
// the indices of the triangles that contain the MAX_ANCHORS anchors.
int triangles[MAX_TRIANGLES+MAX_ANCHORS][6];
//int ** triangles;

// Indexes for each column of "triangles"
#define p0 0
#define p1 1
#define p2 2
#define t0 3
#define t1 4
#define t2 5
#define anchor 0
#define azone MAX_TRIANGLES
// In this way, the triangle that holds the first anchor will
// be triangles[azone][anchor]
//
// **********************************************************************/
// points: Floating-point matrix that holds the points.
// NOTE: Last three elements of "points" will hold the bounds.
float points[N+3][2];
//float ** points;

// Indexes for each column of the "points" table.
#define x 0
#define y 1 

// Indexes for the three bounds
#define bound0 N
#define bound1 N+1
#define bound2 N+2

/************************************************************************/
// nt: number of triangles so far
int nt;

//Auxiliar variable to perform CRC over results
double dt_crc;


/************************************************************************/
// Functions
void bounding_triangle();
float signed_area(int a, int b, int c);
int inside_triangle2(int p, int a, int b,int c);
char insidecirc_test(int pp1, int pp2, int pp3, int pp4);
int legalize_edge_spec(int tt1, int ins_anchors, int current);
void legalize_edge(int tt1, int ins_anchors);
void revisa(int i, int pos);
void write_dt();
void count_dt();
void iterative_write(int tr);
void iterative_count(int tr);

/************************************************************************/
#include <stdio.h>
#include <stdlib.h>

//***********************************************************************
// speccode: OpenMP header file included
#include <omp.h>
//***********************************************************************
// speccode: Main header file 
#include "specEngine.h"
//***********************************************************************

/************************************************************************/
// speccode: CONSTANTES
// speccode: FILAS es el numero de elementos por fila de la matriz triangles
#define FILAS 6

int main()
{

  /************************************************************************/

  //  Local variables
  //
  //  i,j,k: Different indexes of for loops
  int i, j;//, k;

  //  inputfile: File descriptor
  FILE *inputfile;

  //  Local variables for triangulation
  //int triangleid, edge, tmp;
  //    int idx_p1, idx_p2, idx_p3, idx_p4;
  //  Pointers to triangles in "triangles"
  int tr_ptr;//, tt1, tt2;
  //  Pointers to points
  //    int pp1, pp2, pp3, pp4;
  //  Number of anchors inserted so far in the triangulation
  //int ins_anchors;

  //  Temporary variables (to facilitate speculative accesses)
  int temp_tri;
  int temp_p0, temp_p1, temp_p2;

  //  Auxiliar variables to find correct anchor
  float dist, aux;
  int cerc;

  //  Auxiliar variable to perform CRC over results
  //double dt_crc;

  /************************************************************************/
  //  Keeping triangulation structure clean:
  //    Points on triangle are always set in CCW order
  //    Adjacent triangles are called with the index of opposite vertex
  //    Insertions and flips keep in p1p2 the flippable edge
  //
  //    PRECONDITION:
  //    FLoating point exeption occurs if some point in the set is repeated
  /************************************************************************/
  //inicialize();

  //  Opens the input file
  if ( (inputfile = fopen("input.dat", "r")) )
  {
    for(i = 0; i < N; i++)
      fscanf(inputfile, "%f %f", &points[i][x], &points[i][y]);

    fclose(inputfile);
  }
  else
  {
    printf ( "Error reading from file \n " );
    exit(0);
  }
  //for(i = 0; i < N; i++)
  // printf("%f %f\n", points[i][x], points[i][y]);

  /************************************************************************/
  // Beginning of process
  //
  // Find a bounding triangle, initialize "triangles" and make the
  // bounding triangle the first one. 
  bounding_triangle();

  // Initialize index of next triangle to be added
  nt = 1;

  // Initialize the first anchor to the first triangle
  triangles[azone][anchor] = 0;

  /************************************************************************/
  // First loop: inseting only anchors
  /************************************************************************/

  // Beginning of triangulation
  for (i = 0; i < MAX_ANCHORS; i++)
  {
    /**************************************/
    // 0) Find the closest anchor to point i

    dist = (points[0][x]-points[i][x])*(points[0][x]-points[i][x]);
    dist = dist+(points[0][y]-points[i][y])*(points[0][y]-points[i][y]);
    cerc = 0;
    for ( j = 1; j <= i-1; j++)
    {
      aux = (points[j][x]-points[i][x])*(points[j][x]-points[i][x]);
      aux = aux+(points[j][y]-points[i][y])*(points[j][y]-points[i][y]);
      if ( aux < dist)
      {
        dist = aux;
        cerc = j;
      }
    }

    /*************************************/
    // 1) Find the triangle that contains p_i
    //    we start at the triangle containing the closest anchor
    //    at the end of the loop, tr_ptr contains the triangle index
    //    temp_p0, temp_p1 and temp_p2 are its vertices

    tr_ptr = triangles[azone+cerc][anchor];

    while(1)
    {

      temp_p0 = triangles[tr_ptr][p0];
      temp_p1 = triangles[tr_ptr][p1];
      temp_p2 = triangles[tr_ptr][p2];

      if (signed_area(i, temp_p0, temp_p1) < 0)
        tr_ptr = triangles[tr_ptr][t2];
      else
      {
        if (signed_area(i, temp_p1, temp_p2) < 0)
          tr_ptr = triangles[tr_ptr][t0];
        else
        {
          if (signed_area(i, temp_p2, temp_p0) < 0)
            tr_ptr= triangles[tr_ptr][t1];
          else
            break;
        }
      }
    } // end while


    /*************************************/
    // 2) Re-triangulate. 
    // Since i is inside temp_p0, temp_p1 and temp_p2, the new three triangles are
    // i, temp_p0, temp_p1 (stored in nt)
    // i, temp_p1, temp_p2 (stored in nt+1)
    // i, temp_p2, temp_p0 (stored in tr_ptr)

    // Since i is an anchor itself, we first anchor it to some of the three new
    // triangles (i.e. tr_ptr)
    triangles[azone+i][anchor] = tr_ptr;

    // First triangle:
    triangles[nt][p0] = i;
    triangles[nt][p1] = temp_p0;
    triangles[nt][p2] = temp_p1;

    triangles[nt][t1] = nt+1;
    triangles[nt][t2] = tr_ptr;
    temp_tri = triangles[tr_ptr][t2];
    triangles[nt][t0] = temp_tri;

    // Linking back if not in Chull
    if ( temp_tri != NULO )
    {
      for ( j= 0; j <= 2; j++)
        if ( triangles[temp_tri][t0+j] == tr_ptr) break;
      triangles[temp_tri][t0+j] = nt;
    }


    // Second triangle
    triangles[nt+1][p0] = i;
    triangles[nt+1][p1] = temp_p1;
    triangles[nt+1][p2] = temp_p2;

    triangles[nt+1][t1] = tr_ptr;
    triangles[nt+1][t2] = nt;
    temp_tri = triangles[tr_ptr][t0];
    triangles[nt+1][t0] = temp_tri;

    // Linking back if not in Chull
    if ( temp_tri != NULO )
    {
      for ( j= 0; j <= 2; j++)
        if ( triangles[temp_tri][t0+j] == tr_ptr) break;
      triangles[temp_tri][t0+j] = nt+1;
    }


    // Now we modify triangle tr_ptr. No need to relink back!
    triangles[tr_ptr][p0] = i;
    triangles[tr_ptr][p1] = temp_p2;
    triangles[tr_ptr][p2] = temp_p0;
    triangles[tr_ptr][t0] = triangles[tr_ptr][t1]; 
    triangles[tr_ptr][t1] = nt;
    triangles[tr_ptr][t2] = nt+1;

    /*************************************/
    // 3) The anchor temp_p1 is no longer a vertex of tr_ptr 
    //    Reanchor temp_p1 to nt if not steiner point
    if ( temp_p1 < MAX_ANCHORS)
      triangles[azone+temp_p1][anchor] = nt;

    /*************************************/
    // 4) We have added two new triangles 
    nt = nt + 2;

    /*************************************/
    // 5) Legalize the new triangles NO COMPROBADO
    if ( triangles[tr_ptr][t0] != NULO) legalize_edge(tr_ptr,i);
    if ( triangles[nt-2][t0] != NULO) legalize_edge(nt-2,i);
    if ( triangles[nt-1][t0] != NULO) legalize_edge(nt-1,i);

  } // end for(i = 0; i < MAX_ANCHORS; i++)

  //for(i = 0; i < MAX_TRIANGLES; i++)
  // for(j = p0; j <= t2; j++)
  //  printf("%d\n",triangles[i][j]);

  /************************************************************************/
  // Second loop: inseting the remaining points (not anchors) 
  /************************************************************************/

  //***********************************************************************
  // speccode: OMP threading directive
  omp_set_num_threads(threads);
  // speccode: initialize speculation structrures
  specbegin(N-MAX_ANCHORS);
  //specinit();
  //specstart(N-MAX_ANCHORS);
  //***********************************************************************

  //***********************************************************************
  // speccode: Start profiling
  int iteration, ini;
  int value;
  double time_start, time_stop;
  double time_spec=0;

  time_start = get_time();
  //***********************************************************************

  //***********************************************************************
  // speccode: Speculative loop
#pragma omp parallel default(none)                       \
  private(nt, dist, cerc, j, aux, tr_ptr, temp_p0, temp_p1, temp_p2, \
      i, temp_tri, value, ini, iteration, current, tid, retflag) \
  shared(triangles, points, wheel_ns, wheel_ms, wheel, \
      upper_limit, varblock)
  {

#pragma omp for \
    schedule(static)
    //***********************************************************************
    // speccode: Modified loop bounds
    // FOR original:
    //for (i = MAX_ANCHORS; i < N; i++)
    for (tid = 0; tid < threads; tid++) {

      ini = MAX_ANCHORS;
      current = tid;
      //dynScheduler(current);
      iteration = varblock[0][current] + ini;

      if (iteration > upper_limit - 1)
        goto labelEndLoop_1;

      varblock[2][current] = varblock[2][current] + 1;
      //printf("\t\tvarblock[0][current]:%d,current:%d,iteration:%d\n",varblock[0][current],current,iteration);
labelStartIteration_1: //printf("ENDLOOP,iteration:%d,current:%d\n",iteration,current);

      i = iteration;

      //***********************************************************************
      // speccode: nmero de tringulos que hay actualmente
      // e1000:

      nt = (2 * i)+1;
      //***********************************************************************

      //     ************************************
      // 0) Find the closest anchor to point i 
      //
      //
      dist = (points[0][x]-points[i][x])*(points[0][x]-points[i][x]);
      dist = dist+(points[0][y]-points[i][y])*(points[0][y]-points[i][y]);
      cerc = 0;
      for ( j = 1; j < MAX_ANCHORS; j++)
      {
        aux = (points[j][x]-points[i][x])*(points[j][x]-points[i][x]);
        aux = aux+(points[j][y]-points[i][y])*(points[j][y]-points[i][y]);
        if ( aux < dist)
        {
          dist = aux;
          cerc = j;
        }
      }

      //     ************************************
      // 1) Find the triangle that contains p_i
      //    we start at the triangle containing the closest anchor
      //    at the end of the loop, tr_ptr contains the triangle index
      //    temp_p0, temp_p1 and temp_p2 are its vertices

      //***********************************************************************
      // speccode: speculative load. Original line:
      //    tr_ptr = triangles[azone+cerc][anchor];
      //indice = anchor + FILAS * (azone+cerc);
      //if( specload(indice, current, &value, (int *)triangles, &my_IV_tail) == -1)
      if( specload_pointer((void *) &(triangles[azone+cerc][anchor]), sizeof(triangles[azone+cerc][anchor]), current, (void *) &value) == -1)
        earlySquash(1);
      tr_ptr = value;
      //***********************************************************************

      while(1)
      {

        //***********************************************************************
        // speccode: speculative load. Original line:
        //      temp_p0 = triangles[tr_ptr][p0];
        //indice = p0 + FILAS * (tr_ptr);
        //if( specload(indice, current, &value, (int *)triangles, &my_IV_tail) == -1)
        if( specload_pointer((void *) &(triangles[tr_ptr][p0]), sizeof(triangles[tr_ptr][p0]), current, (void *) &value) == -1)
          earlySquash(1);
        temp_p0 = value;

        //      temp_p1 = triangles[tr_ptr][p1];
        //indice = p1 + FILAS * (tr_ptr);
        //if( specload(indice, current, &value, (int *)triangles, &my_IV_tail) == -1)
        if( specload_pointer((void *) &(triangles[tr_ptr][p1]), sizeof(triangles[tr_ptr][p1]), current, (void *) &value) == -1)
          earlySquash(1);
        temp_p1 = value;

        //      temp_p2 = triangles[tr_ptr][p2];
        //indice = p2 + FILAS * (tr_ptr);
        //if( specload(indice, current, &value, (int *)triangles, &my_IV_tail) == -1)
        if( specload_pointer((void *) &(triangles[tr_ptr][p2]), sizeof(triangles[tr_ptr][p2]), current, (void *) &value) == -1)
          earlySquash(1);
        temp_p2 = value;
        //***********************************************************************

        if (signed_area(i, temp_p0, temp_p1) < 0)
        {
          //***********************************************************************
          // speccode: speculative load. Original line:
          //        tr_ptr = triangles[tr_ptr][t2];
          //indice = t2 + FILAS * (tr_ptr);
          //  if( specload(indice, current, &value, (int *)triangles, &my_IV_tail) == -1)
          if( specload_pointer((void *) &(triangles[tr_ptr][t2]), sizeof(triangles[tr_ptr][t2]), current, (void *) &value) == -1)
            earlySquash(1);
          tr_ptr = value;
          //***********************************************************************
        }
        else
        {
          if (signed_area(i, temp_p1, temp_p2) < 0)
          {
            //***********************************************************************
            // speccode: speculative load. Original line:
            //          tr_ptr = triangles[tr_ptr][t0];
            //indice = t0 + FILAS * (tr_ptr);
            //  if( specload(indice, current, &value, (int *)triangles, &my_IV_tail) == -1)
            if( specload_pointer((void *) &(triangles[tr_ptr][t0]), sizeof(triangles[tr_ptr][t0]), current, (void *) &value) == -1)
              earlySquash(1);
            tr_ptr = value;
            //***********************************************************************
          }
          else
          {
            if (signed_area(i, temp_p2, temp_p0) < 0)
            {
              //***********************************************************************
              // speccode: speculative load. Original line:
              //            tr_ptr= triangles[tr_ptr][t1];
              //    indice = t1 + FILAS * (tr_ptr);
              //    if( specload(indice, current, &value, (int *)triangles, &my_IV_tail) == -1)
              if( specload_pointer((void *) &(triangles[tr_ptr][t1]), sizeof(triangles[tr_ptr][t1]), current, (void *) &value) == -1)
                earlySquash(1);
              tr_ptr = value;
              //***********************************************************************
            }
            else
              break;
          }
        }
      } // end while

      //     ************************************
      // 2) Re-triangulate. 
      // Since i is inside temp_p0, temp_p1 and temp_p2, the new three triangles are
      // i, temp_p0, temp_p1 (stored in nt)
      // i, temp_p1, temp_p2 (stored in nt+1)
      // i, temp_p2, temp_p0 (stored in tr_ptr)

      // First triangle:
      //***********************************************************************
      // speccode: speculative store. Original line:
      //    triangles[nt][p0] = i;
      //    triangles[nt][p1] = temp_p0;
      //    triangles[nt][p2] = temp_p1;
      //    triangles[nt][t1] = nt+1;
      //    triangles[nt][t2] = tr_ptr;
      //indice = p0 + FILAS * (nt);
      value = i;
      //specstore(indice, current, value, &my_IV_tail);
      specstore_pointer((void *) &(triangles[nt][p0]), sizeof(triangles[nt][p0]), current, (void *) &value);
      //indice = p1 + FILAS * (nt);
      value=temp_p0;
      //specstore(indice, current, value, &my_IV_tail);
      specstore_pointer((void *) &(triangles[nt][p1]), sizeof(triangles[nt][p1]), current, (void *) &value);
      //indice = p2 + FILAS * (nt);
      value=temp_p1;
      //specstore(indice, current, value, &my_IV_tail);
      specstore_pointer((void *) &(triangles[nt][p2]), sizeof(triangles[nt][p2]), current, (void *) &value);
      //indice = t1 + FILAS * (nt);
      value=nt+1;
      //specstore(indice, current, value, &my_IV_tail);
      specstore_pointer((void *) &(triangles[nt][t1]), sizeof(triangles[nt][t1]), current, (void *) &value);
      //indice = t2 + FILAS * (nt);
      value=tr_ptr;
      //specstore(indice, current, value, &my_IV_tail);
      specstore_pointer((void *) &(triangles[nt][t2]), sizeof(triangles[nt][t2]), current, (void *) &value);
      //***********************************************************************
      //***********************************************************************
      // speccode: speculative load. Original line:
      //    temp_tri = triangles[tr_ptr][t2];
      //indice = t2 + FILAS * (tr_ptr);
      //      if( specload(indice, current, &value, (int *)triangles, &my_IV_tail) == -1)
      if( specload_pointer((void *) &(triangles[tr_ptr][t2]), sizeof(triangles[tr_ptr][t2]), current, (void *) &value) == -1)
        earlySquash(1);
      temp_tri = value;
      //***********************************************************************
      //***********************************************************************
      // speccode: speculative store. Original line:
      //    triangles[nt][t0] = temp_tri;
      //indice = t0 + FILAS * (nt);
      value=temp_tri;
      //specstore(indice, current, value, &my_IV_tail);
      specstore_pointer((void *) &(triangles[nt][t0]), sizeof(triangles[nt][t0]), current, (void *) &value);
      //***********************************************************************

      // Linking back if not in Chull
      if ( temp_tri != NULO )
      {
        for ( j= 0; j <= 2; j++)
        {
          //***********************************************************************
          // speccode: speculative load. Original line:
          //        if ( triangles[temp_tri][t0+j] == tr_ptr) break;
          //indice = t0+j +FILAS * (temp_tri);
          //    if( specload(indice, current, &value, (int *)triangles, &my_IV_tail) == -1)
          if( specload_pointer((void *) &(triangles[temp_tri][t0+j]), sizeof(triangles[temp_tri][t0+j]), current, (void *) &value) == -1)
            earlySquash(1);
          if (value == tr_ptr) break;
          //***********************************************************************
        }
        //***********************************************************************
        // speccode: speculative store. Original line:
        //      triangles[temp_tri][t0+j] = nt;
        //        indice = t0+j + FILAS * (temp_tri);
        value=nt;
        //  specstore(indice, current, value, &my_IV_tail);
        specstore_pointer((void *) &(triangles[temp_tri][t0+j]), sizeof(triangles[temp_tri][t0+j]), current, (void *) &value);
        //***********************************************************************
      }

      // Second triangle
      //***********************************************************************
      // speccode: speculative store. Original line:
      //    triangles[nt+1][p0] = i;
      //    triangles[nt+1][p1] = temp_p1;
      //    triangles[nt+1][p2] = temp_p2;
      //    triangles[nt+1][t1] = tr_ptr;
      //    triangles[nt+1][t2] = nt;
      //      indice = p0 + FILAS * (nt+1);
      value=i;
      //specstore(indice, current, value, &my_IV_tail);
      specstore_pointer((void *) &(triangles[nt+1][p0]), sizeof(triangles[nt+1][p0]), current, (void *) &value);
      //      indice = p1 + FILAS * (nt+1);
      value=temp_p1;
      //specstore(indice, current, value, &my_IV_tail);
      specstore_pointer((void *) &(triangles[nt+1][p1]), sizeof(triangles[nt+1][p1]), current, (void *) &value);
      //    indice = p2 + FILAS * (nt+1);
      value=temp_p2;
      //specstore(indice, current, value, &my_IV_tail);
      specstore_pointer((void *) &(triangles[nt+1][p2]), sizeof(triangles[nt+1][p2]), current, (void *) &value);
      //  indice = t1 + FILAS * (nt+1);
      value=tr_ptr;
      //specstore(indice, current, value, &my_IV_tail);
      specstore_pointer((void *) &(triangles[nt+1][t1]), sizeof(triangles[nt+1][t1]), current, (void *) &value);
      //indice = t2 + FILAS * (nt+1);
      value=nt;
      //specstore(indice, current, value, &my_IV_tail);
      specstore_pointer((void *) &(triangles[nt+1][t2]), sizeof(triangles[nt+1][t2]), current, (void *) &value);
      //***********************************************************************
      //***********************************************************************
      // speccode: speculative load. Original line:
      //    temp_tri = triangles[tr_ptr][t0];
      // indice = t0 +FILAS * (tr_ptr);
      //if( specload(indice, current, &value, (int *)triangles, &my_IV_tail) == -1)
      if( specload_pointer((void *) &(triangles[tr_ptr][t0]), sizeof(triangles[tr_ptr][t0]), current, (void *) &value) == -1)
        earlySquash(1);
      temp_tri = value;
      //***********************************************************************
      //***********************************************************************
      // speccode: speculative store. Original line:
      //    triangles[nt+1][t0] = temp_tri;
      //indice = t0 + FILAS * (nt+1);
      value=temp_tri;
      //specstore(indice, current, value, &my_IV_tail);
      specstore_pointer((void *) &(triangles[nt+1][t0]), sizeof(triangles[nt+1][t0]), current, (void *) &value);
      //***********************************************************************

      // Linking back if not in Chull
      if ( temp_tri != NULO )
      {
        for ( j= 0; j <= 2; j++)
        {
          //***********************************************************************
          // speccode: speculative load. Original line:
          //        if ( triangles[temp_tri][t0+j] == tr_ptr) break;
          //indice = t0+j +FILAS * (temp_tri);
          //if( specload(indice, current, &value, (int *)triangles, &my_IV_tail) == -1)
          if( specload_pointer((void *) &(triangles[temp_tri][t0+j]), sizeof(triangles[temp_tri][t0+j]), current, (void *) &value) == -1)
            earlySquash(1);
          if (value == tr_ptr) break;
          //***********************************************************************
        }
        //***********************************************************************
        // speccode: speculative store. Original line:
        //      triangles[temp_tri][t0+j] = nt+1;
        //        indice = t0+j + FILAS * (temp_tri);
        value=nt+1;
        //specstore(indice, current, value, &my_IV_tail);
        specstore_pointer((void *) &(triangles[temp_tri][t0+j]), sizeof(triangles[temp_tri][t0+j]), current, (void *) &value);
        //***********************************************************************
      }

      // Now we modify triangle tr_ptr. No need to relink back!
      //***********************************************************************
      // speccode: speculative store. Original line:
      //      triangles[tr_ptr][p0] = i;
      //      triangles[tr_ptr][p1] = temp_p2;
      //      triangles[tr_ptr][p2] = temp_p0;
      //indice = p0 + FILAS * (tr_ptr);
      value=i;
      //specstore(indice, current, value, &my_IV_tail);
      specstore_pointer((void *) &(triangles[tr_ptr][p0]), sizeof(triangles[tr_ptr][p0]), current, (void *) &value);
      //        indice = p1 + FILAS * (tr_ptr);
      value=temp_p2;
      //specstore(indice, current, value, &my_IV_tail);
      specstore_pointer((void *) &(triangles[tr_ptr][p1]), sizeof(triangles[tr_ptr][p1]), current, (void *) &value);
      //      indice = p2 + FILAS * (tr_ptr);
      value=temp_p0;
      //specstore(indice, current, value, &my_IV_tail);
      specstore_pointer((void *) &(triangles[tr_ptr][p2]), sizeof(triangles[tr_ptr][p2]), current, (void *) &value);
      //***********************************************************************
      //***********************************************************************
      // speccode: speculative load & store. Original line:
      //      triangles[tr_ptr][t0] = triangles[tr_ptr][t1]; 
      //    indice = t1 + FILAS * (tr_ptr);
      //if( specload(indice, current, &value, (int *)triangles, &my_IV_tail) == -1)
      if( specload_pointer((void *) &(triangles[tr_ptr][t1]), sizeof(triangles[tr_ptr][t1]), current, (void *) &value) == -1)
        earlySquash(1);
      //        indice = t0 + FILAS * (tr_ptr);
      //specstore(indice, current, value, &my_IV_tail);
      specstore_pointer((void *) &(triangles[tr_ptr][t0]), sizeof(triangles[tr_ptr][t0]), current, (void *) &value);
      //***********************************************************************
      //***********************************************************************
      // speccode: speculative store. Original line:
      //      triangles[tr_ptr][t1] = nt;
      //      triangles[tr_ptr][t2] = nt+1;
      //        indice = t1 + FILAS * (tr_ptr);
      value=nt;
      //specstore(indice, current, value, &my_IV_tail);
      specstore_pointer((void *) &(triangles[tr_ptr][t1]), sizeof(triangles[tr_ptr][t1]), current, (void *) &value);
      //        indice = t2 + FILAS * (tr_ptr);
      value=nt+1;
      //specstore(indice, current, value, &my_IV_tail);
      specstore_pointer((void *) &(triangles[tr_ptr][t2]), sizeof(triangles[tr_ptr][t2]), current, (void *) &value);
      //***********************************************************************

      //     ************************************
      // 3) Check if anchors remain in the same triangles (including boundaries)
      // Onlye check if temp_p1 (whick is no longer a vertex of tr_ptr) is an anchor.
      // In that case, reanchor temp_p1 to nt 
      if ( temp_p1 < MAX_ANCHORS)
      {
        //***********************************************************************
        // speccode: speculative store. Original line:
        //      triangles[azone+temp_p1][anchor] = nt;
        //        indice = anchor + FILAS * (azone+temp_p1);
        value=nt;
        //specstore(indice, current, value, &my_IV_tail);
        specstore_pointer((void *) &(triangles[azone+temp_p1][anchor]), sizeof(triangles[azone+temp_p1][anchor]), current, (void *) &value);
        //***********************************************************************
      }

      //     *************************************
      // 4) We have added two new triangles 
      nt = nt + 2;

      //    ***********************************
      // 5) Legalize the new triangles 
      //***********************************************************************
      // speccode: speculative load. Original line:
      //    if ( triangles[tr_ptr][t0] != NULO) legalize_edge(tr_ptr,MAX_ANCHORS-1);
      //      indice = t0 + FILAS * (tr_ptr);
      //if( specload(indice, current, &value, (int *)triangles, &my_IV_tail) == -1)
      if( specload_pointer((void *) &(triangles[tr_ptr][t0]), sizeof(triangles[tr_ptr][t0]), current, (void *) &value) == -1)
        earlySquash(1);
      if ( value != NULO )
      {	
        //legalize_edge_spec(tr_ptr, MAX_ANCHORS-1, &my_IV_tail, current);
        // speccode: early squash
        if ( legalize_edge_spec(tr_ptr, MAX_ANCHORS-1, current) == -1)
          earlySquash(1);
      }
      //***********************************************************************
      //***********************************************************************
      // speccode: speculative load. Original line:
      //    if ( triangles[nt-2][t0] != NULO) legalize_edge(nt-2,MAX_ANCHORS-1);
      //      indice = t0 + FILAS * (nt-2);
      //if( specload(indice, current, &value, (int *)triangles, &my_IV_tail) == -1)
      if( specload_pointer((void *) &(triangles[nt-2][t0]), sizeof(triangles[nt-2][t0]), current, (void *) &value) == -1)
        earlySquash(1);
      if ( value != NULO )
      {
        //legalize_edge_spec(nt-2, MAX_ANCHORS-1, &my_IV_tail, current);
        // speccode: early squash
        if ( legalize_edge_spec(nt-2, MAX_ANCHORS-1, current) == -1)
          earlySquash(1);
      }
      //***********************************************************************
      //***********************************************************************
      // speccode: speculative load. Original line:
      //    if ( triangles[nt-1][t0] != NULO) legalize_edge(nt-1,MAX_ANCHORS-1);
      //      indice = t0 + FILAS * (nt-1);
      //if( specload(indice, current, &value, (int *)triangles, &my_IV_tail) == -1)
      if( specload_pointer((void *) &(triangles[nt-1][t0]), sizeof(triangles[nt-1][t0]), current, (void *) &value) == -1)
        earlySquash(1);
      if ( value != NULO )
      {
        //legalize_edge_spec(nt-1, MAX_ANCHORS-1, &my_IV_tail, current);
        // speccode: early squash
        if ( legalize_edge_spec(nt-1, MAX_ANCHORS-1, current) == -1)
          earlySquash(1);
      }
      //***********************************************************************

      // **********************************************************************
      // speccode: end of thread and commit
      /**********************************************************************             \
      // speccode: end of thread and commit                                               \
      //           First check if block of iterations is done **************/             \
      //labelEndIteration_1:
      //if ((iteration != varblock[1][current] + ini) && (iteration < NITER - 1)) {
      // It is necessary to set the condition with the number of iterations plus 'ini'
      if ((iteration != varblock[1][current] + ini) && (iteration < NITER + ini - 1)) {
        //if ((iteration != varblock[1][current] + ini) && (iteration < N - 1)) {
        iteration = iteration + 1;
        goto labelStartIteration_1;
      }
      /* Thread done, perform the commit */
labelSquash_1:
      retflag = threadend_pointer(&current);
      /* End if job done */
      if (retflag == JOBDONE) goto labelEndLoop_1;
      if (retflag == JOBTODO) {
        /*  Set loop variable */
        iteration = varblock[0][current] + ini;
        varblock[2][current] = varblock[2][current] + 1;
        goto labelStartIteration_1;
      }
labelEndLoop_1:
      ; /*freeWheel(current);*/
      } /*End for initLoop()*/
    } // #pragma omp parallel


    //***********************************************************************
    // speccode: time measurement and display
    time_stop = get_time();
    time_spec += time_stop - time_start;
    fprintf(stderr,"------------------------------------------------\n");
    fprintf(stderr,"speccode: Accumulated time for target loop: %lf seconds.\n", time_spec);

    //***********************************************************************

    //	speccode: free speculation structures
    specend();

    nt = 2*N+1;

    //  Write the triangulation
    //    write_dt();

    //  Calculate CRC of triangulation
    count_dt();

    //***********************************************************************
    //  specepilogue(): Show total number of violations and squashes.
    //  (should be called as last instruction of the application)
    // specepilogue();

    //***********************************************************************
#ifdef TIME_SPECLOAD
    printSpecloadTime();
#endif
#ifdef TIME_SPECSTORE
    printSpecstoreTime();
#endif

#ifdef MEASURE_COUNTERS
    specepilogue();
    // printAccesses();
#endif
    /*for( i = 0; i < MAX_TRIANGLES+MAX_ANCHORS;i++)
      free(triangles[i]);
      free(triangles);
      for( i = 0; i < N+3;i++)
      free(points[i]);
      free(points);*/
    return 0;

    } // END MAIN


    /************************************************************************/
    void bounding_triangle()
    {

      int i, j;
      float M, M3, tmp;

      //  Find M, the max absolut value of any coordinate of every point.

      M = -1.0;
      for (i = 0; i < N; i++)
      {
        tmp = points[i][x];
        if (tmp < 0) tmp = -tmp;
        if (tmp > M) M = tmp;

        tmp = points[i][y];
        if (tmp < 0) tmp = -tmp;
        if (tmp > M) M = tmp;
      }

      //  These formulas are from De Berg et all, Computational Geometry Algorithms,
      //  pg 196.

      M3 = M*3.0;

      //  We store the points that define the bounding triangle.
      //  Point "bound0" is the upper vertex
      points[bound0][x] = 0;
      points[bound0][y] = M3;

      //  Point "bound1" is the left vertex
      points[bound1][x] = -M3;
      points[bound1][y] = -M3;

      //  Point "bound2" is the right vertex
      points[bound2][x] = M3;
      points[bound2][y] = 0;

      //  Initialize triangles[] matrix ( LO INICIALIZO A -1 PARA DIFERENCIAR DEL INDICE 0)
      for(i = 0; i < MAX_TRIANGLES+MAX_ANCHORS; i++)
        for(j = p0; j <= t2; j++)
          triangles[i][j] = NULO;

      //  We now init the triangulation to the triangular bound
      triangles[0][p0] = bound0;
      triangles[0][p1] = bound1;
      triangles[0][p2] = bound2;

      //  Initialization of the remaining elements of the first triangle
      triangles[0][t0] = NULO;
      triangles[0][t1] = NULO;
      triangles[0][t2] = NULO;

    } // end bounding_triangle()

    /************************************************************************/
    float signed_area(int a, int b, int c)
    {

      return ((points[b][x] - points[a][x]) * ( points[c][y] - points[a][y] ) - \
          (points[b][y] - points[a][y]) * ( points[c][x] - points[a][x] ));


    } // end signed_area()

    /************************************************************************/
    void legalize_edge(int tt1, int ins_anchors)
    {

      // Local variables
      int pp0, pp1, pp2, pp3;
      int j, k;
      int tt2, tr_adj_aux, tr_adj;

      pp0 = triangles[tt1][p0];
      pp1 = triangles[tt1][p1];
      pp2 = triangles[tt1][p2];


      // Fourth point is in adjacent triangle through p1--p2

      tt2 = triangles[tt1][t0];

      // Find edge number j through which tt1 is adjacent to tt2
      for ( j = 0; j <= 2; j++)
        if ( triangles[tt2][t0+j] == tt1) break;
      pp3 = triangles[tt2][p0+j];

      if (insidecirc_test(pp0, pp1, pp2, pp3) == 'T')
      {
        // Flip

        // First, store triangle adjacent to tt2 through edge j--j+2
        k = j+1;
        if ( k >= 3 ) k = k-3;
        tr_adj_aux = triangles[tt2][t0+k];

        // Second, store triangle adjacent to tt2 throug edge j--j+1
        k = j+2;
        if ( k >= 3) k = k-3;
        tr_adj = triangles[tt2][t0+k];

        // Third, if not in CH, find index where tr_adj is adjacent to tt2
        if ( tr_adj != NULO )
        {
          for ( k = 0; k <= 2; k++ )
            if (triangles[tr_adj][t0+k] == tt2) break;
          // New external adjacency
          triangles[tr_adj][t0+k] = tt1;
        }
        triangles[tt1][t0] = tr_adj;

        // Idem, toward tt1 through edge p0--p1
        tr_adj = triangles[tt1][t2];
        if (tr_adj != NULO)
        {
          for ( k = 0; k <= 2; k++)
            if (triangles[tr_adj][t0+k] == tt1) break;
          // New external adjacency
          triangles[tr_adj][t0+k] = tt2;
        }
        triangles[tt2][t2] = tr_adj;

        // Fourth, modify tt1 and tt2
        triangles[tt2][t0] = tr_adj_aux;
        triangles[tt2][t1] = tt1;
        triangles[tt1][t2] = tt2;

        triangles[tt2][p0] = pp0;
        triangles[tt2][p1] = pp1;
        triangles[tt2][p2] = pp3;
        triangles[tt1][p1] = pp3;

        // Reanchor (if needed)
        if ( pp1 <= ins_anchors)
        {
          if ( triangles[azone+pp1][anchor] == tt1)
            triangles[azone+pp1][anchor] = tt2;
        }
        if ( pp2 <= ins_anchors)
        {
          //printf("%d\n",pp2);
          if ( triangles[azone+pp2][anchor] == tt2)
            triangles[azone+pp2][anchor] = tt1;
        }

        // Recursive call... fingers crossed
        if ( triangles[tt1][t0] != NULO ) legalize_edge(tt1, ins_anchors);
        if ( triangles[tt2][t0] != NULO ) legalize_edge(tt2, ins_anchors);

      } // end if

    } // end legalize_edge()

    /************************************************************************/
    /************************************************************************/
    int legalize_edge_spec(int tt1,int ins_anchors, int current)
    {

      // Local variables
      int pp0, pp1, pp2, pp3;
      int j, k;
      int tt2, tr_adj_aux, tr_adj;

      int myvalue;
      //int indice;


      //***********************************************************************
      // speccode: speculative load. Original line:
      //  pp0 = triangles[tt1][p0];
      //    indice = p0 + FILAS * (tt1);
      //if( specload(indice, current, &myvalue, (int *)triangles, my_IV_tail) == -1)
      if( specload_pointer((void *) &(triangles[tt1][p0]), sizeof(triangles[tt1][p0]), current, (void *) &myvalue) == -1)
        goto e67891;
      pp0 = myvalue;
      //***********************************************************************
      //***********************************************************************
      // speccode: speculative load. Original line:
      //  pp1 = triangles[tt1][p1];
      //    indice = p1 + FILAS * (tt1);
      //if( specload(indice, current, &myvalue, (int *)triangles, my_IV_tail) == -1)
      if( specload_pointer((void *) &(triangles[tt1][p1]), sizeof(triangles[tt1][p1]), current, (void *) &myvalue) == -1)
        goto e67891;
      pp1 = myvalue;
      //***********************************************************************
      //***********************************************************************
      // speccode: speculative load. Original line:
      //  pp2 = triangles[tt1][p2];
      //  indice = p2 + FILAS * (tt1);
      //if( specload(indice, current, &myvalue, (int *)triangles, my_IV_tail) == -1)
      if( specload_pointer((void *) &(triangles[tt1][p2]), sizeof(triangles[tt1][p2]), current, (void *) &myvalue) == -1)
        goto e67891;
      pp2 = myvalue;
      //***********************************************************************

      // Fourth point is in adjacent triangle through p1--p2

      //***********************************************************************
      // speccode: speculative load. Original line:
      //  tt2 = triangles[tt1][t0];
      //    indice = t0 + FILAS * (tt1);
      //if( specload(indice, current, &myvalue, (int *)triangles, my_IV_tail) == -1)
      if( specload_pointer((void *) &(triangles[tt1][t0]), sizeof(triangles[tt1][t0]), current, (void *) &myvalue) == -1)
        goto e67891;
      tt2 = myvalue;
      //***********************************************************************

      // Find edge number j through which tt1 is adjacent to tt2
      for ( j = 0; j <= 2; j++)
      {
        //***********************************************************************
        // speccode: speculative load. Original line:
        //    if ( triangles[tt2][t0+j] == tt1) break;
        //      indice = t0+j + FILAS * (tt2);
        //  if( specload(indice, current, &myvalue, (int *)triangles, my_IV_tail) == -1)
        if( specload_pointer((void *) &(triangles[tt2][t0+j]), sizeof(triangles[tt2][t0+j]), current, (void *) &myvalue) == -1)
          goto e67891;
        if( myvalue == tt1 ) break;
        //***********************************************************************
      }
      //***********************************************************************
      // speccode: speculative load. Original line:
      //  pp3 = triangles[tt2][p0+j];
      //    indice = p0+j + FILAS * (tt2);
      //if( specload(indice, current, &myvalue, (int *)triangles, my_IV_tail) == -1)
      if( specload_pointer((void *) &(triangles[tt2][p0+j]), sizeof(triangles[tt2][p0+j]), current, (void *) &myvalue) == -1)
        goto e67891;
      pp3 = myvalue;
      //***********************************************************************

      if (insidecirc_test(pp0, pp1, pp2, pp3) == 'T')
      {
        // Flip

        // First, store triangle adjacent to tt2 through edge j--j+2
        k = j+1;
        if ( k >= 3 ) k = k-3;
        //***********************************************************************
        // speccode: speculative load. Original line:
        //    tr_adj_aux = triangles[tt2][t0+k];
        //      indice = t0+k + FILAS * (tt2);
        //  if( specload(indice, current, &myvalue, (int *)triangles, my_IV_tail) == -1)
        if( specload_pointer((void *) &(triangles[tt2][t0+k]), sizeof(triangles[tt2][t0+k]), current, (void *) &myvalue) == -1)
          goto e67891;
        tr_adj_aux = myvalue;
        //***********************************************************************

        // Second, store triangle adjacent to tt2 throug edge j--j+1
        k = j+2;
        if ( k >= 3) k = k-3;
        //***********************************************************************
        // speccode: speculative load. Original line:
        //    tr_adj = triangles[tt2][t0+k];
        //      indice = t0+k + FILAS * (tt2);
        //  if( specload(indice, current, &myvalue, (int *)triangles, my_IV_tail) == -1)
        if( specload_pointer((void *) &(triangles[tt2][t0+k]), sizeof(triangles[tt2][t0+k]), current, (void *) &myvalue) == -1)
          goto e67891;
        tr_adj = myvalue;
        //***********************************************************************

        // Third, if not in CH, find index where tr_adj is adjacent to tt2
        if ( tr_adj != NULO )
        {
          for ( k = 0; k <= 2; k++ )
          {
            //***********************************************************************
            // speccode: speculative load. Original line:
            //        if (triangles[tr_adj][t0+k] == tt2) break;
            //          indice = t0+k + FILAS * (tr_adj);
            //if( specload(indice, current, &myvalue, (int *)triangles, my_IV_tail) == -1)
            if( specload_pointer((void *) &(triangles[tr_adj][t0+k]), sizeof(triangles[tr_adj][t0+k]), current, (void *) &myvalue) == -1)
              goto e67891;
            if( myvalue == tt2 ) break;
            //***********************************************************************
          }
          // New external adjacency
          //***********************************************************************
          // speccode: speculative store. Original line:
          //      triangles[tr_adj][t0+k] = tt1;
          //        indice = t0+k + FILAS * (tr_adj);
          myvalue=tt1;
          //specstore(indice, current, myvalue, my_IV_tail);
          specstore_pointer((void *) &(triangles[tr_adj][t0+k]), sizeof(triangles[tr_adj][t0+k]), current, (void *) &myvalue);
          //***********************************************************************
        }
        //***********************************************************************
        // speccode: speculative store. Original line:
        //    triangles[tt1][t0] = tr_adj;
        //      indice = t0 + FILAS * (tt1);
        myvalue=tr_adj;
        //specstore(indice, current, myvalue, my_IV_tail);
        specstore_pointer((void *) &(triangles[tt1][t0]), sizeof(triangles[tt1][t0]), current, (void *) &myvalue);
        //***********************************************************************

        // Idem, toward tt1 through edge p0--p1
        //***********************************************************************
        // speccode: speculative load. Original line:
        //    tr_adj = triangles[tt1][t2];
        //      indice = t2 + FILAS * (tt1);
        //  if( specload(indice, current, &myvalue, (int *)triangles, my_IV_tail) == -1)
        if( specload_pointer((void *) &(triangles[tt1][t2]), sizeof(triangles[tt1][t2]), current, (void *) &myvalue) == -1)
          goto e67891;
        tr_adj = myvalue;
        //***********************************************************************
        if (tr_adj != NULO)
        {
          for ( k = 0; k <= 2; k++)
          {
            //***********************************************************************
            // speccode: speculative load. Original line:
            //        if (triangles[tr_adj][t0+k] == tt1) break;
            //          indice = t0+k + FILAS * (tr_adj);
            //      if( specload(indice, current, &myvalue, (int *)triangles, my_IV_tail) == -1)
            if( specload_pointer((void *) &(triangles[tr_adj][t0+k]), sizeof(triangles[tr_adj][t0+k]), current, (void *) &myvalue) == -1)
              goto e67891;
            if( myvalue == tt1 ) break;
            //***********************************************************************
          }
          // New external adjacency
          //***********************************************************************
          // speccode: speculative store. Original line:
          //      triangles[tr_adj][t0+k] = tt2;
          //        indice = t0+k + FILAS * (tr_adj);
          myvalue=tt2;
          //specstore(indice, current, myvalue, my_IV_tail);
          specstore_pointer((void *) &(triangles[tr_adj][t0+k]), sizeof(triangles[tr_adj][t0+k]), current, (void *) &myvalue);
          //***********************************************************************
        }
        //***********************************************************************
        // speccode: speculative store. Original line:
        //    triangles[tt2][t2] = tr_adj;
        //      indice = t2 + FILAS * (tt2);
        myvalue=tr_adj;
        //specstore(indice, current, myvalue, my_IV_tail);
        specstore_pointer((void *) &(triangles[tt2][t2]), sizeof(triangles[tt2][t2]), current, (void *) &myvalue);
        //***********************************************************************

        // Fourth, modify tt1 and tt2
        //***********************************************************************
        // speccode: speculative store. Original line:
        //    triangles[tt2][t0] = tr_adj_aux;
        //      indice = t0 + FILAS * (tt2);
        myvalue=tr_adj_aux;
        //specstore(indice, current, myvalue, my_IV_tail);
        specstore_pointer((void *) &(triangles[tt2][t0]), sizeof(triangles[tt2][t0]), current, (void *) &myvalue);
        //***********************************************************************
        //***********************************************************************
        // speccode: speculative store. Original line:
        //    triangles[tt2][t1] = tt1;
        //      indice = t1 + FILAS * (tt2);
        myvalue=tt1;
        //specstore(indice, current, myvalue, my_IV_tail);
        specstore_pointer((void *) &(triangles[tt2][t1]), sizeof(triangles[tt2][t1]), current, (void *) &myvalue);
        //***********************************************************************
        //***********************************************************************
        // speccode: speculative store. Original line:
        //    triangles[tt1][t2] = tt2;
        //      indice = t2 + FILAS * (tt1);
        myvalue=tt2;
        //specstore(indice, current, myvalue, my_IV_tail);
        specstore_pointer((void *) &(triangles[tt1][t2]), sizeof(triangles[tt1][t2]), current, (void *) &myvalue);
        //***********************************************************************

        //***********************************************************************
        // speccode: speculative store. Original line:
        //    triangles[tt2][p0] = pp0;
        //      indice = p0 + FILAS * (tt2);
        myvalue=pp0;
        //specstore(indice, current, myvalue, my_IV_tail);
        specstore_pointer((void *) &(triangles[tt2][p0]), sizeof(triangles[tt2][p0]), current, (void *) &myvalue);
        //***********************************************************************
        //***********************************************************************
        // speccode: speculative store. Original line:
        //    triangles[tt2][p1] = pp1;
        //      indice = p1 + FILAS * (tt2);
        myvalue=pp1;
        //specstore(indice, current, myvalue, my_IV_tail);
        specstore_pointer((void *) &(triangles[tt2][p1]), sizeof(triangles[tt2][p1]), current, (void *) &myvalue);
        //***********************************************************************
        //***********************************************************************
        // speccode: speculative store. Original line:
        //    triangles[tt2][p2] = pp3;
        //      indice = p2 + FILAS * (tt2);
        myvalue=pp3;
        //specstore(indice, current, myvalue, my_IV_tail);
        specstore_pointer((void *) &(triangles[tt2][p2]), sizeof(triangles[tt2][p2]), current, (void *) &myvalue);
        //***********************************************************************
        //***********************************************************************
        // speccode: speculative store. Original line:
        //    triangles[tt1][p1] = pp3;
        //      indice = p1 + FILAS * (tt1);
        myvalue=pp3;
        //specstore(indice, current, myvalue, my_IV_tail);
        specstore_pointer((void *) &(triangles[tt1][p1]), sizeof(triangles[tt1][p1]), current, (void *) &myvalue);
        //***********************************************************************

        // Reanchor (if needed)
        if ( pp1 < ins_anchors)
        {
          //***********************************************************************
          // speccode: speculative load. Original line:
          //      if ( triangles[azone+pp1][anchor] == tt1)
          //        indice = anchor + FILAS * (azone+pp1);
          //    if( specload(indice, current, &myvalue, (int *)triangles, my_IV_tail) == -1)
          if( specload_pointer((void *) &(triangles[azone+pp1][anchor]), sizeof(triangles[azone+pp1][anchor]), current, (void *) &myvalue) == -1)
            goto e67891;
          if( myvalue == tt1 ) 
          {
            //***********************************************************************
            //***********************************************************************
            // speccode: speculative store. Original line:
            //        triangles[azone+pp1][anchor] = tt2;
            //          indice = anchor + FILAS * (azone+pp1);
            myvalue=tt2;
            //specstore(indice, current, myvalue, my_IV_tail);
            specstore_pointer((void *) &(triangles[azone+pp1][anchor]), sizeof(triangles[azone+pp1][anchor]), current, (void *) &myvalue);
            //***********************************************************************
          }
        }
        if ( pp2 < ins_anchors)
        {
          //***********************************************************************
          // speccode: speculative load. Original line:
          //      if ( triangles[azone+pp2][anchor] == tt2)
          //        indice = anchor + FILAS * (azone+pp2);
          //    if( specload(indice, current, &myvalue, (int *)triangles, my_IV_tail) == -1)
          if( specload_pointer((void *) &(triangles[azone+pp2][anchor]), sizeof(triangles[azone+pp2][anchor]), current, (void *) &myvalue) == -1)
            goto e67891;
          if( myvalue == tt2 ) 
          {
            //***********************************************************************
            //***********************************************************************
            // speccode: speculative store. Original line:
            //        triangles[azone+pp2][anchor] = tt1;
            //          indice = anchor + FILAS * (azone+pp2);
            myvalue=tt1;
            //    specstore(indice, current, myvalue, my_IV_tail);
            specstore_pointer((void *) &(triangles[azone+pp2][anchor]), sizeof(triangles[azone+pp2][anchor]), current, (void *) &myvalue);
            //***********************************************************************
          }
        }

        // Recursive call... fingers crossed
        //***********************************************************************
        // speccode: speculative load. Original line:
        //    if ( triangles[tt1][t0] != NULO ) legalize_edge(tt1, ins_anchors);
        //      indice = t0 + FILAS * (tt1);
        //  if( specload(indice, current, &myvalue, (int *)triangles, my_IV_tail) == -1)
        if( specload_pointer((void *) &(triangles[tt1][t0]), sizeof(triangles[tt1][t0]), current, (void *) &myvalue) == -1)
          goto e67891;
        if ( myvalue != NULO )
        {
          //legalize_edge_spec(tt1, ins_anchors, my_IV_tail, current);
          // speccode: early squash
          if ( legalize_edge_spec(tt1, ins_anchors, current) == -1)
            goto e67891;
        }
        //***********************************************************************
        //***********************************************************************
        // speccode: speculative load. Original line:
        //    if ( triangles[tt2][t0] != NULO ) legalize_edge(tt2, ins_anchors);
        //      indice = t0 + FILAS * (tt2);
        //  if( specload(indice, current, &myvalue, (int *)triangles, my_IV_tail) == -1)
        if( specload_pointer((void *) &(triangles[tt2][t0]), sizeof(triangles[tt2][t0]), current, (void *) &myvalue) == -1)
          goto e67891;
        if ( myvalue != NULO )
        {
          //legalize_edge_spec(tt2, ins_anchors, my_IV_tail, current);
          // speccode: early squash
          if ( legalize_edge_spec(tt2, ins_anchors, current) == -1)
            goto e67891;
        }
        //***********************************************************************

      } // end if

      //***********************************************************************
      // speccode: No se han producido violaciones de dependencia
      return 0;
      //***********************************************************************

      //***********************************************************************
      // speccode: si llegamos aqu se ha producido algn squashed
e67891:
      return -1;
      //**********************************************************************

    } // end legalize_edge_spec()


    /************************************************************************/
    char insidecirc_test(int pp1, int pp2, int pp3, int pp4)
    {

      double ccx, ccy;
      double r, d;
      double A,B,C,D,E,F,G;

      // Calculate the circumcenter of p1 p2 p3
      // Function circumcenter( p1, p2, p3, &cc ) embedded here.
      //
      // Computed using the fast algorithm (low number of mults) found on
      //  pg. 189 of O'Rourke

      A = points[pp2][x] - points[pp1][x];
      B = points[pp2][y] - points[pp1][y];
      C = points[pp3][x] - points[pp1][x];
      D = points[pp3][y] - points[pp1][y];

      E = A*(points[pp1][x] + points[pp2][x]) + B*(points[pp1][y] + points[pp2][y]);

      F = C*(points[pp1][x] + points[pp3][x]) + D*(points[pp1][y] + points[pp3][y]);

      G = 2*(A*(points[pp3][y]-points[pp2][y]) - B*(points[pp3][x] - points[pp2][x]));

      if (G == 0)
        return 'T';

      ccx = (D*E - B*F)/G;

      ccy = (A*F - C*E)/G;

      // Calculate distance from p to cc
      d = (points[pp4][x]-ccx)*(points[pp4][x]-ccx) + (points[pp4][y]-ccy)*(points[pp4][y]-ccy);

      // Calculate radius
      r = (ccx-points[pp1][x])*(ccx-points[pp1][x]) + (ccy-points[pp1][y])*(ccy-points[pp1][y]);

      // Dist less or equal to radius? --> inside
      if ( d<=r)
        return 'T';
      else
        return 'F'; 

    } // end insidecirc_test()

    /************************************************************************/
    int inside_triangle2(int p, int a,int b,int c)
    {

      if ( (signed_area(a,b,p)>=0) && (signed_area(b,c,p) >= 0) && (signed_area(a,p,c) >= 0))
        return 1;
      else
        return 0;

    }

    /************************************************************************/
    // int i - iteracion en la que estamos -
    // int pos - num identificando desde donde se ha llamado -
    void revisa(int i, int pos)
    {
      int tr_ptr;
      //int tr;
      int j;
      int temp_p0, temp_p1, temp_p2;

      // First, initialize the CRC. It will serve as accumulator in iterative_write().
      dt_crc = 0;

      for ( j = 0; j < MAX_ANCHORS ; j++)
      {
        tr_ptr = triangles[azone+j][p0];

        temp_p0 = triangles[tr_ptr][p0];
        temp_p1 = triangles[tr_ptr][p1];
        temp_p2 = triangles[tr_ptr][p2];

        if ((inside_triangle2(j,temp_p0,temp_p1,temp_p2) == 0))
        {
          printf("ERROR en reanchoring punto %d anchor %d en llamada num %d",i,j,pos);
          write_dt();
          break;
        }
      }

    }

    /************************************************************************/
    void write_dt()
    {

      iterative_write(1);

    } // end write_dt()

    /************************************************************************/
    void count_dt()
    {
      dt_crc = 0;

      iterative_count(1);
      // print the CRC
      printf("-------------------------------------------------------\n");
      printf("CRC: %f\n", dt_crc);
      printf("-------------------------------------------------------\n");

    }

    /************************************************************************/
    void iterative_write(int tr)
    {

      for (tr = 0; tr < nt; tr++)
      {

        printf(" %d\n",triangles[tr][p0]+1);
        printf(" %d\n",triangles[tr][p1]+1);
        printf(" %d\n",triangles[tr][p2]+1);
        //printf("%f %f\n",points[triangles[tr][p0]][x], points[triangles[tr][p0]][y]);
        //printf("%f %f\n",points[triangles[tr][p1]][x], points[triangles[tr][p1]][y]);
        //printf("%f %f\n",points[triangles[tr][p2]][x], points[triangles[tr][p2]][y]);

      }

    } // end iterative_write()

    /************************************************************************/
    void iterative_count(int tr)
    {

      for (tr = 0; tr < nt; tr++)
      {
        //    Add point locations to dt_crc
        dt_crc = dt_crc + points[triangles[tr][p0]][x] + points[triangles[tr][p0]][y];
        dt_crc = dt_crc + points[triangles[tr][p1]][x] + points[triangles[tr][p1]][y];
        dt_crc = dt_crc + points[triangles[tr][p2]][x] + points[triangles[tr][p2]][y];
        dt_crc = dt_crc + points[triangles[tr][p0]][x] + points[triangles[tr][p0]][y];
      }

    }

    /************************************************************************/
    void write_triangle(int tri, int report)
    {

      printf("------------------------\n");
      printf("write_triangle, report from call %d\n",report);
      printf("  Index: %d\n",tri);
      printf("  Point p0 is %d coord (%f ; %f)\n", triangles[p0][tri], points[x][triangles[p0][tri]],points[y][triangles[p0][tri]]);
      printf("  Point p1 is %d coord (%f ; %f)\n", triangles[p1][tri], points[x][triangles[p1][tri]],points[y][triangles[p1][tri]]);
      printf("  Point p2 is %d coord (%f ; %f)\n", triangles[p2][tri], points[x][triangles[p2][tri]],points[y][triangles[p2][tri]]);
      printf("  Linked through t0 to %d", triangles[t0][tri]);
      printf("  Linked through t1 to %d", triangles[t1][tri]);
      printf("  Linked through t2 to %d", triangles[t2][tri]);
      printf("write_triangle: end of report");
      printf("------------------------");

    }

