// ============================================================
// TREEDEFS: include file with common definitions for TREECODE.
// ============================================================
 
// --------------------------------------------------------------
// Fundamental constants for treecode, not to be altered lightly.
// --------------------------------------------------------------
 
// NULL        value denoting pointer to nothing.
// NDIM        number of spatial dimensions.
// NSUBC       number of descendents per cell.
// NQUAD       number of independent quadrupole components.
 
#include <stdlib.h>
#include <stdio.h>
#define NDIM 3
//#define NSUBC 2^NDIM
#define NSUBC 8
#define NQUAD 2*NDIM-1
 
// -------------------------------------------------------
// Treecode control parameters, input from parameter file.
// -------------------------------------------------------
 
#define false 0
#define true 1
 
// MXBODY      maximum number of bodies allowed.
// MXCELL      maximum number of cells allowed.
// INCELL      index of first cell in arrays.
// MXNODE      maximum number of nodes (bodies + cells).
 
#define MXBODY 8192
#define MXCELL MXBODY
#define INCELL MXBODY + 1
#define MXNODE MXBODY + MXCELL
 
// Function declarations

void initcd();
void advpos(double dt);
void advvel(double dt);
void diags();
void terror(char *msg);
void accel();
void inparm();
void getbds();
void begout();
void outhdr();
void outtim();
void outlog();
void outdag();
void outerr(char *msg);
void outcpu();
void putbds(int final);
void putmas(FILE *file);
void putpos(FILE *file);
void putvel(FILE *file);
void stpout();
void mktree();
void expbox();
void ldtree();
void ldbody(int p);
void hackcm();
void bflist(int *ind);

int mkcell();
int sbindx(int p, int q);
double cputimer();
