
#include <stdio.h>
#include <stdlib.h>

//***********************************************************************
// speccode: OpenMP header file included
#include <omp.h>
//***********************************************************************
//***********************************************************************
// speccode: Main header file with all common variables
#include "specEngine.h"
//************************************************************************

/*******************************************/
/* User parameters set by the programmer:  */
/*******************************************/
#define NITER 1000000

/*--------------------------------------------------------------------------*/

int main() {

	double value;

	//***********************************************************************
	// speccode: OMP threading directive
	omp_set_num_threads(threads);
	// speccode: initialize speculation structures
	specbegin(NITER);
	//***********************************************************************
	
	// speccode: Speculative loop
	int iteration, ini;

	double sumatorio = 3.123124;

	/************************************************************************/
	// speccode: Start profiling
	double spec_time=0;
	double spec_start, spec_stop;
	spec_start = get_time();
	/***********************************************************************/

#pragma omp parallel default(none)	\
	private(value,iteration, ini, current, tid, retflag)    \
	shared(sumatorio,wheel_ns, wheel_ms, wheel, upper_limit, varblock)
{

#pragma omp for \
	schedule(static)
	
	//for ( P = 0 ; P < NITER ; P++ )
	for (tid = 0; tid <= threads - 1; tid++) {
 
		ini = 0;
		current = tid;
		iteration = varblock[0][current]+ini;
		// If there is not work to this thread, go out.
		if (iteration > upper_limit - 1)
			goto labelSquash_1;
		// Increment the number of attempts for this iteration.
		varblock[2][current] = varblock[2][current]+1;
		
labelStartIteration_1: 
		//  For loop body code

		value = (iteration % 2) + 0.0003*iteration;

		// sumatorio = sumatorio + iteration;
		specredadd((unsigned char *) &sumatorio, sizeof(sumatorio), current, (unsigned char *) &value);

//labelEndIteration_1:
		if ((iteration != varblock[1][current] + ini) && (iteration < NITER - 1)) {
			iteration = iteration + 1;
			goto labelStartIteration_1;
		}
		
		/* Thread done, perform the commit */
labelSquash_1:

		retflag = threadend_pointer(&current);
		printf("sum=%lf,it=%d\n",sumatorio,iteration);
		
		/* End if job done */
		if (retflag == JOBDONE) goto labelEndLoop_1;
		if (retflag == JOBTODO) {
		
			/*  Set loop variable */
			iteration = varblock[0][current] + ini;
			varblock[2][current] = varblock[2][current] + 1;
			goto labelStartIteration_1;
		}

labelEndLoop_1:
		; /*freeWheel(current);*/
	
	} /*End of the FOR loop*/

} // #pragma omp parallel

	/************************************************************************/
	// speccode: Stop measuring time
	spec_stop = get_time();
	spec_time += spec_stop - spec_start;
	/************************************************************************/

//	speccode: free speculation structures
	specend();

	/****************************/
	/* Printing execution times */
	/****************************/
	printf("Sumatorio = %lf.\n", sumatorio);
	fprintf(stderr,"Speculative time: %lf seconds.\n", spec_time);


#ifdef TIME_SPECLOAD
 printSpecloadTime();
#endif
#ifdef TIME_SPECSTORE
 printSpecstoreTime();
#endif

#ifdef MEASURE_COUNTERS
// specepilogue();
 printAccesses();
#endif

	return 0;
}
