//
// Regression test to check the correct transformation for a min reduction, 
//  where 'min' is a local variable
//

#include <stdio.h>
#include <stdlib.h>

//***********************************************************************
// speccode: OpenMP header file included
#include <omp.h>
//***********************************************************************
//***********************************************************************
// speccode: Main header file with all common variables
#include "specEngine.h"
//***********************************************************************

/*******************************************/
/* User parameters set by the programmer:  */
/*******************************************/
// Array size 
#define MAX 100
// Number of iterations of the loop
#define NITER 1000000
//#define NITER 1

// Global data structure
int array[MAX];
/*--------------------------------------------------------------------------*/

int main() {

  /* Local variables:
   * 	P represents the number of iteration
   *		Q represents an the array index
   */
  int aux, i;
  FILE *file;
  int pre_min = 0;
  int min = 0;

  /* Open and read the file.  */
  if ((file = fopen("rand1000000.in", "r")) == NULL) {
    printf("Error opening the file\n ");
    exit(0);
  } else {
    fscanf(file, "%d", &aux);
    for (i = 0; i < MAX; i++) {
      array[i] = aux;
      fscanf(file, "%d", &aux);
    }
    fclose(file);
  }

  int x,y,z;

  /*  the elements of the array.  */
  for (i = 0; i < MAX; i++) {
    x = array[i];
    y = array[i] + 2;
    z = x + y;
    if (pre_min > z) pre_min = z;
  }
  printf(" Max of the elements of the array (before parallelization) = %d\n", pre_min);

  /* Start profiling.  */
  double seq_init = 0;
  double seq_end = 0;
  double seq_time = 0;
  seq_init = get_time();


  omp_set_num_threads(threads);
  specbegin(MAX);

  int iteration, ini, nIterations;
  int value;
  // The loop is not parallelizable because indexes of 
  // the array elements depends on the input values from a file
#pragma omp parallel for default(none) \
  private(x,y,z,iteration, ini, nIterations, current, tid, retflag, value) \
  shared(min,array,wheel_ns, wheel_ms, wheel, upper_limit, varblock)
  //for ( P = 0; P < MAX; P++) {
  for (tid = 0; tid <= threads - 1; tid++) {

    ini = 0;
    current = tid;
    iteration = varblock[0][current] + ini;
    nIterations = MAX;

    if (iteration > upper_limit - 1)
      goto labelSquash_1;

    varblock[2][current] = varblock[2][current] + 1;
labelStartIteration_1:

    x = array[iteration];
    y = array[iteration] + 2;
    z = x + y; 
    //if (min > z) min = z;
    value = z;
    specredmin ( (void *) &(min), sizeof(min), current, (void *) &value);

    //labelEndIteration_1:
    if ((iteration != varblock[1][current] + ini) && (iteration < nIterations - 1) ) {
      iteration = iteration + 1;
      goto labelStartIteration_1;
    }

    /* Thread done, perform the commit */
labelSquash_1:
    retflag = threadend_pointer (&current);
    /* End if job done */
    if (retflag == JOBDONE)
      goto labelEndLoop_1;
    if (retflag == JOBTODO) {
      /*  Set loop variable */
      iteration = varblock[0][current] + ini;
      varblock[2][current] = varblock[2][current] + 1;
      goto labelStartIteration_1;
    }
labelEndLoop_1:
    ; /*freeWheel(current); */
  }  /*End for initLoop() */
  //} // #pragma omp parallel

  //	}


/* Stop profiling.  */
seq_end = get_time();
seq_time += seq_end - seq_init;

/********************/
/* Writing results  */
/********************/
printf(" Min of the elements of the array (after parallelization) = %d\n", min);

/****************************/
/* Printing execution times */
/****************************/
fprintf(stderr, "Execution time: %f seg. \n", seq_time);

return 0;
}


