#!/bin/bash

# Check the input and set the variable THREADS
check_input () {
    if test $# != 1
    then
        echo "$0: format: $0 threads"
        echo -e "      example: $0 16\n"
        exit -1
    fi

    THREADS=$1

    echo "-------------------------------------------------------"
    echo "$0: running experiment with $THREADS threads."
    echo 
}

# run if user hits control-c
control_c()
{
    echo -en "\n*** Ouch! Exiting ***\n"
    echo "Canceling the job ${RUN}..."
    scancel $RUN
    echo "Job cancelled successfully"
    exit $?
}


# Execute the benchmark and save the result in the corresponding directory
execute () {

    if [ -z ${APP+x} ]
    then
        echo "APP is unset"
        exit -1
        #else
        #echo "APP is set to '$APP'"
    fi
    if [ -z ${THREADS+x} ]
    then
        echo "THREADS is unset"
        exit -1
        #else
        #echo "THREADS is set to '$THREADS'"
    fi

    if [ $THREADS -lt 10 ]
    then 
        FILENAME=$APP-p0$THREADS
    else
        FILENAME=$APP-p$THREADS
    fi 

    mkdir -p Results

    # Generate "run.slurm"
    echo "#! /bin/bash" > run.slurm

    echo "#SBATCH --job-name=${FILENAME}" >> run.slurm
    echo "#SBATCH --output=Results/${FILENAME}.out" >> run.slurm
    echo "#SBATCH --error=Results/${FILENAME}.err" >> run.slurm
    echo "#SBATCH --partition=trasgo" >> run.slurm
    echo "#SBATCH --nodelist=heracles" >> run.slurm
    echo "#SBATCH --nodes=1" >> run.slurm
    echo "#SBATCH --ntasks-per-node=1" >> run.slurm
    echo "#SBATCH --cpus-per-task=$THREADS" >> run.slurm
    echo "#SBATCH --exclusive" >> run.slurm

    echo -e "\nsrun /usr/bin/time -p ./$APP \n" >> run.slurm

    # Submit "run.slurm"
    cwdfront
    RUN=`sbatch run.slurm | cut -d " " -f 4`

    # trap keyboard interrupt (control-c)
    trap control_c SIGINT

    # Waiting for the run.slurm.o* file to be created
    echo "Experiment $RUN queued."
    while ! test -e Results/${FILENAME}.err
    do
        sleep 1 ;
    done

    # Waiting for the run.slurm.o* file to be created
    echo "Experiment $RUN running."
    squeue | grep $RUN > tt
    while test -s tt
    do
        sleep 1 ;
        squeue | grep $RUN > tt
    done
    rm tt

    echo "Experiment $RUN finished."

    # If the APP is not the sequential version, results are compared with it
    echo "$APP" | grep sequential > temp.txt
    if ! [ -s temp.txt ]
    then
        diff Results/${FILENAME}.out ../01*/Results/*.out > diff.txt
        if [ -s diff.txt ]
        then
            echo "$0: Results differ." 
            echo "diff.txt contains the differences (output of diff command)"
        else
            rm diff.txt
            rm run.slurm
        fi
    else
        rm run.slurm
    fi
    rm temp.txt

    cwdhome

}

