/* Information about plugins and header that can be used 
* could be find in the directory 'gcc -print-file-name=plugin' */

#include <gcc-plugin.h>
#include <plugin-version.h>
#include <coretypes.h>
#include <gimple.h>
#include <tree.h>
#include <tree-flow.h>
#include <tree-pass.h>
#include "langhooks.h"

#include <stdio.h>

#include "specprag.h"
#include "debug-specprag.h"
#include "spec-functions.h"
#include "handle-stmts.h"

/* We could choose how to access to varblock:
	-FALSE: We implement accesses to varblock as readings and writings 
					from a 1-dimensional array -> XXX: Deprecated!
	-TRUE: We implement accesses to varblock as function calls
*/	
#define use_varblock_function true
/* XXX: Using FALSE could produce errors, since the 'varblock1d' variable
created hasn't got the context of the loop parallelized.  */

int value_struct_num = 0;

/* Global array to maintain the name of each speculative variable.  */
char * spec_variables_name[10];
/* Global variable to save the number of speculative variables.  */
int spec_variables_num;
/* Global array to maintain each speculative variable.  */
tree spec_variables[10];

/* Global variables to maintain the 'value' variable for each kind of variable.  */
tree value_vars[10];

/* Global array to control which functions are copied because they contain
speculative variables and are called inside the speculative loop.  */
Function function_spec;
All_Functions all_functions;


/* Global variable to store the number of iterations of the FOR loop.  */
int nIt = maxiter; // maxiter is set within the Makefile.
/* Global variable to keep the gimple_stmt_iterator in the
beginning of the main() function.  */
gimple_stmt_iterator gsi_init;

/* Global flag to control that the engine functions 
inserted before the FOR loop.  */
bool firstLoopDone = false;

/* Global flag to detect the use of the 'specstart()' function.  */
bool specstart_detected = false;

/* Global variable to keep the reference to the FOR loop to parallelize
and add the OMP clauses to it.  */
gimple omp_for_loop;
/* Global variable to keep the gimple_stmt_iterator in the 
beginning of the FOR loop.  */
gimple_stmt_iterator gsi_for_loop;

/* Global variable to keep the lhs in a possible reading from or writing
into a speculative variable.  */
tree saved_lhs;


/***************************************************************/
/* Global flag to detect when it is necessary to compare labels, 
in the context of the detection of a 'max reduction'.  */
bool check_label = false;

/* Global variable to keep the label (tree) for the true branch of a IF-condition.
It is used to identify 'max reduction's.  */
tree cond_true_label = NULL_TREE;

/* Global variables to keep the lhs and the rhs of an IF-condition.  */
tree cond_lhs, cond_rhs;

/*****************************************************************/


/* Global variables for common contructs that are shared for multiple 
functions in this code (different sections in the source code that we want
to parallelize):

	int current;
	labelSquash
*/
tree current_t; 
tree current_t_alone; /* To use without the operator '&'.  */
tree current_t_addr;	/* To use with the operator '&'.  */
tree value;						/* To use with specload_ and specstore_pointer.  */
tree pre_if_var;			/* To use before if conditions in 
														getFirstBlock(), and specload_pointer().  */

tree specstart_iterations;  /* Got from specstart(iterations).  */

//TODO: We dont use getFirstBlock() anymore!!

/* To save the number of iterations and use it in specstart().  */
//TODO: Remove this. NOT USED. 
tree nIterations;			
gimple nIter_ini, nIter_1;

tree labelSquash_decl;
gimple labelSquash;

tree varblock1d; /* The 1-dimension version of varblock.  */

/* The gate is a callback tripped just before exec is executed
	If the gate returns true, exec is executed, 
	otherwise it is skipped.  */
static bool 
specprag_gate(void)
{
	return true;
}

/* Given a type TYPE, this function returns the name of the type.  */

static const char *
get_type_name (tree type)
{
  if (! TYPE_NAME (type))
    return NULL;

  if (TREE_CODE (TYPE_NAME (type)) == IDENTIFIER_NODE)
    return IDENTIFIER_POINTER (TYPE_NAME (type));
  else if (TREE_CODE (TYPE_NAME (type)) == TYPE_DECL
	   && DECL_NAME (TYPE_NAME (type)))
    return IDENTIFIER_POINTER (DECL_NAME (TYPE_NAME (type)));
  else
    return NULL;
}

/* TODO: REMOVE, CURRENTLY UNUSED.   */
static tree
set_value_type (const char * type_name)
{
	
	// INTEGER 
	if ( strcmp (type_name, "int") == 0) 
	{
		return integer_type_node;
	} 
	else if	(strcmp (type_name, "unsigned int") == 0) 
	{
		return unsigned_type_node;
	}
	else if	(strcmp (type_name, "long int") == 0) 
	{
		return long_integer_type_node;
	}
	else if	(strcmp (type_name, "long unsigned int") == 0) 
	{
		return long_unsigned_type_node;
	}
	else if	(strcmp (type_name, "long long int") == 0) 
	{
		return long_long_integer_type_node;
	}
	else if	(strcmp (type_name, "long long unsigned int") == 0) 
	{
		return long_long_unsigned_type_node;
	}
	else if (strcmp (type_name, "short int") == 0)
	{
		return short_integer_type_node;
	}
	else if (strcmp (type_name, "short unsigned int") == 0)
	{
		return short_unsigned_type_node;
	}

	
	// DOUBLE
	else if	(strcmp (type_name, "double") == 0) 
	{
		return double_type_node;
	}
	else if	(strcmp (type_name, "long double") == 0) 
	{
		return long_double_type_node;
	}
	
	// FLOAT
	else if	(strcmp (type_name, "float") == 0) 
	{
		return float_type_node;
	}
	
	// CHAR
	else if (strcmp (type_name, "char") == 0)
	{
		return char_type_node;
	}
	else if (strcmp (type_name, "unsigned char") == 0)
	{
		return unsigned_char_type_node;
	}
	else if (strcmp (type_name, "signed char") == 0)
	{
		return signed_char_type_node;
	}
	
	// By default integer_type_node
	else 
	{
		printf("Type of the speculative variable not recognized. ");
		printf("By default it is set to integer. ");
		printf("This value will be identified in each load and store. \n");
		return integer_type_node;
	}
}


/* Function to add a varblock access 
	(using the one-dimension version of varblock). 
	
	* varblock_element is D.xxx6, the value from varblock[i][j], ready to use.
	* varblock_pointer is D.xxx5, the reference node necessary to write into a varblock position

	In order to write a varblock position, after calling this function it is necessary to add
	the mem_ref using 'build_simple_mem_ref(*varblock_pointer)'. 
	Example:	*varblock_pointer = something.
	

	To obtain the element of an multidimensional array, we have to transform the access:
	   varblock[i][j] = * (varblock_1d + (i* SIZE_DIM_J) + j )
	   varblock[1][3] = * (varblock_1d + (1 * max_upper_limit+8) + 3)
	
	   In GIMPLE, we express this statement as:
	
	   D.xxx1 = i * SIZE_DIM_J;
	   D.xxx2 = j + D.xxx1;
	   D.xxx3 = 4 * D.xxx2;            // 4 is the size of an integer.
	   D.xxx4 = (unsigned int) D.xxx3; // Without this, it does not work.
	   D.xxx5 = varblock1d + D.xxx4;
	   D.xxx6 = * D.xxx5;
	
	   Now, we can use D.xxx6 in other statement, or as parameter in a function.
*/
static gimple_seq
varblock_access (tree * varblock_element, tree * varblock_pointer,
									int index_i, tree index_j)
{
	gimple_seq varblock_seq;
	varblock_seq = gimple_seq_alloc ();

	/* * (varblock + (i*(max_upper_limit+8)) + j ) */
	
	/* D.xxx1 = i * SIZE_DIM_J;  */
	tree dxxx1 = build_decl (UNKNOWN_LOCATION, VAR_DECL, NULL, integer_type_node);
	tree tmp_expr_i = build2 (MULT_EXPR, integer_type_node, 
														build_int_cst (integer_type_node, index_i),
														build_int_cst (integer_type_node, max_upper_limit + 8));
	gimple g_tmp_expr_i = gimple_build_assign	(dxxx1, tmp_expr_i);
	//gsi_insert_after (ptr_gsi, g_tmp_expr_i, GSI_NEW_STMT);
	gimple_seq_add_stmt (&varblock_seq, g_tmp_expr_i);
 	
	/* D.xxx2 = j + D.xxx1;  */
	tree dxxx2 = build_decl (UNKNOWN_LOCATION, VAR_DECL, NULL, integer_type_node);
	tree tmp_expr_j = build2 (PLUS_EXPR, integer_type_node,
														index_j, dxxx1);
	gimple g_tmp_expr_j = gimple_build_assign	(dxxx2, tmp_expr_j);
	//gsi_insert_after (ptr_gsi, g_tmp_expr_j, GSI_NEW_STMT);
	gimple_seq_add_stmt (&varblock_seq, g_tmp_expr_j);

	/* D.xxx3 = 4 * D.xxx2;  */
	tree dxxx3 = build_decl (UNKNOWN_LOCATION, VAR_DECL, NULL, integer_type_node);
	tree tmp_mult4 = build2 (MULT_EXPR, integer_type_node, 
													build_int_cst (integer_type_node, 4),
													dxxx2);
	gimple g_tmp_mult4 = gimple_build_assign (dxxx3, tmp_mult4);
	//gsi_insert_after (ptr_gsi, g_tmp_mult4, GSI_NEW_STMT);
	gimple_seq_add_stmt (&varblock_seq, g_tmp_mult4);
	

	/* D.xxx4 = (unsigned int) D.xxx3;  */
	tree tmp_size = fold_convert (sizetype, dxxx3);
	tree dxxx4 = build_decl (UNKNOWN_LOCATION, VAR_DECL, NULL, sizetype);
	gimple g_tmp_size = gimple_build_assign (dxxx4, tmp_size);
	//gsi_insert_after(ptr_gsi, g_tmp_size, GSI_NEW_STMT);
	gimple_seq_add_stmt (&varblock_seq, g_tmp_size);

	/* D.xxx5 = varblock1d + Dxxx4;  */
	tree dxxx5 = build_decl (UNKNOWN_LOCATION, VAR_DECL, NULL, integer_ptr_type_node);
	tree tmp_varblock = fold_build2 (POINTER_PLUS_EXPR, integer_ptr_type_node,
															varblock1d,
															dxxx4
//														size_int (4) // We already multiply by 4 before.
															);
	gimple g_tmp_varblock = gimple_build_assign	(dxxx5, tmp_varblock);
	//gsi_insert_after (ptr_gsi, g_tmp_varblock, GSI_NEW_STMT);
	gimple_seq_add_stmt (&varblock_seq, g_tmp_varblock);
	
	/* varblock_element = * D.xxx5;  */
	tree tmp_varblock_ptr = build_simple_mem_ref (dxxx5);
	*varblock_pointer = dxxx5;

	gimple varblock1d_tmp = gimple_build_assign(*varblock_element, tmp_varblock_ptr);
	
	//gsi_insert_after (ptr_gsi, varblock1d_tmp, GSI_NEW_STMT);
	gimple_seq_add_stmt (&varblock_seq, varblock1d_tmp);

	//debug_gimple_seq (varblock_seq);
	//gsi_insert_seq_after (ptr_gsi, varblock_seq, GSI_CONTINUE_LINKING);
	//gsi_insert_seq_after (ptr_gsi, varblock_seq, GSI_NEW_STMT);
	
	//return dxxx5;
	return varblock_seq;
}


/* Function to add a variable as OpenMP clause of a particular kind:
OMP_CLAUSE_PRIVATE or OMP_CLAUSE_SHARED  .*/

static void
add_omp_clause(tree var, enum omp_clause_code clause_kind)
{	
	/* Create a new clause  */
	tree clause = build_omp_clause (gimple_location(omp_for_loop), clause_kind);
	/* Add the variable to the clause  */
	OMP_CLAUSE_DECL (clause) = var;	
	/* Add this clause to the chain of clauses  */
	OMP_CLAUSE_CHAIN (clause) = gimple_omp_parallel_clauses (omp_for_loop);
	/* Finally, set the new clause to the pragma  */
	gimple_omp_parallel_set_clauses (omp_for_loop, clause);
}


/* Function to create variables shared for different blocks in the code.  */
static void
initialize_variables(void)
{
	
	/* Creating artificial variables to use in 'self-assignments' 
	before a function call:

		current.0 = current;
		function (current.0);
	
	*/
	current_t_alone = build_decl (gimple_location(omp_for_loop), VAR_DECL,
																		get_identifier ("current.0"),
																		integer_type_node);
	/* Add variable as OMP clause.  */
	add_omp_clause(current_t_alone, OMP_CLAUSE_PRIVATE);
	
	DECL_COMMON (current_t_alone) = 1;
	TREE_USED (current_t_alone) = 1;
	DECL_IGNORED_P (current_t_alone) = 1;
	
	/***********/
	/* current */
	/***********/
	current_t = build_decl (gimple_location(omp_for_loop), VAR_DECL,
														get_identifier ("current"),
														integer_type_node);
	/* Add variable as OMP clause.  */
	add_omp_clause(current_t, OMP_CLAUSE_PRIVATE);
	
	DECL_COMMON (current_t) = 1;
	TREE_USED (current_t) = 1;
	TREE_PUBLIC (current_t) = 1;
	
	//layout_decl(current_t, false);
	//rest_of_decl_compilation(current_t, 1, 0);
	
	/* These options are mandatory. If not -> ERROR at using '&current'.  */
	TREE_STATIC (current_t) = 1;
	TREE_ADDRESSABLE (current_t) = 1;
	
	/* &current */
	/* TODO: See what the difference is.  */
	current_t_addr = build_fold_addr_expr (current_t);
	//current_t_addr = build1 (ADDR_EXPR, integer_type_node, current_t);


	/****************/
	/*		value			*/
	/****************/
	printf("Number of speculative variables = %d\n", spec_variables_num);
	
	tree value_type = integer_type_node; //Default type.
	const char * type_name;
	tree spec_variable_type;
	
	/* Depending on the type of each speculative variable, we create a different 
	'value' variable, and this is saved in the global variable 'value_vars'.  */
	int i;
	for (i = 0; i < spec_variables_num; i++) {
		
		spec_variable_type = TREE_TYPE (spec_variables [i]);
		while (TREE_CODE (spec_variable_type) == ARRAY_TYPE) {
			spec_variable_type = TREE_TYPE (spec_variable_type);
		}
		//debug_tree (spec_variable_type);

		type_name = get_type_name (spec_variable_type);
		printf("\t Variable : %s. Type: %s.\n", spec_variables_name[i], type_name);	
		
		/* Set the correct type for the speculative variable.  */
		
		/* If we detect a pointer type, get the original type.  */
		if (TREE_CODE (spec_variable_type) == POINTER_TYPE) {
			printf("Pointer detected!\n");
			
			// Support for multiple pointers
			while (TREE_CODE (spec_variable_type) == POINTER_TYPE)
				spec_variable_type = TREE_TYPE (spec_variable_type);
		}	
		
		//value_type = set_value_type (type_name); //ERROR if we have a pointer.
		value_type = spec_variable_type;

		/* Set a different name for each 'value' variable.  */
		char * name;
		int ss = asprintf(&name, "%s%d", "value_", i);

		/* Create the 'value' variable for each speculative variable.  */
		value_vars[i] = build_decl (gimple_location(omp_for_loop), VAR_DECL,
														get_identifier(name),
														value_type);
		add_omp_clause(value_vars[i], OMP_CLAUSE_PRIVATE);
		//DECL_INITIAL (value_vars[i]) = 0;
		DECL_COMMON (value_vars[i]) = 1;
		TREE_USED (value_vars[i]) = 1;
		TREE_PUBLIC (value_vars[i]) = 1;
	
		// This option is mandatory. If not -> ERROR at using '&value'.
		// However, this makes imposible to use 'value' in the IF condition.
		TREE_STATIC (value_vars[i]) = 1; 
		// This option also makes imposible to use 'value' in the IF condition.
		TREE_ADDRESSABLE (value_vars[i]) = 1;	
	
		layout_decl(value_vars[i], false);
		rest_of_decl_compilation(value_vars[i], 1, 0);

	}
	
	/*******************/
	/*	pre_if_var 		 */
	/*******************/
	/* TODO: Check if it is correct just to build the variable and
			not to add it as declaration.  */
	pre_if_var = build_decl (gimple_location(omp_for_loop), VAR_DECL, 
																		get_identifier ("pre_if_var"),
																		//NULL,
																		integer_type_node);
	DECL_IGNORED_P (pre_if_var) = 1;
	add_omp_clause (pre_if_var, OMP_CLAUSE_PRIVATE);
	
	/****************/
	/* labelSquash  */
	/****************/
	labelSquash_decl = create_artificial_label (UNKNOWN_LOCATION);
	labelSquash = gimple_build_label (labelSquash_decl);


	/****************/
	/* nIterations  */
	/****************/
	nIterations = build_decl (UNKNOWN_LOCATION, VAR_DECL, 
															get_identifier("nIterations"),
															integer_type_node);
	DECL_COMMON (nIterations) = 1;
	TREE_USED (nIterations) = 1;
	TREE_PUBLIC (nIterations) = 1;

}



/* Function to add the engine's private and shared vars as OpenMP clauses.  */

static void
add_engine_vars(gimple stmt, enum omp_clause_code clause_kind)
{
	int i;
	tree var;
	const char ** engine_vars; 		/* The name of the variables.  */
	const int * engine_vars_type;	/* The type of these variables.  */
	int engine_vars_length;				/* Number of variables, to control the loop.  */
	
	/* This function serves to add the private, and also the shared variables.  */
	switch(clause_kind)
	{
		case OMP_CLAUSE_PRIVATE:
			/* TO DO NOTHING: Currently there are not private variables to add in this point.
			engine_vars = engine_private_vars;
			engine_vars_length = sizeof(engine_private_vars)/sizeof(const char * const);
			engine_vars_type = engine_private_types;
			*/
			break;
		
		case OMP_CLAUSE_SHARED:
			engine_vars = engine_shared_vars;
			engine_vars_length = sizeof(engine_shared_vars)/sizeof(const char * const);
			engine_vars_type = engine_shared_types;
			break;
		
		default: return;
	}
	
	/* Put a new clause por each private or shared variable.  */
	for (i = 0; i < engine_vars_length; i++) 
	{	
		/* Create the new var, with its corresponding type.  */
		/* FIXME: Using trees instead of chars avoids to do this trick.  */
		switch(engine_vars_type[i])
		{
			case INTEGER_TYPE:
				var = build_decl (gimple_location(stmt), VAR_DECL,
														get_identifier(engine_vars[i]),
														integer_type_node);
				break;
			case CHAR_TYPE:
				var = build_decl (gimple_location(stmt), VAR_DECL,
														get_identifier(engine_vars[i]),
														char_type_node);
				break;
			default:
				var = build_decl (gimple_location(stmt), VAR_DECL,
														get_identifier(engine_vars[i]),
														integer_type_node);
		}
	
		/* Create a new clause  */
		add_omp_clause(var, clause_kind);
	}
}


/* Function to extract the speculative variables defined by the clause. 
	
		It stores speculative variables in two global variables:

		- const char * spec_variables_name[]: Name of each speculative variable.
		- tree spec_variables[]: Tree object representing each speculative variable.  */

static void
extract_speculative_clauses (tree clauses)
{
	tree c; /* For clauses iteration.  */
	int i=0;
	
	/* TODO: Clean the global variables for each FOR loop.  */

	/* Traversing each clause in the pragma.  */
	for (c = clauses; c; c = OMP_CLAUSE_CHAIN(c))
	{
		if ( OMP_CLAUSE_CODE(c) == OMP_CLAUSE_SPECULATIVE )
		{
			//printf("Extracting speculative clause: %s\n", var_name_in_clause(c));
			spec_variables_name[i] = (char *) var_name_in_clause(c);
			spec_variables[i] = OMP_CLAUSE_DECL(c);
			i++;
			spec_variables_num = i;
		}
	}

}


/* Helper function to 'walk_tree()' called to search for speculative readings 
or writings of the speculative variable.  */
static tree
search_spec (tree *tp, int *ws, void *data) {
	
	tree t = *tp;
	int i;

	/* We search for the variable reference, instead for the variable, because we want
	global variables, not local.  */
	for ( i = 0; i < spec_variables_num; i++)
	{
		if (operand_equal_p (t, spec_variables[i], OEP_PURE_SAME)) {
			printf("\tSpeculative variable detected: '%s'\n", IDENTIFIER_POINTER(DECL_NAME(t)));
			return t;
		}
	}
	return NULL_TREE;

}

/***********************************************************************************************/
/* TODO: CURRENTLY UNUSED
Helper function to 'walk_gimple_seq()' called to detect the use of speculative variables
in prototyped functions.  */
static tree
search_spec_gimple (gimple_stmt_iterator *ptr_gsi, bool *handled_ops_p,
		 struct walk_stmt_info *wi)
{
	return NULL_TREE;
}

/* TODO: CURRENTLY UNUSED
Helper function to 'walk_gimple_seq()' called to detect the use of speculative variables
in prototyped functions.  */
static tree
search_spec_gimple_ops(tree *tp, int *walk_subtrees, void *data)
{
	tree t = *tp;
	int i;
	
	/* We search for the variable reference, instead for the variable, because we want
	global variables, not local.  */
	for ( i = 0; i < spec_variables_num; i++)
	{
		if (operand_equal_p (t, spec_variables[i], OEP_PURE_SAME)) {
			printf("\tSpeculative variable detected: '%s'\n", IDENTIFIER_POINTER(DECL_NAME(t)));
			//debug_tree (t);
			return t;
		}
	}
	return NULL_TREE;
}
/***********************************************************************************************/


/* Function that adds the functions specinit at the beginning of the code.  */

static void
insert_specinit ()
{
	tree funcdecl;
	/****************************************************
	* Inserting specinit() in the beginning of the code
	*****************************************************/
	/* Build the function prototype:
	*			int * specinit()
	*/
	//tree specinit_type = build_function_type_list (void_type_node, 
	tree specinit_type = build_function_type_list (integer_ptr_type_node, 
																			void_type_node, NULL_TREE);
	
	/* Create the function declaration.  */
	funcdecl = build_fn_decl ("specinit", specinit_type);
	
	/* Create the variable to gather the blocks from varblock.  */
	/* FIXME: If there are several loops, this line produces an error, because
		the variable is declared in the location of one single loop. If the variable
	is used in another loop, it results in an error, probably in 
	verify_gimple_assign (from tree-cfg.c).  
	
	SOLUTIONS: 
		* Create a different variable for each FOR loop.
		* Create variables correctly, setting the scope and handle the cfg.
	*/
	
	/* XXX: If the variables has not got the context of the loop cannot be used in
	the parallelized loop.  */
	varblock1d = build_decl (UNKNOWN_LOCATION, VAR_DECL, 
	//varblock1d = build_decl (gimple_location(omp_for_loop), VAR_DECL, 
														 get_identifier("varblock1d"), integer_ptr_type_node);
	DECL_COMMON (varblock1d) = 1;
	TREE_USED (varblock1d) = 1;
	TREE_PUBLIC (varblock1d) = 1;
	//TREE_STATIC (varblock1d) = 1; 
	//TREE_ADDRESSABLE (varblock1d) = 1;	
	layout_decl(varblock1d, false);
 	rest_of_decl_compilation(varblock1d, 1, 0);
	
	/* TODO: Remove this, because specinit() is added before the loop is detected.  */
	//add_omp_clause (varblock1d, OMP_CLAUSE_SHARED);

	/* Build the GIMPLE_CALL and insert it.  */
	gimple specinit_call = gimple_build_call (funcdecl, 0);
	gimple_call_set_lhs (specinit_call, varblock1d);
	gsi_insert_before(&gsi_init, specinit_call, GSI_SAME_STMT);

	/*-----------------------------------------------------------------------------*/
	/* FIXME: We have to create a global variable in order to get this working well.  */
	/* Only call specinit() if it is not called before. Therefore, CALL IT ONCE.  
		bool specinit_called = false //GLOBAL VARIABLE
		if (!specinit_called) {
			specinit();
			specinit_called = true;
		}	
	*/
/*	
	tree specinit_called = build_decl (UNKNOWN_LOCATION, VAR_DECL, 
																						get_identifier("specinit_called"),
																						boolean_type_node);
	// Build the if condition and the labels.
	tree label_true_decl = create_artificial_label (UNKNOWN_LOCATION);
	gimple label_true = gimple_build_label (label_true_decl);
	
	tree label_false_decl = create_artificial_label (UNKNOWN_LOCATION);
	gimple label_false = gimple_build_label (label_false_decl);
	
	gimple if_stmt = gimple_build_cond (NE_EXPR, specinit_called, 
																			build_int_cst (boolean_type_node, 1), //true == 1
																			label_true_decl, label_false_decl);
	gimple set_to_called = gimple_build_assign (specinit_called, build_int_cst (boolean_type_node, 1));
	
	//Insert if (specinit_called != true).
	gsi_insert_before(&gsi_init, if_stmt, GSI_NEW_STMT);
	
	//Insert label_true.
	gsi_insert_after (&gsi_init, label_true, GSI_NEW_STMT);
	// Insert specinit();.
	gsi_insert_after (&gsi_init, specinit_call, GSI_NEW_STMT);
	// Insert specinit_called = true.
	gsi_insert_after (&gsi_init, set_to_called, GSI_NEW_STMT);

	// Insert label_false.
	gsi_insert_after (&gsi_init, label_false, GSI_NEW_STMT);
*/
	/*-----------------------------------------------------------------------------*/
	
	/* Malloc for varblock1d */
	//int* varblock1d = (int*) malloc(sizeof(int) * 3*(max_upper_limit+8));
	//tree varblock_malloc = built_in_decls[BUILT_IN_MALLOC];
	//gimple g_varblock_malloc = gimple_build_call (varblock_malloc, 1, 
	//															build_int_cst(integer_type_node, 4*3*(max_upper_limit+8))); 
	
	/* varblock1d = malloc  */
	//gimple_call_set_lhs (g_varblock_malloc, varblock1d);
	
	//gsi_insert_before (&gsi_init, g_varblock_malloc, GSI_SAME_STMT);

}


/* Function that 'specend()' at the end of the FOR loop.  */
static void
insert_specend(gimple_stmt_iterator * ptr_gsi)
{
	tree funcdecl;
	
	//printf("Inserting 'specend'\n");
	//gimple stmt = gsi_stmt (*ptr_gsi);
	//debug_gimple_stmt (stmt);
	
	gimple_seq seq = gimple_body(current_function_decl);
	gimple_stmt_iterator gsi = gsi_last_main(seq);

	/****************************************************
	* Inserting specend() after the FOR loop
	*****************************************************/
	/* Build the function prototype:
	*			void specend()
	*/
	tree specend_type = build_function_type_list (void_type_node, 
																			void_type_node, NULL_TREE);
	
	/* Create the function declaration.  */
	funcdecl = build_fn_decl ("specend", specend_type);

	/* Build the GIMPLE_CALL and insert it.  */
	gimple specend_call = gimple_build_call (funcdecl, 0);
	
	// 'specend()' inserted just after the loop parallelized. Problem with nested loops.
	//gsi_insert_after(ptr_gsi, specend_call, GSI_SAME_STMT);
	
	// 'specend()' inserted before the last stmt - a return.
	//gsi_insert_before(&gsi, specend_call, GSI_SAME_STMT);

};


/* Function that inserts the engine's functions needed to
*  execute speculatively a loop.  */

static void
insert_engine_functions (gimple_stmt_iterator * ptr_gsi, int nIt)
{
	tree funcdecl;

	//TODO: Remove this because it is now implemented in other part.
	/*****************************************
	* Inserting omp_set_num_threads(threads)
	*****************************************/
	/* Set the number of threads */
	tree nthreads = build_int_cst (integer_type_node, threads);
	
	/* Build the function prototype:
	* 		void omp_set_num_threads(int) 
	*/
	tree void_int_fn_type = build_function_type_list (void_type_node, 
														integer_type_node, NULL_TREE);
	
	/* Create the function declaration.  */
	funcdecl = build_fn_decl ("omp_set_num_threads", void_int_fn_type);

	/* Build the GIMPLE_CALL and insert it.  */
	gimple omp_set_call = gimple_build_call (funcdecl, 1, nthreads);
	gsi_insert_before (ptr_gsi, omp_set_call, GSI_SAME_STMT);

	

	/*****************************
	* Inserting specstart(nIt)  
	******************************/
	
	/* Inserting statements to set the correct number of iterations.  */
	gsi_insert_before (ptr_gsi, nIter_ini, GSI_SAME_STMT);
	gsi_insert_before (ptr_gsi, nIter_1, GSI_SAME_STMT);
	
	/* Build the function prototype:
	*			void specstart (int)  
	*/
	tree specstart_type = build_function_type_list (void_type_node,
															integer_type_node, NULL_TREE);

	/* Create the function declaration.  */
	funcdecl = build_fn_decl ("specbegin", specstart_type);

	/* Build the GIMPLE_CALL and insert it.  */
	gimple specstart_call = gimple_build_call (funcdecl, 1, build_int_cst(integer_type_node, nIt));
	//gimple specstart_call = gimple_build_call (funcdecl, 1, nIterations);
	gsi_insert_before(ptr_gsi, specstart_call, GSI_SAME_STMT);
	//insert_printf(ptr_gsi, "Valor de nIterations = %d\n", nIterations);

}



/* Function to replace a reading from an speculative variable for a 
*  speculative load operation.  */

static void
replace_speculative_load (gimple_stmt_iterator *ptr_gsi, tree lhs, tree variable, gimple stmt, int pos)
{	

	tree specload, earlySquash;
	tree size;
	tree int_ptr_variable;
	tree label_false_decl;
	int i;
	tree * value_ptr;
	
	bool new_value_created = false; //Boolean variable to control when we create a new 'value' variable.
	tree new_value;

	/* Create the type (unsigned char *).  */
	tree unsigned_char_ptr_type_node = build_pointer_type (unsigned_char_type_node);
	
	/* Create the parameters:
		- current
		- value 
		- sizeof(variable)
		- (unsigned char*) &value
		- (unsigned char*) &variable
	*/

	/***********/
	/* current */
	/***********/
	/* It is already created at the beginning of the plugin as global variable.  */
	
	/*********/
	/* value */
	/*********/
	/* It is already created at the beginning of the plugin as global variable.  
	NOW, we select the correct 'value' variable depending on the type of the 
	speculative variable detected.  */
	
	
	/* If we find that the speculative variable is a COMPONENT_REF (Structs), 
	create a new 'value' with the correct type.  */
	/* TODO: What's happen with address reference, etc.  */
	

	tree var_s = variable;
	while (TREE_OPERAND_LENGTH (var_s) > 0) { 
		var_s = TREE_OPERAND(var_s, 0);
	}
	
	if (TREE_CODE(TREE_TYPE (var_s)) == RECORD_TYPE) {
		for (i = 0; i < spec_variables_num; i++)
			if (operand_equal_p (var_s, spec_variables[i], OEP_PURE_SAME)) {
				
				new_value_created = true;
				
				/* Create a new 'value' variable for each load.  */
				char * name;
				int ss = asprintf(&name, "%s%d", "value_struct_", value_struct_num);
				value_struct_num++;
				
				tree value_type = TREE_TYPE (variable);
				tree new_value = build_decl (gimple_location(omp_for_loop), VAR_DECL,
														get_identifier(name),
														value_type);
				add_omp_clause (new_value, OMP_CLAUSE_PRIVATE);
				DECL_COMMON (new_value) = 1;
				TREE_USED (new_value) = 1;
				TREE_PUBLIC (new_value) = 1;
	
				// This option is mandatory. If not -> ERROR at using '&value'.
				// However, this makes imposible to use 'value' in the IF condition.
				TREE_STATIC (new_value) = 1; 
				// This option also makes imposible to use 'value' in the IF condition.
				TREE_ADDRESSABLE (new_value) = 1;	
	
				layout_decl (new_value, false);
				rest_of_decl_compilation (new_value, 1, 0);
				
				value_ptr = &new_value; // Set the new 'value' for this spec-load
		}
	}
	
	/* If we have a MEM_REF, create a new 'value'.  */
	if (TREE_CODE(variable) == MEM_REF) {
		new_value_created = true;
				
		/* Create a new 'value' variable.  */
		tree value_type = TREE_TYPE (variable);
		tree new_value = build_decl (gimple_location(omp_for_loop), VAR_DECL,
													get_identifier("value_mem"),
													value_type);
		add_omp_clause (new_value, OMP_CLAUSE_PRIVATE);
		DECL_COMMON (new_value) = 1;
		TREE_USED (new_value) = 1;
		TREE_PUBLIC (new_value) = 1;
	
		// This option is mandatory. If not -> ERROR at using '&value'.
		// However, this makes imposible to use 'value' in the IF condition.
		TREE_STATIC (new_value) = 1; 
		// This option also makes imposible to use 'value' in the IF condition.
		TREE_ADDRESSABLE (new_value) = 1;	
	
		layout_decl (new_value, false);
		rest_of_decl_compilation (new_value, 1, 0);
				
		value_ptr = &new_value; // Set the new 'value' for this spec-load

	}

	/* If there is not a new 'value' created, use one of the previously created.  */
	if (!new_value_created) {
		tree var = variable;
		
		if (TREE_CODE (variable) == ARRAY_REF) {
			 var = TREE_OPERAND(variable, 0);
			
			/* An ARRAY_REF could have more than one dimension. We recursively resolve the
				several dimensions until obtaining the VAR_DECL that represents the variable.  */
			while (TREE_CODE(var) != VAR_DECL)
				var = TREE_OPERAND (var, 0);
		}	
	
		/* Check each speculative variable and select the correct 'value' variable.  */
		for (i = 0; i < spec_variables_num; i++) {
			if (operand_equal_p (var, spec_variables[i], OEP_PURE_SAME)) {
				value_ptr = &value_vars[i];
			}
		}
	} // if new_value_created


	/********************/
	/* sizeof(variable)	*/
	/********************/
	/* Get the size of the variable for the second argument of specload_pointer.  */
	/* TODO: Distinguish between arrays and single variables.  */
	/* In arrays we search for the size of an array element. 'variable' is an 'array_ref' */
	tree variable_element_size = TYPE_SIZE_UNIT( TREE_TYPE (variable));
	
	/*********************************/
	/* (unsigned char *) &variable.  */
	/*********************************/
	/* To apply the operator '&', 'variable has to be the bit ADDRESSABLE activated. 
	If the variable is GLOBAL, the bit STATIC is set up, and if the bit ADDRESSABLE is not also set up, 
	we cannot use the address operation.

	If the variable is LOCAL, the bit STATIC is not set up, and GCC allows to use the '&' operator 
	without more transformations.
	
	Therefore: 
	 If a variable is LOCAL do not activate the bit ADDRESSABLE. 
	 If a variable is GLOBAL (TREE_STATIC) activate the bit ADDRESSABLE.  */ 
	
	if ( 	TREE_CODE(variable) == VAR_DECL && 
				TREE_STATIC (variable) == 1 &&
				TREE_ADDRESSABLE(variable)==0 ) 
	{
		TREE_ADDRESSABLE (variable) = 1;
	}
	
	else if ( TREE_CODE(variable) == ARRAY_REF ) 
	{
		/* An ARRAY_REF could have more than one dimension. We recursively resolve the
		several dimensions until obtaining the VAR_DECL that represents the variable.  */
		tree variable_array = TREE_OPERAND (variable, 0);
		
		while (TREE_CODE(variable_array) != VAR_DECL)
			variable_array = TREE_OPERAND (variable_array, 0);
		
		if ( TREE_STATIC (variable_array) == 1 &&
				 TREE_ADDRESSABLE (variable_array) == 0 )
		{
			TREE_ADDRESSABLE (variable_array) = 1; 
		}
	}

	/* TODO: Check this alternative to the above lines.  Add the ARRAY_REF distinction.  */
	//tree variable_new = copy_node (variable);
	//TREE_ADDRESSABLE (variable_new) = 1;


	/* This eliminates support for pointers.  */
	/* If the variables is not recognized as VAR_DECL, or ARRAY_REF, exit.  */
	/*if ((TREE_CODE(variable) != VAR_DECL && TREE_CODE(variable) != ARRAY_REF))
	{
		printf("Speculative variable is not an ARRAY_REF or a VAR_DECL\n");
		debug_tree(variable);
		exit(-1);
	}*/
	
	tree variable_addr = build_fold_addr_expr (variable);
	tree pre_unsigned_char_ptr_var = fold_convert (unsigned_char_ptr_type_node, variable_addr);
	
	/* TODO: Check if this patch is correct. It is only necessary when the speculative variable is 
	* a global variable. Other solution is to not doing the fold_convert, and directly 
	* build the assignment with 'variable_addr' .  */ 
	/* This macro is necessary for global variables. Local variables are not affected.  */
	STRIP_USELESS_TYPE_CONVERSION (pre_unsigned_char_ptr_var);

	tree unsigned_char_ptr_var = build_decl (gimple_location(omp_for_loop), VAR_DECL, 
																								//get_identifier("specload_address"),
																								NULL, 
																								unsigned_char_ptr_type_node);
	TREE_USED (unsigned_char_ptr_var) = 1;
	DECL_IGNORED_P (unsigned_char_ptr_var) = 1;
	//add_omp_clause (unsigned_char_ptr_var, OMP_CLAUSE_PRIVATE);

	gimple aux_stmt = gimple_build_assign (unsigned_char_ptr_var, pre_unsigned_char_ptr_var);
	gsi_insert_before (ptr_gsi, aux_stmt, GSI_SAME_STMT);

	
	/****************************/
	/* (unsigned char *) &value */
	/****************************/
	tree value_addr = build_fold_addr_expr (*value_ptr);
	tree pre_unsigned_char_ptr_var2 = fold_convert (unsigned_char_ptr_type_node, value_addr);
	tree unsigned_char_ptr_var2 = build_decl (gimple_location(omp_for_loop), VAR_DECL, 
																								//get_identifier("specload_value"),
																								NULL, 
																								unsigned_char_ptr_type_node);
	TREE_USED (unsigned_char_ptr_var2) = 1;
	DECL_IGNORED_P (unsigned_char_ptr_var2) = 1;
	//add_omp_clause (unsigned_char_ptr_var2, OMP_CLAUSE_PRIVATE); 
	
	gimple aux_stmt2 = gimple_build_assign (unsigned_char_ptr_var2, pre_unsigned_char_ptr_var2);
	gsi_insert_before (ptr_gsi, aux_stmt2, GSI_SAME_STMT);



	/* Build the function prototype for specload:
	* 		int specload_pointer (unsigned char *, sizeof, int, unsigned char *)    */
	tree specload_pointer_type = build_function_type_list (integer_type_node,
																unsigned_char_ptr_type_node,
																integer_type_node,
																integer_type_node,
																unsigned_char_ptr_type_node,
																NULL_TREE);

	/* Create the function declaration, the function call and a temporal variable:
		
			D.xxxx = specload_pointer(..);
			if ( D.xxxx == -1 )
	*/
	tree specload_pointer = build_fn_decl ("specload_pointer", specload_pointer_type);
	TREE_ADDRESSABLE (specload_pointer) = 1;
	TREE_USED (specload_pointer) = 1;
	TREE_PUBLIC (specload_pointer) = 1;
	TREE_STATIC (specload_pointer) = 1;
	TREE_NOTHROW (specload_pointer) = 0;

	gimple specload_pointer_call = gimple_build_call (specload_pointer, 4, 
																	unsigned_char_ptr_var,
																	variable_element_size,
																	current_t_alone,
																	unsigned_char_ptr_var2);

	
	/* Insert 'current.x = current' before the function call.  */
	//gimple current_assign0 = gimple_build_assign (current_t_alone, current_t);
	//gsi_insert_before (ptr_gsi, current_assign0, GSI_SAME_STMT);
	/* D.xxxx = specload_pointer(...);.  */
	gimple_call_set_lhs (specload_pointer_call, pre_if_var);
	gsi_insert_before (ptr_gsi, specload_pointer_call, GSI_SAME_STMT);
	
	
	/* Now, create and insert the if condition.  
			D.xxxx = specload_pointer();
			if ( D.xxxx == -1 )
				goto labelSquash; 
			lhs = value	
	*/
	
	/* Build the if condition and the labels.  */
	
	label_false_decl = create_artificial_label (UNKNOWN_LOCATION);
	gimple label_false = gimple_build_label (label_false_decl);
	
	gimple if_stmt = gimple_build_cond (EQ_EXPR, pre_if_var, 
																			build_int_cst (integer_type_node, -1),
																			labelSquash_decl, label_false_decl);
	
	/* Insert if (D.xxxx == -1).  */
	gsi_insert_before(ptr_gsi, if_stmt, GSI_SAME_STMT);
	
	/* Insert label_false.  */
	gsi_insert_before (ptr_gsi, label_false, GSI_SAME_STMT);
	
	/* Modify the original line with the speculative load: LHS = value + X;  */
	// Before that, create a variable, because value is ADDRESSABLE:
	tree pre_value_type = TREE_TYPE (variable);
	tree pre_value = build_decl (gimple_location(omp_for_loop), VAR_DECL,
														NULL, 
														pre_value_type);
	add_omp_clause(pre_value, OMP_CLAUSE_PRIVATE);
	TREE_USED (pre_value) = 1;
	DECL_IGNORED_P (pre_value) = 1;
	
	//	D.xxxx = value_ptr;
	gimple g_pre_value = gimple_build_assign (pre_value, *value_ptr);
	gsi_insert_before (ptr_gsi, g_pre_value, GSI_SAME_STMT);

	//	LHS = D.xxxx + X;
	gimple_set_op (stmt, pos, pre_value);
	
}


/* Function to replace a writing into an speculative variable for a 
*  speculative store operation.  */

static void
replace_speculative_store (gimple_stmt_iterator *ptr_gsi, tree variable, tree rhs, gimple stmt)
{
	tree specstore_pointer, funcdecl;
	int i;
	
	tree * value_ptr;
	bool new_value_created = false; //Boolean variable to control when we create a new 'value' variable.
	tree new_value;
	
	/* Create some of the parameters:
		- current
		- value   
		- sizeof(variable)
		- (unsigned char *) &value
		- (unsigned char *) &variable
	*/

	
	/***********/
	/* current */
	/***********/
	/* It is already created at the beginning of the plugin as global variable.  */

	/*********/
	/* value */
	/*********/
	/* It is already created at the beginning of the plugin as global variable.  
		NOW, we select the correct 'value' variable depending on the type of the 
	speculative variable detected.  */
		
	/* If we find that the speculative variable is a COMPONENT_REF (Structs), 
	create a new 'value' with the correct type.  */
	/* TODO: What's happen with address reference, etc.  */
	
	tree var_s = variable;
	while (TREE_OPERAND_LENGTH (var_s) > 0) { 
		var_s = TREE_OPERAND(var_s, 0);
	}
	
	if (TREE_CODE(TREE_TYPE (var_s)) == RECORD_TYPE) {
		for (i = 0; i < spec_variables_num; i++)
			if (operand_equal_p (var_s, spec_variables[i], OEP_PURE_SAME)) {
				
				new_value_created = true;
				
				/* Create a new 'value' variable for each store.  */
				char * name;
				int ss = asprintf(&name, "%s%d", "value_struct_", value_struct_num);
				value_struct_num++;

				tree value_type = TREE_TYPE (variable);
				tree new_value = build_decl (gimple_location(omp_for_loop), VAR_DECL,
														get_identifier(name),
														value_type);
				add_omp_clause (new_value, OMP_CLAUSE_PRIVATE);
				DECL_COMMON (new_value) = 1;
				TREE_USED (new_value) = 1;
				TREE_PUBLIC (new_value) = 1;
	
				// This option is mandatory. If not -> ERROR at using '&value'.
				// However, this makes imposible to use 'value' in the IF condition.
				TREE_STATIC (new_value) = 1; 
				// This option also makes imposible to use 'value' in the IF condition.
				TREE_ADDRESSABLE (new_value) = 1;	
	
				layout_decl (new_value, false);
				rest_of_decl_compilation (new_value, 1, 0);
				
				value_ptr = &new_value; // Set the new 'value' for this spec-load
		}
	}
	
	/* If we have a MEM_REF, create a new 'value'.  */
	if (TREE_CODE(variable) == MEM_REF) {
		new_value_created = true;
				
		/* Create a new 'value' variable.  */
		tree value_type = TREE_TYPE (variable);
		tree new_value = build_decl (gimple_location(omp_for_loop), VAR_DECL,
													get_identifier("value_mem"),
													value_type);
		add_omp_clause (new_value, OMP_CLAUSE_PRIVATE);
		DECL_COMMON (new_value) = 1;
		TREE_USED (new_value) = 1;
		TREE_PUBLIC (new_value) = 1;
	
		// This option is mandatory. If not -> ERROR at using '&value'.
		// However, this makes imposible to use 'value' in the IF condition.
		TREE_STATIC (new_value) = 1; 
		// This option also makes imposible to use 'value' in the IF condition.
		TREE_ADDRESSABLE (new_value) = 1;	
	
		layout_decl (new_value, false);
		rest_of_decl_compilation (new_value, 1, 0);
				
		value_ptr = &new_value; // Set the new 'value' for this spec-load

	}

	/* If there is not a new 'value' created, use one of the previously created.  */
	if (!new_value_created) {
		tree var = variable;
		
		if (TREE_CODE (variable) == ARRAY_REF) {
			 var = TREE_OPERAND(variable, 0);
			
			/* An ARRAY_REF could have more than one dimension. We recursively resolve the
				several dimensions until obtaining the VAR_DECL that represents the variable.  */
			while (TREE_CODE(var) != VAR_DECL)
				var = TREE_OPERAND (var, 0);
		}	
	
		/* Check each speculative variable and select the correct 'value' variable.  */
		for (i = 0; i < spec_variables_num; i++) {
			if (operand_equal_p (var, spec_variables[i], OEP_PURE_SAME)) {
				value_ptr = &value_vars[i];
			}
		}
	} // if new_value_created

	
	/********************/
	/* sizeof(variable)	*/
	/********************/
	/* Get the size of the variable for the second argument of specload_pointer.  */
	/* TODO: Distinguish between arrays and single variables.  */
	/* In arrays we search for the size of an array element. 'variable' is an 'array_ref' */
	tree variable_element_size = TYPE_SIZE_UNIT( TREE_TYPE(variable));


	/*********************************/
	/* (unsigned char *) &variable.  */
	/*********************************/
	/* To apply the operator '&', 'variable has to be the bit ADDRESSABLE activated. 
	If the variable is GLOBAL, the bit STATIC is set up, and if the bit ADDRESSABLE is not also set up, 
	we cannot use the address operation.

	If the variable is LOCAL, the bit STATIC is not set up, and GCC allows to use the '&' operator 
	without more transformations.
	
	Therefore: 
	 If a variable is LOCAL do not activate the bit ADDRESSABLE. 
	 If a variable is GLOBAL (TREE_STATIC) activate the bit ADDRESSABLE.  */ 
	
	if ( 	TREE_CODE(variable) == VAR_DECL && 
				TREE_STATIC (variable) == 1 &&
				TREE_ADDRESSABLE(variable)==0 ) 
	{
		TREE_ADDRESSABLE (variable) = 1;
	}
	
	else if ( TREE_CODE(variable) == ARRAY_REF ) 
	{
		/* An ARRAY_REF could have more than one dimension. We recursively resolve the
		several dimensions until obtaining the VAR_DECL that represents the variable.  */
		tree variable_array = TREE_OPERAND (variable, 0);
		
		while (TREE_CODE(variable_array) != VAR_DECL)
			variable_array = TREE_OPERAND (variable_array, 0);
		
		if ( TREE_STATIC (variable_array) == 1 &&
				 TREE_ADDRESSABLE (variable_array) == 0 )
		{
			TREE_ADDRESSABLE (variable_array) = 1; 
		}
	}
	
	/* An alternative to the above lines. Add the ARRAY_REF distinction.  */
	//tree variable_new = copy_node (variable);
	//TREE_ADDRESSABLE (variable_new) = 1;
	
	/* This eliminates support for pointers.  */
	/* If the variables is not recognized as VAR_DECL, or ARRAY_REF, exit.  */
	/*if ((TREE_CODE(variable) != VAR_DECL && TREE_CODE(variable) != ARRAY_REF))
	{
		printf("Speculative variable is not an ARRAY_REF or a VAR_DECL\n");
		debug_tree(variable);
		exit(-1);
	}*/
	
	/* Create the type (unsigned char *).  */
	tree unsigned_char_ptr_type_node = build_pointer_type (unsigned_char_type_node);
	
	tree variable_addr = build_fold_addr_expr (variable);
	tree pre_unsigned_char_ptr_var = fold_convert (unsigned_char_ptr_type_node, variable_addr);
	
	/* TODO: Check if this patch is correct. It is only necessary when the speculative variable is 
	* a global variable. Other solution is to not doing the fold_convert, and directly 
	* build the assignment with 'variable_addr' .  */ 
	/* This macro is necessary for global variables. Local variables are not affected.  */
	STRIP_USELESS_TYPE_CONVERSION (pre_unsigned_char_ptr_var);

	tree unsigned_char_ptr_var = build_decl (gimple_location(omp_for_loop), VAR_DECL, 
																								//get_identifier ("specstore_address"),
																								NULL,
																								unsigned_char_ptr_type_node);
	TREE_USED (unsigned_char_ptr_var) = 1;
	DECL_IGNORED_P (unsigned_char_ptr_var) = 1;
	//add_omp_clause (unsigned_char_ptr_var, OMP_CLAUSE_PRIVATE);

	gimple aux_stmt = gimple_build_assign (unsigned_char_ptr_var, pre_unsigned_char_ptr_var);
	gsi_insert_before (ptr_gsi, aux_stmt, GSI_SAME_STMT);
	
	/****************************/
	/* (unsigned char *) &value */
	/****************************/
	tree value_addr = build_fold_addr_expr (*value_ptr);
	tree pre_unsigned_char_ptr_var2 = fold_convert (unsigned_char_ptr_type_node, value_addr);
	tree unsigned_char_ptr_var2 = build_decl (gimple_location(omp_for_loop), VAR_DECL, 
																								//get_identifier ("specstore_value"),
																								NULL, 
																								unsigned_char_ptr_type_node);
	TREE_USED (unsigned_char_ptr_var2) = 1;
	DECL_IGNORED_P (unsigned_char_ptr_var2) = 1;
	//add_omp_clause (unsigned_char_ptr_var2, OMP_CLAUSE_PRIVATE);
	
	gimple aux_stmt2 = gimple_build_assign (unsigned_char_ptr_var2, pre_unsigned_char_ptr_var2);
	gsi_insert_before (ptr_gsi, aux_stmt2, GSI_SAME_STMT);


	/* Assign RHS to 'value' variable.  */
	/* Create a new auxiliar variable as intermediate step to set the 'value':
			lhs_new = RHS;
			value = lhs_new;
		*/

	tree lhs_new_type = TREE_TYPE (*value_ptr);
	tree lhs_new = build_decl (gimple_location(omp_for_loop), VAR_DECL,
														NULL, 
														lhs_new_type);
	add_omp_clause(lhs_new, OMP_CLAUSE_PRIVATE);
	TREE_USED (lhs_new) = 1;
	DECL_IGNORED_P (lhs_new) = 1;
		
	/* Insert lhs.1 = RHS;.  */
	tree rhs_tree = gimple_assign_rhs_to_tree (stmt);
	gimple g_lhs_new = gimple_build_assign (lhs_new, rhs_tree);
	gsi_insert_before (ptr_gsi, g_lhs_new, GSI_SAME_STMT);
	
	/* Insert value = lhs.  */
	gimple g_lhs_new2 = gimple_build_assign (*value_ptr, lhs_new);
	gsi_insert_before (ptr_gsi, g_lhs_new2, GSI_SAME_STMT);
	

	/* Build the function prototype:
	* 		void specstore(unsigned char*, int, int, unsigned char*)    */
	tree specstore_pointer_type = build_function_type_list (void_type_node,
																	unsigned_char_ptr_type_node,
																	integer_type_node,
																	integer_type_node,
																	unsigned_char_ptr_type_node,
																	NULL_TREE);
																
	/* Create the function declaration and insert it */
	specstore_pointer = build_fn_decl ("specstore_pointer", specstore_pointer_type);
	gimple specstore_pointer_call = gimple_build_call (specstore_pointer, 4,
																			unsigned_char_ptr_var,
																			variable_element_size, 
																			current_t_alone,
																			unsigned_char_ptr_var2);
	
	/* Insert 'current.x = current' before the function call.  */
	//gimple current_assign0 = gimple_build_assign (current_t_alone, current_t);
	//gsi_insert_before (ptr_gsi, current_assign0, GSI_SAME_STMT);
	
	gsi_insert_after (ptr_gsi, specstore_pointer_call, GSI_SAME_STMT);
	
	/* Remove the original writing into the speculative variables.  */
	gsi_remove (ptr_gsi, true);

}


/* Function to replace a 'max reduction' for the speculative max reduction operation.
'variable' is the variable where we store the maximum.  */

static void
replace_speculative_max (gimple_stmt_iterator *ptr_gsi, tree variable, tree rhs, gimple stmt)
{
	printf("Replacing max reduction\n");

	tree specredmax, funcdecl;
	int i;
	
	tree * value_ptr;
	bool new_value_created = false; //Boolean variable to control when we create a new 'value' variable.
	tree new_value;
	

	/* First, remove the GIMPLE_COND.  */
	gimple_stmt_iterator gsi_copy;
	gsi_copy = * ptr_gsi;
	gimple if_stmt = gsi_stmt (gsi_copy); 

	// Find the GIMPLE_COND
	while (gimple_code(if_stmt) != GIMPLE_COND ) {
		gsi_prev(&gsi_copy);
		if_stmt = gsi_stmt (gsi_copy);
	}
	
	// Remove the GIMPLE_COND
	gsi_remove (&gsi_copy, true);
	// Remove the following GIMPLE_LABEL
	//if (gimple_code (gsi_stmt(gsi_copy)) == GIMPLE_LABEL) gsi_remove (&gsi_copy, true);


	/* Create some of the parameters:
		- current
		- value   
		- sizeof(variable)
		- (unsigned char *) &value
		- (unsigned char *) &variable
	*/

	
	/***********/
	/* current */
	/***********/
	/* It is already created at the beginning of the plugin as global variable.  */

	/*********/
	/* value */
	/*********/
	/* It is already created at the beginning of the plugin as global variable.  
		NOW, we select the correct 'value' variable depending on the type of the 
	speculative variable detected.  */
	
	/* TODO: At the moment, this function only works with VAR_DECL variables.  */
	
	
	tree var = variable;
	if (TREE_CODE (variable) == ARRAY_REF) {
		 var = TREE_OPERAND(variable, 0);
			
		/* An ARRAY_REF could have more than one dimension. We recursively resolve the
			several dimensions until obtaining the VAR_DECL that represents the variable.  */
		while (TREE_CODE(var) != VAR_DECL)
			var = TREE_OPERAND (var, 0);
	}	
	
	/* Check each speculative variable and select the correct 'value' variable.  */
	for (i = 0; i < spec_variables_num; i++) {
		if (operand_equal_p (var, spec_variables[i], OEP_PURE_SAME)) {
			value_ptr = &value_vars[i];
		}
	}

	/********************/
	/* sizeof(variable)	*/
	/********************/
	/* Get the size of the variable for the second argument of specload_pointer.  */
	/* TODO: Distinguish between arrays and single variables.  */
	/* In arrays we search for the size of an array element. 'variable' is an 'array_ref' */
	tree variable_element_size = TYPE_SIZE_UNIT( TREE_TYPE(variable));


	/*********************************/
	/* (unsigned char *) &variable.  */
	/*********************************/
	/* To apply the operator '&', 'variable has to be the bit ADDRESSABLE activated. 
	If the variable is GLOBAL, the bit STATIC is set up, and if the bit ADDRESSABLE is not also set up, 
	we cannot use the address operation.

	If the variable is LOCAL, the bit STATIC is not set up, and GCC allows to use the '&' operator 
	without more transformations.
	
	Therefore: 
	 If a variable is LOCAL do not activate the bit ADDRESSABLE. 
	 If a variable is GLOBAL (TREE_STATIC) activate the bit ADDRESSABLE.  */ 
	
	if ( 	TREE_CODE(variable) == VAR_DECL && 
				TREE_STATIC (variable) == 1 &&
				TREE_ADDRESSABLE(variable)==0 ) 
	{
		TREE_ADDRESSABLE (variable) = 1;
	}
	
	else if ( TREE_CODE(variable) == ARRAY_REF ) 
	{
		/* An ARRAY_REF could have more than one dimension. We recursively resolve the
		several dimensions until obtaining the VAR_DECL that represents the variable.  */
		tree variable_array = TREE_OPERAND (variable, 0);
		
		while (TREE_CODE(variable_array) != VAR_DECL)
			variable_array = TREE_OPERAND (variable_array, 0);
		
		if ( TREE_STATIC (variable_array) == 1 &&
				 TREE_ADDRESSABLE (variable_array) == 0 )
		{
			TREE_ADDRESSABLE (variable_array) = 1; 
		}
	}
	
	/* An alternative to the above lines. Add the ARRAY_REF distinction.  */
	//tree variable_new = copy_node (variable);
	//TREE_ADDRESSABLE (variable_new) = 1;
	
	/* This eliminates support for pointers.  */
	/* If the variables is not recognized as VAR_DECL, or ARRAY_REF, exit.  */
	/*if ((TREE_CODE(variable) != VAR_DECL && TREE_CODE(variable) != ARRAY_REF))
	{
		printf("Speculative variable is not an ARRAY_REF or a VAR_DECL\n");
		debug_tree(variable);
		exit(-1);
	}*/

	
	/* Create the type (unsigned char *).  */
	tree unsigned_char_ptr_type_node = build_pointer_type (unsigned_char_type_node);
	
	tree variable_addr = build_fold_addr_expr (variable);
	tree pre_unsigned_char_ptr_var = fold_convert (unsigned_char_ptr_type_node, variable_addr);
	
	/* TODO: Check if this patch is correct. It is only necessary when the speculative variable is 
	* a global variable. Other solution is to not doing the fold_convert, and directly 
	* build the assignment with 'variable_addr' .  */ 
	/* This macro is necessary for global variables. Local variables are not affected.  */
	STRIP_USELESS_TYPE_CONVERSION (pre_unsigned_char_ptr_var);

	tree unsigned_char_ptr_var = build_decl (gimple_location(omp_for_loop), VAR_DECL, 
																								//get_identifier ("specstore_address"),
																								NULL,
																								unsigned_char_ptr_type_node);
	TREE_USED (unsigned_char_ptr_var) = 1;
	DECL_IGNORED_P (unsigned_char_ptr_var) = 1;
	//add_omp_clause (unsigned_char_ptr_var, OMP_CLAUSE_PRIVATE);

	gimple aux_stmt = gimple_build_assign (unsigned_char_ptr_var, pre_unsigned_char_ptr_var);
	gsi_insert_before (ptr_gsi, aux_stmt, GSI_SAME_STMT);


	/****************************/
	/* (unsigned char *) &value */
	/****************************/
	tree value_addr = build_fold_addr_expr (*value_ptr);
	tree pre_unsigned_char_ptr_var2 = fold_convert (unsigned_char_ptr_type_node, value_addr);
	tree unsigned_char_ptr_var2 = build_decl (gimple_location(omp_for_loop), VAR_DECL, 
																								//get_identifier ("specstore_value"),
																								NULL, 
																								unsigned_char_ptr_type_node);
	TREE_USED (unsigned_char_ptr_var2) = 1;
	DECL_IGNORED_P (unsigned_char_ptr_var2) = 1;
	//add_omp_clause (unsigned_char_ptr_var2, OMP_CLAUSE_PRIVATE);
	
	gimple aux_stmt2 = gimple_build_assign (unsigned_char_ptr_var2, pre_unsigned_char_ptr_var2);
	gsi_insert_before (ptr_gsi, aux_stmt2, GSI_SAME_STMT);

	/* Assign RHS to 'value' variable.  */
	/* Create a new auxiliar variable as intermediate step to set the 'value':
			lhs_new = RHS;
			value = lhs_new;
		*/

	tree lhs_new_type = TREE_TYPE (*value_ptr);
	tree lhs_new = build_decl (gimple_location(omp_for_loop), VAR_DECL,
														NULL, 
														lhs_new_type);
	add_omp_clause(lhs_new, OMP_CLAUSE_PRIVATE);
	TREE_USED (lhs_new) = 1;
	DECL_IGNORED_P (lhs_new) = 1;
		
	/* Insert lhs.1 = RHS;.  */
	tree rhs_tree = gimple_assign_rhs_to_tree (stmt);
	gimple g_lhs_new = gimple_build_assign (lhs_new, rhs_tree);
	gsi_insert_before (ptr_gsi, g_lhs_new, GSI_SAME_STMT);
	
	/* Insert value = lhs.  */
	gimple g_lhs_new2 = gimple_build_assign (*value_ptr, lhs_new);
	gsi_insert_before (ptr_gsi, g_lhs_new2, GSI_SAME_STMT);
	

	/* Build the function prototype:
	* 		void specstore(unsigned char*, int, int, unsigned char*)    */
	tree specredmax_type = build_function_type_list (void_type_node,
																	unsigned_char_ptr_type_node,
																	integer_type_node,
																	integer_type_node,
																	unsigned_char_ptr_type_node,
																	NULL_TREE);
																
	/* Create the function declaration and insert it */
	specredmax = build_fn_decl ("specredmax", specredmax_type);
	gimple specredmax_call = gimple_build_call (specredmax, 4,
																			unsigned_char_ptr_var,
																			variable_element_size, 
																			current_t_alone,
																			unsigned_char_ptr_var2);
	
	/* Insert 'current.x = current' before the function call.  */
	//gimple current_assign0 = gimple_build_assign (current_t_alone, current_t);
	//gsi_insert_before (ptr_gsi, current_assign0, GSI_SAME_STMT);
	
	gsi_insert_after (ptr_gsi, specredmax_call, GSI_SAME_STMT);
	
	/* Remove the original writing into the speculative variable.  */
	gsi_remove (ptr_gsi, true);



}





/* Function that replaces the old FOR loop for the new one. 

	This function changes the index, initial and final value, the step, 
	and also adds new statements before the body of the old FOR loop.  
	
	PRE: *ptr_gsi points to a GIMPLE_OMP_FOR statement.  */

static void
replace_for_loop (gimple_stmt_iterator *ptr_gsi)
{

	int i, step = 0;

	gimple stmt = gsi_stmt (*ptr_gsi);
	gimple_seq for_body	= gimple_omp_body(stmt); /* Get the body of the FOR loop.  */
	

	printf("Replacing original FOR loop for a speculative version...\n\n");
	//debug_gimple_stmt (stmt);
	//debug_gimple_seq(for_body);
	
	/* Get the pointer to the first statement of the FOR loop.  */
	gimple_stmt_iterator gsi_first_stmt = gsi_start (for_body);

	/* Get the pointer to the last statement of the FOR loop.  */
	gimple_stmt_iterator gsi_last_stmt = gsi_last (for_body);


	/* Creating some parameters and variables:
		- current
		- retflag
		- endLoop
		- endReturn
		- varblock[][]
	*/


	/***********/
	/* current */
	/***********/
	/* It is already created at the beginning of the plugin as global variable.  */
	

	/******************/
	/* 		retflag			*/
	/******************/
	tree retflag = build_decl (gimple_location(omp_for_loop), VAR_DECL,
															get_identifier("retflag"),
															integer_type_node);
	
	add_omp_clause(retflag, OMP_CLAUSE_PRIVATE);

	DECL_COMMON (retflag) = 1;
	TREE_USED (retflag) = 1;
	TREE_PUBLIC (retflag) = 1;

	/********************/
	/* 		endLoop				*/
	/********************/
	/* TODO: Remove every single reference to endLoop and its statements.  */
	tree endLoop = build_decl (UNKNOWN_LOCATION, VAR_DECL,
															get_identifier ("endLoop"),
															integer_type_node);
	
	DECL_COMMON (endLoop) = 1;
	TREE_USED (endLoop) = 1;
	TREE_PUBLIC (endLoop) = 1;
	
	/********************/
	/* 		endReturn			*/
	/********************/
	/* TODO: Remove every single reference to endReturn and its statements.  */
	tree endReturn = build_decl (UNKNOWN_LOCATION, VAR_DECL,
															get_identifier ("endReturn"),
															integer_type_node);
	
	DECL_COMMON (endReturn) = 1;
	TREE_USED (endReturn) = 1;
	TREE_PUBLIC (endReturn) = 1;
	
	
	/****************/
	/* varblock[][]	*/
	/****************/
	/* Definition of varblock: #define varblock[3][max_upper_limit + 8]. 
	
		We implement accesses to varblock[][] as accesses to an 1-dimensional 
		array varblock1d[], returned by specinit() function:  
		array2d[i][j] = *(array1d + (i * SIZE_DIM_J) + j )
	*/

	/* TODO: REMOVE THIS CODE WHEN IT IS SURE THAT WE ARE NOT NEED TO USE THIS.  */
	/* To implement a reading and a writing into this variable we
	implement a couple a functions:
			
			int varblock_reading (i,j)
			void varblock_writing (i,j,value)
	
	*/
	
	/* Build the function prototype for varblock_reading(): 
	* 		int varblock_reading (int, int).  */
	tree varblock_reading_type = build_function_type_list (integer_type_node,
																		integer_type_node,
																		integer_type_node, NULL_TREE);

	/* Create the function declaration.  */
	tree varblock_reading = build_fn_decl ("varblock_reading", varblock_reading_type);
	TREE_ADDRESSABLE (varblock_reading) = 1;
	TREE_USED (varblock_reading) = 1;
	TREE_PUBLIC (varblock_reading) = 1;
	TREE_STATIC (varblock_reading) = 1;
	TREE_NOTHROW (varblock_reading) = 0;
	

	/* Build the function prototype for varblock_writing():
	*			void varblock_writing (int, int, value).  */
	tree varblock_writing_type = build_function_type_list (void_type_node, 
																	integer_type_node,
																	integer_type_node,
																	integer_type_node,
																	NULL_TREE);

	/* Create the function declaration.  */
	tree varblock_writing = build_fn_decl ("varblock_writing", varblock_writing_type);
	TREE_ADDRESSABLE (varblock_writing) = 1;
	TREE_USED (varblock_writing) = 1;
	TREE_PUBLIC (varblock_writing) = 1;
	TREE_STATIC (varblock_writing) = 1;
	TREE_NOTHROW (varblock_writing) = 0;


	/**************************/
	/* Creating real labels.  */
	/**************************/

	/* labelStartIteration.  */
	tree labelStartIteration_decl = create_artificial_label (UNKNOWN_LOCATION);
	gimple labelStartIteration = gimple_build_label (labelStartIteration_decl);
			
	/* labelEndIteration.  */
	tree labelEndIteration_decl = create_artificial_label (UNKNOWN_LOCATION);
	gimple labelEndIteration = gimple_build_label (labelEndIteration_decl);

	/* labelEndLoop.  */
	tree labelEndLoop_decl = create_artificial_label (UNKNOWN_LOCATION);
	gimple labelEndLoop = gimple_build_label (labelEndLoop_decl);
	
	/* labelSquash.  */
	// Created as global variable because it is used from 'replace_speculative_load'
	//tree labelSquash_decl = create_artificial_label (UNKNOWN_LOCATION);
	//gimple labelSquash = gimple_build_label (labelSquash_decl);


	/**************************************************/
	/* Creating artificial labels for if statements.  */
	/**************************************************/
		
	/* label_false_upper_limit.  */
	tree label_false_upper_limit_decl = create_artificial_label (UNKNOWN_LOCATION);
	gimple label_false_upper_limit = gimple_build_label (label_false_upper_limit_decl);

	/* label_true_endLoop.  */
	tree label_true_endLoop_decl = create_artificial_label (UNKNOWN_LOCATION);
	gimple label_true_endLoop = gimple_build_label (label_true_endLoop_decl);

	/* label_endReturn.  */
	tree label_endReturn_decl = create_artificial_label (UNKNOWN_LOCATION);
	gimple label_endReturn = gimple_build_label (label_endReturn_decl);
	
	/* label_true_iteration_1.  */
	tree label_true_iteration_1_decl = create_artificial_label (UNKNOWN_LOCATION);
	gimple label_true_iteration_1 = gimple_build_label (label_true_iteration_1_decl);
	/* label_true_iteration_2.  */
	tree label_true_iteration_2_decl = create_artificial_label (UNKNOWN_LOCATION);
	gimple label_true_iteration_2 = gimple_build_label (label_true_iteration_2_decl);

	/* label_false_JOBDONE.  */
	tree label_false_JOBDONE_decl = create_artificial_label (UNKNOWN_LOCATION); 
	gimple label_false_JOBDONE =gimple_build_label (label_false_JOBDONE_decl); 

	/* label_true_JOBTODO.  */
	tree label_true_JOBTODO_decl = create_artificial_label (UNKNOWN_LOCATION); 
	gimple label_true_JOBTODO = gimple_build_label (label_true_JOBTODO_decl); 



	/* End of declarations and variables.  */
	/*************************************************/


	/* Get the number of iterations to pass */
	for (i = 0; i < gimple_omp_for_collapse (stmt); i++)
	{
		/* Get the variable involved in the loop control.  */
		tree index = gimple_omp_for_index	(stmt,i);
		/* Get the initial and the final value of the variable involved in the loop control.  */
		tree initial = gimple_omp_for_initial (stmt,i);
		tree final = gimple_omp_for_final (stmt,i);

		/* Get the increment expression to calculate the step.  */
		tree incr = gimple_omp_for_incr (stmt,i);
		tree step_tree = TREE_OPERAND (incr, 1);
		
		/* Get the operator in the final condition.  */
		enum tree_code final_condition = gimple_omp_for_cond (stmt, i);

		/* TODO: Remove this code. It only works if the final expression is an "integer_cst", 
		not if the expression is like "j < (aux - 1)" or simply "j < aux".  */
		
		/* Check the increment expression and set the number of iterations.  */
		/*if (TREE_CODE(incr) == PLUS_EXPR)
		{
			tree step_tree = TREE_OPERAND(incr,1);
			step = TREE_INT_CST_LOW (step_tree);
			nIt = (TREE_INT_CST_LOW (final) - TREE_INT_CST_LOW (initial) + 1) / step;

			//printf ("This loop has a step of: %d\n", step);
			//printf ("Number of loop iterations: %d\n", nIt);		
		}	
		else if (TREE_CODE(incr) == MINUS_EXPR)
		{
			tree step_tree = TREE_OPERAND(incr,1);
			step = TREE_INT_CST_LOW (step_tree);
			nIt = (TREE_INT_CST_LOW (initial) - TREE_INT_CST_LOW (final) + 1) / step;

			//printf ("This loop has a step of: %d\n", step);
			//printf ("Number of loop iterations: %d\n", nIt);
		}
		else {
			// Set the number of iterations to nIt.
			nIt = TREE_INT_CST_LOW (final) - TREE_INT_CST_LOW (initial);
			//printf ("Number of loop iterations: %d\n", nIt);
		}
		*/		


		/********************************/
		/*     REPLACING THE FOR LOOP.  */
		/********************************/
		
		/* Change the FOR loop by this: 
					
				for ( tid = 0; tid <= threads - 1; tid++) .  */

		/* Set the new variable involved in the loop control.  */
		/* new_index  = tid.  */
		tree new_index = build_decl (gimple_location(omp_for_loop), VAR_DECL, 
																	get_identifier ("tid"),
																	integer_type_node);
													
		add_omp_clause (new_index, OMP_CLAUSE_PRIVATE);

		DECL_COMMON (new_index) = 1;
		TREE_USED (new_index) = 1;
		TREE_PUBLIC (new_index) = 1;
		
		gimple_omp_for_set_index (stmt, i, new_index);

		/* Set the new initial value: 

				tid = 0
		*/
		tree new_initial = build_int_cst (integer_type_node, 0);
		gimple_omp_for_set_initial (stmt, i, new_initial);
			
		/* Set the new final value:
				
			tid <= threads - 1 
		*/
		tree nthreads = build_int_cst (integer_type_node, threads-1);
		gimple_omp_for_set_final (stmt, i, nthreads);
			
		/* Set the new step.  */
		tree tid_update = build2(PLUS_EXPR, integer_type_node, 
															new_index, 
															build_int_cst (integer_type_node, 1));
		
		gimple_omp_for_set_incr (stmt, i, tid_update);
			
		/* Set the new condition.  */
		gimple_omp_for_set_cond (stmt, i, LE_EXPR);


		/*****************************************************************/
		/* ADDING NEW STATEMENTS BEFORE THE BODY OF THE 'OLD' FOR LOOP: 
					
				ini = 'initial value of the original loop';
				current = tid;
				iteration = varblock[0][current] + ini; 
				
				if (iteration > upper_limit - 1 + ini) //upper_limit = NITER from specstart(NITER)
					goto labelEndLoop;

				varblock[2][current] = varblock[2][current] + 1;
					
				labelStartIteration: 
					
				--> Old Loop code
					
		******************************************************************/
		
		/***********************************************/	
		/* ini = 'initial value of the original loop'; */
		/***********************************************/
		tree ini = build_decl (gimple_location(omp_for_loop), VAR_DECL, 
														get_identifier ("ini"),
														integer_type_node);
		
		add_omp_clause(ini, OMP_CLAUSE_PRIVATE);

		//gimple ini_assign = gimple_build_assign (ini, build_int_cst (integer_type_node, 0));
		gimple ini_assign = gimple_build_assign (ini, initial);
		

		/************************************************/
		/* current = tid;																*/
		/************************************************/
		//gimple current_tid = gimple_build_assign (current_t_alone, new_index);
		gimple current_tid = gimple_build_assign (current_t, new_index);


		/******************************************
			iteration = varblock[0][current] + ini;

			D.xxxx = varblock[0][current];
			iteration = D.xxxx + ini;
		*******************************************/
		
		/* D.xxxx = varblock[0][current].  */
		tree pre_varblock1d_0 = build_decl (gimple_location(omp_for_loop), VAR_DECL,
																//get_identifier ("pre_varblock1d_0"),
																NULL,
																integer_type_node);
		DECL_IGNORED_P (pre_varblock1d_0) = 0;
		//add_omp_clause (pre_varblock1d_0, OMP_CLAUSE_PRIVATE);														
		
		tree write_varblock_0;
		gimple_seq varblock_seq_0;
		gimple varblock_reading_call_0;
		
		/* Implemented as readings from and writings into 1-dimensional array.  */
		if (!use_varblock_function) {
			/* Get the sequence of statements that we need to add to access to varblock[0][current]
				in the 1-dimension version (varblock_1d).  */
			varblock_seq_0 = varblock_access (&pre_varblock1d_0, &write_varblock_0,
																							0, current_t_alone);

		/* Implemented as a function call.  */																							
		} else {
			varblock_reading_call_0 = gimple_build_call (varblock_reading, 2,
																				build_int_cst (integer_type_node, 0), current_t_alone);
			gimple_call_set_lhs (varblock_reading_call_0, pre_varblock1d_0);
		}
		
		/* iteration = D.xxxx + ini.  */
		tree tmp_expr = build2 (PLUS_EXPR, integer_type_node, 
														pre_varblock1d_0, ini);
		gimple varblock_0 = gimple_build_assign (index, tmp_expr);


		/****************************************************************** 
			niter = NITER;  // It could be an expression as (j-1) (NITER == upper_limit)
			niter_1 = niter - 1;
			niter_ini = niter_1 + ini
				if (iteration > niter_ini) {
					goto labelEndLoop;
				}
		******************************************************************/
		tree niter = build_decl (gimple_location (omp_for_loop), VAR_DECL, 
															get_identifier("niter"),
															integer_type_node);
		DECL_IGNORED_P (niter) = 1;
		add_omp_clause (niter, OMP_CLAUSE_PRIVATE);			
 		
		tree niter_1 = build_decl (gimple_location (omp_for_loop), VAR_DECL, 
															NULL,
															integer_type_node);
		DECL_IGNORED_P (niter_1) = 1;
		add_omp_clause (niter_1, OMP_CLAUSE_PRIVATE);	
		
		tree niter_minus_1 = build2 (MINUS_EXPR, integer_type_node, 
														niter, build_int_cst(integer_type_node, 1));
		
		tree niter_ini = build_decl (gimple_location (omp_for_loop), VAR_DECL, 
															NULL,
															integer_type_node);
		DECL_IGNORED_P (niter_ini) = 1;
		add_omp_clause (niter_ini, OMP_CLAUSE_PRIVATE);	

		tree niter_plus_ini = build2 (PLUS_EXPR, integer_type_node,
														niter_1, ini);
		
		/* Before using specstart_iterations, we need to added to the loop 
		as a SHARED variable ONLY IF IT IS A ARTIFICIAL VARIABLE. Otherwise, OpenMP already add
		the final condition of the loop.  
			int L = NITER;                   int L = NITER + 1;
			specstart(L);                    specstart (L-1);	 // This creates an artificial variable before the stmt.
			for (i = 0; i < L; i++)          for (i = 0; i < L-1; i++).  */
		if (TREE_CODE (specstart_iterations) == VAR_DECL && DECL_NAME(specstart_iterations) == NULL) {
			add_omp_clause (specstart_iterations, OMP_CLAUSE_SHARED); // Only with artificial because this stmt
																																// fails when iterations is JUST a VAR_DECL 
																																// (and not an expression).
		}
		else if (TREE_CODE (specstart_iterations) == VAR_DECL && DECL_NAME(specstart_iterations) != NULL); //Do nothing
		else if (TREE_CODE (specstart_iterations) == PARM_DECL); //Do nothing, because if we do as with a VAR_DECL, it fails.
		else if (TREE_CODE (specstart_iterations) == INTEGER_CST); //Do nothing
		else { // Possibility of another kind of code not considered.
			printf("TREE_CODE not recognized in the parameter passed to 'specstart()'\n");
			debug_tree (specstart_iterations);
			exit(-1);
		}	
		
		gimple g_niter = gimple_build_assign (niter, specstart_iterations);
		gimple g_niter_1 = gimple_build_assign (niter_1, niter_minus_1);
		gimple g_niter_ini = gimple_build_assign (niter_ini, niter_plus_ini);
		
		/* if (iteration > niter_1) goto labelEndLoop;  */
		gimple if_upper_limit = gimple_build_cond (GT_EXPR, index, niter_ini,
																		labelEndLoop_decl, label_false_upper_limit_decl);
		
		/****************************************************
			varblock[2][current] = varblock[2][current] + 1;

			D.xxxx = varblock[2][current];
			D.xxxx = D.xxxx + 1;
			varblock[2][current] = D.xxxx;
		*****************************************************/
		
		tree pre_varblock_2 = build_decl (gimple_location(omp_for_loop), VAR_DECL, 
																		//get_identifier ("pre_varblock_2"),
																		NULL,
																		integer_type_node);
		DECL_IGNORED_P (pre_varblock_2) = 1;																		
		//add_omp_clause (pre_varblock_2, OMP_CLAUSE_PRIVATE);

		/* Common variables.  */
		tree tmp_expr_2;
		gimple varblock_2;
		
		/* Variables for the 1-dimensional array option.  */
		gimple varblock_2b;
		gimple_seq varblock_seq_2;
		tree tmp_expr_2_plus;
		
		/* Variables for the 'function call' option.  */
		gimple varblock_reading_call_2;
		gimple varblock_writing_call_2;
		
		/* Implemented as readings from and writings into 1-dimensional array.  */
		if (!use_varblock_function) {
			/* pre_varblock_2 = varblock[2][current].  */
			
			/* Get the sequence of statements that we need to add to access to varblock[2][current]
				in the 1-dimension version (varblock_1d). 
				In write_varblock_2 it is returned the node that points to the 
				original 2-dimensional varblock position
			*/
			tree write_varblock_2;
			varblock_seq_2 = varblock_access (&pre_varblock_2, &write_varblock_2,
																							2, current_t_alone);

			/* pre_varblock_2_JOBTODO_plus = pre_varblock_2_JOBTODO + 1.  */
			tree tmp_expr_2_plus = build_decl (gimple_location (omp_for_loop), VAR_DECL, 
																			//get_identifier ("pre_varblock1d_2_JOBTODO_plus"),
																			NULL,
																			integer_type_node);
			DECL_IGNORED_P (tmp_expr_2_plus) = 1;
			//add_omp_clause(tmp_expr_2_plus, OMP_CLAUSE_PRIVATE);

			tmp_expr_2 = build2 (PLUS_EXPR, integer_type_node, 
														pre_varblock_2, build_int_cst(integer_type_node, 1));

			varblock_2b = gimple_build_assign (tmp_expr_2_plus, tmp_expr_2);

			/* varblock[2][current] = pre_varblock_2_JOBTODO_plus.  */
			varblock_2 = gimple_build_assign ( build_simple_mem_ref(write_varblock_2), 
																											tmp_expr_2_plus);

		/* Implemented as a function call.  */
		} else {
			/* varblock[2][current].  */
			varblock_reading_call_2 = gimple_build_call (varblock_reading, 2, 
																							build_int_cst (integer_type_node, 2),
																							current_t_alone);
			/* D.xxx2 = varblock[2][current].  */
													
			gimple_call_set_lhs (varblock_reading_call_2, pre_varblock_2);
		
			/* D.xxx3 = D.xxx2 + 1.  */
			tmp_expr_2 = build2 (PLUS_EXPR, integer_type_node, 
														pre_varblock_2, build_int_cst(integer_type_node, 1));

			varblock_2 = gimple_build_assign (pre_varblock_2, tmp_expr_2);
		
			/* varblock[2][current] = D.xxx3.  */
			varblock_writing_call_2 = gimple_build_call (varblock_writing, 3,
																		build_int_cst (integer_type_node, 2), 
																		current_t_alone, 
																		pre_varblock_2);
		}


		
		/*************************************
				END OF THE STATEMENTS PRE-LOOP
		*************************************/
		

		/********************************** 
					INSERTING CODE PRE-LOOP
		***********************************/
		
		/* ini = 'initial value of the original loop'.  */
		gsi_insert_before (&gsi_first_stmt, ini_assign, GSI_NEW_STMT);
		
		/* current = tid.  */
		gsi_insert_after (&gsi_first_stmt, current_tid, GSI_NEW_STMT);
		
		/* Insert 'current.x = current' before reading from varblock[x][current].  */
		gimple current_assign_tid = gimple_build_assign (current_t_alone, current_t);
		gsi_insert_after(&gsi_first_stmt, current_assign_tid, GSI_NEW_STMT);	
		
		/* D.xxxx = varblock[0][current]; 
			iteration = D.xxxx + ini;  */
		if (!use_varblock_function) {
			gsi_insert_seq_after (&gsi_first_stmt, varblock_seq_0, GSI_CONTINUE_LINKING);
			gsi_insert_after (&gsi_first_stmt, varblock_0, GSI_NEW_STMT);
		} else {
			gsi_insert_after (&gsi_first_stmt, varblock_reading_call_0, GSI_NEW_STMT);
			gsi_insert_after (&gsi_first_stmt, varblock_0, GSI_NEW_STMT);
		}
		
		/*	niter = NITER;  // It could be an expression as (j-1) (NITER == upper_limit)
				niter_1 = niter - 1;
				if (iteration > niter_1) {
					goto labelEndLoop;
				}  */
		gsi_insert_after (&gsi_first_stmt, g_niter, GSI_NEW_STMT);
		gsi_insert_after (&gsi_first_stmt, g_niter_1, GSI_NEW_STMT);
		gsi_insert_after (&gsi_first_stmt, g_niter_ini, GSI_NEW_STMT);
		gsi_insert_after (&gsi_first_stmt, if_upper_limit, GSI_NEW_STMT);
		/* label_false from IF.  */
		gsi_insert_after (&gsi_first_stmt, label_false_upper_limit, GSI_NEW_STMT);
		
		/* Insert 'current.x = current' before reading from varblock[x][current].  */
		//gimple current_assign0 = gimple_build_assign (current_t_alone, current_t);
		//gsi_insert_after(&gsi_first_stmt, current_assign0, GSI_NEW_STMT);
		
		/*	D.xxxx = varblock[2][current];
			D.xxxx = D.xxxx + 1;
			varblock[2][current] = D.xxxx;  */
		if (!use_varblock_function) {
			gsi_insert_seq_after (&gsi_first_stmt, varblock_seq_2, GSI_CONTINUE_LINKING);
			gsi_insert_after (&gsi_first_stmt, varblock_2b , GSI_NEW_STMT);
			gsi_insert_after (&gsi_first_stmt, varblock_2 , GSI_NEW_STMT);
		} else {
			//insert_printf(&gsi_first_stmt, "Current = %d\n", current_t_alone);
			gsi_insert_after (&gsi_first_stmt, varblock_reading_call_2, GSI_NEW_STMT);
			gsi_insert_after (&gsi_first_stmt, varblock_2 , GSI_NEW_STMT);
			gsi_insert_after (&gsi_first_stmt, varblock_writing_call_2, GSI_NEW_STMT);
		}
	
		/* DEBUG: CHECK IF varblock[2][current] is really modified.  */
		//tree arg_ii = build_int_cst (integer_type_node, 2);
		//gimple varblock_reading_call_2 = gimple_build_call (varblock_reading, 2, 
		//																arg_ii, current_t_alone);
		//gsi_insert_after (&gsi_first_stmt, varblock_reading_call_2, GSI_NEW_STMT);

		/* label labelStartIteration.  */
		gsi_insert_after (&gsi_first_stmt, labelStartIteration, GSI_NEW_STMT);


		/*****************************************************************/
		/* ADDING NEW STATEMENTS AFTER THE BODY OF THE 'OLD' FOR LOOP: 

		--> old loop code	
		
				if (endLoop || endReturn) {
					freeWheel(current);
					goto labelEndLoop;
				}

		labelEndIteration:
				
				if ( (iteration != varblock[1][current] + ini) 
							&& (iteration < NITER - 1) ) {
						
						iteration = iteration + 1;
						goto labelStartIteration;

				}
		
		labelSquash:
				
				retflag = threadend_pointer (&current);
				
				if (retflag == JOBDONE) goto labelEndLoop;
				if (retflag == JOBTODO) {

					iteration = varblock[0][current] + ini;
					varblock[2][current] = varblock[2][current] + 1;
					goto labelStartIteration;

				}

		labelEndLoop:
				
				; 
					
		******************************************************************/
		
		/*****************************
			if (endLoop || endReturn) {
					freeWheel(current);
					goto labelEndLoop;
				}
		******************************/		
		
		/* TODO: Currently, it is not used ''endLoop'' and ''endReturn''.  */
		
		/* Build the function prototype for freeWheel(): 
		* 	void freeWheel (int).  */
		//tree freeWheel_type = build_function_type_list (void_type_node,
		//																			integer_type_node, NULL_TREE);

		/* Create the function declaration and the function call.  */
		//tree freeWheel = build_fn_decl ("freeWheel", freeWheel_type);
		//TREE_ADDRESSABLE (freeWheel) = 1;
		//TREE_USED (freeWheel) = 1;
		//TREE_PUBLIC (freeWheel) = 1;
		//TREE_STATIC (freeWheel) = 1;
		//TREE_NOTHROW (freeWheel) = 0;
	
		//gimple freeWheel_call = gimple_build_call (freeWheel, 1, current_t_alone);
		
		/* goto labelEndLoop.  */
		//gimple goto_labelEndLoop = gimple_build_goto (labelEndLoop_decl);


		/* The IF-statement is represented in GIMPLE as following: 		
	
			if ( endLoop != 0 )
						goto label_true_endLoop;
			 		else
			 			goto label_endReturn;
			label_endReturn:
					if (endReturn != 0)
						got label_true_endLoop;
					else
						goto labelEndIteration;
			
			label_true_endLoop:
					freeWheel(current);
					goto labelEndLoop;
		*/
			
		//gimple if_endLoop = gimple_build_cond (NE_EXPR, endLoop, 
		//																	build_int_cst (integer_type_node, 0),
		//																	label_true_endLoop_decl, label_endReturn_decl);
		
		//gimple if_endReturn = gimple_build_cond (NE_EXPR, endReturn,
		//																	build_int_cst (integer_type_node, 0),
		//																	label_true_endLoop_decl, labelEndIteration_decl);
		

		/****************************************************
				if ( (iteration != varblock[1][current] + ini) 
							&& (iteration < NITER - 1) ) {
						
						iteration = iteration + 1;
						goto labelStartIteration;

				}
			labelSquash:	


		is represented in GIMPLE as:


				D.xxx1 = varblock[1][current];
				D.xxx2 = D.xxx1 + ini;

				if ( iteration != D.xxx2 ) 
					goto label_true_iteration_1;
				else
					goto labelSquash;

			label_true_iteration_1:
				for_ITERS = NITER - 1
				if ( iteration < for_ITERS )
					goto label_true_iteration_2;
				else
					goto labelSquash;

			label_true_iteration_2:
				iteration = iteration + 1;
				goto labelStartIteration;

			labelSquash:
		******************************************************/
		
		/* D.xxx1 = varblock[1][current].  */
		tree pre_varblock_1 = build_decl (gimple_location(omp_for_loop), VAR_DECL,
																//get_identifier ("pre_varblock_1"),
																NULL,
																integer_type_node);
		DECL_IGNORED_P (pre_varblock_1) = 1;														
		//add_omp_clause (pre_varblock_1, OMP_CLAUSE_PRIVATE);														
		
		tree write_varblock_1;
		gimple_seq varblock_seq_1;
		gimple varblock_reading_call_1;
	
		/* Implemented as readings from and writings into 1-dimensional array.  */
		if (!use_varblock_function) {
			/* Get the sequence of statements that we need to add to access to varblock[0][current]
				in the 1-dimension version (varblock_1d).  */
			varblock_seq_1 = varblock_access (&pre_varblock_1, &write_varblock_1,
																							1, current_t_alone);
			
		/* Implemented as a function call.  */
		} else {
			varblock_reading_call_1 = gimple_build_call (varblock_reading, 2,
																		build_int_cst (integer_type_node, 1), current_t_alone);
			gimple_call_set_lhs (varblock_reading_call_1, pre_varblock_1);
		}

		/* D.xxx2 = D.xxx1 + ini.  */
		tree pre_varblock_1_ini = build_decl (gimple_location(omp_for_loop), VAR_DECL,
													//get_identifier ("pre_varblock_1_ini"),
													NULL,
													integer_type_node);
		DECL_IGNORED_P (pre_varblock_1_ini) = 1;											
		//add_omp_clause (pre_varblock_1_ini, OMP_CLAUSE_PRIVATE);
		
		tree pre_varblock_1_expr = build2 (PLUS_EXPR, integer_type_node, 
																				pre_varblock_1, ini);

		gimple varblock_1 = gimple_build_assign (pre_varblock_1_ini, pre_varblock_1_expr);



		/* IF-Statement_1.  */
		gimple if_endIteration_1 = gimple_build_cond (NE_EXPR, index, 
																			pre_varblock_1_ini,
																			label_true_iteration_1_decl, labelSquash_decl);
		
		/* for_ITERS = NITER - 1 + ini.  */
		tree final_niter_1 = build_decl (gimple_location (omp_for_loop), VAR_DECL, 
															NULL,
															integer_type_node);
		DECL_IGNORED_P (final_niter_1) = 1;
		add_omp_clause (final_niter_1, OMP_CLAUSE_PRIVATE);	
		
		tree final_niter_minus_1 = build2 (MINUS_EXPR, integer_type_node, 
														niter, build_int_cst(integer_type_node, 1));
		
		tree final_niter_ini = build_decl (gimple_location (omp_for_loop), VAR_DECL, 
															NULL,
															integer_type_node);
		DECL_IGNORED_P (final_niter_ini) = 1;
		add_omp_clause (final_niter_ini, OMP_CLAUSE_PRIVATE);	

		tree final_niter_plus_ini = build2 (PLUS_EXPR, integer_type_node,
														final_niter_1, ini);

		gimple g_final_niter_1 = gimple_build_assign (final_niter_1, final_niter_minus_1);
		gimple g_final_niter_ini = gimple_build_assign (final_niter_ini, final_niter_plus_ini);
		/**************************************************************************************/
		
		/* IF-Statement_2.  */
		gimple if_endIteration_2 = gimple_build_cond (LT_EXPR, index,
																			final_niter_ini,
																			label_true_iteration_2_decl, labelSquash_decl);
		
		/* iteration = iteration + 1.  */
		tree iteration_plus_one_expr = build2 (PLUS_EXPR, integer_type_node, 
																			//index, build_int_cst (integer_type_node, 1));
																			index, step_tree);
		gimple iteration_plus_one = gimple_build_assign (index, iteration_plus_one_expr);

		/* goto labelStartIteration.  */
		gimple goto_labelStartIteration = gimple_build_goto (labelStartIteration_decl);


		/***********************************************************
		labelSquash:
				
				retflag = threadend_pointer (&current);
				
				if (retflag == JOBDONE) goto labelEndLoop;
				if (retflag == JOBTODO) {

					iteration = varblock[0][current] + ini;
					varblock[2][current] = varblock[2][current] + 1;
					goto labelStartIteration;

				}

		is represented in GIMPLE AS		

				retflag = threadend_pointer (&current);

				if (retflag == JOBDONE) goto labelEndLoop; else goto label_false_JOBDONE;
		
		label_false_JOBDONE:
				if (retflag == JOBTODO) goto label_true_JOBTODO; else goto labelEndLoop;
		
		label_true_JOBTODO:
				D.xxx1 = varblock[0][current];
				iteration = D.xxx1 + ini;

				D.xxx2 = varblock[2][current];
				D.xxx3 = D.xxx2 + 1;
				varblock[2][current] = D.xxx3;

				goto labelStartIteration;

				
		labelEndLoop:

		***************************************************************/
	
		/* Build the function prototype for threadend_pointer(): 
		* 	int threadend_pointer (*int).  */
		tree threadend_pointer_type = build_function_type_list (integer_type_node,
																					integer_ptr_type_node, 
																					NULL_TREE);

		/* Create the function declaration and the function call.  */
		tree threadend_pointer = build_fn_decl ("threadend_pointer", threadend_pointer_type);
		TREE_ADDRESSABLE (threadend_pointer) = 1;
		TREE_USED (threadend_pointer) = 1;
		TREE_PUBLIC (threadend_pointer) = 1;
		TREE_STATIC (threadend_pointer) = 1;
		TREE_NOTHROW (threadend_pointer) = 0;
	
		gimple threadend_pointer_call = gimple_build_call (threadend_pointer, 1, current_t_addr);
	
		/* retflag = threadend_pointer (&current).  */
		gimple_call_set_lhs (threadend_pointer_call, retflag);

		
		/* if (retflag == JOBDONE) goto labelEndLoop; else goto label_false_JOBDONE.  */
		gimple if_JOBDONE = gimple_build_cond (EQ_EXPR, retflag, 
																			// JOBDONE = '0', and the char '0' is the integer 48. 
																			build_int_cst (integer_type_node, 48),
																			labelEndLoop_decl, label_false_JOBDONE_decl);

		/* if (retflag == JOBTODO) goto label_true_JOBTODO; else goto labelEndLoop.  */
		gimple if_JOBTODO = gimple_build_cond (EQ_EXPR, retflag,
																			// JOBTODO = '1', and the char '0' is the integer 49. 
																			build_int_cst (integer_type_node, 49),
																			label_true_JOBTODO_decl, labelEndLoop_decl);

		/*******************************	
		D.xxx1 = varblock[0][current];
				iteration = D.xxx1 + ini;  
		********************************/
		
		/* D.xxx1 = varblock[0][current].  */
		tree pre_varblock_0_JOBTODO = build_decl (gimple_location(omp_for_loop), VAR_DECL,
																//get_identifier ("pre_varblock_0_JOBTODO"),
																NULL,
																integer_type_node);
		DECL_IGNORED_P (pre_varblock_0_JOBTODO) = 1;														
		//add_omp_clause (pre_varblock_0_JOBTODO, OMP_CLAUSE_PRIVATE);														
		
		tree write_varblock_0_JOBTODO;
		gimple_seq varblock_seq_0_JOBTODO;
		gimple varblock_reading_call_0_JOBTODO;
		
		/* Implemented as readings from and writings into 1-dimensional array.  */
		if (!use_varblock_function) {
			/* Get the sequence of statements that we need to add to access to varblock[0][current]
				in the 1-dimension version (varblock_1d).  */
			varblock_seq_0_JOBTODO = varblock_access (&pre_varblock_0_JOBTODO, &write_varblock_0_JOBTODO,
																							0, current_t_alone);
		
		/* Implemented as a function call.  */																							
		} else {
			varblock_reading_call_0_JOBTODO = gimple_build_call (varblock_reading, 2,
																							build_int_cst (integer_type_node, 0), 
																							current_t_alone);
			gimple_call_set_lhs (varblock_reading_call_0_JOBTODO, pre_varblock_0_JOBTODO);
		}
		

		/* iteration = D.xxx1 + ini.  */
		tree tmp_expr_JOBTODO = build2 (PLUS_EXPR, integer_type_node, 
														pre_varblock_0_JOBTODO, ini);

		gimple varblock_0_JOBTODO = gimple_build_assign (index, tmp_expr_JOBTODO);



		/************************************
		D.xxx2 = varblock[2][current];
				D.xxx3 = D.xxx2 + 1;
				varblock[2][current] = D.xxx3;  
		*************************************/

		tree pre_varblock_2_JOBTODO = build_decl (gimple_location(omp_for_loop), VAR_DECL, 
																		//get_identifier ("pre_varblock_2_JOBTODO"),
																		NULL,
																		integer_type_node);
		DECL_IGNORED_P (pre_varblock_2_JOBTODO) = 1;																		
		//add_omp_clause (pre_varblock_2_JOBTODO, OMP_CLAUSE_PRIVATE);

		/* Common variables.  */
		tree tmp_expr_2_JOBTODO;
		gimple varblock_2_JOBTODO;
		
		/* Variables for the 1-dimensional array option.  */
		gimple varblock_2b_JOBTODO;
		gimple_seq varblock_seq_2_JOBTODO;
		tree tmp_expr_2_JOBTODO_plus;
		
		/* Variables for the 'function call' option.  */
		gimple varblock_reading_call_2_JOBTODO;
		gimple varblock_writing_call_JOBTODO;
		
		/* Implemented as readings from and writings into 1-dimensional array.  */
		if (!use_varblock_function) {
			/* pre_varblock1d_2_JOBTODO = varblock[2][current].  */
			
			/* Get the sequence of statements that we need to add to access to varblock[2][current]
				in the 1-dimension version (varblock_1d). 
				In write_varblock_2 it is returned the node that points to the 
				original 2-dimensional varblock position
			*/
			tree write_varblock_2_JOBTODO;
			varblock_seq_2_JOBTODO = varblock_access (&pre_varblock_2_JOBTODO, &write_varblock_2_JOBTODO,
																							2, current_t_alone);

			/* pre_varblock_2_JOBTODO_plus = pre_varblock_2_JOBTODO + 1.  */
			tree tmp_expr_2_JOBTODO_plus = build_decl (gimple_location (omp_for_loop), VAR_DECL, 
																			//get_identifier ("pre_varblock1d_2_JOBTODO_plus"),
																			NULL,
																			integer_type_node);
			DECL_IGNORED_P (tmp_expr_2_JOBTODO_plus) = 1;
			//add_omp_clause(tmp_expr_2_JOBTODO_plus, OMP_CLAUSE_PRIVATE);

			tmp_expr_2_JOBTODO = build2 (PLUS_EXPR, integer_type_node, 
														pre_varblock_2_JOBTODO, build_int_cst(integer_type_node, 1));

			varblock_2b_JOBTODO = gimple_build_assign (tmp_expr_2_JOBTODO_plus, tmp_expr_2_JOBTODO);

			/* varblock[2][current] = pre_varblock_2_JOBTODO_plus.  */
			varblock_2_JOBTODO = gimple_build_assign ( build_simple_mem_ref(write_varblock_2_JOBTODO), 
																											tmp_expr_2_JOBTODO_plus);

		/* Implemented as a function call.  */
		} else {
			/* varblock[2][current].  */
			varblock_reading_call_2_JOBTODO = gimple_build_call (varblock_reading, 2, 
																							build_int_cst (integer_type_node, 2),
																							current_t_alone);
			/* D.xxx2 = varblock[2][current].  */
													
			gimple_call_set_lhs (varblock_reading_call_2_JOBTODO, pre_varblock_2_JOBTODO);
		
			/* D.xxx3 = D.xxx2 + 1.  */
			tmp_expr_2_JOBTODO = build2 (PLUS_EXPR, integer_type_node, 
														pre_varblock_2_JOBTODO, build_int_cst(integer_type_node, 1));

			varblock_2_JOBTODO = gimple_build_assign (pre_varblock_2_JOBTODO, tmp_expr_2_JOBTODO);
		
			/* varblock[2][current] = D.xxx3.  */
			varblock_writing_call_JOBTODO = gimple_build_call (varblock_writing, 3,
																		build_int_cst (integer_type_node, 2), 
																		current_t_alone, 
																		pre_varblock_2_JOBTODO);
		}

		/********************************** 
					INSERTING CODE POST-LOOP
		***********************************/
		

		/* 		if ( endLoop != 0 )
						goto label_true_endLoop;
			 		else
			 			goto label_endReturn;
			label_endReturn:
					if (endReturn != 0)
						got label_true_endLoop;
					else
						goto labelEndIteration;
			
			label_true_endLoop:
					freeWheel(current);
					goto labelEndLoop;
		*/
		
		/* TODO: Currently, it is not used ''endLoop'' and ''endReturn''.  */
		//gsi_insert_after (&gsi_last_stmt, if_endLoop, GSI_NEW_STMT);
		//gsi_insert_after (&gsi_last_stmt, label_endReturn, GSI_NEW_STMT);
		//gsi_insert_after (&gsi_last_stmt, if_endReturn, GSI_NEW_STMT);
		//gsi_insert_after (&gsi_last_stmt, label_true_endLoop, GSI_NEW_STMT);
		/* Insert 'current.x = current' before the function call.  */
		//gimple current_assign1 = gimple_build_assign (current_t_alone, current_t);
		//gsi_insert_after (&gsi_last_stmt, current_assign1, GSI_NEW_STMT);
		//gsi_insert_after (&gsi_last_stmt, freeWheel_call, GSI_NEW_STMT);
		//gsi_insert_after (&gsi_last_stmt, goto_labelEndLoop, GSI_NEW_STMT);


		/* label labelEndIteration.  */
		gsi_insert_after (&gsi_last_stmt, labelEndIteration, GSI_NEW_STMT);
		/*	
				D.xxx1 = varblock[1][current];
				D.xxx2 = D.xxx1 + ini;
		*/
		/* Insert 'current.x = current' before the function call.  */
		//gimple current_assign2 = gimple_build_assign (current_t_alone, current_t);
		//gsi_insert_after (&gsi_last_stmt, current_assign2, GSI_NEW_STMT);
		
		if (!use_varblock_function) {
			gsi_insert_seq_after (&gsi_last_stmt, varblock_seq_1, GSI_CONTINUE_LINKING);
		} else {
			gsi_insert_after (&gsi_last_stmt, varblock_reading_call_1, GSI_NEW_STMT);
		}

		gsi_insert_after (&gsi_last_stmt, varblock_1, GSI_NEW_STMT);
		/*
				if ( iteration != D.xxx2 ) 
					goto label_true_iteration_1;
				else
					goto labelSquash;

		label_true_iteration_1:
				for_ITERS = NITER - 1 + ini
				if ( iteration < for_ITERS )
					goto label_true_iteration_2;
				else
					goto labelSquash;

		label_true_iteration_2:
				iteration = iteration + 1;
				goto labelStartIteration;

		*/
		gsi_insert_after (&gsi_last_stmt, if_endIteration_1, GSI_NEW_STMT);
		gsi_insert_after (&gsi_last_stmt, label_true_iteration_1, GSI_NEW_STMT);
		gsi_insert_after (&gsi_last_stmt, g_final_niter_1, GSI_NEW_STMT);
		gsi_insert_after (&gsi_last_stmt, g_final_niter_ini, GSI_NEW_STMT);
		gsi_insert_after (&gsi_last_stmt, if_endIteration_2, GSI_NEW_STMT);
		gsi_insert_after (&gsi_last_stmt, label_true_iteration_2, GSI_NEW_STMT);
		gsi_insert_after (&gsi_last_stmt, iteration_plus_one, GSI_NEW_STMT);
		gsi_insert_after (&gsi_last_stmt, goto_labelStartIteration, GSI_NEW_STMT);


		/* label labelSquash.  */
		gsi_insert_after (&gsi_last_stmt, labelSquash, GSI_NEW_STMT);
		
		/*	retflag = threadend_pointer (&current).  */
		gsi_insert_after (&gsi_last_stmt, threadend_pointer_call, GSI_NEW_STMT);
		
		/* Insert 'current.x = current' after the function call to commit the changes on current.x.  */
		gimple current_assign_retflag = gimple_build_assign (current_t_alone, current_t);
		gsi_insert_after (&gsi_last_stmt, current_assign_retflag, GSI_NEW_STMT);
		
		/*	if (retflag == JOBDONE) goto labelEndLoop; else goto label_false_JOBDONE;
		
		label_false_JOBDONE:
				if (retflag == JOBTODO) goto label_true_JOBTODO; else goto labelEndLoop;
		
		label_true_JOBTODO:
				D.xxx1 = varblock[0][current];
				iteration = D.xxx1 + ini;

				D.xxx2 = varblock[2][current];
				D.xxx3 = D.xxx2 + 1;
				varblock[2][current] = D.xxx3;

				goto labelStartIteration;
		*/
		gsi_insert_after (&gsi_last_stmt, if_JOBDONE, GSI_NEW_STMT);
		gsi_insert_after (&gsi_last_stmt, label_false_JOBDONE, GSI_NEW_STMT);
		gsi_insert_after (&gsi_last_stmt, if_JOBTODO, GSI_NEW_STMT);
		gsi_insert_after (&gsi_last_stmt, label_true_JOBTODO, GSI_NEW_STMT);
		/* Insert 'current.x = current' before the function call.  */
		//gimple current_assign3 = gimple_build_assign (current_t_alone, current_t);
		//gsi_insert_after (&gsi_last_stmt, current_assign3, GSI_NEW_STMT);
			
		if (!use_varblock_function) {
			gsi_insert_seq_after (&gsi_last_stmt, varblock_seq_0_JOBTODO, GSI_CONTINUE_LINKING);
			gsi_insert_after (&gsi_last_stmt, varblock_0_JOBTODO, GSI_NEW_STMT);
			gsi_insert_seq_after (&gsi_last_stmt, varblock_seq_2_JOBTODO, GSI_CONTINUE_LINKING);
			gsi_insert_after (&gsi_last_stmt, varblock_2b_JOBTODO , GSI_NEW_STMT);
			gsi_insert_after (&gsi_last_stmt, varblock_2_JOBTODO , GSI_NEW_STMT);
		} else {
			gsi_insert_after (&gsi_last_stmt, varblock_reading_call_0_JOBTODO, GSI_NEW_STMT);
			gsi_insert_after (&gsi_last_stmt, varblock_0_JOBTODO, GSI_NEW_STMT);
			gsi_insert_after (&gsi_last_stmt, varblock_reading_call_2_JOBTODO, GSI_NEW_STMT);
			gsi_insert_after (&gsi_last_stmt, varblock_2_JOBTODO , GSI_NEW_STMT);
			gsi_insert_after (&gsi_last_stmt, varblock_writing_call_JOBTODO, GSI_NEW_STMT);
		}

		gsi_insert_after (&gsi_last_stmt, goto_labelStartIteration, GSI_NEW_STMT);
		

		/* label labelEndLoop.  */
		gsi_insert_after (&gsi_last_stmt, labelEndLoop, GSI_NEW_STMT);
		gsi_insert_after (&gsi_last_stmt, gimple_build_nop(), GSI_NEW_STMT);


	
	}
} // End of 'replace_for_loop()'




/* TODO: This needs to be updated withh the changes performed in 'scan_parallel_body'.  */
/* Helper function for to walk inside a function which contains spec-reads and -stores.

   Callback for walk_gimple_stmt used to scan the body of
	 an OpenMP parallel FOR loop.  */
static tree
scan_function_body (gimple_stmt_iterator *ptr_gsi, bool *handled_ops_p,
		 struct walk_stmt_info *wi)
{
	
	tree lhs, rhs;
	tree variable;
	gimple stmt = gsi_stmt (*ptr_gsi);
	const char * variable_name;
	
	int i, step = 0;

	//printf("Statement in FOR body:\n");
	//debug_gimple_stmt(stmt);
	//printf("\n");

	
	/* Searching for ocurrences of any of the speculative variables.  */
	/* TODO: Detect every kind of possible situation.  */
	switch (gimple_code (stmt))
	{
	
	/*****************/
	/* GIMPLE_FOR.  */
	/*****************/
	case GIMPLE_OMP_FOR:
		//printf("This is a GIMPLE_FOR\n");
		//replace_for_loop (ptr_gsi);
		break; // case GIMPLE_OMP_FOR


	/******************/
	/* GIMPLE_ASSIGN. */
	/******************/
	case GIMPLE_ASSIGN:
		//printf("This is a GIMPLE_ASSIGN\n");
		
		lhs = gimple_assign_lhs(stmt);
		rhs = gimple_assign_rhs1(stmt);

		//debug_tree (lhs);
		//debug_tree (rhs);


		/* Check the RHS. Searching for readings.  */
		/* TODO: Detect every kind of possible situation.  */
		/* TODO: What's happen whether rhs = v[i] + something;  */
		switch(TREE_CODE(rhs))
		{
		
		case VAR_DECL:
			/* Check if the variable found matches with one of the speculative variables.  */
			for (i = 0; i < spec_variables_num; i++)
			{
				if (operand_equal_p (rhs, spec_variables[i], OEP_PURE_SAME))
				{
					printf("Reading from speculative variable:  ");
					fflush(stdout);
					print_gimple_stmt (stdout, stmt, 0, 0);
					/* TODO: Check the value of 'index'.  */
					replace_speculative_load_in_function (ptr_gsi, lhs, rhs, current_t_alone);
				}
			}	
			break;

		case ARRAY_REF:
			//printf("RHS: This is an ARRAY_REF\n");
			
			/* In v[i], variable is v, and index is i.  */
			variable = TREE_OPERAND(rhs,0);
			//tree index = TREE_OPERAND(rhs,1);
			
			/* An ARRAY_REF could have more than one dimension. We recursively resolve the
			several dimensions until obtaining the VAR_DECL that represents the variable.  */
			while (TREE_CODE(variable) != VAR_DECL)
				variable = TREE_OPERAND (variable, 0);
			
			/* Check if the variable found matches with one of the speculative variables.  */
			for (i = 0; i < spec_variables_num; i++)
			{
				if (operand_equal_p (variable, spec_variables[i], OEP_PURE_SAME))
				{
					printf("Reading from speculative variable:  ");
					fflush(stdout);
					print_gimple_stmt (stdout, stmt, 0, 0);
					replace_speculative_load_in_function (ptr_gsi, lhs, rhs, current_t_alone);
				}
			}
			break;

		/* TODO:	*/
		case ADDR_EXPR:
				break;
		case PARM_DECL:
		case FIELD_DECL:
		case COMPONENT_REF:
		case MEM_REF:
				break;

		} // switch TREE_CODE(rhs)


		/* Check the LHS. Searching for writings.  */
		/* TODO: Detect every kind of possible situation.  */
		switch(TREE_CODE(lhs))
		{
		
		case VAR_DECL:
			//printf("LHS: This is an VAR_DECL\n");
			
			/* Check if the variable found matches with one of the speculative variables.  */
			for (i = 0; i < spec_variables_num; i++)
			{
				if (spec_variables[i] == NULL)
					printf("There are not any speculative variable detected.\n");
			
				if (operand_equal_p (lhs, spec_variables[i], OEP_PURE_SAME))
				{	
					printf("Writing into speculative variable:  ");
					fflush(stdout);
					print_gimple_stmt (stdout, stmt, 0, 0);
					/* TODO: Check the value of 'index'.  */
					replace_speculative_store_in_function(ptr_gsi, lhs, rhs, current_t_alone);
				}
			}
			break;

		case ARRAY_REF:
			//printf("LHS: This is an ARRAY_REF\n");
			variable = TREE_OPERAND(lhs,0);
			
			/* An ARRAY_REF could have more than one dimension. We recursively resolve the
			several dimensions until obtaining the VAR_DECL that represents the variable.  */
			while (TREE_CODE(variable) != VAR_DECL)
				variable = TREE_OPERAND (variable, 0);

			/* Check if the variable found matches with one of the speculative variables.  */
			for (i = 0; i < spec_variables_num; i++)
			{
				if (operand_equal_p (variable, spec_variables[i], OEP_PURE_SAME))
				{
					printf("Writing into speculative variable:  ");
					fflush(stdout);
					print_gimple_stmt (stdout, stmt, 0, 0);
					replace_speculative_store_in_function(ptr_gsi, lhs, rhs, current_t_alone);
				}
			}
			break;
		
		
		/* TODO:	*/
		case ADDR_EXPR:
				break;
		case PARM_DECL:
		case FIELD_DECL:
		case COMPONENT_REF:
		case MEM_REF:
				break;
		} // switch TREE_CODE(lhs)


		break; // case GIMPLE_ASSIGN


	/***********************************************/
	/* GIMPLE_CALL.                                */
	/* Looking for recursive or nested functions.  */
	/***********************************************/
	case GIMPLE_CALL:
		printf("Searching speculative readings or writings inside the function called: ");
		fflush(stdout);
		print_gimple_stmt (stdout, stmt, 0, 0);

		tree fn_decl = gimple_call_fndecl (stmt);
		const char * fn_name = IDENTIFIER_POINTER (DECL_NAME(fn_decl));
		
		/* TODO: In this point, all original functions has GIMPLIFY version, because we are
		in a copied version. Add support for nested functions.  */
		/* Search in the function body if there is some use of speculative variables. */
		
		/* If the function detected is the same the copied: Recursion. 
			We only need to replace the function call.  */
		if (is_A_Recursive_Function (&function_spec, current_function_decl, fn_decl)) {
		// TODO: Add the detection of nested functions.  */
			
			/* Check if this function was already copied and get its position.  */	
			int func_already_copied = -1;
			for (i = 0; i < function_spec.used; i++) {
				if (strcmp (function_spec.name[i], fn_name) == 0) 
					func_already_copied = i; // position of the copied function.
			}
			
			/* Create a copy of the function, but not the body. The copy of the body is deferred
			until the plugin has a GIMPLE version of the function being copied. 
			
			If the function was already copied, only get the tree representing the new function. */

			tree new_function;
			if (func_already_copied == -1) {
				new_function = create_copy_function(fn_decl, current_t_alone);
				insert_Function(&function_spec, fn_name, new_function);
			} else {
				new_function = function_spec.copied_function [func_already_copied];
			}

			/* Replace this gimple call by the new one.  */
			int num_args = gimple_call_num_args (stmt);
			
			gimple new_call = gimple_build_call (new_function, num_args + 1);
			
			/* Set arguments as the original function.  */
			int i;
			for (i = 0; i < num_args; i++) {
				gimple_call_set_arg (new_call, i, gimple_call_arg(stmt,i));	
			}
			
			/* Set as lhs an artificial variable to read the return value
			 of the speculative version of the function.  */
			tree new_lhs = build_decl (UNKNOWN_LOCATION, VAR_DECL, get_identifier("pre_if_var"), integer_type_node);
			gimple_call_set_lhs (new_call, new_lhs);
			
			/* If the original function has a left side, set it.  */
			//if (gimple_call_lhs (stmt) != NULL)
			//	gimple_call_set_lhs (new_call, gimple_call_lhs (stmt));

			/* Set 'current' as last argument. Here we dont use current_t_alone, 
				because we are inside a function.	We need to recovery the 'current'
				parameter of the function.  */
			tree param;
			for (param = DECL_ARGUMENTS (current_function_decl); param; param = DECL_CHAIN (param)) {
				const char * param_name = IDENTIFIER_POINTER (DECL_NAME(param));
				if (strcmp (param_name, "current") == 0) 
					gimple_call_set_arg (new_call, num_args, param);
			}

			/* Build the if-construction:
					if (function_spec (arg1, current) == -1)
						earlySquash(1);
			*/
			tree label_false_decl = create_artificial_label (UNKNOWN_LOCATION);
			gimple label_false = gimple_build_label (label_false_decl);
	
			gimple if_stmt = gimple_build_cond (EQ_EXPR, new_lhs, 
													build_int_cst (integer_type_node, -1),
													labelSquash_decl, label_false_decl);		

			// Insert the new function call and remove the old one
			printf("\tReplacing sequential function call by the copied version\n");
			/* Insert if (function_spec () == -1).  */
			gsi_insert_after(ptr_gsi, new_call, GSI_SAME_STMT);
			gsi_remove(ptr_gsi, true);
			gsi_insert_after(ptr_gsi, if_stmt, GSI_NEW_STMT);
			/* Insert label_false.  */
			gsi_insert_after (ptr_gsi, label_false, GSI_NEW_STMT);


		}

		break; // case GIMPLE_CALL
		
	
	/*****************/
	/* GIMPLE_BIND.  */
	/*****************/
	case GIMPLE_BIND:
		//printf("This is a GIMPLE_BIND\n");
		break;
	
	/****************/
	/* GIMPLE_ASM.  */
	/****************/
	case GIMPLE_ASM:
		//printf("This is a GIMPLE_ASM\n");
		break;
	
	/******************/
	/* GIMPLE_LABEL.  */
	/******************/
	case GIMPLE_LABEL:
		//printf("This is a GIMPLE_LABEL\n");
		break;
	
	/*****************/
	/* GIMPLE_COND.  */
	/*****************/
	case GIMPLE_COND:
		//printf("This is a GIMPLE_COND\n");
		break;

	/*****************/	
	/* GIMPLE_GOTO.  */
	/*****************/
	case GIMPLE_GOTO:
		//printf("This is a GIMPLE_GOTO\n");
		break;
	
	/****************/
	/* GIMPLE_NOP.  */
	/****************/
	case GIMPLE_NOP:
		//printf("This is a GIMPLE_NOP\n");
		break;
	
	/******************/
	/* GIMPLE_RETURN. */
	/******************/
	case GIMPLE_RETURN:
		//printf("This is a GIMPLE_RETURN\n");
		break;

	default:
		printf("THIS IS A NON-RECOGNIZED GIMPLE STATEMENT\n");
		print_gimple_stmt (stdout, stmt, 0, 0);
	
	} // switch (gimple_code(stmt))
	
	return NULL_TREE;
}



/* Helper function for to walk inside a function which contains spec-reads and -stores.

   Callback for operands in the body of an OpenMP FOR loop.  */

static tree
scan_function_ops(tree *tp, int *walk_subtrees, void *data)
{
	return NULL_TREE;
}











/* Helper function for handle_omp_parallel.

   Callback for walk_gimple_stmt used to scan the body of
	 an OpenMP parallel FOR loop.  */

/* TODO: It is not necessary a FOR loop. Fix this situation */

static tree
scan_omp_parallel_body(gimple_stmt_iterator *ptr_gsi, bool *handled_ops_p,
		 struct walk_stmt_info *wi)
{
	
	tree lhs, rhs, rhs1, rhs2, rhs3;
	tree variable;
	tree fn_decl;
	const char * fn_name;
	const char * variable_name;
	
	gimple stmt = gsi_stmt (*ptr_gsi);
	
	int rhs_op, i, step = 0;
	
/*
	printf("Statement in FOR body:\n");
	debug_gimple_stmt(stmt);
	printf("\n");
*/
	
	/* Searching for ocurrences of any of the speculative variables.  */
	/* TODO: Detect every kind of possible situation.  */
	switch (gimple_code (stmt))
	{
	
	/*****************/
	/* GIMPLE_COND.  */
	/*****************/
	case GIMPLE_COND:
		//printf("This is a GIMPLE_COND\n");
		
		/* Search for a MAX reduction expression:
			if (max < a) { max = a; }  */
		if (gimple_cond_code(stmt)  == LT_EXPR ) {
			
			cond_lhs = gimple_cond_lhs (stmt);
			cond_rhs = gimple_cond_rhs (stmt);
			
			/* Check if the lhs of the expression is a speculative variable.  */
			/* TODO: Detect every kind of possible situation.  */
			switch(TREE_CODE(cond_lhs))
			{
		
			case VAR_DECL:
				/* Check if the variable found matches with one of the speculative variables.  */
				for (i = 0; i < spec_variables_num; i++)
				{	
					if (operand_equal_p (cond_lhs, spec_variables[i], OEP_PURE_SAME))
					{	
						/* TODO: Check the value of 'index'.  */
						if (TREE_CODE(TREE_TYPE(cond_lhs)) != POINTER_TYPE) {
							printf("Comparing a speculative variable:  ");
							fflush(stdout);
							print_gimple_stmt (stdout, stmt, 0, 0);

							/* Check whether the body of the GIMPLE_COND contains max = a.  */
							cond_true_label = gimple_cond_true_label(stmt);
							//debug_tree (cond_true_label);
							check_label = true;

						}

					}
				}
				break;
			} // switch TREE_CODE(cond_lhs)

		}	// if (LT_EXPR)
		break;

	/******************/
	/* GIMPLE_LABEL.  */
	/******************/
	case GIMPLE_LABEL:
		//printf("This is a GIMPLE_LABEL\n");
		if (check_label) { // This comparison is set to avoid work
			
			/* Get the label and compare with the label from the GIMPLE_COND.  */
			tree label_tree = gimple_label_label (stmt);
			if (simple_cst_equal(cond_true_label, label_tree) == 1) {
			//if (operand_equal_p (cond_true_label, label_tree, OEP_PURE_SAME)) {
				printf("True label detected\n");
				
				// Get the next statement, which should be the GIMPLE_ASSIGN that we find
				gimple_stmt_iterator gsi_copy;
				gsi_copy = * ptr_gsi;
				gsi_next(&gsi_copy); // Advance the pointer, so we do not need to deal with this statement in the next iteration
				gimple stmt_assign = gsi_stmt (gsi_copy);
				//debug_gimple_stmt (stmt_assign);
				
				if (gimple_code(stmt_assign) == GIMPLE_ASSIGN && gimple_num_ops (stmt_assign) == 2 ) {
					
					tree assign_lhs = gimple_assign_lhs(stmt_assign);
					tree assign_rhs;
					
					// Checking for each RHS. Searching for readings.
					int ops = gimple_num_ops(stmt_assign);
					for (rhs_op = 1; rhs_op <= (ops - 1); rhs_op++) 
					{
						assign_rhs = gimple_op (stmt_assign, rhs_op);
						
						// Check if the LHS and the RHS are the same
						if ( operand_equal_p (cond_rhs, assign_rhs, OEP_PURE_SAME) &&
								operand_equal_p (cond_lhs, assign_lhs, OEP_PURE_SAME) ) {

								printf("Same lhs and rhs\n");
								// TODO: Apply MAX reduction
								replace_speculative_max(&gsi_copy, assign_lhs, assign_rhs, stmt_assign);
						}
					}
					
				}
				
			}
			
			// Label already checked
			check_label = false;
		}
		break;


	/*****************/
	/* GIMPLE_FOR.  */
	/*****************/
	case GIMPLE_OMP_FOR:
		//printf("This is a GIMPLE_FOR\n");
		replace_for_loop (ptr_gsi);

		break; // case GIMPLE_OMP_FOR


	/******************/
	/* GIMPLE_ASSIGN. */
	/******************/
	case GIMPLE_ASSIGN:
		//printf("This is a GIMPLE_ASSIGN\n");
		
		lhs = gimple_assign_lhs(stmt);
		//rhs = gimple_assign_rhs1(stmt);
		//debug_gimple_stmt(stmt);
		//debug_tree(lhs);
		//debug_tree(rhs);
		
		int ops = gimple_num_ops (stmt);
		
		/* Checking for each RHS. Searching for readings.  */
		for (rhs_op = 1; rhs_op <= (ops - 1); rhs_op++) 
		{
			rhs = gimple_op (stmt, rhs_op);
			
			/* TODO: Detect every kind of possible situation.  */
			switch(TREE_CODE(rhs))
			{
		
			case VAR_DECL:
				/* Check if the variable found matches with one of the speculative variables.  */
				for (i = 0; i < spec_variables_num; i++)
				{	
					if (operand_equal_p (rhs, spec_variables[i], OEP_PURE_SAME))
					{	
						/* TODO: Check the value of 'index'.  */
						if (TREE_CODE(TREE_TYPE(rhs)) != POINTER_TYPE) {
							
							/* Detect a situation of 'sum reduction'.  */
							if (operand_equal_p (lhs, rhs, OEP_PURE_SAME) && 
									gimple_assign_rhs_code(stmt)){
									
									printf("Sum Reduction:   ");
									fflush(stdout);
									print_gimple_stmt (stdout, stmt, 0, 0);
									replace_speculative_sum (ptr_gsi, lhs, rhs, stmt);

							} else {

								printf("Reading from speculative variable:  ");
								fflush(stdout);
								print_gimple_stmt (stdout, stmt, 0, 0);
								replace_speculative_load (ptr_gsi, lhs, rhs, stmt, rhs_op);
							}	
						}

						/* If we have a POINTER_TYPE, the reading or writing is deferred
						Save the lhs to detect a possible reading or writing.  */
						else if (TREE_CODE (TREE_TYPE(rhs)) == POINTER_TYPE) {
							saved_lhs = lhs;
						}

					}
				}
			
				/* If we detects the saved_lhs, replace it by the new lhs.  */
				if (saved_lhs != NULL_TREE && operand_equal_p (rhs, saved_lhs, OEP_PURE_SAME)) {
					saved_lhs = lhs;
				}

				break;

			case ARRAY_REF:
				//printf("RHS: This is an ARRAY_REF\n");
			
				/* In v[i], variable is v, and index is i.  */
				variable = TREE_OPERAND(rhs,0);
				//tree index = TREE_OPERAND(rhs,1);
			
				/* An ARRAY_REF could have more than one dimension. We recursively resolve the
				several dimensions until obtaining the VAR_DECL that represents the variable.  */
				while (TREE_CODE(variable) != VAR_DECL && TREE_CODE(variable) != PARM_DECL)
					variable = TREE_OPERAND (variable, 0);
			
				/* Check if the variable found matches with one of the speculative variables.  */
				for (i = 0; i < spec_variables_num; i++)
				{
					if (operand_equal_p (variable, spec_variables[i], OEP_PURE_SAME))
					{
						printf("Reading from speculative variable:  ");
						fflush(stdout);
						print_gimple_stmt (stdout, stmt, 0, 0);
						replace_speculative_load(ptr_gsi, lhs, rhs, stmt, rhs_op);
					}
				}
				break;
		
			/* TODO: A MEM_REF also could be a regular speculative (or not speculative) load, not a deferred load.  */
			/* If we detect a MEM_REF, compare with the saved_lhs. If they match, this 
			is a speculative_load.  */
			case MEM_REF:
				//debug_tree (rhs);
				//debug_tree (saved_lhs);
				//debug_tree (lhs);
				
				if (saved_lhs != NULL && operand_equal_p (TREE_OPERAND(rhs,0), saved_lhs, OEP_PURE_SAME)) {
					
					/* If lhs is POINTER_TYPE, we have another indirection (for example in int **ptr).  */
					if (TREE_CODE (TREE_TYPE (lhs)) == POINTER_TYPE) {
						saved_lhs = lhs;
					}	else {
						printf("Reading from speculative variable:  ");
						fflush(stdout);
						print_gimple_stmt (stdout, stmt, 0, 0);
							
						replace_speculative_load(ptr_gsi, lhs, rhs, stmt, rhs_op);
					}	
				}
				break;


			/* TODO:	*/
			case ADDR_EXPR:
				break;
			case PARM_DECL:
			case FIELD_DECL:
			case COMPONENT_REF:
				break;

			} // switch TREE_CODE(rhs)
		}


		/* Check the LHS. Searching for writings.  */
		/* TODO: Detect every kind of possible situation.  */
		switch(TREE_CODE(lhs))
		{
		
		case VAR_DECL:
			//printf("LHS: This is an VAR_DECL\n");
			
			/* Check if the variable found matches with one of the speculative variables.  */
			for (i = 0; i < spec_variables_num; i++)
			{
				if (spec_variables[i] == NULL)
					printf("There are not any speculative variable detected.\n");
			
				if (operand_equal_p (lhs, spec_variables[i], OEP_PURE_SAME))
				{	
					printf("Writing into speculative variable:  ");
					fflush(stdout);
					print_gimple_stmt (stdout, stmt, 0, 0);
					/* TODO: Check the value of 'index'.  */
					replace_speculative_store(ptr_gsi, lhs, rhs, stmt);
				}
			}
			break;

		case ARRAY_REF:
			//printf("LHS: This is an ARRAY_REF\n");
			variable = TREE_OPERAND(lhs,0);
			
			/* An ARRAY_REF could have more than one dimension. We recursively resolve the
			several dimensions until obtaining the VAR_DECL that represents the variable.  */
			while (TREE_CODE(variable) != VAR_DECL && TREE_CODE(variable) != PARM_DECL)
				variable = TREE_OPERAND (variable, 0);

			/* Check if the variable found matches with one of the speculative variables.  */
			for (i = 0; i < spec_variables_num; i++) 
			{
				if (spec_variables[i] == NULL)
					printf("There are not any speculative variable detected.\n");
					
				if (operand_equal_p (variable, spec_variables[i], OEP_PURE_SAME))
				{
					printf("Writing into speculative variable:  ");
					fflush(stdout);
					print_gimple_stmt (stdout, stmt, 0, 0);
					replace_speculative_store(ptr_gsi, lhs, rhs, stmt);
				}
			}
			break;
		
		/* TODO: A MEM_REF also could be a regular speculative (or not speculative) store, not a deferred store.  */
		/* If we detect a MEM_REF, compare with the saved_lhs. If they match, this 
		is a speculative_store.  */
		case MEM_REF:
				//debug_tree (lhs);
				//debug_tree (saved_lhs);
				if (saved_lhs != NULL && operand_equal_p (TREE_OPERAND(lhs,0), saved_lhs, OEP_PURE_SAME)) {

					printf("Writing into speculative variable:  ");
					fflush(stdout);
					print_gimple_stmt (stdout, stmt, 0, 0);
					
					replace_speculative_store(ptr_gsi, lhs, rhs, stmt);
				}
				break;	
		
		/* TODO:	*/
		case ADDR_EXPR:
				break;
		case PARM_DECL:
		case FIELD_DECL:
		case COMPONENT_REF:
				break;
		} // switch TREE_CODE(lhs)


		break; // case GIMPLE_ASSIGN


	/*****************/
	/* GIMPLE_CALL.  */
	/*****************/
	case GIMPLE_CALL:
		
		break; //FIXME: Currently this functionality is unplugged.

		fn_decl = gimple_call_fndecl (stmt);
		fn_name = IDENTIFIER_POINTER (DECL_NAME (fn_decl));
		
		/* Discard spec engine functions.  */
		if ( strcmp (fn_name, "varblock_reading") == 0 || 
				strcmp (fn_name, "varblock_writing") == 0 ||
				strcmp (fn_name, "threadend_pointer") == 0 ) break;
		
		printf("Searching speculative readings or writings inside the function called: ");
		fflush(stdout);
		print_gimple_stmt (stdout, stmt, 0, 0);
		
	
		//debug_gimple_seq (gimple_body (fn_decl));
		//debug_tree (DECL_SAVED_TREE (fn_decl));

		/* TODO: If the function is processed before we find the function call (using a prototype), 
		it is already in GIMPLE, and DECL_SAVED_TREE is empty.
		The gimple_body is also empty, because the function is only declared, not defined.  */
		tree variable_detected = walk_tree (&DECL_SAVED_TREE (fn_decl), search_spec, NULL, NULL);
		
		// TODO: This recovers the GIMPLE_SEQ of a function already processed, due to a prototype.
		/*gimple var_detected;
		if (!variable_detected) {
			//gimple_seq gs = recovery_function_body (&all_functions, fn_decl);
			//debug_gimple_seq (gs);
			//var_detected = walk_gimple_seq (gs, search_spec_gimple, search_spec_gimple_ops, NULL);
		}*/
		
		
		/* Search in the function body if there is some use of speculative variables. */
		//if (strcmp (fn_name, "function_name") == 0) {	
		if (variable_detected) {
			
			/* Check if this function was already copied.  */
			int func_already_copied = -1;
			for (i = 0; i < function_spec.used; i++) {
				if (strcmp (function_spec.name[i], fn_name) == 0) 
					func_already_copied = i; // position of the copied function.
			}
			
			/* Create a copy of the function, but not the body. The copy of the body is deferred
			until the plugin has a GIMPLE version of the function being copied. 
			
			If the function was already copied, only get the tree representing the new function. */

			tree new_function;
			if (func_already_copied == -1) {
				new_function = create_copy_function(fn_decl, current_t_alone);
				insert_Function(&function_spec, fn_name, new_function);
			} else {
				new_function = function_spec.copied_function [func_already_copied];
			}

			
			// Replace this gimple call by the new one.
			int num_args = gimple_call_num_args (stmt);
			
			gimple new_call = gimple_build_call (new_function, num_args + 1);
			
			// Set arguments as the original function
			for (i = 0; i < num_args; i++) {
				gimple_call_set_arg (new_call, i, gimple_call_arg(stmt,i));	
			}
			
			/* Set as lhs an artificial variable to read the return value
			 of the speculative version of the function.  */
			gimple_call_set_lhs (new_call, pre_if_var);
			//if (gimple_call_lhs (stmt) != NULL)
			//	gimple_call_set_lhs (new_call, gimple_call_lhs (stmt));

			// Set 'current' as last argument.
			gimple_call_set_arg (new_call, num_args, current_t_alone);

			/* Build the if-construction:
					if (function_spec (arg1, current) == -1)
						earlySquash(1);
			*/
			tree label_false_decl = create_artificial_label (UNKNOWN_LOCATION);
			gimple label_false = gimple_build_label (label_false_decl);
	
			gimple if_stmt = gimple_build_cond (EQ_EXPR, pre_if_var, 
													build_int_cst (integer_type_node, -1),
													labelSquash_decl, label_false_decl);										

			// Insert the new function call and remove the old one
			printf("\tReplacing sequential function call by the copied version\n");		
			
			/* Insert if (function_spec () == -1).  */
			gsi_insert_after(ptr_gsi, new_call, GSI_SAME_STMT);
			gsi_remove(ptr_gsi, true);
			gsi_insert_after(ptr_gsi, if_stmt, GSI_NEW_STMT);
			/* Insert label_false.  */
			gsi_insert_after (ptr_gsi, label_false, GSI_NEW_STMT);

		}

		break; // case GIMPLE_CALL
		

	/*****************/
	/* GIMPLE_BIND.  */
	/*****************/
	case GIMPLE_BIND:
		//printf("This is a GIMPLE_BIND\n");
		break;
	
	/****************/
	/* GIMPLE_ASM.  */
	/****************/
	case GIMPLE_ASM:
		//printf("This is a GIMPLE_ASM\n");
		break;
	
	/*****************/	
	/* GIMPLE_GOTO.  */
	/*****************/
	case GIMPLE_GOTO:
		//printf("This is a GIMPLE_GOTO\n");
		break;
	
	/****************/
	/* GIMPLE_NOP.  */
	/****************/
	case GIMPLE_NOP:
		//printf("This is a GIMPLE_NOP\n");
		break;
		
	default:
		printf("THIS IS A NON-RECOGNIZED GIMPLE STATEMENT\n");
		print_gimple_stmt (stdout, stmt, 0, 0);
	
	} // switch (gimple_code(stmt))
	
	return NULL_TREE;
}



/* Helper function for handle_omp_parallel.

   Callback for operands in the body of an OpenMP FOR loop.  */

static tree
scan_omp_parallel_ops(tree *tp, int *walk_subtrees, void *data)
{
	return NULL_TREE;
}


/* Function to handle the OMP PARALLEL directive.  
		'*ptr_gsi' points to the beginning of the OMP PARALLEL FOR loop.  */

/*	Currently, the speculative engine developed by the group only works with 
*	the FOR directive
**/

static void
handle_omp_parallel(gimple for_stmt, gimple_stmt_iterator *ptr_gsi)
{
	
	tree clauses;
	gimple_seq for_body;
	gimple stmt;
	int i;

	gimple_stmt_iterator gsi;

	gimple_seq for_loop;
	gimple for_loop_stmt;
	
	/* Saving the original position of the FOR loop.  */
	gsi_for_loop = * ptr_gsi;

	/* Insert specinit at the beginning of the code, 
		only if this is the first loop detected.  */
	/* XXX: Currently, specinit is added when the main function is detected!.  */
	/*if (!firstLoopDone) {
		insert_specinit ();
	}*/
	
	/* Insert specend at the end of the code.  */
	insert_specend(ptr_gsi);

	/* Extracting clauses from the pragma.  */
	clauses = gimple_omp_parallel_clauses(for_stmt);	
	
	/* DEBUG: print the clauses.  */
	print_clauses(clauses);
	
	/* First clearing spec_variables array and value_vars.  */
	memset(&spec_variables[0], 0, sizeof(spec_variables));
	memset(&value_vars[0], 0, sizeof(value_vars));
	
	saved_lhs = NULL_TREE;
	
	/* Extract speculative variables defined by the clause.  */
	extract_speculative_clauses (clauses);

	/* If there is not any speculative variables (pointed using the
		'speculative' clause), we do not augment the code.  */
	if (spec_variables[0] == NULL) {
		printf("This PARALLEL construction is not speculatively parallelized\n");
		printf("Aborting speculative parallelization\n");
		return;
	}

	/* It is possible that the programmer write the 'speculative' clause in a 
		PARALLEL directive without the FOR/LOOP directive.  */
	/* Is there a FOR directive after the PARALLEL directive detected?.  */
	if ( !forAfterParallel(for_stmt) ) {
		printf("There is NOT a FOR directive after the PARALLEL one\n");
		printf("Aborting speculative parallelization\n");
		return;
	}
	
	/* If in 'scan_stmt', we don't detect the use of 'specstart()', exiting the
	speculative parallelization, because we don't know the number of iterations.  */
	/* TODO: Maybe a better solution is to set by default the maximum number of iterations
	passed as argument in the makefile.  */
	if (!specstart_detected) {
		printf("There is NOT a 'specstart()' function before the OMP pragma\n");
		printf("Aborting speculative parallelization\n");
		return;
	}

	/* Add engine's PRIVATE variables to the clauses */
	//add_engine_vars(for_stmt, OMP_CLAUSE_PRIVATE);
	/* Add engine's SHARED variables to the clauses */
	add_engine_vars(for_stmt, OMP_CLAUSE_SHARED);
	
	/* Initialize some variables that we want to add in multiple 
	sections of the original source code.  */
	/* TODO: Check if there are problems for multiple loops.  */
	//if (!firstLoopDone) {
		initialize_variables();
	//}		
	

	/* Get the body of the FOR loop.  */
	for_body = gimple_omp_body(for_stmt);
	
	/* Walk along the sequence of gimple statements of the FOR body.  */
	/* After this function call, we know the number of iterations,
			stores in the global variable 'nIt'.  */
	/* TODO: Fix it. Now we dont know the number of iterations. Maybe is better to remove its use 
	of the ''insert_engine_functions()'' function.  */
	walk_gimple_seq (for_body, scan_omp_parallel_body, scan_omp_parallel_ops, NULL);
	
	/* Insert engine's functions before the pragma only with
	the first FOR loop detected.  */
	/* XXX: Currently unused -> REMOVE IT.  */
	if (!firstLoopDone) {
		printf("Adding engine's functions pre-loop\n");
		//TODO: Now specstart is added by the user and omp_set_num_thread() is added when 
		//specstart() GIMPLE CALL is detected.
		//insert_engine_functions(ptr_gsi, nIt); 
		firstLoopDone = true;
	}	
}


/* Function to handle the OMP FOR directive.  
		It replaces the original 'FOR' loop for another with different
		index, initial, final and incremental expressions. It also adds
		some code lines before and after the original loop body.  */

static void
handle_omp_for(gimple for_stmt, gimple_stmt_iterator *ptr_gsi)
{
	//debug_gimple_stmt (for_stmt);
	return;
}


/* Function to add 'omp_set_num_threads()' before the specstart() function
call.  */
static void
insert_omp_threads (gimple stmt, gimple_stmt_iterator *ptr_gsi)
{
	tree funcdecl;
	
	/*****************************************
	* Inserting omp_set_num_threads(threads)
	*****************************************/
	/* Set the number of threads */
	tree nthreads = build_int_cst (integer_type_node, threads);
	
	/* Build the function prototype:
	* 		void omp_set_num_threads(int) 
	*/
	tree void_int_fn_type = build_function_type_list (void_type_node, 
														integer_type_node, NULL_TREE);
	
	/* Create the function declaration.  */
	funcdecl = build_fn_decl ("omp_set_num_threads", void_int_fn_type);

	/* Build the GIMPLE_CALL and insert it.  */
	gimple omp_set_call = gimple_build_call (funcdecl, 1, nthreads);
	gsi_insert_before (ptr_gsi, omp_set_call, GSI_SAME_STMT);

	return;
}

/* Helper function for the plugin exec.

   Callback for walk_gimple_stmt used to scan the stmt pointed
	 by the gimple_stmt_iterator.  */

static tree
scan_stmt( gimple_stmt_iterator *ptr_gsi, bool *handled_ops_p,
		 struct walk_stmt_info *wi)
{
	tree lhs, rhs, call_fndecl;
	gimple stmt = gsi_stmt (*ptr_gsi);
	const char * function_decl_name = " ";
/*	
	//if (strcmp ( IDENTIFIER_POINTER(DECL_NAME(current_function_decl)), "readLoop") == 0) { 
		printf("New Statement:\n");
		debug_gimple_stmt(stmt);
		printf("\n");
	//}
*/

	/* We only want to find OpenMP directives at this time.  */
	switch (gimple_code (stmt))
	{
	case GIMPLE_OMP_PARALLEL:	
		printf("\nOpenMP pragma omp PARALLEL detected in line %d!\n", gimple_lineno(stmt));
		/* Save the stmt as global variable.  */
		/* TODO: Change the code to use the global instead the local in handle_omp_parallel.  */
		omp_for_loop = stmt;
		/* Handle the OMP PARALLEL.  */
		handle_omp_parallel(stmt, ptr_gsi);
		break;
		
	case GIMPLE_OMP_FOR:
		//printf("OpenMP pragma omp FOR detected in line %d!\n", gimple_lineno(stmt));
		/* Handle the OMP FOR.  */
		//handle_omp_for(stmt, ptr_gsi); 
		break;
	
	case GIMPLE_CALL:
		
		call_fndecl = gimple_call_fndecl (stmt);
		if (call_fndecl != NULL) {
		
			function_decl_name = IDENTIFIER_POINTER(DECL_NAME(gimple_call_fndecl(stmt)));
		
			//printf("Function call detected: %s\n", function_decl_name);
			if ( strcmp (function_decl_name, "specbegin") == 0) { 
				specstart_iterations = gimple_call_arg(stmt,0);
				//debug_tree (specstart_iterations); 
				insert_omp_threads (stmt, ptr_gsi);
				specstart_detected = true;
			}
		}
		break;

	case GIMPLE_ASSIGN:
		//print_gimple_stmt (stdout, stmt, 0, TDF_RAW);
		//debug_gimple_stmt(stmt);
		//tree rhs = gimple_assign_rhs_to_tree(stmt);
		//tree lhs = gimple_assign_lhs (stmt);
		//debug_tree(lhs);
		//tree rhs = gimple_assign_rhs1 (stmt);
		//debug_tree(rhs);
		break;
	}
	
	return NULL_TREE;

}

/* Helper function for the plugin exec.

   Callback for walk_gimple_stmt used to scan the stmt pointed
	 by the gimple_stmt_iterator.  */

static tree
scan_op(tree *tp, int *walk_subtrees, void *data)

{
	return NULL_TREE;
}



/* The callback that we have registered in the pass definition
* It is tripped when a compiler event is met.  */

gimple_seq body, body_copy;

static unsigned
specprag_exec(void)
{

	unsigned i;
	gimple stmt;
	gimple_stmt_iterator gsi; /* GIMPLE statement iterator.  */
	
	
	/* Get the name of the current function declaration.  */
	const char * function_decl_name = " ";
	function_decl_name = IDENTIFIER_POINTER (DECL_NAME(current_function_decl));
	printf("\n*****************************************\n");
	printf("Analyzing '%s()'...\n", function_decl_name);
	printf("*****************************************\n");
	
	/* TODO: Save the name and the body of this function. This is done to recover the 
	body of the functions prototyped.  */
	//insert_All_Functions (&all_functions, current_function_decl);

	/* FIXME: I dont know why at the end of the process, in each position is stored the 
	 gimple_seq body of the last function.  */
/*if (all_functions.used == 5) {
		for (i = 0; i < all_functions.used; i++) {
			printf("Function Name [%d]: %s\n", i, all_functions.name[i]);
			gimple_seq gss = all_functions.func_body[i];
			debug_gimple_seq (gss);
		}
	}*/

	/* Search if the current function is one of the marked functions to be copied.
	and get the original body.  */
	search_Original_Function (&function_spec, current_function_decl);
	
	/* Search if the current function is the copied version, and assign the body 
	of the original function.  */
	if (search_Copied_Function (&function_spec, current_function_decl)) {
		
		/* After copying the body of the function, transform it.  */
		set_context_local_variables (current_function_decl);
		replace_parm_decls (current_function_decl);
		
		/* We need to create some variables and labels in the copied function.  */
		prepare_copied_function (current_function_decl);
		
		/* Traverse the body of the new function to replace speculative loads and stores. */
		walk_gimple_seq (gimple_body(current_function_decl), 
										scan_function_body, scan_function_ops, NULL);
	}
	

	gimple_seq seq = gimple_body (current_function_decl);

  /* If this function is `main', save the first statement.  */
  if (DECL_NAME (current_function_decl)
		&& MAIN_NAME_P (DECL_NAME (current_function_decl))
		&& DECL_FILE_SCOPE_P (current_function_decl))
	{
    printf("Adding specinit() at the beginning of the function.\n");
			
		/* Get the gsi of the first statement of main. 'gsi_init' is a global variable.  */
		gsi_init = gsi_init_main (seq);

		/* Add the specinit() function.  */
		//insert_specinit();
	}

	/* Traverse the body of the current function.  */
	walk_gimple_seq (seq, scan_stmt, scan_op, NULL);
  
	/*for (gsi = gsi_start (seq); !gsi_end_p (gsi); gsi_next (&gsi))
	{
			stmt = gsi_stmt(gsi);
			printf("Plugin working in statement...:\n");
			debug_gimple_stmt(stmt);
	}*/


	printf("\nPlugin finished in function '%s()'.\n\n", function_decl_name);
	return 0;
}


/* Plugin specification.  */

static struct gimple_opt_pass specprag_pass =
{
  {
    GIMPLE_PASS,
    "specprag",                           /* name */
    specprag_gate,                        /* gate */
    specprag_exec,                        /* execute */
    NULL,                                 /* sub */
    NULL,                                 /* next */
    0,                                    /* static_pass_number */
    TV_NONE,                              /* tv_id */
    PROP_gimple_any,	//PROP_cfg | PROP_ssa,                  /* properties_required */
    0,                                    /* properties_provided */
    0,                                    /* properties_destroyed */
    0,                                    /* todo_flags_start */
    TODO_dump_func												/* todo_flags_finish */
  }
};


/* Return 0 on success or error code on failure.  */

int 
plugin_init(struct plugin_name_args *info, /* Argument information.  */
								struct plugin_gcc_version *ver) /* Version info of GCC.  */

{
	/* Used to tell the plugin-framework about where we want to be called in
	* the set of all passes. This is located in tree-pass.h. */
	struct register_pass_info pass;
	printf("\nPlugin initialized...\n\n");
	
	/* Validate the plugin with the correct version of GCC.  */
	if (strncmp(ver->basever, specprag_ver.basever, strlen("4.6.2")))
		return -1; /* Incorrect version of GCC */
	
	/* It is also possible to validate our plugin through plugin_default_version_check().  */
	//if (!plugin_default_version_check (ver, &gcc_version))
  //  return 1;

	/* Initialize the dynamic array to hold the functions being copied.  */
	init_Function (&function_spec);
	init_All_Functions (&all_functions);

	/* Setup the info to register with GCC telling when we want to be called and 
	* to what GCC should call, when it's time to be called.  */
	pass.pass = &specprag_pass.pass;
	
	/* Get called before GCC did the lower OpenMP pass (lower_omp) 
	* In this case, after the first mudflap1 pass. 	*/
	pass.reference_pass_name = "mudflap1";
	pass.ref_pass_instance_number = 1;
	pass.pos_op = PASS_POS_INSERT_AFTER;
	
	/* Tell GCC that we want to be called after the first mudflap1 pass.  */
	register_callback("specprag", PLUGIN_PASS_MANAGER_SETUP, NULL, &pass);

	/* Tell GCC some information about us, just for use in --help and --version.  */
	register_callback("specprag", PLUGIN_INFO, NULL, &specprag_info);

	/* Successful initialization.  */
		return 0;
		
}
