#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

/* Funtion to get the current time.  */
double gettime()
{
   struct timeval tv;
   gettimeofday(&tv, (void *)0);
   return (tv.tv_sec + 1.0e-6*tv.tv_usec);
}

/*******************************************/
/* User parameters set by the programmer:  */
/*******************************************/
// Array size 
#define MAX 100
// Number of iterations of the loop
#define NITER 1000000
//#define NITER 1


/*--------------------------------------------------------------------------*/

int main() {

	/* Local variables:
	* 	P represents the number of iteration
	*		Q represents an the array index
	*/
	int P, Q, i;
	
	/* Variables to change in order to test different type nodes.  */
	double array[MAX], aux, pre_sum, sum;
	FILE *file;
	
	pre_sum = 0;
	sum = 0;

	
	/* Open and read the file.  */
	if ((file = fopen("rand1000000.in", "r")) == NULL) {
		printf("Error opening the file\n ");
		exit(0);
	} else {
		fscanf(file, "%lf", &aux);
		for (i = 0; i < MAX; i++) {
			array[i] = aux;
			fscanf(file, "%lf", &aux);
		}
		fclose(file);
	}

	/* Sum the elements of the array.  */
	for (i = 0; i < MAX; i++)
		pre_sum = pre_sum + array[i];
	printf(" Sum of the elements of the array (before processing) = %f\n", pre_sum);
	
	/* Start profiling.  */
	double seq_init = 0;
	double seq_end = 0;
	double seq_time = 0;
	seq_init = gettime();

#ifndef SEQUENTIAL
	specstart(NITER);
#endif

	// The loop is not parallelizable because indexes of 
	// the array elements depends on the input values from a file
	#pragma omp parallel for default(none) \
		private(Q,P,aux) \
		speculative(array)
	for ( P = 0 ; P < NITER ; P++ ) {
	
		Q = P % (MAX) + 1;
		aux = array[Q-1];

		Q = (4 * (int)aux) % (MAX) + 1;
		array[Q-1] = aux;
	}
	
	/* Stop profiling.  */
	seq_end = gettime();
	seq_time += seq_end - seq_init;

	/********************/
	/* Writing results  */
	/********************/
	sum = 0;
	for (i = 0; i < MAX; i++)
		sum = sum + array[i];
	printf(" Sum of the elements of the array (after sequential execution) =  %f\n", sum);
	
	/****************************/
	/* Printing execution times */
	/****************************/
	fprintf(stderr, "Execution time: %f seg. \n", seq_time);
	
	/* Printing array elements.  */
	//for (i=0; i < MAX; i++)
	//	printf("array[%d]=%d\n", i, array[i]);

	return 0;
}


