#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#include <omp.h>
#include "specEngine.h"

/* Funtion to get the current time.  */
double gettime()
{
   struct timeval tv;
   gettimeofday(&tv, (void *)0);
   return (tv.tv_sec + 1.0e-6*tv.tv_usec);
}

/*******************************************/
/* User parameters set by the programmer:  */
/*******************************************/
// Array size 
#define MAX 100
// Number of iterations of the loop
#define NITER 1000000
//#define NITER 1

// Global data structure
int array[MAX];
int spec_array[3];

#define index 1
#define index2 2

/*--------------------------------------------------------------------------*/

int main() {

	/* Local variables:
	* 	P represents the number of iteration
	*		Q represents an the array index
	*/
	int P, Q, aux, i;
	FILE *file;
	int pre_sum = 0;
	int sum = 0;
	int aux_fscanf;
	
	specinit();

	/* Open and read the file.  */
	if ((file = fopen("rand1000000.in", "r")) == NULL) {
		printf("Error opening the file\n ");
		exit(0);
	} else {
		aux_fscanf = fscanf(file, "%d", &aux);
		for (i = 0; i < MAX; i++) {
			array[i] = aux;
			aux_fscanf = fscanf(file, "%d", &aux);
		}
		fclose(file);
	}

	/* Sum the elements of the array.  */
	for (i = 0; i < MAX; i++)
		pre_sum = pre_sum + array[i];
	printf(" Sum of the elements of the array (before processing) = %d\n", pre_sum);


	/* Initializations.  */
	spec_array[0] = 0;
	spec_array[1] = 0;
	spec_array[2] = 0;

	/* Start profiling.  */
	double seq_init = 0;
	double seq_end = 0;
	double seq_time = 0;
	seq_init = gettime();

	
	omp_set_num_threads(4);
	specstart(NITER);

	int value = 0;
	int iteration, ini, nIterations;
	// The loop is not parallelizable because indexes of 
	// the array elements depends on the input values from a file
//	#pragma omp parallel for default(none) \
//		private(Q,P,aux,array) \
//		speculative(spec_array)
	#pragma omp parallel for default(none) schedule(static) \
		private(Q,P,aux,iteration, ini, value, current, tid, retflag) \
		firstprivate(array) \
		lastprivate(array) \
		shared(spec_array, wheel_ns, wheel_ms, wheel, upper_limit, varblock)
	//{
	//#pragma omp for schedule(static)
	//for ( P = 0 ; P < NITER ; P++ ) {
	for (tid = 0; tid <= threads-1; tid++) {
		
		ini = 0;
		current = tid;
		iteration = varblock[0][current] + ini;

		if (iteration > upper_limit - 1)
				goto labelEndLoop_1;
		
		varblock[2][current] = varblock[2][current] + 1;

labelStartIteration_1: 	
		Q = iteration % (MAX) + 1;
		
		/***********************************************************************/
		// speccode: speculative store. Original line:
		//     array[Q-1] = aux;
		//			spec_array[0] = Q;
		//
		value = Q;
		specstore_pointer((unsigned char *) &(spec_array[0]), sizeof (spec_array[0]), 
														current, (unsigned char *) &value);
		/***********************************************************************/

		Q = (4 * iteration) % (MAX) + 1;
		
		/**********************************************************************/ 
		// speccode: speculative load. Original line:
		//		array[Q-1] = spec_array[0];
		if (
		specload_pointer((unsigned char *) &(spec_array[0]), sizeof (spec_array[0]),
												current, (unsigned char *) &value) == -1)
			earlySquash(1);
		array[Q-1] = value;
		/***********************************************************************/
		/************************************************/
		/* speccode: end of thread and commit						*/
		/************************************************/

labelEndIteration_1:
		if ((iteration != varblock[1][current] + ini) && (iteration < NITER - 1)) {
			iteration = iteration + 1;
			goto labelStartIteration_1;
		}
		
		/* Thread done, perform the commit */
labelSquash_1:
		retflag = threadend_pointer(&current);
		
		/* End if job done */
		if (retflag == JOBDONE) goto labelEndLoop_1;
		if (retflag == JOBTODO) {
		
			/*  Set loop variable */
			iteration = varblock[0][current] + ini;
			varblock[2][current] = varblock[2][current] + 1;
			goto labelStartIteration_1;
		}

labelEndLoop_1:
		; /*freeWheel(current);*/


	}
	//}
	
	/* Stop profiling.  */
	seq_end = gettime();
	seq_time += seq_end - seq_init;

	/********************/
	/* Writing results  */
	/********************/
	sum = 0;
	for (i = 0; i < MAX; i++)
		sum = sum + array[i];
	printf(" Sum of the elements of the array (after speculative execution) =  %d\n", sum);
	
	/****************************/
	/* Printing execution times */
	/****************************/
	fprintf(stderr, "Execution time: %f seg. \n", seq_time);
	
	/* Printing array elements.  */
	//for (i=0; i < MAX; i++)
	//	printf("array[%d]=%d\n", i, array[i]);

	return 0;
}


