#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#include <omp.h>

#include "specEngine.h"

/* Funtion to get the current time.  */
double gettime()
{
   struct timeval tv;
   gettimeofday(&tv, (void *)0);
   return (tv.tv_sec + 1.0e-6*tv.tv_usec);
}

/*******************************************/
/* User parameters set by the programmer:  */
/*******************************************/
// Array size 
#define MAX 100
// Number of iterations of the loop
#define NITER 100
//#define NITER 1

// Global data structure
int array[MAX];

/*--------------------------------------------------------------------------*/

int main() {

	/* Local variables:
	* 	P represents the number of iteration
	*		Q represents an the array index
	*/
	int i, j, k, Q, aux;
	FILE *file;
	int pre_sum = 0;
	int sum = 0;
	int aux_fscanf;
	
	/* Initialize data structures needed to speculative parallelization.  */
	specinit();

	/* Open and read the file.  */
	if ((file = fopen("rand1000000.in", "r")) == NULL) {
		printf("Error opening the file\n ");
		exit(0);
	} else {
		aux_fscanf = fscanf(file, "%d", &aux);
		for (i = 0; i < MAX; i++) {
			array[i] = aux;
			aux_fscanf = fscanf(file, "%d", &aux);
		}
		fclose(file);
	}

	/* Sum the elements of the array.  */
	for (i = 0; i < MAX; i++)
		pre_sum = pre_sum + array[i];
	printf(" Sum of the elements of the array (before processing) = %d\n", pre_sum);
	
	/* Start profiling.  */
	double seq_init = 0;
	double seq_end = 0;
	double seq_time = 0;
	seq_init = gettime();
	

	int value = 0;
	int iteration, ini;

	// The loop is not parallelizable because indexes of 
	// the array elements depends on the input values from a file
	for (	i = 0; i < NITER; i++) {
		for ( j = 0 ; j < (i-1) ; j++ ) {
		
		//***********************************************************************
		// speccode: OMP threading directive
		omp_set_num_threads(threads);
		// speccode: initialize speculation structures
		specstart(j-1);
		//***********************************************************************

		#pragma omp parallel for default(none) schedule(static) \
			private(Q, aux, k, iteration, ini, value, current, tid, retflag, nonSpeculative) \
			shared (array, wheel_ns, wheel_ms, wheel, upper_limit, varblock) \
			shared (i,j)
			//for (k = 0; k < (j-1) ; j++) 
			for (tid = 0; tid <= threads-1; tid++)
			{
				ini = 0;
				current = tid;
				iteration = varblock[0][current]+ini;
				// If there is not work to this thread, go out.
				if (iteration > upper_limit - 1)
					goto labelEndLoop_1;
				// Increment the number of attempts for this iteration.
				varblock[2][current] = varblock[2][current]+1;

labelStartIteration_1:
				//k = iteration;
				Q = iteration % (MAX) + 1; 
				
				/**********************************************************************/ 
				// speccode: speculative load. Original line:
				//	aux = array[Q-1];
				if (specload_pointer((unsigned char *) &(array[Q-1]), sizeof (array[Q-1]),
												current, (unsigned char *) &value) == -1)
					earlySquash(1);
				aux = value;
				/***********************************************************************/
				
				Q = (4 * aux) % (MAX) + 1;

				/***********************************************************************/
				// speccode: speculative store. Original line:
				//     array[Q-1] = aux;
				value = aux;
				specstore_pointer((unsigned char *) &(array[Q-1]), sizeof (array[Q-1]), 
														current, (unsigned char *) &value);
				/***********************************************************************/

labelEndIteration_1:
				if ((iteration != varblock[1][current] + ini) && (iteration < j-1 - 1)) {
					iteration = iteration + 1;
					goto labelStartIteration_1;
				}
		
				/* Thread done, perform the commit */
labelSquash_1:
				retflag = threadend_pointer(&current);
		
				/* End if job done */
				if (retflag == JOBDONE) goto labelEndLoop_1;
				if (retflag == JOBTODO) {
		
					/*  Set loop variable */
					iteration = varblock[0][current] + ini;
					varblock[2][current] = varblock[2][current] + 1;
					goto labelStartIteration_1;
				}

labelEndLoop_1:
				; /*freeWheel(current);*/

			} //end of k
		} // end of j
	} // end of i

	/* Stop profiling.  */
	seq_end = gettime();
	seq_time += seq_end - seq_init;

	/********************/
	/* Writing results  */
	/********************/
	sum = 0;
	for (i = 0; i < MAX; i++)
		sum = sum + array[i];
	printf(" Sum of the elements of the array (after sequential execution) =  %d\n", sum);
	
	/****************************/
	/* Printing execution times */
	/****************************/
	fprintf(stderr, "Execution time: %f seg. \n", seq_time);
	
	/* Printing array elements.  */
	//for (i=0; i < MAX; i++)
	//	fprintf(stderr,"array[%d]=%d\n", i, array[i]);

	return 0;
}


