
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

//***********************************************************************
// speccode: OpenMP header file included
#include <omp.h>
//***********************************************************************
//***********************************************************************
// speccode: Main header file with all common variables
#include "specEngine.h"
//************************************************************************

/*******************************************/
/* User parameters set by the programmer:  */
/*******************************************/
// Array size 
#define MAX 100
// Number of iterations of the parallel loop
#define NITER 1000000
//#define NITER 1

/* Funtion to get the current time.  */
double gettime()
{
   struct timeval tv;
   gettimeofday(&tv, (void *)0);
   return (tv.tv_sec + 1.0e-6*tv.tv_usec);
}

// Global data structure
int array[MAX][MAX];

/*--------------------------------------------------------------------------*/

int main() {

	/* Local variables:
	* 	P represents the number of iteration
	*		Q represents an the array index
	*/
	int P, Q, aux, i, j;
	FILE *file;
	int sum_parallel = 0;
	int sum_sequential = 0;
	int pre_sum = 0;
	int array_sequential[MAX][MAX];
	int aux_fscanf;

	/* Initialize data structures needed to speculative parallelization.  */
	specinit();

	/* Open and read the file.  */
	if ((file = fopen("rand1000000.in", "r")) == NULL) {
		printf("Error opening the file\n ");
		exit(0);
	} else {
		aux_fscanf = fscanf(file, "%d", &aux);
		for (i = 0; i < MAX; i++) {
			for (j = 0; j < MAX; j++) {
			array[i][j] = aux;
			array_sequential[i][j] = aux;
			aux_fscanf = fscanf(file, "%d", &aux);
			}
		}
		fclose(file);
	}

	/* Sum the elements of the array.  */
	for (i = 0; i < MAX; i++)
		for (j = 0; j < MAX; j++)
		pre_sum = pre_sum + array[i][j];
	printf(" Sum of the elements of the array (before processing) = %d\n", pre_sum);
		
	/* Print number of threads to be set  */
	printf(" Execution with %d threads.\n\n", threads);
		
	// The loop is not parallelizable because indexes of 
	// the array elements depends on the input values from a file

	//***********************************************************************
	// speccode: OMP threading directive
	omp_set_num_threads(threads);
	// speccode: initialize speculation structures
	specstart(NITER);
	//***********************************************************************
	
	// speccode: Speculative loop
	int iteration, ini, nIterations;
	int value = 0;
	int linear = 0;
	int itera_parallel = 0;

	/************************************************************************/
	// speccode: Start profiling
	double spec_time=0;
	double spec_start, spec_stop;
	spec_start = gettime();
	/***********************************************************************/
#define p0 0
#define p1 1
#define p2 2

#pragma omp parallel default(none)                           \
	private(Q,P,aux,linear, iteration, ini, value, \
					 current, tid, retflag, stdout, nonSpeculative)    \
	shared(array, wheel_ns, wheel_ms,    \
					 wheel, upper_limit, varblock) \
	reduction(+:itera_parallel)
{

#pragma omp for \
	schedule(static)
	
	//for ( P = 0 ; P < NITER ; P++ )
	for (tid = 0; tid <= threads - 1; tid++) {
	
		ini = 0;
		
		if (getFirstBlock(&current, &nonSpeculative) == 0) {
				goto labelEndLoop_1;
		}
		
		iteration = varblock[0][current] + ini;
		varblock[2][current] = varblock[2][current] + 1;

labelStartIteration_1: 
		//  For loop body code
						
		itera_parallel++; 
		Q = iteration % (MAX) + 1 - 3; // -3 because we need to compensate the 'p2+Q' sum (p2=2)

		/**********************************************************************/ 
		// speccode: speculative load. Original line:
		//    aux = array[Q-1];
		//
		// linear = (Q - 1);
		if (
		specload_pointer((unsigned char *) &(array[p2+Q][p0]), sizeof (array[p2+Q][p0]),
												current, (unsigned char *) &value) == -1)
			earlySquash(1);
		aux = value;
		/***********************************************************************/

		Q = (4 * aux) % (MAX) + 1 - 3; // -3 because we need to compensate the 'p2+Q' sum (p2=2)

		/***********************************************************************/
		// speccode: speculative store. Original line:
		//     array[Q-1] = aux;
		//
		value = aux;
		specstore_pointer((unsigned char *) &(array[p2+Q][p0]), sizeof (array[p2+Q][p0]), 
														current, (unsigned char *) &value);
		/***********************************************************************/

		/************************************************/
		/* speccode: end of thread and commit						*/
		/************************************************/

labelEndIteration_1:
		if ((iteration != varblock[1][current] + ini) && (iteration < NITER - 1)) {
			iteration = iteration + 1;
			goto labelStartIteration_1;
		}
		
		/* Thread done, perform the commit */
labelSquash_1:
		retflag = threadend_pointer(&current, &nonSpeculative);
		
		/* End if job done */
		if (retflag == JOBDONE) goto labelEndLoop_1;
		if (retflag == JOBTODO) {
		
			/*  Set loop variable */
			iteration = varblock[0][current] + ini;
			varblock[2][current] = varblock[2][current] + 1;
			goto labelStartIteration_1;
		}

labelEndLoop_1:
		; /*freeWheel(current);*/
	
	} /*End of the FOR loop*/

} // #pragma omp parallel

	/************************************************************************/
	// speccode: Stop measuring time
	spec_stop = gettime();
	spec_time += spec_stop - spec_start;
	/************************************************************************/

	/*****************************************************/
	/* Executing the loop in sequential to compare times */
	/*****************************************************/
	double seq_init = 0;
	double seq_end = 0;
	double seq_time = 0;
	int itera_sequential = 0;

	seq_init = gettime();
	for (P = 0; P < NITER; P++) {

		itera_sequential++;
		Q = P % (MAX) + 1 - 3;
		//aux = array_sequential[Q+2][Q - 1];
		aux = array_sequential[p2+Q][p0];

		Q = (4 * aux) % (MAX) + 1 - 3;
		//array_sequential[Q+2][Q - 1] = aux;
		array_sequential[p2+Q][p0] = aux;
		//printf("[%d][%d], ", Q-1, Q-1);
	}
		
	seq_end = gettime();
	seq_time += seq_end - seq_init;


	/****************************/
	/* Printing execution times */
	/****************************/
	printf("Sequential time: %f seg. ", seq_time);
	printf("Speculative time: %lf seconds.\n", spec_time);
	printf("Loss rate: %f times longer than sequential time\n\n", (spec_time/seq_time));

	/********************/
	/* Writing results  */
	/********************/
	sum_parallel=0;
	for (i = 0; i < MAX; i++)
		for (j = 0; j < MAX; j++)
		sum_parallel = sum_parallel + array[i][j];
	printf(" Sum of the elements of the array (after speculative execution) = %d\n", sum_parallel);

	sum_sequential=0;
	for (i = 0; i < MAX; i++)
		for (j = 0; j < MAX; j++)
		sum_sequential = sum_sequential + array_sequential[i][j];
	printf(" Sum of the elements of the array (after sequential execution) =  %d\n", sum_sequential);

	if (sum_parallel != sum_sequential) {
		printf("ERROR:\n");
		for (i = 0; i < MAX; i++) {
			for (j = 0; j < MAX; j++) {
			if (array[i][j] != array_sequential[i][j])
				printf("array[%d][%d]=%d, array_sequential[%d][%d]=%d\n", i, j, array[i][j], i, j, array_sequential[i][j]);
			}
		}
		printf("\n");
	}

	/* Printing array elements.  */
	//for (i=0; i < MAX; i++)
	//	printf("array[%d]=%d\n", i, array[i]);

	
	printf("Iterations done in parallel: %d\n", itera_parallel);
	printf("Iterations done in sequential: %d\n", itera_sequential);

	return 0;
}
