#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

//***********************************************************************
// speccode: OpenMP header file included
#include <omp.h>
//***********************************************************************
//***********************************************************************
// speccode: Main header file with all common variables
#include "specEngine.h"
//************************************************************************

/* Funtion to get the current time.  */
double gettime()
{
   struct timeval tv;
   gettimeofday(&tv, (void *)0);
   return (tv.tv_sec + 1.0e-6*tv.tv_usec);
}

/*******************************************/
/* User parameters set by the programmer:  */
/*******************************************/
// Array size 
#define MAX 100
// Number of iterations of the loop
#define NITER 1000000	//1.000.000
//#define NITER 1

// Global data structure
int a = 0;

/*--------------------------------------------------------------------------*/

int main() {

	/* Local variables:
	* 	P represents the number of iteration
	*		Q represents an the array index
	*/
	int P, Q, aux, i;
	FILE *file;
	int pre_sum = 0;
	int sum = 0;
	
	/* Initialize data structures needed to speculative parallelization.  */
	specinit();

	//***********************************************************************
	// speccode: OMP threading directive
	omp_set_num_threads(threads);
	// speccode: initialize speculation structures
	specstart(NITER);
	//***********************************************************************
	
	// speccode: Speculative loop
	int iteration, ini, nIterations;
	int value = 0;
	
	/* Start profiling.  */
	double seq_init = 0;
	double seq_end = 0;
	double seq_time = 0;
	seq_init = gettime();

	// The loop is not parallelizable because indexes of 
	// the array elements depends on the input values from a file
	#pragma omp parallel for default(none) schedule(static) \
		private (Q,P,aux) \
		private (iteration, ini, value, current, tid, retflag) \
		shared  (a, wheel_ns, wheel_ms, wheel, upper_limit, varblock)
	//for ( P = 0 ; P < NITER ; P++ ) {
	for (tid = 0; tid <= threads - 1; tid++) {
	
		ini = 0;
		current = tid;	
		iteration = varblock[0][current] + ini;

		if (iteration > upper_limit - 1)
				goto labelEndLoop_1;
		
		varblock[2][current] = varblock[2][current] + 1;

labelStartIteration_1: 

		Q = iteration % (MAX) + 1;

		//*************************************************************
		// speculative load. Original line:
		//		aux = Q - 1 + a;
		if (specload_pointer((unsigned char *) &a, sizeof (a),
												current, (unsigned char *) &value) == -1)
			earlySquash(1);
		aux = value + Q - 1;
		//*************************************************************
		
		Q = (4 * aux) % (MAX) + 1;

		//**************************************************************
		// speccode: speculative store. Original line:
		//     a = Q;
		//
		value = Q;
		specstore_pointer((unsigned char *) &a, sizeof (a), 
														current, (unsigned char *) &value);
		//*****************************************************************

labelEndIteration_1:
		if ((iteration != varblock[1][current] + ini) && (iteration < NITER - 1)) {
			iteration = iteration + 1;
			goto labelStartIteration_1;
		}
		
		/* Thread done, perform the commit */
labelSquash_1:
		retflag = threadend_pointer(&current);
		
		/* End if job done */
		if (retflag == JOBDONE) goto labelEndLoop_1;
		if (retflag == JOBTODO) {
		
			/*  Set loop variable */
			iteration = varblock[0][current] + ini;
			varblock[2][current] = varblock[2][current] + 1;
			goto labelStartIteration_1;
		}

labelEndLoop_1:
		; /*freeWheel(current);*/

	} //for loop
	
	/* Stop profiling.  */
	seq_end = gettime();
	seq_time += seq_end - seq_init;

	/********************/
	/* Writing results  */
	/********************/
	printf("Exit value of a = %d\n", a);
	
	/****************************/
	/* Printing execution times */
	/****************************/
	fprintf(stderr, "Execution time: %f seg. \n", seq_time);
	
	return 0;
}


