#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

/* Funtion to get the current time.  */
double gettime()
{
   struct timeval tv;
   gettimeofday(&tv, (void *)0);
   return (tv.tv_sec + 1.0e-6*tv.tv_usec);
}

/*******************************************/
/* User parameters set by the programmer:  */
/*******************************************/
// Array size 
#define MAX 100
// Number of iterations of the loop
#define NITER 1000000
//#define NITER 1

// Global data structure
int a[MAX];
struct mystruct {
		int vv[MAX][MAX];
	};

struct mystruct process (struct mystruct d) {
	
	int P, Q, aux;
	// The loop is not parallelizable because indexes of 
	// the array elements depends on the input values from a file
#ifndef SEQUENTIAL
	specstart(NITER);
#endif	
	#pragma omp parallel for default(none) schedule(static) \
		private(P, Q, aux) \
		shared (d) \
		speculative(a)
	for ( P = 0 ; P < NITER ; P++ ) {
		
		Q = P % (MAX) + 1;
		d.vv[P%MAX][P%MAX] = Q;
		aux = a[Q-1] + d.vv[P%MAX][P%MAX];
		
		Q = (4 * aux) % (MAX) + 1;
		a[Q-1] = aux;
		
	}
	
	return d;
}


/*--------------------------------------------------------------------------*/

int main() {

	/* Local variables:
	* 	P represents the number of iteration
	*		Q represents an the array index
	*/
	int i, j;
	FILE *file;
	int pre_sum = 0;
	int sum = 0;
	
	struct mystruct d;

	
	/* Initialize speculative vars.  */
	for (i = 0; i<MAX; i++) {
		a[i] = i;
		for (j = 0; j < MAX; j++) {
			d.vv[i][j] = i + j;
		}
	}
	
	/* Start profiling.  */
	double seq_init = 0;
	double seq_end = 0;
	double seq_time = 0;
	seq_init = gettime();
	
	struct mystruct d2 = process(d);	
	d2 = process (d);
	
	/* Stop profiling.  */
	seq_end = gettime();
	seq_time += seq_end - seq_init;

	/********************/
	/* Writing results  */
	/********************/
	
	int sum_a = 0;
	int sum_d = 0;

	for (i = 0; i < MAX; i++) {
		sum_a = sum_a + a[i];
		for ( j = 0; j < MAX; j++) {
			sum_d = sum_d + d2.vv[i][j];
		}
	}
	
	printf("Sum of values of a[ ] = %d\n", sum_a);
	printf("Sum of values of d.vv[ ][ ] = %d\n", sum_d);
	
	/****************************/
	/* Printing execution times */
	/****************************/
	fprintf(stderr, "Execution time: %f seg. \n", seq_time);
	
	return 0;
}


