/* TEST: Check if the statement with a binary RHS is correctly replaced. 

The speculative variables is local, so the compiler does not creates copies and we do not 
have to replace the entire RHS by the return value of specload. What the plugin has to do
is replace the correct RHS operand (by position) in the original statement.  

With global variable:
	aux_int.2 = aux_int; 	// Just one operand in the RHS. Plugin can replace it completely.
	aux = aux_int.2 + X;

With local variable:
	aux = aux_int + X;	// Plugin has to replace the correct operand.

Compiler creates copies, because a variable with the bit ADDRESSABLE set up cannot stay
in the RHS of a statement with more operands.

*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

/* Funtion to get the current time.  */
double gettime()
{
   struct timeval tv;
   gettimeofday(&tv, (void *)0);
   return (tv.tv_sec + 1.0e-6*tv.tv_usec);
}

/*******************************************/
/* User parameters set by the programmer:  */
/*******************************************/
// Array size 
#define MAX 4
// Number of iterations of the loop
#define NITER 6000

/*--------------------------------------------------------------------------*/

int main() {

	int aux_int = 3;
	int array[MAX];
	int aux, i, j;

	FILE *file;

	/* Open and read the file.  */
	if ((file = fopen("data.in", "r")) == NULL) {
		printf("Error opening the file\n ");
		exit(0);
	} else {
		j = fscanf(file, "%d", &aux);
		for (i = 0; i < MAX; i++) {
			array[i] = aux;
			j = fscanf(file, "%d", &aux);
		}
		fclose(file);
	}

	// Values of elements before processing:
	printf("Before:\n");
	printf("aux = %d\n", aux);

	/* Start profiling.  */
	double seq_init = 0;
	double seq_end = 0;
	double seq_time = 0;
	seq_init = gettime();
	
#ifndef SEQUENTIAL
	specstart(NITER);
#endif
	// The loop is not parallelizable because indexes of 
	// the array elements depends on the input values from a file
	#pragma omp parallel for schedule(static) \
		private(i) \
		lastprivate(aux) \
		shared(array) \
		speculative(aux_int)
	for ( i = 0 ; i < NITER ; i++ ) {
		aux = aux_int + array[i%4]/(i+1);
	}
	
	/* Stop profiling.  */
	seq_end = gettime();
	seq_time += seq_end - seq_init;

	/********************/
	/* Writing results  */
	/********************/
	// Values of elements before processing:
	printf("After:\n");
	printf("aux = %d\n",aux);
	
	/****************************/
	/* Printing execution times */
	/****************************/
	fprintf(stderr,"Execution time: %f seg. \n", seq_time);
	

	return 0;
}


