/* TEST: Check the correction of a load from a local speculative variable.

A statetement like this
	
	aux = aux_int,

where 'aux_int' is a speculative variable, is processed by the plugin and produces the 
following lines:

	specload (aux_int, value);
	aux = value;

If 'aux_int' is GLOBAL, there is no problem, because actually the compiler generates this:

	aux_int.1 = aux_int;
	specload (aux_int, value);
	aux_int.1 = value;
	aux = aux_int.1;

But if the variable is local, this auxiliar variables is not created, and the new assignment
is incorrect, because 'aux' is ADDRESSABLE. We first need to create a copy of the variable 
without this bit, and then assign it to 'value'.

*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

/* Funtion to get the current time.  */
double gettime()
{
   struct timeval tv;
   gettimeofday(&tv, (void *)0);
   return (tv.tv_sec + 1.0e-6*tv.tv_usec);
}

/*******************************************/
/* User parameters set by the programmer:  */
/*******************************************/
// Array size 
#define MAX 4
// Number of iterations of the loop
#define NITER 6000

/*--------------------------------------------------------------------------*/

int main() {

	int aux_int = 3;
	int array[MAX];
	int aux, i, j;

	FILE *file;

	/* Open and read the file.  */
	if ((file = fopen("data.in", "r")) == NULL) {
		printf("Error opening the file\n ");
		exit(0);
	} else {
		j = fscanf(file, "%d", &aux);
		for (i = 0; i < MAX; i++) {
			array[i] = aux;
			j = fscanf(file, "%d", &aux);
		}
		fclose(file);
	}

	// Values of elements before processing:
	printf("Before:\n");
	printf("aux = %d\n", aux);

	/* Start profiling.  */
	double seq_init = 0;
	double seq_end = 0;
	double seq_time = 0;
	seq_init = gettime();
	
#ifndef SEQUENTIAL
	specstart(NITER);
#endif
	// The loop is not parallelizable because indexes of 
	// the array elements depends on the input values from a file
	#pragma omp parallel for schedule(static) \
		private(i) \
		lastprivate(aux) \
		shared(array) \
		speculative(aux_int)
	for ( i = 0 ; i < NITER ; i++ ) {
		aux = aux_int;
	}
	
	/* Stop profiling.  */
	seq_end = gettime();
	seq_time += seq_end - seq_init;

	/********************/
	/* Writing results  */
	/********************/
	// Values of elements before processing:
	printf("After:\n");
	printf("aux = %d\n",aux);
	
	/****************************/
	/* Printing execution times */
	/****************************/
	fprintf(stderr,"Execution time: %f seg. \n", seq_time);
	

	return 0;
}


