/* TEST: Check if the statement with a binary RHS is correctly replaced. We have 
a speculative store into a local variable, with two operands as RHS. This test checks if the store is 
performed using the whole RHS, and not only one operand of the RHS.

With global variable:
	aux.2 = aux;
	aux_int.3 = D.3222 + aux.2;
	aux_int = aux_int.3; // Just one operand in the RHS. Plugin can replace it completely.

With local variable:
	aux_int = D.3222 + aux.1;

Compiler creates copies, because a variable with the bit ADDRESSABLE set up cannot stay
in the RHS of a statement with more operands.

*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

/* Funtion to get the current time.  */
double gettime()
{
   struct timeval tv;
   gettimeofday(&tv, (void *)0);
   return (tv.tv_sec + 1.0e-6*tv.tv_usec);
}

/*******************************************/
/* User parameters set by the programmer:  */
/*******************************************/
// Array size 
#define MAX 4
// Number of iterations of the loop
#define NITER 6000

/*--------------------------------------------------------------------------*/

int main() {

	int aux_int = 3;
	int array[MAX];
	int aux, i, j;

	FILE *file;

	/* Open and read the file.  */
	if ((file = fopen("data.in", "r")) == NULL) {
		printf("Error opening the file\n ");
		exit(0);
	} else {
		j = fscanf(file, "%d", &aux);
		for (i = 0; i < MAX; i++) {
			array[i] = aux;
			j = fscanf(file, "%d", &aux);
		}
		fclose(file);
	}

	// Values of elements before processing:
	printf("Before:\n");
	printf("aux_int = %d\n", aux_int);

	/* Start profiling.  */
	double seq_init = 0;
	double seq_end = 0;
	double seq_time = 0;
	seq_init = gettime();
	
#ifndef SEQUENTIAL
	specstart(NITER);
#endif
	// The loop is not parallelizable because indexes of 
	// the array elements depends on the input values from a file
	#pragma omp parallel for schedule(static) \
		private(i) \
		shared(array,aux) \
		speculative(aux_int)
	for ( i = 0 ; i < NITER ; i++ ) {
		aux_int = aux + array[i%4]/(i+1);
	}
	
	/* Stop profiling.  */
	seq_end = gettime();
	seq_time += seq_end - seq_init;

	/********************/
	/* Writing results  */
	/********************/
	// Values of elements before processing:
	printf("After:\n");
	printf("aux_int = %d\n",aux_int);
	
	/****************************/
	/* Printing execution times */
	/****************************/
	fprintf(stderr,"Execution time: %f seg. \n", seq_time);
	

	return 0;
}


