/* TEST: Check the correction of doing several speculative loads, and a speculative store in 
the same line:

	aux_int = aux_int2 + aux_int2,

where 'aux_int' and 'aux_int2' are speculative.

*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

/* Funtion to get the current time.  */
double gettime()
{
   struct timeval tv;
   gettimeofday(&tv, (void *)0);
   return (tv.tv_sec + 1.0e-6*tv.tv_usec);
}

/*******************************************/
/* User parameters set by the programmer:  */
/*******************************************/
// Array size 
#define MAX 4
// Number of iterations of the loop
#define NITER 6000

/*--------------------------------------------------------------------------*/

int main() {

	int aux_int = 3;
	int aux_int2 = 4;
	int array[MAX];
	int aux, i, j;

	FILE *file;

	/* Open and read the file.  */
	if ((file = fopen("data.in", "r")) == NULL) {
		printf("Error opening the file\n ");
		exit(0);
	} else {
		j = fscanf(file, "%d", &aux);
		for (i = 0; i < MAX; i++) {
			array[i] = aux;
			j = fscanf(file, "%d", &aux);
		}
		fclose(file);
	}

	// Values of elements before processing:
	printf("Before:\n");
	printf("aux_int = %d\n", aux_int);
	printf("aux_int2 = %d\n", aux_int2);

	/* Start profiling.  */
	double seq_init = 0;
	double seq_end = 0;
	double seq_time = 0;
	seq_init = gettime();
	
#ifndef SEQUENTIAL
	specstart(NITER);
#endif
	// The loop is not parallelizable because indexes of 
	// the array elements depends on the input values from a file
	#pragma omp parallel for schedule(static) \
		private(i) \
		shared(array) \
		speculative(aux_int, aux_int2)
	for ( i = 0 ; i < NITER ; i++ ) {
		aux_int = aux_int2 + aux_int2;
		aux_int2 = array[i%4];
	}
	
	/* Stop profiling.  */
	seq_end = gettime();
	seq_time += seq_end - seq_init;

	/********************/
	/* Writing results  */
	/********************/
	// Values of elements before processing:
	printf("After:\n");
	printf("aux_int = %d\n",aux_int);
	printf("aux_int2 = %d\n",aux_int2);
	
	/****************************/
	/* Printing execution times */
	/****************************/
	fprintf(stderr,"Execution time: %f seg. \n", seq_time);
	

	return 0;
}


