#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

/* Funtion to get the current time.  */
double gettime()
{
   struct timeval tv;
   gettimeofday(&tv, (void *)0);
   return (tv.tv_sec + 1.0e-6*tv.tv_usec);
}

/*******************************************/
/* User parameters set by the programmer:  */
/*******************************************/
// Array size 
#define MAX 4
// Number of iterations of the loop
#define NITER 6000

/*--------------------------------------------------------------------------*/

int main() {

	int array[MAX];
	double aux_double = 3.435;
	int aux, i, j;

	FILE *file;

	/* Open and read the file.  */
	if ((file = fopen("data.in", "r")) == NULL) {
		printf("Error opening the file\n ");
		exit(0);
	} else {
		j = fscanf(file, "%d", &aux);
		for (i = 0; i < MAX; i++) {
			array[i] = aux;
			j = fscanf(file, "%d", &aux);
		}
		fclose(file);
	}

	// Values of elements before processing:
	printf("Before:\n");
	printf("aux_double = %f\n",aux_double);

	/* Start profiling.  */
	double seq_init = 0;
	double seq_end = 0;
	double seq_time = 0;
	seq_init = gettime();

#ifndef SEQUENTIAL
	specstart(NITER);
#endif
	// The loop is not parallelizable because indexes of 
	// the array elements depends on the input values from a file
	#pragma omp parallel for default(none) schedule(static) \
		private(i,j,aux) \
		shared(array) \
		speculative(aux_double)
	for ( i = 0 ; i < NITER ; i++ ) {
		
		aux = array[i%4]/(i+1) + aux_double;
		aux_double = aux;
	}
	
	/* Stop profiling.  */
	seq_end = gettime();
	seq_time += seq_end - seq_init;

	/********************/
	/* Writing results  */
	/********************/
	// Values of elements before processing:
	printf("After:\n");
	printf("aux_double = %f\n",aux_double);
	
	/****************************/
	/* Printing execution times */
	/****************************/
	fprintf(stderr,"Execution time: %f seg. \n", seq_time);
	

	return 0;
}


