#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

/* Funtion to get the current time.  */
double gettime()
{
   struct timeval tv;
   gettimeofday(&tv, (void *)0);
   return (tv.tv_sec + 1.0e-6*tv.tv_usec);
}

/*******************************************/
/* User parameters set by the programmer:  */
/*******************************************/
// Array size 
#define MAX 100
// Number of iterations of the loop
#define NITER 1000000
//#define NITER 1

// Global data structure
int array[MAX];
int a[MAX];
int aa[MAX][MAX];
char b[MAX];
double c[MAX];
struct mystruct {
	struct innerstruct {
		int iv[MAX];
		int ivv[MAX][MAX];
	} id;
	int v[MAX];
	int vv[MAX][MAX];
	double vd[MAX];
	char vc[MAX];
} d;

/*--------------------------------------------------------------------------*/

int main() {

	/* Local variables:
	* 	P represents the number of iteration
	*		Q represents an the array index
	*/
	int P, Q, aux, i;
	FILE *file;
	int pre_sum = 0;
	int sum = 0;
	
	int aux_int, aux_int2, Q_int, Q_int2;
	int aux_struct, Q_struct;
	char aux_char, Q_char;
	double aux_double, Q_double;


	/* Open and read the file.  */
	if ((file = fopen("rand1000000.in", "r")) == NULL) {
		printf("Error opening the file\n ");
		exit(0);
	} else {
		fscanf(file, "%d", &aux);
		for (i = 0; i < MAX; i++) {
			array[i] = aux;
			fscanf(file, "%d", &aux);
		}
		fclose(file);
	}

	/* Initialize speculative vars.  */
	for (i = 0; i<MAX; i++) {
		a[i] = 0;
		aa[i][0] = 0;
		b[i] = 0;
		c[i] = 0;
		d.v[i] = 0;
		d.vv[i][0] = 0;
	}

	/* Sum the elements of the array.  */
	for (i = 0; i < MAX; i++)
		pre_sum = pre_sum + array[i];
	printf(" Sum of the elements of the array (before processing) = %d\n", pre_sum);
	
	/* Start profiling.  */
	double seq_init = 0;
	double seq_end = 0;
	double seq_time = 0;
	seq_init = gettime();

/* FIRST LOOP.  */
#ifndef SEQUENTIAL
	specstart(NITER);
#endif	
	// The loop is not parallelizable because indexes of 
	// the array elements depends on the input values from a file
	#pragma omp parallel for default(none) \
		private(Q,P,aux) \
		speculative(array)
	for ( P = 1; P <= NITER ; P++ ) {

		Q = P % (MAX) + 1;
		aux = array[Q-1];

		Q = (4 * aux) % (MAX) + 1;
		array[Q-1] = aux;
	}


/* SECOND LOOP.  */
	// The loop is not parallelizable because indexes of 
	// the array elements depends on the input values from a file
#ifndef SEQUENTIAL
	specstart(NITER);
#endif	
	#pragma omp parallel for default(none) schedule(static) \
		private(P, Q_int, Q_int2, Q_char, Q_double, Q_struct, \
		aux_int, aux_int2, aux_char, aux_double, aux_struct) \
		speculative(a,aa,b,c,d)
	for ( P = 0 ; P < NITER ; P++ ) {
		
		Q_int = P % (MAX) + 1;
		Q_int2 = P % (MAX) + 2;
		Q_char = P + 2;
		Q_double = P % (MAX) + 1;
		Q_struct = P % (MAX) + 2;
		
		aux_int = Q_int - 1 + a[P%MAX];
		aux_int2 = Q_int - 2 + aa[P%MAX][0];
		aux_char = (Q_char + b[P%MAX] - 16) % 255;
		aux_double = Q_double + c[P%MAX]; 
		aux_struct = Q_struct + d.id.ivv[P%MAX][0];

		
		Q_int = (4 * aux_int) % (MAX) + 1;
		Q_int2 = (3 * aux_int2) % (MAX) + 2;
		Q_char = (aux_char + 1) % 255;
		Q_double = (aux_double * 3)/100;
		Q_struct = (2 * aux_struct) % (MAX) + 1;
		
		a[P%MAX] = Q_int;
		aa[P%MAX][0] = Q_int2;
		b[P%MAX] = Q_char;
		c[P%MAX] = Q_double;
		d.id.ivv[P%MAX][0] = Q_struct;
	}


	/* Stop profiling.  */
	seq_end = gettime();
	seq_time += seq_end - seq_init;
	
	/********************/
	/* Writing results  */
	/********************/
	
	// First loop
	sum = 0;
	for (i = 0; i < MAX; i++)
		sum = sum + array[i];
	printf(" Sum of the elements of the array (after sequential execution) =  %d\n", sum);

	// Second loop
	int sum_a = 0;
	int sum_a2 = 0;
	char sum_b = 0;
	double sum_c = 0;
	int sum_d = 0;

	for (i = 0; i < MAX; i++) {
		sum_a = sum_a + a[i];
		sum_a2 = sum_a2 + aa[i][0];
		sum_b = sum_b + b[i]; 
		sum_c = sum_c + c[i]; 
		sum_d = sum_d + d.id.ivv[i][0];
	}
	
	sum_b = sum_b - 300 % 255;

	printf("Sum of values of a[ ] = %d\n", sum_a);
	printf("Sum of values of aa[ ][ ] = %d\n", sum_a2);
	printf("Sum of values of b[ ] = %c\n", sum_b);
	printf("Sum of values of c[ ] = %f\n", sum_c);
	printf("Sum of values of d.id.ivv[ ][ ] = %d\n", sum_d);

	/****************************/
	/* Printing execution times */
	/****************************/
	fprintf(stderr, "Execution time: %f seg. \n", seq_time);
	
	/* Printing array elements.  */
	//for (i=0; i < MAX; i++)
	//	printf("array[%d]=%d\n", i, array[i]);

	return 0;
}


