/* TEST: Check if the plugin handles correctly different operations over
a structure. Some previous versions of the plugin shows errors when we copy two structures 
and also write/read in a field which is an array:

 -> Warning: size of 'value_struct' is already "sizeof(tt)"; not changing to "sizeof(int)", where 
 int the size of each element of the array vv in the structure tt.

	Example:
			
			struct tt {
				int v;
				int vv[MAX];
			} tt1, tt2
			
			for (...) {
				
				vv[..][..] = ...;

				tt1 = tt2;
			}
*/


#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include "specEngine.h"

/* Funtion to get the current time.  */
double gettime()
{
   struct timeval tv;
   gettimeofday(&tv, (void *)0);
   return (tv.tv_sec + 1.0e-6*tv.tv_usec);
}

/*******************************************/
/* User parameters set by the programmer:  */
/*******************************************/
// Array size 
#define MAX 100
// Number of iterations of the loop
#define NITER 1000000
//#define NITER 1

// Global data structure
struct mystruct {
	double v;
	int vv[MAX][MAX];
};

struct mystruct d = {2};
struct mystruct d2 = {45};
/*--------------------------------------------------------------------------*/

int main() {

	/* Local variables:
	* 	P represents the number of iteration
	*		Q represents an the array index
	*/
	int P, i;
	int aux_struct, Q_struct;
	FILE *file;
	int pre_sum = 0;
	int sum = 0;
	
	specinit();

	/* Initialize speculative vars.  */
//	d.v = 2;
//	d2.v = 45;
	for (i = 0; i<MAX; i++) {
		d.vv[i][0] = 0;
		d2.vv[i][0] = 0;
	}

	/* Start profiling.  */
	double seq_init = 0;
	double seq_end = 0;
	double seq_time = 0;
	seq_init = gettime();
	
	
	// The loop is not parallelizable because indexes of 
	// the array elements depends on the input values from a file
	omp_set_num_threads(4);
	specstart(NITER);
	
	int value = 0;
	struct mystruct value_struct;
	int iteration, ini, nIterations;

	#pragma omp parallel for default(none) schedule(static) \
		private(P, Q_struct, aux_struct, iteration, ini, value, value_struct, current, tid, retflag) \
		shared (d, d2, wheel_ns, wheel_ms, wheel, upper_limit, varblock) 	\
//	for ( P = 0 ; P < NITER ; P++ ) {
	for (tid = 0; tid <= threads-1; tid++) {
		
		ini = 0;
		current = tid;
		iteration = varblock[0][current] + ini;

		if (iteration > upper_limit - 1)
				goto labelEndLoop_1;
		
		varblock[2][current] = varblock[2][current] + 1;

labelStartIteration_1: 	
		Q_struct = iteration % (MAX) + 2;
		
		/***********************************************************************/
		// speccode: speculative load. Original line:
		// 	aux_struct = Q_struct + d.vv[P%MAX][0];
		if ( specload_pointer ( (unsigned char *) &(d.vv[iteration%MAX][0]), sizeof(d.vv[iteration%MAX][0]),
															current, (unsigned char *) &value) == -1 )
			earlySquash(1);
		aux_struct = Q_struct + value;
		
		Q_struct = (2 * aux_struct) % (MAX) + 1;
		
		/***********************************************************************/
		// speccode: speculative store. Original line:
    //		d.vv[P%MAX][0] = Q_struct;
		value = Q_struct;
		specstore_pointer ( (unsigned char *) &(d.vv[iteration%MAX][0]), sizeof(d.vv[iteration%MAX][0]),
															current, (unsigned char *) &value); 
		
		
		//d2 = d;
		if ( specload_pointer ( (unsigned char *) &(d), sizeof(d),
															current, (unsigned char *) &value_struct) == -1 )
				earlySquash(1);
		
		specstore_pointer ( (unsigned char *) &(d2), sizeof(d2),
															current, (unsigned char *) &value_struct); 

		/************************************************/
		/* speccode: end of thread and commit						*/
		/************************************************/

labelEndIteration_1:
		if ((iteration != varblock[1][current] + ini) && (iteration < NITER - 1)) {
			iteration = iteration + 1;
			goto labelStartIteration_1;
		}
		
		/* Thread done, perform the commit */
labelSquash_1:
		retflag = threadend_pointer(&current);
		
		/* End if job done */
		if (retflag == JOBDONE) goto labelEndLoop_1;
		if (retflag == JOBTODO) {
		
			/*  Set loop variable */
			iteration = varblock[0][current] + ini;
			varblock[2][current] = varblock[2][current] + 1;
			goto labelStartIteration_1;
		}

labelEndLoop_1:
		; /*freeWheel(current);*/
	}
	
	/* Stop profiling.  */
	seq_end = gettime();
	seq_time += seq_end - seq_init;

	/********************/
	/* Writing results  */
	/********************/
	
	int sum_d = 0;
	int sum_d2 = 0;
	
	for (i = 0; i < MAX; i++) {
		sum_d = sum_d + d.vv[i][0];
		sum_d2 = sum_d2 + d2.vv[i][0];
	}
	
	printf("Sum of values of d.vv[ ][ ] = %d\n", sum_d);
	printf("Sum of values of d2.vv[ ][ ] = %d\n", sum_d2);
	
	/****************************/
	/* Printing execution times */
	/****************************/
	fprintf(stderr, "Execution time: %f seg. \n", seq_time);
	
	return 0;
}


