/* TEST: Check if the plugin handles correctly different operations over
a structure. Some previous versions of the plugin shows errors when we copy two structures 
and also write/read in a field which is an array:

 -> Warning: size of 'value_struct' is already "sizeof(tt)"; not changing to "sizeof(int)", where 
 int the size of each element of the array vv in the structure tt.

	Example:
			
			struct tt {
				int v;
				int vv[MAX];
			} tt1, tt2
			
			for (...) {
				
				vv[..][..] = ...;

				tt1 = tt2;
			}
*/


#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

/* Funtion to get the current time.  */
double gettime()
{
   struct timeval tv;
   gettimeofday(&tv, (void *)0);
   return (tv.tv_sec + 1.0e-6*tv.tv_usec);
}

/*******************************************/
/* User parameters set by the programmer:  */
/*******************************************/
// Array size 
#define MAX 100
// Number of iterations of the loop
#define NITER 1000000
//#define NITER 1

// Global data structure
struct mystruct {
	double v;
	int vv[MAX][MAX];
};

struct mystruct d = {2};
struct mystruct d2 = {45};
/*--------------------------------------------------------------------------*/

int main() {

	/* Local variables:
	* 	P represents the number of iteration
	*		Q represents an the array index
	*/
	int P, i;
	int aux_struct, Q_struct;
	FILE *file;
	int pre_sum = 0;
	int sum = 0;
	
	/* Initialize speculative vars.  */
//	d.v = 2;
//	d2.v = 45;
	for (i = 0; i<MAX; i++) {
		d.vv[i][0] = 0;
		d2.vv[i][0] = 0;
	}

	/* Start profiling.  */
	double seq_init = 0;
	double seq_end = 0;
	double seq_time = 0;
	seq_init = gettime();
	
	
	// The loop is not parallelizable because indexes of 
	// the array elements depends on the input values from a file
#ifndef SEQUENTIAL
	specstart(NITER);
#endif	
	#pragma omp parallel for default(none) schedule(static) \
		private(P, Q_struct, aux_struct) \
		speculative(d, d2)
	for ( P = 0 ; P < NITER ; P++ ) {
		
		Q_struct = P % (MAX) + 2;
		
		aux_struct = Q_struct + d.vv[P%MAX][0];
		//aux_struct = Q_struct + d.v;
		
		Q_struct = (2 * aux_struct) % (MAX) + 1;
		
		d.vv[P%MAX][0] = Q_struct;
		//d.v = Q_struct;
		d2 = d;
	}
	
	/* Stop profiling.  */
	seq_end = gettime();
	seq_time += seq_end - seq_init;

	/********************/
	/* Writing results  */
	/********************/
	
	int sum_d = 0;
	int sum_d2 = 0;
	
	for (i = 0; i < MAX; i++) {
		sum_d = sum_d + d.vv[i][0];
		sum_d2 = sum_d2 + d2.vv[i][0];
	}
	
	printf("Sum of values of d.vv[ ][ ] = %d\n", sum_d);
	printf("Sum of values of d2.vv[ ][ ] = %d\n", sum_d2);
	
	/****************************/
	/* Printing execution times */
	/****************************/
	fprintf(stderr, "Execution time: %f seg. \n", seq_time);
	
	return 0;
}


